# SPDX-FileCopyrightText: 2018-2022 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

BLENDER_IOR = 1.45
BLENDER_SPECULAR = 0.5
BLENDER_SPECULAR_TINT = 0.0

BLENDER_GLTF_SPECIAL_COLLECTION = "glTF_not_exported"

LIGHTS = {
    "POINT": "point",
    "SUN": "directional",
    "SPOT": "spot"
}
