# SPDX-FileCopyrightText: 2024 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# The purpose of this list is to present the permissions to be picked up by translation.

# The initial list of permissions is the one defined in the manifest schema
# (https://developer.blender.org/docs/features/extensions/schema/).

permissions = [
    "camera",
    "clipboard",
    "files",
    "microphone",
    "network",
]
