/* SPDX-FileCopyrightText: 2011-2022 Blender Foundation
 *
 * SPDX-License-Identifier: Apache-2.0 */

#pragma once

#include "util/types.h"  // IWYU pragma: export

#include "util/math_base.h"  // IWYU pragma: export

#include "util/math_int2.h"  // IWYU pragma: export
#include "util/math_int3.h"  // IWYU pragma: export
#include "util/math_int4.h"  // IWYU pragma: export
#include "util/math_int8.h"  // IWYU pragma: export

#include "util/math_float2.h"  // IWYU pragma: export
#include "util/math_float4.h"  // IWYU pragma: export
#include "util/math_float8.h"  // IWYU pragma: export

#include "util/math_float3.h"  // IWYU pragma: export

#include "util/rect.h"  // IWYU pragma: export
