# SPDX-FileCopyrightText: 2019-2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Do not edit this file. This file is auto generated from:
# ./tools/utils_doc/rna_manual_reference_updater.py

# autopep8: off
import bpy

url_manual_prefix = "https://docs.blender.org/manual/{:s}/{:d}.{:d}/".format(
    bpy.utils.manual_language_code(),
    *bpy.app.version[:2],
)

url_manual_mapping = (
    ("bpy.types.movietrackingsettings.refine_intrinsics_tangential_distortion*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-refine-intrinsics-tangential-distortion"),
    ("bpy.types.greasepencillineartmodifier.use_offset_towards_custom_camera*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-offset-towards-custom-camera"),
    ("bpy.types.greasepencillineartmodifier.use_overlap_edge_type_support*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-overlap-edge-type-support"),
    ("bpy.types.movietrackingsettings.refine_intrinsics_radial_distortion*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-refine-intrinsics-radial-distortion"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_max_trappedair*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-max-trappedair"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_min_trappedair*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-min-trappedair"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_max_wavecrest*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-max-wavecrest"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_min_wavecrest*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-min-wavecrest"),
    ("bpy.types.greasepencillineartmodifier.use_image_boundary_trimming*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-image-boundary-trimming"),
    ("bpy.types.movietrackingsettings.refine_intrinsics_principal_point*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-refine-intrinsics-principal-point"),
    ("bpy.types.greasepencillineartmodifier.use_face_mark_keep_contour*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-face-mark-keep-contour"),
    ("bpy.types.clothcollisionsettings.vertex_group_object_collisions*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-vertex-group-object-collisions"),
    ("bpy.types.gpencilsculptsettings.use_automasking_material_active*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-gpencilsculptsettings-use-automasking-material-active"),
    ("bpy.types.gpencilsculptsettings.use_automasking_material_stroke*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-gpencilsculptsettings-use-automasking-material-stroke"),
    ("bpy.types.cycleslightsettings.use_multiple_importance_sampling*", "render/cycles/light_settings.html#bpy-types-cycleslightsettings-use-multiple-importance-sampling"),
    ("bpy.types.fileassetselectparams.instance_collections_on_append*", "editors/asset_browser.html#bpy-types-fileassetselectparams-instance-collections-on-append"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_max_energy*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-max-energy"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_min_energy*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-min-energy"),
    ("bpy.types.greasepencillineartmodifier.use_face_mark_boundaries*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-face-mark-boundaries"),
    ("bpy.types.movietrackingsettings.refine_intrinsics_focal_length*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-refine-intrinsics-focal-length"),
    ("bpy.types.rigidbodyconstraint.rigidbodyconstraint.use_breaking*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-rigidbodyconstraint-use-breaking"),
    ("bpy.types.clothcollisionsettings.vertex_group_self_collisions*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-vertex-group-self-collisions"),
    ("bpy.types.cyclesrendersettings.preview_denoising_input_passes*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoising-input-passes"),
    ("bpy.types.cyclesrendersettings.preview_denoising_start_sample*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoising-start-sample"),
    ("bpy.types.fluiddomainsettings.sndparticle_sampling_trappedair*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-sampling-trappedair"),
    ("bpy.types.greasepencillineartmodifier.shadow_region_filtering*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-shadow-region-filtering"),
    ("bpy.types.fileassetselectparams.instance_collections_on_link*", "editors/asset_browser.html#bpy-types-fileassetselectparams-instance-collections-on-link"),
    ("bpy.types.fluiddomainsettings.sndparticle_sampling_wavecrest*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-sampling-wavecrest"),
    ("bpy.types.gpencilsculptsettings.use_automasking_layer_active*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-gpencilsculptsettings-use-automasking-layer-active"),
    ("bpy.types.gpencilsculptsettings.use_automasking_layer_stroke*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-gpencilsculptsettings-use-automasking-layer-stroke"),
    ("bpy.types.greasepencillineartmodifier.use_intersection_match*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-intersection-match"),
    ("bpy.types.materiallineart.use_intersection_priority_override*", "render/materials/line_art.html#bpy-types-materiallineart-use-intersection-priority-override"),
    ("bpy.types.preferencesedit.show_only_selected_curve_keyframes*", "editors/preferences/animation.html#bpy-types-preferencesedit-show-only-selected-curve-keyframes"),
    ("bpy.types.preferencesfilepaths.use_auto_save_temporary_files*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-auto-save-temporary-files"),
    ("bpy.types.preferencesinput.view_rotate_sensitivity_turntable*", "editors/preferences/navigation.html#bpy-types-preferencesinput-view-rotate-sensitivity-turntable"),
    ("bpy.types.rigidbodyconstraint.use_override_solver_iterations*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-use-override-solver-iterations"),
    ("bpy.types.toolsettings.use_gpencil_vertex_select_mask_stroke*", "grease_pencil/modes/vertex_paint/introduction.html#bpy-types-toolsettings-use-gpencil-vertex-select-mask-stroke"),
    ("bpy.types.toolsettings.use_grease_pencil_multi_frame_editing*", "grease_pencil/multiframe.html#bpy-types-toolsettings-use-grease-pencil-multi-frame-editing"),
    ("bpy.types.toolsettings.use_transform_correct_face_attributes*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-use-transform-correct-face-attributes"),
    ("bpy.types.view3doverlay.use_gpencil_onion_skin_active_object*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-onion-skin-active-object"),
    ("bpy.types.brushcurvessculptsettings.interpolate_point_count*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-interpolate-point-count"),
    ("bpy.types.cyclesrendersettings.adaptive_scrambling_distance*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-adaptive-scrambling-distance"),
    ("bpy.types.cyclesrendersettings.preview_adaptive_min_samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-adaptive-min-samples"),
    ("bpy.types.greasepencillineartmodifier.use_back_face_culling*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-back-face-culling"),
    ("bpy.types.greasepencillineartmodifier.use_intersection_mask*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-intersection-mask"),
    ("bpy.types.greasepencillineartmodifier.use_invert_collection*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-invert-collection"),
    ("bpy.types.greasepencillineartmodifier.use_invert_silhouette*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-invert-silhouette"),
    ("bpy.types.preferencesfilepaths.show_hidden_files_datablocks*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-show-hidden-files-datablocks"),
    ("bpy.types.rendersettings.compositor_denoise_preview_quality*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-compositor-denoise-preview-quality"),
    ("bpy.types.rendersettings.use_sequencer_override_scene_strip*", "editors/video_sequencer/preview/sidebar.html#bpy-types-rendersettings-use-sequencer-override-scene-strip"),
    ("bpy.types.toolsettings.use_annotation_project_only_selected*", "interface/annotate_tool.html#bpy-types-toolsettings-use-annotation-project-only-selected"),
    ("bpy.types.toolsettings.use_transform_correct_keep_connected*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-use-transform-correct-keep-connected"),
    ("bpy.types.brushgpencilsettings.use_auto_remove_fill_guides*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-use-auto-remove-fill-guides"),
    ("bpy.types.clothsettings.internal_compression_stiffness_max*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-compression-stiffness-max"),
    ("bpy.types.cyclesrendersettings.preview_denoising_prefilter*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoising-prefilter"),
    ("bpy.types.cyclesrendersettings.preview_scrambling_distance*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-scrambling-distance"),
    ("bpy.types.fluiddomainsettings.sndparticle_potential_radius*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-potential-radius"),
    ("bpy.types.greasepencillineartmodifier.light_contour_object*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-light-contour-object"),
    ("bpy.types.greasepencillineartmodifier.silhouette_filtering*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-silhouette-filtering"),
    ("bpy.types.greasepencillineartmodifier.use_crease_on_smooth*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-crease-on-smooth"),
    ("bpy.types.greasepencillineartmodifier.use_face_mark_invert*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-face-mark-invert"),
    ("bpy.types.objectlineart.use_intersection_priority_override*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart-use-intersection-priority-override"),
    ("bpy.types.preferencesedit.grease_pencil_euclidean_distance*", "editors/preferences/editing.html#bpy-types-preferencesedit-grease-pencil-euclidean-distance"),
    ("bpy.types.preferencesedit.grease_pencil_manhattan_distance*", "editors/preferences/editing.html#bpy-types-preferencesedit-grease-pencil-manhattan-distance"),
    ("bpy.types.preferencesinput.mouse_emulate_3_button_modifier*", "editors/preferences/input.html#bpy-types-preferencesinput-mouse-emulate-3-button-modifier"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_strength*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-strength"),
    ("bpy.types.clothsettings.vertex_group_structural_stiffness*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-vertex-group-structural-stiffness"),
    ("bpy.types.cyclesrendersettings.film_transparent_roughness*", "render/cycles/render_settings/film.html#bpy-types-cyclesrendersettings-film-transparent-roughness"),
    ("bpy.types.cyclesrendersettings.preview_adaptive_threshold*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-adaptive-threshold"),
    ("bpy.types.fluiddomainsettings.openvdb_cache_compress_type*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-openvdb-cache-compress-type"),
    ("bpy.types.fluiddomainsettings.sndparticle_bubble_buoyancy*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-bubble-buoyancy"),
    ("bpy.types.fluiddomainsettings.sndparticle_combined_export*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-combined-export"),
    ("bpy.types.fluiddomainsettings.use_collision_border_bottom*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-bottom"),
    ("bpy.types.fluiddomainsettings.vector_scale_with_magnitude*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-scale-with-magnitude"),
    ("bpy.types.greasepencillineartmodifier.stroke_depth_offset*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-stroke-depth-offset"),
    ("bpy.types.greasepencillineartmodifier.use_crease_on_sharp*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-crease-on-sharp"),
    ("bpy.types.movietrackingstabilization.use_2d_stabilization*", "movie_clip/tracking/clip/sidebar/stabilization/panel.html#bpy-types-movietrackingstabilization-use-2d-stabilization"),
    ("bpy.types.preferencesedit.keyframe_new_interpolation_type*", "editors/preferences/animation.html#bpy-types-preferencesedit-keyframe-new-interpolation-type"),
    ("bpy.types.preferencesedit.use_auto_keyframe_insert_needed*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-auto-keyframe-insert-needed"),
    ("bpy.types.preferencesedit.use_fcurve_high_quality_drawing*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-fcurve-high-quality-drawing"),
    ("bpy.types.rendersettings.compositor_denoise_final_quality*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-compositor-denoise-final-quality"),
    ("bpy.types.sequencertimelineoverlay.waveform_display_style*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-waveform-display-style"),
    ("bpy.types.spacegrapheditor.use_auto_lock_translation_axis*", "editors/graph_editor/introduction.html#bpy-types-spacegrapheditor-use-auto-lock-translation-axis"),
    ("bpy.types.toolsettings.annotation_stroke_placement_view2d*", "interface/annotate_tool.html#bpy-types-toolsettings-annotation-stroke-placement-view2d"),
    ("bpy.types.toolsettings.annotation_stroke_placement_view3d*", "interface/annotate_tool.html#bpy-types-toolsettings-annotation-stroke-placement-view3d"),
    ("bpy.ops.object.simulation_nodes_cache_calculate_to_frame*", "physics/simulation_nodes.html#bpy-ops-object-simulation-nodes-cache-calculate-to-frame"),
    ("bpy.types.brushcurvessculptsettings.density_add_attempts*", "sculpt_paint/curves_sculpting/brushes/density_curves.html#bpy-types-brushcurvessculptsettings-density-add-attempts"),
    ("bpy.types.brushgpencilsettings.use_random_press_strength*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-strength"),
    ("bpy.types.cyclesmaterialsettings.use_bump_map_correction*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-use-bump-map-correction"),
    ("bpy.types.cyclesrendersettings.preview_denoising_quality*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoising-quality"),
    ("bpy.types.cyclesrendersettings.preview_denoising_use_gpu*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoising-use-gpu"),
    ("bpy.types.fluiddomainsettings.use_collision_border_front*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-front"),
    ("bpy.types.fluiddomainsettings.use_collision_border_right*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-right"),
    ("bpy.types.greasepencillineartmodifier.shadow_camera_near*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-shadow-camera-near"),
    ("bpy.types.greasepencillineartmodifier.shadow_camera_size*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-shadow-camera-size"),
    ("bpy.types.preferencesedit.collection_instance_empty_size*", "editors/preferences/editing.html#bpy-types-preferencesedit-collection-instance-empty-size"),
    ("bpy.types.sequencertimelineoverlay.waveform_display_type*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-waveform-display-type"),
    ("bpy.types.toolsettings.snap_angle_increment_3d_precision*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-angle-increment-3d-precision"),
    ("bpy.types.toolsettings.use_gpencil_project_only_selected*", "grease_pencil/modes/draw/stroke_placement.html#bpy-types-toolsettings-use-gpencil-project-only-selected"),
    ("bpy.types.unifiedpaintsettings.use_unified_input_samples*", "sculpt_paint/brush/stroke.html#bpy-types-unifiedpaintsettings-use-unified-input-samples"),
    ("bpy.types.view3doverlay.use_normals_constant_screen_size*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-normals-constant-screen-size"),
    ("bpy.types.brushgpencilsettings.random_saturation_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-random-saturation-factor"),
    ("bpy.types.brushgpencilsettings.use_settings_postprocess*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-settings-postprocess"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_radius*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-radius"),
    ("bpy.types.cyclesmaterialsettings.use_transparent_shadow*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-use-transparent-shadow"),
    ("bpy.types.cyclesobjectsettings.use_adaptive_subdivision*", "render/cycles/object_settings/adaptive_subdiv.html#bpy-types-cyclesobjectsettings-use-adaptive-subdivision"),
    ("bpy.types.cyclesrendersettings.debug_use_spatial_splits*", "render/cycles/render_settings/performance.html#bpy-types-cyclesrendersettings-debug-use-spatial-splits"),
    ("bpy.types.cyclesrendersettings.guiding_training_samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-guiding-training-samples"),
    ("bpy.types.cyclesrendersettings.light_sampling_threshold*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-light-sampling-threshold"),
    ("bpy.types.cyclesrendersettings.rolling_shutter_duration*", "render/cycles/render_settings/motion_blur.html#bpy-types-cyclesrendersettings-rolling-shutter-duration"),
    ("bpy.types.cyclesrendersettings.volume_preview_step_rate*", "render/cycles/render_settings/volumes.html#bpy-types-cyclesrendersettings-volume-preview-step-rate"),
    ("bpy.types.fileassetselectparams.asset_library_reference*", "editors/asset_browser.html#bpy-types-fileassetselectparams-asset-library-reference"),
    ("bpy.types.fluiddomainsettings.sndparticle_update_radius*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-update-radius"),
    ("bpy.types.fluiddomainsettings.use_collision_border_back*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-back"),
    ("bpy.types.fluiddomainsettings.use_collision_border_left*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-left"),
    ("bpy.types.greasepencillineartmodifier.shadow_camera_far*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-shadow-camera-far"),
    ("bpy.types.greasepencillineartmodifier.use_custom_camera*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-custom-camera"),
    ("bpy.types.greasepencillineartmodifier.use_light_contour*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-light-contour"),
    ("bpy.types.preferencesedit.use_anim_channel_group_colors*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-anim-channel-group-colors"),
    ("bpy.types.preferencesedit.use_keyframe_insert_available*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-keyframe-insert-available"),
    ("bpy.types.preferencesexperimental.show_asset_debug_info*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-show-asset-debug-info"),
    ("bpy.types.preferencesfilepaths.use_scripts_auto_execute*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-scripts-auto-execute"),
    ("bpy.types.preferencesinput.ndof_show_guide_orbit_center*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-show-guide-orbit-center"),
    ("bpy.types.rendersettings_simplify_gpencil_view_modifier*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-view-modifier"),
    ("bpy.types.sequencertimelineoverlay.show_strip_tag_color*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-strip-tag-color"),
    ("bpy.types.spaceoutliner.use_filter_object_grease_pencil*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-grease-pencil"),
    ("bpy.types.view3doverlay.use_gpencil_multiedit_line_only*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-multiedit-line-only"),
    ("bpy.ops.grease_pencil.vertex_color_brightness_contrast*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-vertex-color-brightness-contrast"),
    ("bpy.types.brushcurvessculptsettings.interpolate_length*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-interpolate-length"),
    ("bpy.types.brushcurvessculptsettings.interpolate_radius*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-interpolate-radius"),
    ("bpy.types.brushgpencilsettings.use_random_press_radius*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-radius"),
    ("bpy.types.brushgpencilsettings.use_settings_stabilizer*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-settings-stabilizer"),
    ("bpy.types.clothsettings.internal_compression_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-compression-stiffness"),
    ("bpy.types.clothsettings.internal_tension_stiffness_max*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-tension-stiffness-max"),
    ("bpy.types.collection.use_lineart_intersection_priority*", "scene_layout/collections/collections.html#bpy-types-collection-use-lineart-intersection-priority"),
    ("bpy.types.colormanagedsequencercolorspacesettings.name*", "render/color_management.html#bpy-types-colormanagedsequencercolorspacesettings-name"),
    ("bpy.types.cyclesrendersettings.max_transparent_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-max-transparent-bounces"),
    ("bpy.types.cyclesrendersettings.min_transparent_bounces*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-min-transparent-bounces"),
    ("bpy.types.fluiddomainsettings.use_collision_border_top*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-collision-border-top"),
    ("bpy.types.geometrynodecamerainfogeometrynodecamerainfo*", "modeling/geometry_nodes/input/scene/camera_info.html#bpy-types-geometrynodecamerainfogeometrynodecamerainfo"),
    ("bpy.types.gpencilsculptsettings.intersection_threshold*", "grease_pencil/modes/draw/tools/trim.html#bpy-types-gpencilsculptsettings-intersection-threshold"),
    ("bpy.types.gpencilsculptsettings.use_automasking_stroke*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-gpencilsculptsettings-use-automasking-stroke"),
    ("bpy.types.gpencilsculptsettings.use_multiframe_falloff*", "grease_pencil/multiframe.html#bpy-types-gpencilsculptsettings-use-multiframe-falloff"),
    ("bpy.types.greasepencillineartmodifier.crease_threshold*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-crease-threshold"),
    ("bpy.types.greasepencillineartmodifier.smooth_tolerance*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-smooth-tolerance"),
    ("bpy.types.greasepencillineartmodifier.use_intersection*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-intersection"),
    ("bpy.types.movietrackingsettings.use_keyframe_selection*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-use-keyframe-selection"),
    ("bpy.types.preferencesexperimental.override_auto_resync*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-override-auto-resync"),
    ("bpy.types.preferencesfilepaths.animation_player_preset*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-animation-player-preset"),
    ("bpy.types.preferencesfilepaths.i18n_branches_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-i18n-branches-directory"),
    ("bpy.types.preferencesfilepaths.render_output_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-render-output-directory"),
    ("bpy.types.rendersettings.simplify_gpencil_antialiasing*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-antialiasing"),
    ("bpy.types.sequencertimelineoverlay.show_strip_duration*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-strip-duration"),
    ("bpy.types.sequencertoolsettings.snap_to_strips_preview*", "editors/video_sequencer/preview/controls/snapping.html#bpy-types-sequencertoolsettings-snap-to-strips-preview"),
    ("bpy.types.spaceoutliner.use_filter_lib_override_system*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-lib-override-system"),
    ("bpy.types.toolsettings.gpencil_stroke_placement_view3d*", "grease_pencil/modes/draw/stroke_placement.html#bpy-types-toolsettings-gpencil-stroke-placement-view3d"),
    ("bpy.types.toolsettings.use_annotation_stroke_endpoints*", "interface/annotate_tool.html#bpy-types-toolsettings-use-annotation-stroke-endpoints"),
    ("bpy.types.toolsettings.use_gpencil_select_mask_segment*", "grease_pencil/modes/sculpting/introduction.html#bpy-types-toolsettings-use-gpencil-select-mask-segment"),
    ("bpy.types.toolsettings.use_transform_pivot_point_align*", "scene_layout/object/tools/tool_settings.html#bpy-types-toolsettings-use-transform-pivot-point-align"),
    ("bpy.types.view3doverlay.use_gpencil_show_material_name*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-show-material-name"),
    ("bpy.types.animvizmotionpaths.show_keyframe_action_all*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-show-keyframe-action-all"),
    ("bpy.types.brush.show_multiplane_scrape_planes_preview*", "sculpt_paint/sculpting/brushes/multiplane_scrape.html#bpy-types-brush-show-multiplane-scrape-planes-preview"),
    ("bpy.types.brushcurvessculptsettings.interpolate_shape*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-interpolate-shape"),
    ("bpy.types.clothsettings.internal_spring_max_diversion*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-spring-max-diversion"),
    ("bpy.types.cyclesmaterialsettings.volume_interpolation*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-volume-interpolation"),
    ("bpy.types.cyclesrendersettings.denoising_input_passes*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-denoising-input-passes"),
    ("bpy.types.cyclesrendersettings.film_transparent_glass*", "render/cycles/render_settings/film.html#bpy-types-cyclesrendersettings-film-transparent-glass"),
    ("bpy.types.cyclesrendersettings.offscreen_dicing_scale*", "render/cycles/render_settings/subdivision.html#bpy-types-cyclesrendersettings-offscreen-dicing-scale"),
    ("bpy.types.fluiddomainsettings.sndparticle_bubble_drag*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-bubble-drag"),
    ("bpy.types.keyingsetpath.use_insertkey_override_needed*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-use-insertkey-override-needed"),
    ("bpy.types.keyingsetpath.use_insertkey_override_visual*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-use-insertkey-override-visual"),
    ("bpy.types.linestylegeometrymodifier_backbonestretcher*", "render/freestyle/view_layer/line_style/modifiers/geometry/backbone_stretcher.html#bpy-types-linestylegeometrymodifier-backbonestretcher"),
    ("bpy.types.linestylegeometrymodifier_sinusdisplacement*", "render/freestyle/view_layer/line_style/modifiers/geometry/sinus_displacement.html#bpy-types-linestylegeometrymodifier-sinusdisplacement"),
    ("bpy.types.preferencesedit.grease_pencil_default_color*", "editors/preferences/editing.html#bpy-types-preferencesedit-grease-pencil-default-color"),
    ("bpy.types.preferencesedit.grease_pencil_eraser_radius*", "editors/preferences/editing.html#bpy-types-preferencesedit-grease-pencil-eraser-radius"),
    ("bpy.types.preferencesfilepaths.render_cache_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-render-cache-directory"),
    ("bpy.types.preferencesinput.ndof_orbit_center_selected*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-orbit-center-selected"),
    ("bpy.types.preferencesinput.ndof_show_guide_orbit_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-show-guide-orbit-axis"),
    ("bpy.types.preferencesinput.use_mouse_emulate_3_button*", "editors/preferences/input.html#bpy-types-preferencesinput-use-mouse-emulate-3-button"),
    ("bpy.types.preferencesinput.use_numeric_input_advanced*", "editors/preferences/input.html#bpy-types-preferencesinput-use-numeric-input-advanced"),
    ("bpy.types.preferencessystem.shader_compilation_method*", "editors/preferences/system.html#bpy-types-preferencessystem-shader-compilation-method"),
    ("bpy.types.preferencessystem.use_edit_mode_smooth_wire*", "editors/preferences/viewport.html#bpy-types-preferencessystem-use-edit-mode-smooth-wire"),
    ("bpy.types.sequencertoolsettings.snap_to_current_frame*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-to-current-frame"),
    ("bpy.types.sequencertoolsettings.snap_to_retiming_keys*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-to-retiming-keys"),
    ("bpy.types.view3doverlay.sculpt_mode_face_sets_opacity*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-types-view3doverlay-sculpt-mode-face-sets-opacity"),
    ("bpy.ops.object.geometry_nodes_input_attribute_toggle*", "modeling/modifiers/generate/geometry_nodes.html#bpy-ops-object-geometry-nodes-input-attribute-toggle"),
    ("bpy.types.animvizmotionpaths.show_keyframe_highlight*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-show-keyframe-highlight"),
    ("bpy.types.brushcurvessculptsettings.minimum_distance*", "sculpt_paint/curves_sculpting/brushes/density_curves.html#bpy-types-brushcurvessculptsettings-minimum-distance"),
    ("bpy.types.brushcurvessculptsettings.points_per_curve*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-points-per-curve"),
    ("bpy.types.brushgpencilsettings.pen_subdivision_steps*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-pen-subdivision-steps"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_hue*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-hue"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_sat*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-sat"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_val*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-val"),
    ("bpy.types.clothsettings.internal_spring_normal_check*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-spring-normal-check"),
    ("bpy.types.clothsettings.vertex_group_shear_stiffness*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-vertex-group-shear-stiffness"),
    ("bpy.types.colormanageddisplaysettings.display_device*", "render/color_management.html#bpy-types-colormanageddisplaysettings-display-device"),
    ("bpy.types.colormanagedviewsettings.use_curve_mapping*", "render/color_management.html#bpy-types-colormanagedviewsettings-use-curve-mapping"),
    ("bpy.types.colormanagedviewsettings.use_white_balance*", "render/color_management.html#bpy-types-colormanagedviewsettings-use-white-balance"),
    ("bpy.types.cyclesrendersettings.sample_clamp_indirect*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-sample-clamp-indirect"),
    ("bpy.types.cyclesrendersettings.use_adaptive_sampling*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-adaptive-sampling"),
    ("bpy.types.cyclesrendersettings.use_preview_denoising*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-preview-denoising"),
    ("bpy.types.fluiddomainsettings.color_ramp_field_scale*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-color-ramp-field-scale"),
    ("bpy.types.fluiddomainsettings.use_adaptive_timesteps*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-adaptive-timesteps"),
    ("bpy.types.fluiddomainsettings.use_dissolve_smoke_log*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-dissolve-smoke-log"),
    ("bpy.types.freestylelineset.select_suggestive_contour*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-suggestive-contour"),
    ("bpy.types.gpencillayer.annotation_onion_before_color*", "interface/annotate_tool.html#bpy-types-gpencillayer-annotation-onion-before-color"),
    ("bpy.types.gpencillayer.annotation_onion_before_range*", "interface/annotate_tool.html#bpy-types-gpencillayer-annotation-onion-before-range"),
    ("bpy.types.gpencillayer.use_annotation_onion_skinning*", "interface/annotate_tool.html#bpy-types-gpencillayer-use-annotation-onion-skinning"),
    ("bpy.types.linestylegeometrymodifier_polygonalization*", "render/freestyle/view_layer/line_style/modifiers/geometry/polygonization.html#bpy-types-linestylegeometrymodifier-polygonalization"),
    ("bpy.types.preferencesedit.sculpt_paint_overlay_color*", "editors/preferences/editing.html#bpy-types-preferencesedit-sculpt-paint-overlay-color"),
    ("bpy.types.preferencesedit.use_keyframe_insert_needed*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-keyframe-insert-needed"),
    ("bpy.types.preferencesexperimental.use_asset_indexing*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-use-asset-indexing"),
    ("bpy.types.preferencesexperimental.use_viewport_debug*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-use-viewport-debug"),
    ("bpy.types.preferencesfilepaths.show_recent_locations*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-show-recent-locations"),
    ("bpy.types.preferencesfilepaths.show_system_bookmarks*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-show-system-bookmarks"),
    ("bpy.types.preferencesinput.ndof_lock_camera_pan_zoom*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-lock-camera-pan-zoom"),
    ("bpy.types.preferencesinput.touchpad_scroll_direction*", "editors/preferences/input.html#bpy-types-preferencesinput-touchpad-scroll-direction"),
    ("bpy.types.preferencessystem.network_connection_limit*", "editors/preferences/system.html#bpy-types-preferencessystem-network-connection-limit"),
    ("bpy.types.preferencesview.use_filter_brushes_by_tool*", "interface/window_system/regions.html#bpy-types-preferencesview-use-filter-brushes-by-tool"),
    ("bpy.types.preferencesview.use_text_render_subpixelaa*", "editors/preferences/interface.html#bpy-types-preferencesview-use-text-render-subpixelaa"),
    ("bpy.types.sequencertimelineoverlay.show_strip_offset*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-strip-offset"),
    ("bpy.types.sequencertimelineoverlay.show_strip_source*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-strip-source"),
    ("bpy.types.spacesequenceeditor.show_transform_preview*", "editors/video_sequencer/preview/header.html#bpy-types-spacesequenceeditor-show-transform-preview"),
    ("bpy.types.toolsettings.use_gpencil_automerge_strokes*", "grease_pencil/modes/draw/introduction.html#bpy-types-toolsettings-use-gpencil-automerge-strokes"),
    ("bpy.types.toolsettings.use_keyframe_insert_keyingset*", "editors/timeline.html#bpy-types-toolsettings-use-keyframe-insert-keyingset"),
    ("bpy.types.toolsettings.use_proportional_edit_objects*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-use-proportional-edit-objects"),
    ("bpy.types.view3doverlay.gpencil_vertex_paint_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-vertex-paint-opacity"),
    ("bpy.ops.outliner.liboverride_troubleshoot_operation*", "files/linked_libraries/library_overrides.html#bpy-ops-outliner-liboverride-troubleshoot-operation"),
    ("bpy.ops.view3d.edit_mesh_extrude_move_shrink_fatten*", "modeling/meshes/editing/face/extrude_faces_normal.html#bpy-ops-view3d-edit-mesh-extrude-move-shrink-fatten"),
    ("bpy.types.brushgpencilsettings.active_smooth_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-active-smooth-factor"),
    ("bpy.types.brushgpencilsettings.extend_stroke_factor*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-extend-stroke-factor"),
    ("bpy.types.brushgpencilsettings.use_random_press_hue*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-hue"),
    ("bpy.types.brushgpencilsettings.use_random_press_sat*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-sat"),
    ("bpy.types.brushgpencilsettings.use_random_press_val*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-val"),
    ("bpy.types.brushgpencilsettings.use_settings_outline*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-settings-outline"),
    ("bpy.types.brushgpencilsettings.use_stroke_random_uv*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-stroke-random-uv"),
    ("bpy.types.clothcollisionsettings.self_impulse_clamp*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-self-impulse-clamp"),
    ("bpy.types.clothcollisionsettings.use_self_collision*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-use-self-collision"),
    ("bpy.types.cyclesmaterialsettings.homogeneous_volume*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-homogeneous-volume"),
    ("bpy.types.cyclesobjectsettings.is_caustics_receiver*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-is-caustics-receiver"),
    ("bpy.types.cyclesrendersettings.adaptive_min_samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-adaptive-min-samples"),
    ("bpy.types.cyclesrendersettings.debug_bvh_time_steps*", "render/cycles/render_settings/performance.html#bpy-types-cyclesrendersettings-debug-bvh-time-steps"),
    ("bpy.types.cyclesrendersettings.distance_cull_margin*", "render/cycles/render_settings/simplify.html#bpy-types-cyclesrendersettings-distance-cull-margin"),
    ("bpy.types.cyclesrendersettings.rolling_shutter_type*", "render/cycles/render_settings/motion_blur.html#bpy-types-cyclesrendersettings-rolling-shutter-type"),
    ("bpy.types.cyclesrendersettings.sample_subset_length*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-sample-subset-length"),
    ("bpy.types.cyclesrendersettings.transmission_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-transmission-bounces"),
    ("bpy.types.cyclesworldsettings.sample_map_resolution*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-sample-map-resolution"),
    ("bpy.types.fluiddomainsettings.display_interpolation*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-display-interpolation"),
    ("bpy.types.fluiddomainsettings.gridlines_cell_filter*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-gridlines-cell-filter"),
    ("bpy.types.fluiddomainsettings.gridlines_color_field*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-gridlines-color-field"),
    ("bpy.types.fluiddomainsettings.gridlines_lower_bound*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-gridlines-lower-bound"),
    ("bpy.types.fluiddomainsettings.gridlines_range_color*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-gridlines-range-color"),
    ("bpy.types.fluiddomainsettings.gridlines_upper_bound*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-gridlines-upper-bound"),
    ("bpy.types.freestylelineset.select_material_boundary*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-material-boundary"),
    ("bpy.types.gpencillayer.annotation_onion_after_color*", "interface/annotate_tool.html#bpy-types-gpencillayer-annotation-onion-after-color"),
    ("bpy.types.gpencillayer.annotation_onion_after_range*", "interface/annotate_tool.html#bpy-types-gpencillayer-annotation-onion-after-range"),
    ("bpy.types.gpencilsculptsettings.use_scale_thickness*", "grease_pencil/modes/edit/stroke_menu.html#bpy-types-gpencilsculptsettings-use-scale-thickness"),
    ("bpy.types.greasepencillineartmodifier.source_camera*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-source-camera"),
    ("bpy.types.greasepencillineartmodifier.use_edge_mark*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-edge-mark"),
    ("bpy.types.greasepencillineartmodifier.use_face_mark*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-face-mark"),
    ("bpy.types.preferencesedit.fcurve_new_auto_smoothing*", "editors/preferences/animation.html#bpy-types-preferencesedit-fcurve-new-auto-smoothing"),
    ("bpy.types.preferencesfilepaths.use_file_compression*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-file-compression"),
    ("bpy.types.preferencesinput.use_mouse_depth_navigate*", "editors/preferences/navigation.html#bpy-types-preferencesinput-use-mouse-depth-navigate"),
    ("bpy.types.preferencesinput.use_rotate_around_active*", "editors/preferences/navigation.html#bpy-types-preferencesinput-use-rotate-around-active"),
    ("bpy.types.preferencessystem.texture_collection_rate*", "editors/preferences/system.html#bpy-types-preferencessystem-texture-collection-rate"),
    ("bpy.types.preferencessystem.use_overlay_smooth_wire*", "editors/preferences/viewport.html#bpy-types-preferencessystem-use-overlay-smooth-wire"),
    ("bpy.types.rendersettings_simplify_gpencil_shader_fx*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-shader-fx"),
    ("bpy.types.rendersettings_simplify_gpencil_view_fill*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-view-fill"),
    ("bpy.types.sculpt.use_automasking_boundary_face_sets*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-boundary-face-sets"),
    ("bpy.types.sequencertoolsettings.snap_to_frame_range*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-to-frame-range"),
    ("bpy.types.sequencertoolsettings.snap_to_hold_offset*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-to-hold-offset"),
    ("bpy.types.spacenodeoverlay.show_reroute_auto_labels*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-reroute-auto-labels"),
    ("bpy.types.spacespreadsheet.show_internal_attributes*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-show-internal-attributes"),
    ("bpy.types.toolsettings.use_mesh_automerge_and_split*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-use-mesh-automerge-and-split"),
    ("bpy.types.view3doverlay.use_gpencil_fade_gp_objects*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-fade-gp-objects"),
    ("bpy.types.view3doverlay.use_gpencil_show_directions*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-show-directions"),
    ("bpy.ops.grease_pencil.bake_grease_pencil_animation*", "grease_pencil/animation/tools.html#bpy-ops-grease-pencil-bake-grease-pencil-animation"),
    ("bpy.ops.scene.view_layer_remove_unused_lightgroups*", "render/layers/passes.html#bpy-ops-scene-view-layer-remove-unused-lightgroups"),
    ("bpy.types.animvizmotionpaths.show_keyframe_numbers*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-show-keyframe-numbers"),
    ("bpy.types.animvizmotionpaths.use_camera_space_bake*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-use-camera-space-bake"),
    ("bpy.types.brush.cloth_constraint_softbody_strength*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-constraint-softbody-strength"),
    ("bpy.types.brush.elastic_deform_volume_preservation*", "sculpt_paint/sculpting/brushes/elastic_deform.html#bpy-types-brush-elastic-deform-volume-preservation"),
    ("bpy.types.brush.use_automasking_boundary_face_sets*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-boundary-face-sets"),
    ("bpy.types.brushcurvessculptsettings.minimum_length*", "sculpt_paint/curves_sculpting/brushes/grow_shrink_curves.html#bpy-types-brushcurvessculptsettings-minimum-length"),
    ("bpy.types.brushgpencilsettings.fill_simplify_level*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-simplify-level"),
    ("bpy.types.brushgpencilsettings.random_value_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-random-value-factor"),
    ("bpy.types.brushgpencilsettings.use_collide_strokes*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-use-collide-strokes"),
    ("bpy.types.brushgpencilsettings.use_jitter_pressure*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-jitter-pressure"),
    ("bpy.types.brushgpencilsettings.use_random_press_uv*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-random-press-uv"),
    ("bpy.types.brushgpencilsettings.use_settings_random*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-settings-random"),
    ("bpy.types.brushgpencilsettings.vertex_color_factor*", "grease_pencil/modes/draw/tool_settings/color.html#bpy-types-brushgpencilsettings-vertex-color-factor"),
    ("bpy.types.clothcollisionsettings.collision_quality*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-collision-quality"),
    ("bpy.types.clothcollisionsettings.self_distance_min*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-self-distance-min"),
    ("bpy.types.clothsettings.internal_spring_max_length*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-spring-max-length"),
    ("bpy.types.clothsettings.internal_tension_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-internal-tension-stiffness"),
    ("bpy.types.collection.lineart_intersection_priority*", "scene_layout/collections/collections.html#bpy-types-collection-lineart-intersection-priority"),
    ("bpy.types.collection.lineart_use_intersection_mask*", "scene_layout/collections/collections.html#bpy-types-collection-lineart-use-intersection-mask"),
    ("bpy.types.colormanagedinputcolorspacesettings.name*", "editors/image/image_settings.html#bpy-types-colormanagedinputcolorspacesettings-name"),
    ("bpy.types.cyclesmaterialsettings.emission_sampling*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-emission-sampling"),
    ("bpy.types.cyclesrendersettings.denoising_prefilter*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-denoising-prefilter"),
    ("bpy.types.cyclesrendersettings.preview_dicing_rate*", "render/cycles/render_settings/subdivision.html#bpy-types-cyclesrendersettings-preview-dicing-rate"),
    ("bpy.types.cyclesrendersettings.sample_clamp_direct*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-sample-clamp-direct"),
    ("bpy.types.cyclesrendersettings.scrambling_distance*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-scrambling-distance"),
    ("bpy.types.cyclesrendersettings.use_surface_guiding*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-surface-guiding"),
    ("bpy.types.cyclesworldsettings.volume_interpolation*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-volume-interpolation"),
    ("bpy.types.fluiddomainsettings.mesh_particle_radius*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-particle-radius"),
    ("bpy.types.fluiddomainsettings.sndparticle_boundary*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-boundary"),
    ("bpy.types.fluiddomainsettings.sndparticle_life_max*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-life-max"),
    ("bpy.types.fluiddomainsettings.sndparticle_life_min*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-sndparticle-life-min"),
    ("bpy.types.fluiddomainsettings.sys_particle_maximum*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-sys-particle-maximum"),
    ("bpy.types.fluiddomainsettings.use_bubble_particles*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-use-bubble-particles"),
    ("bpy.types.freestylelineset.select_external_contour*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-external-contour"),
    ("bpy.types.geometrynodeforeachgeometryelementoutput*", "modeling/geometry_nodes/utilities/for_each_geometry_zone.html#bpy-types-geometrynodeforeachgeometryelementoutput"),
    ("bpy.types.greasepencillineartmodifier.use_material*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-material"),
    ("bpy.types.linestylegeometrymodifier_simplification*", "render/freestyle/view_layer/line_style/modifiers/geometry/simplification.html#bpy-types-linestylegeometrymodifier-simplification"),
    ("bpy.types.materialgpencilstyle.use_overlap_strokes*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-use-overlap-strokes"),
    ("bpy.types.movietrackingtrack.use_grayscale_preview*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-grayscale-preview"),
    ("bpy.types.nodetreeinterfacesocket.hide_in_modifier*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-hide-in-modifier"),
    ("bpy.types.nodetreeinterfacesocketvector.dimensions*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocketvector-dimensions"),
    ("bpy.types.object.shadow_terminator_geometry_offset*", "render/eevee/object_settings/object_data.html#bpy-types-object-shadow-terminator-geometry-offset"),
    ("bpy.types.preferencesedit.keyframe_new_handle_type*", "editors/preferences/animation.html#bpy-types-preferencesedit-keyframe-new-handle-type"),
    ("bpy.types.preferencesedit.use_insertkey_xyz_to_rgb*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-insertkey-xyz-to-rgb"),
    ("bpy.types.preferencesedit.use_text_edit_auto_close*", "editors/preferences/editing.html#bpy-types-preferencesedit-use-text-edit-auto-close"),
    ("bpy.types.preferencesexperimental.use_cycles_debug*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-use-cycles-debug"),
    ("bpy.types.preferencesfilepaths.temporary_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-temporary-directory"),
    ("bpy.types.preferencesinput.mouse_double_click_time*", "editors/preferences/input.html#bpy-types-preferencesinput-mouse-double-click-time"),
    ("bpy.types.preferencesinput.use_multitouch_gestures*", "editors/preferences/input.html#bpy-types-preferencesinput-use-multitouch-gestures"),
    ("bpy.types.preferencesview.filebrowser_display_type*", "editors/preferences/interface.html#bpy-types-preferencesview-filebrowser-display-type"),
    ("bpy.types.preferencesview.show_addons_enabled_only*", "editors/preferences/addons.html#bpy-types-preferencesview-show-addons-enabled-only"),
    ("bpy.types.rendersettings.compositor_denoise_device*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-compositor-denoise-device"),
    ("bpy.types.sequencertimelineoverlay.show_strip_name*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-strip-name"),
    ("bpy.types.sequencertimelineoverlay.show_thumbnails*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-thumbnails"),
    ("bpy.types.spacespreadsheet.geometry_component_type*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-geometry-component-type"),
    ("bpy.types.toolsettings.use_gpencil_weight_data_add*", "grease_pencil/modes/draw/introduction.html#bpy-types-toolsettings-use-gpencil-weight-data-add"),
    ("bpy.types.view3doverlay.sculpt_curves_cage_opacity*", "sculpt_paint/curves_sculpting/introduction.html#bpy-types-view3doverlay-sculpt-curves-cage-opacity"),
    ("bpy.types.view3doverlay.texture_paint_mode_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-texture-paint-mode-opacity"),
    ("bpy.types.viewlayer.use_pass_transmission_indirect*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-transmission-indirect"),
    ("bpy.ops.mesh.customdata_custom_splitnormals_clear*", "modeling/meshes/properties/custom_data.html#bpy-ops-mesh-customdata-custom-splitnormals-clear"),
    ("bpy.ops.sequencer.retiming_add_freeze_frame_slide*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-add-freeze-frame-slide"),
    ("bpy.types.bakesettings.use_pass_ambient_occlusion*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-ambient-occlusion"),
    ("bpy.types.brush.surface_smooth_shape_preservation*", "sculpt_paint/sculpting/brushes/smooth.html#bpy-types-brush-surface-smooth-shape-preservation"),
    ("bpy.types.brush.use_cloth_pin_simulation_boundary*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-use-cloth-pin-simulation-boundary"),
    ("bpy.types.brushcurvessculptsettings.scale_uniform*", "sculpt_paint/curves_sculpting/brushes/grow_shrink_curves.html#bpy-types-brushcurvessculptsettings-scale-uniform"),
    ("bpy.types.brushgpencilsettings.show_fill_boundary*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-show-fill-boundary"),
    ("bpy.types.clothsettings.compression_stiffness_max*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-compression-stiffness-max"),
    ("bpy.types.colormanagedsequencercolorspacesettings*", "render/color_management.html#bpy-types-colormanagedsequencercolorspacesettings"),
    ("bpy.types.colormanagedviewsettings.view_transform*", "render/color_management.html#bpy-types-colormanagedviewsettings-view-transform"),
    ("bpy.types.cyclesmaterialsettings.volume_step_rate*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-volume-step-rate"),
    ("bpy.types.cyclesobjectsettings.is_caustics_caster*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-is-caustics-caster"),
    ("bpy.types.cyclesrendersettings.adaptive_threshold*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-adaptive-threshold"),
    ("bpy.types.cyclesrendersettings.camera_cull_margin*", "render/cycles/render_settings/simplify.html#bpy-types-cyclesrendersettings-camera-cull-margin"),
    ("bpy.types.cyclesrendersettings.debug_use_hair_bvh*", "render/cycles/render_settings/performance.html#bpy-types-cyclesrendersettings-debug-use-hair-bvh"),
    ("bpy.types.cyclesrendersettings.use_volume_guiding*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-volume-guiding"),
    ("bpy.types.fluiddomainsettings.export_manta_script*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-export-manta-script"),
    ("bpy.types.fluiddomainsettings.fractions_threshold*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-fractions-threshold"),
    ("bpy.types.fluiddomainsettings.particle_band_width*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-band-width"),
    ("bpy.types.fluiddomainsettings.particle_randomness*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-randomness"),
    ("bpy.types.fluiddomainsettings.use_adaptive_domain*", "physics/fluid/type/domain/gas/adaptive_domain.html#bpy-types-fluiddomainsettings-use-adaptive-domain"),
    ("bpy.types.fluiddomainsettings.use_resumable_cache*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-use-resumable-cache"),
    ("bpy.types.fluiddomainsettings.use_spray_particles*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-use-spray-particles"),
    ("bpy.types.fluiddomainsettings.vector_display_type*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-display-type"),
    ("bpy.types.freestylelineset.select_by_image_border*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-by-image-border"),
    ("bpy.types.freestylesettings.kr_derivative_epsilon*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-kr-derivative-epsilon"),
    ("bpy.types.geometrynodecurveprimitivebeziersegment*", "modeling/geometry_nodes/curve/primitives/bezier_segment.html#bpy-types-geometrynodecurveprimitivebeziersegment"),
    ("bpy.types.geometrynodecurveprimitivequadrilateral*", "modeling/geometry_nodes/curve/primitives/quadrilateral.html#bpy-types-geometrynodecurveprimitivequadrilateral"),
    ("bpy.types.geometrynodeforeachgeometryelementinput*", "modeling/geometry_nodes/utilities/for_each_geometry_zone.html#bpy-types-geometrynodeforeachgeometryelementinput"),
    ("bpy.types.greasepencillineartmodifier.use_contour*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-contour"),
    ("bpy.types.keyingset.use_insertkey_override_needed*", "animation/keyframes/keying_sets.html#bpy-types-keyingset-use-insertkey-override-needed"),
    ("bpy.types.keyingset.use_insertkey_override_visual*", "animation/keyframes/keying_sets.html#bpy-types-keyingset-use-insertkey-override-visual"),
    ("bpy.types.linestylegeometrymodifier_perlinnoise1d*", "render/freestyle/view_layer/line_style/modifiers/geometry/perlin_noise_1d.html#bpy-types-linestylegeometrymodifier-perlinnoise1d"),
    ("bpy.types.linestylegeometrymodifier_perlinnoise2d*", "render/freestyle/view_layer/line_style/modifiers/geometry/perlin_noise_2d.html#bpy-types-linestylegeometrymodifier-perlinnoise2d"),
    ("bpy.types.materialgpencilstyle.alignment_rotation*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-alignment-rotation"),
    ("bpy.types.materialgpencilstyle.use_stroke_holdout*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-use-stroke-holdout"),
    ("bpy.types.movietrackingplanetrack.use_auto_keying*", "movie_clip/tracking/clip/sidebar/track/plane_track.html#bpy-types-movietrackingplanetrack-use-auto-keying"),
    ("bpy.types.movietrackingsettings.use_tripod_solver*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingsettings-use-tripod-solver"),
    ("bpy.types.nodesmodifier.simulation_bake_directory*", "modeling/modifiers/generate/geometry_nodes.html#bpy-types-nodesmodifier-simulation-bake-directory"),
    ("bpy.types.object.shadow_terminator_shading_offset*", "render/cycles/object_settings/object_data.html#bpy-types-object-shadow-terminator-shading-offset"),
    ("bpy.types.preferencesedit.fcurve_unselected_alpha*", "editors/preferences/animation.html#bpy-types-preferencesedit-fcurve-unselected-alpha"),
    ("bpy.types.preferencesedit.use_auto_keying_warning*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-auto-keying-warning"),
    ("bpy.types.preferencesexperimental.use_eevee_debug*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-use-eevee-debug"),
    ("bpy.types.preferencesexperimental.use_undo_legacy*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental-use-undo-legacy"),
    ("bpy.types.preferencesfilepaths.use_relative_paths*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-relative-paths"),
    ("bpy.types.preferencesfilepaths.use_tabs_as_spaces*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-tabs-as-spaces"),
    ("bpy.types.preferencesinput.ndof_orbit_center_auto*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-orbit-center-auto"),
    ("bpy.types.preferencesinput.ndof_orbit_sensitivity*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-orbit-sensitivity"),
    ("bpy.types.preferencesinput.pressure_threshold_max*", "editors/preferences/input.html#bpy-types-preferencesinput-pressure-threshold-max"),
    ("bpy.types.preferencessystem.sequencer_proxy_setup*", "editors/preferences/system.html#bpy-types-preferencessystem-sequencer-proxy-setup"),
    ("bpy.types.preferencessystem.use_studio_light_edit*", "editors/preferences/lights.html#bpy-types-preferencessystem-use-studio-light-edit"),
    ("bpy.types.preferencesview.gizmo_size_navigate_v3d*", "editors/preferences/viewport.html#bpy-types-preferencesview-gizmo-size-navigate-v3d"),
    ("bpy.types.preferencesview.view2d_grid_spacing_min*", "editors/preferences/animation.html#bpy-types-preferencesview-view2d-grid-spacing-min"),
    ("bpy.types.rendersettings.simplify_child_particles*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-child-particles"),
    ("bpy.types.rendersettings.use_high_quality_normals*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-use-high-quality-normals"),
    ("bpy.types.scenegpencil.antialias_threshold_render*", "render/cycles/render_settings/grease_pencil.html#bpy-types-scenegpencil-antialias-threshold-render"),
    ("bpy.types.sculpt.automasking_start_normal_falloff*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-automasking-start-normal-falloff"),
    ("bpy.types.sequencerpreviewoverlay.show_annotation*", "editors/video_sequencer/preview/display/overlays.html#bpy-types-sequencerpreviewoverlay-show-annotation"),
    ("bpy.types.sequencerpreviewoverlay.show_safe_areas*", "editors/video_sequencer/preview/display/overlays.html#bpy-types-sequencerpreviewoverlay-show-safe-areas"),
    ("bpy.types.sequencertoolsettings.snap_ignore_muted*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-ignore-muted"),
    ("bpy.types.sequencertoolsettings.snap_ignore_sound*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-ignore-sound"),
    ("bpy.types.spaceoutliner.use_filter_case_sensitive*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-case-sensitive"),
    ("bpy.types.spaceoutliner.use_filter_object_content*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-content"),
    ("bpy.types.spacesequenceeditor.show_gizmo_navigate*", "editors/video_sequencer/preview/display/gizmos.html#bpy-types-spacesequenceeditor-show-gizmo-navigate"),
    ("bpy.types.toolsettings.use_proportional_connected*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-use-proportional-connected"),
    ("bpy.types.toolsettings.use_proportional_projected*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-use-proportional-projected"),
    ("bpy.types.unifiedpaintsettings.use_stroke_random**", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-use-stroke-random"),
    ("bpy.types.view3doverlay.gpencil_grid_subdivisions*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-grid-subdivisions"),
    ("bpy.types.view3doverlay.vertex_paint_mode_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-vertex-paint-mode-opacity"),
    ("bpy.types.viewlayer.use_pass_cryptomatte_material*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-cryptomatte-material"),
    ("bpy.ops.grease_pencil.vertex_group_normalize_all*", "grease_pencil/modes/weight_paint/weights_menu.html#bpy-ops-grease-pencil-vertex-group-normalize-all"),
    ("bpy.ops.view3d.edit_mesh_extrude_individual_move*", "modeling/meshes/editing/face/extrude_faces.html#bpy-ops-view3d-edit-mesh-extrude-individual-move"),
    ("bpy.ops.view3d.edit_mesh_extrude_manifold_normal*", "modeling/meshes/tools/extrude_manifold.html#bpy-ops-view3d-edit-mesh-extrude-manifold-normal"),
    ("bpy.types.brushcurvessculptsettings.curve_length*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-curve-length"),
    ("bpy.types.brushcurvessculptsettings.curve_radius*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-curve-radius"),
    ("bpy.types.brushcurvessculptsettings.density_mode*", "sculpt_paint/curves_sculpting/brushes/density_curves.html#bpy-types-brushcurvessculptsettings-density-mode"),
    ("bpy.types.brushgpencilsettings.pen_smooth_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-pen-smooth-factor"),
    ("bpy.types.brushgpencilsettings.random_hue_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-random-hue-factor"),
    ("bpy.types.colormanagedviewsettings.white_balance*", "render/color_management.html#bpy-types-colormanagedviewsettings-white-balance"),
    ("bpy.types.cyclescurverendersettings.subdivisions*", "render/cycles/render_settings/hair.html#bpy-types-cyclescurverendersettings-subdivisions"),
    ("bpy.types.cyclesmaterialsettings.volume_sampling*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings-volume-sampling"),
    ("bpy.types.cyclesobjectsettings.use_deform_motion*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-use-deform-motion"),
    ("bpy.types.cyclesobjectsettings.use_distance_cull*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-use-distance-cull"),
    ("bpy.types.cyclesrendersettings.ao_bounces_render*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-ao-bounces-render"),
    ("bpy.types.cyclesrendersettings.denoising_quality*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-denoising-quality"),
    ("bpy.types.cyclesrendersettings.denoising_use_gpu*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-denoising-use-gpu"),
    ("bpy.types.cyclesrendersettings.min_light_bounces*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-min-light-bounces"),
    ("bpy.types.cyclesrendersettings.pixel_filter_type*", "render/cycles/render_settings/film.html#bpy-types-cyclesrendersettings-pixel-filter-type"),
    ("bpy.types.cyclesrendersettings.use_animated_seed*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-animated-seed"),
    ("bpy.types.cyclesrendersettings.use_distance_cull*", "render/cycles/render_settings/simplify.html#bpy-types-cyclesrendersettings-use-distance-cull"),
    ("bpy.types.cyclesrendersettings.use_layer_samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-layer-samples"),
    ("bpy.types.cyclesrendersettings.use_sample_subset*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-sample-subset"),
    ("bpy.types.cyclesworldsettings.homogeneous_volume*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-homogeneous-volume"),
    ("bpy.types.fluiddomainsettings.cache_frame_offset*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-frame-offset"),
    ("bpy.types.fluiddomainsettings.delete_in_obstacle*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-delete-in-obstacle"),
    ("bpy.types.fluiddomainsettings.fractions_distance*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-fractions-distance"),
    ("bpy.types.fluiddomainsettings.mesh_concave_lower*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-concave-lower"),
    ("bpy.types.fluiddomainsettings.mesh_concave_upper*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-concave-upper"),
    ("bpy.types.fluiddomainsettings.openvdb_data_depth*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-openvdb-data-depth"),
    ("bpy.types.fluiddomainsettings.use_dissolve_smoke*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-dissolve-smoke"),
    ("bpy.types.fluiddomainsettings.use_flip_particles*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-flip-particles"),
    ("bpy.types.fluiddomainsettings.use_foam_particles*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-use-foam-particles"),
    ("bpy.types.fluiddomainsettings.viscosity_exponent*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-viscosity-exponent"),
    ("bpy.types.fluideffectorsettings.surface_distance*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-surface-distance"),
    ("bpy.types.fluidflowsettings.density_vertex_group*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-density-vertex-group"),
    ("bpy.types.fluidflowsettings.use_initial_velocity*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-initial-velocity"),
    ("bpy.types.geometrynodetree.is_type_grease_pencil*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-type-grease-pencil"),
    ("bpy.types.greasepencillineartmodifier.use_crease*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-crease"),
    ("bpy.types.greasepencillineartmodifier.use_shadow*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-shadow"),
    ("bpy.types.linestylegeometrymodifier_guidinglines*", "render/freestyle/view_layer/line_style/modifiers/geometry/guiding_lines.html#bpy-types-linestylegeometrymodifier-guidinglines"),
    ("bpy.types.linestylegeometrymodifier_spatialnoise*", "render/freestyle/view_layer/line_style/modifiers/geometry/spatial_noise.html#bpy-types-linestylegeometrymodifier-spatialnoise"),
    ("bpy.types.linestylethicknessmodifier_calligraphy*", "render/freestyle/view_layer/line_style/modifiers/thickness/calligraphy.html#bpy-types-linestylethicknessmodifier-calligraphy"),
    ("bpy.types.materiallineart.use_material_mask_bits*", "render/materials/line_art.html#bpy-types-materiallineart-use-material-mask-bits"),
    ("bpy.types.movietrackingdopesheet.use_invert_sort*", "movie_clip/tracking/dope_sheet.html#bpy-types-movietrackingdopesheet-use-invert-sort"),
    ("bpy.types.nodetreeinterfacesocket*.default_value*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-default-value"),
    ("bpy.types.nodetreeinterfacesocket.structure_type*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-structure-type"),
    ("bpy.types.object.shadow_terminator_normal_offset*", "render/eevee/object_settings/object_data.html#bpy-types-object-shadow-terminator-normal-offset"),
    ("bpy.types.preferencesedit.node_use_insert_offset*", "editors/preferences/editing.html#bpy-types-preferencesedit-node-use-insert-offset"),
    ("bpy.types.preferencesedit.use_cursor_lock_adjust*", "editors/preferences/editing.html#bpy-types-preferencesedit-use-cursor-lock-adjust"),
    ("bpy.types.preferencesedit.use_mouse_depth_cursor*", "editors/preferences/editing.html#bpy-types-preferencesedit-use-mouse-depth-cursor"),
    ("bpy.types.preferencesfilepaths.file_preview_type*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-file-preview-type"),
    ("bpy.types.preferencesfilepaths.texture_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-texture-directory"),
    ("bpy.types.preferencesinput.drag_threshold_tablet*", "editors/preferences/input.html#bpy-types-preferencesinput-drag-threshold-tablet"),
    ("bpy.types.preferencesinput.ndof_panx_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-panx-invert-axis"),
    ("bpy.types.preferencesinput.ndof_pany_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-pany-invert-axis"),
    ("bpy.types.preferencesinput.ndof_panz_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-panz-invert-axis"),
    ("bpy.types.preferencesinput.ndof_rotx_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-rotx-invert-axis"),
    ("bpy.types.preferencesinput.ndof_roty_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-roty-invert-axis"),
    ("bpy.types.preferencesinput.ndof_rotz_invert_axis*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-rotz-invert-axis"),
    ("bpy.types.preferencessystem.gpu_preferred_device*", "editors/preferences/system.html#bpy-types-preferencessystem-gpu-preferred-device"),
    ("bpy.types.preferencesview.use_weight_color_range*", "editors/preferences/editing.html#bpy-types-preferencesview-use-weight-color-range"),
    ("bpy.types.rendersettings_simplify_gpencil_onplay*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-onplay"),
    ("bpy.types.rigidbodyconstraint.breaking_threshold*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-breaking-threshold"),
    ("bpy.types.sculpt.use_automasking_cavity_inverted*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-cavity-inverted"),
    ("bpy.types.sequencercacheoverlay.show_cache_final*", "editors/video_sequencer/sequencer/sidebar/cache.html#bpy-types-sequencercacheoverlay-show-cache-final"),
    ("bpy.types.spaceclipeditor.use_manual_calibration*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-use-manual-calibration"),
    ("bpy.types.spacedopesheeteditor.show_pose_markers*", "animation/markers.html#bpy-types-spacedopesheeteditor-show-pose-markers"),
    ("bpy.types.spaceimageoverlay.show_grid_background*", "editors/uv/overlays.html#bpy-types-spaceimageoverlay-show-grid-background"),
    ("bpy.types.spacenodeeditor.show_gizmo_active_node*", "editors/compositor.html#bpy-types-spacenodeeditor-show-gizmo-active-node"),
    ("bpy.types.spacenodeoverlay.show_named_attributes*", "modeling/geometry_nodes/inspection.html#bpy-types-spacenodeoverlay-show-named-attributes"),
    ("bpy.types.spaceoutliner.use_filter_object_camera*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-camera"),
    ("bpy.types.spaceoutliner.use_filter_object_others*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-others"),
    ("bpy.types.spacesequenceeditor.overlay_frame_type*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-overlay-frame-type"),
    ("bpy.types.spacesequenceeditor.show_strip_overlay*", "editors/video_sequencer/sequencer/display.html#bpy-types-spacesequenceeditor-show-strip-overlay"),
    ("bpy.types.spaceuveditor.custom_grid_subdivisions*", "editors/uv/overlays.html#bpy-types-spaceuveditor-custom-grid-subdivisions"),
    ("bpy.types.toolsettings.proportional_edit_falloff*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-proportional-edit-falloff"),
    ("bpy.types.toolsettings.snap_playhead_second_step*", "editors/timeline.html#bpy-types-toolsettings-snap-playhead-second-step"),
    ("bpy.types.toolsettings.use_edge_path_live_unwrap*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-use-edge-path-live-unwrap"),
    ("bpy.types.toolsettings.use_gpencil_draw_additive*", "grease_pencil/modes/draw/introduction.html#bpy-types-toolsettings-use-gpencil-draw-additive"),
    ("bpy.types.toolsettings.use_snap_backface_culling*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-backface-culling"),
    ("bpy.types.toolsettings.use_transform_data_origin*", "scene_layout/object/tools/tool_settings.html#bpy-types-toolsettings-use-transform-data-origin"),
    ("bpy.types.unifiedpaintsettings.saturation_jitter*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-saturation-jitter"),
    ("bpy.types.unifiedpaintsettings.use_random_press**", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-use-random-press"),
    ("bpy.types.unifiedpaintsettings.use_unified_color*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-use-unified-color"),
    ("bpy.types.view3doverlay.sculpt_mode_mask_opacity*", "sculpt_paint/sculpting/editing/mask.html#bpy-types-view3doverlay-sculpt-mode-mask-opacity"),
    ("bpy.types.view3doverlay.show_camera_passepartout*", "editors/3dview/sidebar.html#bpy-types-view3doverlay-show-camera-passepartout"),
    ("bpy.types.view3doverlay.use_gpencil_fade_objects*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-fade-objects"),
    ("bpy.types.view3doverlay.viewer_attribute_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-viewer-attribute-opacity"),
    ("bpy.types.viewlayer.use_pass_transmission_direct*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-transmission-direct"),
    ("bpy.ops.grease_pencil.stroke_reset_vertex_color*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-stroke-reset-vertex-color"),
    ("bpy.ops.mesh.customdata_custom_splitnormals_add*", "modeling/meshes/properties/custom_data.html#bpy-ops-mesh-customdata-custom-splitnormals-add"),
    ("bpy.ops.node.interface_item_unlink_panel_toggle*", "interface/controls/nodes/groups.html#bpy-ops-node-interface-item-unlink-panel-toggle"),
    ("bpy.ops.outliner.collection_indirect_only_clear*", "render/layers/introduction.html#bpy-ops-outliner-collection-indirect-only-clear"),
    ("bpy.ops.sequencer.retiming_add_transition_slide*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-add-transition-slide"),
    ("bpy.types.animvizmotionpaths.show_frame_numbers*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-show-frame-numbers"),
    ("bpy.types.brush.use_automasking_cavity_inverted*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-cavity-inverted"),
    ("bpy.types.brushgpencilsettings.fill_extend_mode*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-extend-mode"),
    ("bpy.types.brushgpencilsettings.pen_smooth_steps*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-pen-smooth-steps"),
    ("bpy.types.brushgpencilsettings.show_fill_extend*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-show-fill-extend"),
    ("bpy.types.brushgpencilsettings.use_collide_only*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-use-collide-only"),
    ("bpy.types.cycleslightsettings.is_caustics_light*", "render/cycles/light_settings.html#bpy-types-cycleslightsettings-is-caustics-light"),
    ("bpy.types.cyclesrendersettings.max_subdivisions*", "render/cycles/render_settings/subdivision.html#bpy-types-cyclesrendersettings-max-subdivisions"),
    ("bpy.types.cyclesrendersettings.preview_denoiser*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-denoiser"),
    ("bpy.types.cyclesrendersettings.sampling_pattern*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-sampling-pattern"),
    ("bpy.types.cyclesrendersettings.volume_max_steps*", "render/cycles/render_settings/volumes.html#bpy-types-cyclesrendersettings-volume-max-steps"),
    ("bpy.types.cyclesrendersettings.volume_step_rate*", "render/cycles/render_settings/volumes.html#bpy-types-cyclesrendersettings-volume-step-rate"),
    ("bpy.types.cyclesworldsettings.is_caustics_light*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-is-caustics-light"),
    ("bpy.types.dynamicpaintbrushsettings.wave_factor*", "physics/dynamic_paint/brush.html#bpy-types-dynamicpaintbrushsettings-wave-factor"),
    ("bpy.types.editbone.bbone_handle_use_scale_start*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-use-scale-start"),
    ("bpy.types.fluiddomainsettings.cache_data_format*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-data-format"),
    ("bpy.types.fluiddomainsettings.cache_frame_start*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-frame-start"),
    ("bpy.types.fluiddomainsettings.cache_mesh_format*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-mesh-format"),
    ("bpy.types.fluiddomainsettings.display_thickness*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-display-thickness"),
    ("bpy.types.fluiddomainsettings.flame_smoke_color*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flame-smoke-color"),
    ("bpy.types.fluiddomainsettings.mesh_smoothen_neg*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-smoothen-neg"),
    ("bpy.types.fluiddomainsettings.mesh_smoothen_pos*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-smoothen-pos"),
    ("bpy.types.fluiddomainsettings.simulation_method*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-simulation-method"),
    ("bpy.types.fluiddomainsettings.use_speed_vectors*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-use-speed-vectors"),
    ("bpy.types.fluiddomainsettings.vector_show_mac_x*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-show-mac-x"),
    ("bpy.types.fluiddomainsettings.vector_show_mac_y*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-show-mac-y"),
    ("bpy.types.fluiddomainsettings.vector_show_mac_z*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-show-mac-z"),
    ("bpy.types.fluideffectorsettings.velocity_factor*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-velocity-factor"),
    ("bpy.types.freestylelineset.select_by_collection*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-by-collection"),
    ("bpy.types.freestylelineset.select_by_edge_types*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-by-edge-types"),
    ("bpy.types.freestylelineset.select_by_face_marks*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-by-face-marks"),
    ("bpy.types.geometrynodeinputcurvehandlepositions*", "modeling/geometry_nodes/curve/read/curve_handle_positions.html#bpy-types-geometrynodeinputcurvehandlepositions"),
    ("bpy.types.greasepencillayer.use_locked_material*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-use-locked-material"),
    ("bpy.types.greasepencillayer.use_viewlayer_masks*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-use-viewlayer-masks"),
    ("bpy.types.greasepencillineartmodifier.use_cache*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-cache"),
    ("bpy.types.greasepencillineartmodifier.use_loose*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-use-loose"),
    ("bpy.types.linestyle*modifier_distancefromcamera*", "render/freestyle/view_layer/line_style/modifiers/color/distance_from_camera.html#bpy-types-linestyle-modifier-distancefromcamera"),
    ("bpy.types.linestyle*modifier_distancefromobject*", "render/freestyle/view_layer/line_style/modifiers/color/distance_from_object.html#bpy-types-linestyle-modifier-distancefromobject"),
    ("bpy.types.linestylegeometrymodifier_2dtransform*", "render/freestyle/view_layer/line_style/modifiers/geometry/2d_transform.html#bpy-types-linestylegeometrymodifier-2dtransform"),
    ("bpy.types.linestylegeometrymodifier_beziercurve*", "render/freestyle/view_layer/line_style/modifiers/geometry/bezier_curve.html#bpy-types-linestylegeometrymodifier-beziercurve"),
    ("bpy.types.materialgpencilstyle.use_fill_holdout*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-use-fill-holdout"),
    ("bpy.types.materiallineart.intersection_priority*", "render/materials/line_art.html#bpy-types-materiallineart-intersection-priority"),
    ("bpy.types.movietrackingplanetrack.image_opacity*", "movie_clip/tracking/clip/sidebar/track/plane_track.html#bpy-types-movietrackingplanetrack-image-opacity"),
    ("bpy.types.nodetreeinterfacepanel.default_closed*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacepanel-default-closed"),
    ("bpy.types.nodetreeinterfacesocket.default_input*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-default-input"),
    ("bpy.types.nodetreeinterfacesocket.menu_expanded*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-menu-expanded"),
    ("bpy.types.particlesettings.use_parent_particles*", "physics/particles/emitter/render.html#bpy-types-particlesettings-use-parent-particles"),
    ("bpy.types.preferencesfilepaths.text_editor_args*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-text-editor-args"),
    ("bpy.types.preferencesfilepaths.use_filter_files*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-filter-files"),
    ("bpy.types.preferencesinput.drag_threshold_mouse*", "editors/preferences/input.html#bpy-types-preferencesinput-drag-threshold-mouse"),
    ("bpy.types.preferencesinput.ndof_navigation_mode*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-navigation-mode"),
    ("bpy.types.preferencesinput.use_auto_perspective*", "editors/preferences/navigation.html#bpy-types-preferencesinput-use-auto-perspective"),
    ("bpy.types.preferencesinput.use_drag_immediately*", "editors/preferences/input.html#bpy-types-preferencesinput-use-drag-immediately"),
    ("bpy.types.preferencesinput.use_mouse_continuous*", "editors/preferences/input.html#bpy-types-preferencesinput-use-mouse-continuous"),
    ("bpy.types.preferencessystem.audio_mixing_buffer*", "editors/preferences/system.html#bpy-types-preferencessystem-audio-mixing-buffer"),
    ("bpy.types.preferencessystem.audio_sample_format*", "editors/preferences/system.html#bpy-types-preferencessystem-audio-sample-format"),
    ("bpy.types.preferencessystem.use_gpu_subdivision*", "editors/preferences/viewport.html#bpy-types-preferencessystem-use-gpu-subdivision"),
    ("bpy.types.preferencessystem.vbo_collection_rate*", "editors/preferences/system.html#bpy-types-preferencessystem-vbo-collection-rate"),
    ("bpy.types.preferencesview.pie_animation_timeout*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-animation-timeout"),
    ("bpy.types.preferencesview.use_text_antialiasing*", "editors/preferences/interface.html#bpy-types-preferencesview-use-text-antialiasing"),
    ("bpy.types.rigidbodyconstraint.solver_iterations*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-solver-iterations"),
    ("bpy.types.rigidbodyobject.collision_collections*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-collision-collections"),
    ("bpy.types.sceneeevee.use_shadow_jitter_viewport*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-use-shadow-jitter-viewport"),
    ("bpy.types.sculpt.automasking_start_normal_limit*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-automasking-start-normal-limit"),
    ("bpy.types.sculpt.use_automasking_boundary_edges*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-boundary-edges"),
    ("bpy.types.sequenceeditor.use_overlay_frame_lock*", "editors/video_sequencer/preview/sidebar.html#bpy-types-sequenceeditor-use-overlay-frame-lock"),
    ("bpy.types.sequencerpreviewoverlay.show_metadata*", "editors/video_sequencer/preview/display/overlays.html#bpy-types-sequencerpreviewoverlay-show-metadata"),
    ("bpy.types.sequencertimelineoverlay.show_fcurves*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-fcurves"),
    ("bpy.types.sequencertoolsettings.snap_to_borders*", "editors/video_sequencer/preview/controls/snapping.html#bpy-types-sequencertoolsettings-snap-to-borders"),
    ("bpy.types.sequencertoolsettings.snap_to_markers*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-snap-to-markers"),
    ("bpy.types.spaceclipeditor.use_grayscale_preview*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-use-grayscale-preview"),
    ("bpy.types.spaceoutliner.use_filter_object_empty*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-empty"),
    ("bpy.types.spaceoutliner.use_filter_object_light*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-light"),
    ("bpy.types.spacesequenceeditor.proxy_render_size*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-proxy-render-size"),
    ("bpy.types.spacespreadsheetrowfilter.column_name*", "editors/spreadsheet.html#bpy-types-spacespreadsheetrowfilter-column-name"),
    ("bpy.types.toolsettings.gpencil_stroke_snap_mode*", "grease_pencil/modes/draw/stroke_placement.html#bpy-types-toolsettings-gpencil-stroke-snap-mode"),
    ("bpy.types.toolsettings.snap_elements_individual*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-elements-individual"),
    ("bpy.types.toolsettings.snap_playhead_frame_step*", "editors/timeline.html#bpy-types-toolsettings-snap-playhead-frame-step"),
    ("bpy.types.toolsettings.use_keyframe_cycle_aware*", "editors/timeline.html#bpy-types-toolsettings-use-keyframe-cycle-aware"),
    ("bpy.types.toolsettings.use_keyframe_insert_auto*", "editors/timeline.html#bpy-types-toolsettings-use-keyframe-insert-auto"),
    ("bpy.types.unifiedpaintsettings.use_color_jitter*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-use-color-jitter"),
    ("bpy.types.view3doverlay.show_relationship_lines*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-relationship-lines"),
    ("bpy.types.view3doverlay.show_sculpt_curves_cage*", "sculpt_paint/curves_sculpting/introduction.html#bpy-types-view3doverlay-show-sculpt-curves-cage"),
    ("bpy.types.view3doverlay.use_gpencil_canvas_xray*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-canvas-xray"),
    ("bpy.types.view3doverlay.use_gpencil_fade_layers*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-fade-layers"),
    ("bpy.types.viewlayer.use_pass_cryptomatte_object*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-cryptomatte-object"),
    ("bpy.types.viewlayer.use_pass_transmission_color*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-transmission-color"),
    ("bpy.types.viewlayereevee.use_pass_volume_direct*", "render/layers/passes.html#bpy-types-viewlayereevee-use-pass-volume-direct"),
    ("bpy.ops.armature.rigify_apply_selection_colors*", "addons/rigging/rigify/metarigs.html#bpy-ops-armature-rigify-apply-selection-colors"),
    ("bpy.ops.ed.lib_id_generate_preview_from_object*", "editors/asset_browser.html#bpy-ops-ed-lib-id-generate-preview-from-object"),
    ("bpy.ops.grease_pencil.material_lock_unselected*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-lock-unselected"),
    ("bpy.ops.grease_pencil.stroke_merge_by_distance*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-stroke-merge-by-distance"),
    ("bpy.ops.paint.vertex_color_brightness_contrast*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-brightness-contrast"),
    ("bpy.types.brush.use_automasking_boundary_edges*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-boundary-edges"),
    ("bpy.types.brushcurvessculptsettings.add_amount*", "sculpt_paint/curves_sculpting/brushes/add_curves.html#bpy-types-brushcurvessculptsettings-add-amount"),
    ("bpy.types.brushgpencilsettings.fill_layer_mode*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-layer-mode"),
    ("bpy.types.brushgpencilsettings.random_strength*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-random-strength"),
    ("bpy.types.brushgpencilsettings.simplify_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-simplify-factor"),
    ("bpy.types.clothcollisionsettings.impulse_clamp*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-impulse-clamp"),
    ("bpy.types.clothcollisionsettings.self_friction*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-self-friction"),
    ("bpy.types.clothcollisionsettings.use_collision*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-use-collision"),
    ("bpy.types.clothsettings.uniform_pressure_force*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-uniform-pressure-force"),
    ("bpy.types.collection.lineart_intersection_mask*", "scene_layout/collections/collections.html#bpy-types-collection-lineart-intersection-mask"),
    ("bpy.types.compositornodetree.use_viewer_border*", "compositing/sidebar.html#bpy-types-compositornodetree-use-viewer-border"),
    ("bpy.types.cyclesobjectsettings.use_camera_cull*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-use-camera-cull"),
    ("bpy.types.cyclesobjectsettings.use_motion_blur*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-use-motion-blur"),
    ("bpy.types.cyclesrendersettings.diffuse_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-diffuse-bounces"),
    ("bpy.types.cyclesrendersettings.preview_samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-preview-samples"),
    ("bpy.types.cyclesrendersettings.use_camera_cull*", "render/cycles/render_settings/simplify.html#bpy-types-cyclesrendersettings-use-camera-cull"),
    ("bpy.types.cyclesworldsettings.volume_step_size*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-volume-step-size"),
    ("bpy.types.dynamicpaintbrushsettings.wave_clamp*", "physics/dynamic_paint/brush.html#bpy-types-dynamicpaintbrushsettings-wave-clamp"),
    ("bpy.types.editbone.bbone_handle_use_ease_start*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-use-ease-start"),
    ("bpy.types.ffmpegsettings.ffmpeg_prores_profile*", "render/output/properties/output.html#bpy-types-ffmpegsettings-ffmpeg-prores-profile"),
    ("bpy.types.fluiddomainsettings.color_ramp_field*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-color-ramp-field"),
    ("bpy.types.fluiddomainsettings.guide_vel_factor*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-guide-vel-factor"),
    ("bpy.types.fluideffectorsettings.use_plane_init*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-use-plane-init"),
    ("bpy.types.freestylelineset.collection_negation*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-collection-negation"),
    ("bpy.types.freestylelineset.face_mark_condition*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-face-mark-condition"),
    ("bpy.types.freestylelineset.select_ridge_valley*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-ridge-valley"),
    ("bpy.types.freestylelinestyle.material_boundary*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-material-boundary"),
    ("bpy.types.freestylelinestyle.use_split_pattern*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-split-pattern"),
    ("bpy.types.freestylesettings.use_view_map_cache*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-use-view-map-cache"),
    ("bpy.types.geometrynodecurvehandletypeselection*", "modeling/geometry_nodes/curve/read/handle_type_selection.html#bpy-types-geometrynodecurvehandletypeselection"),
    ("bpy.types.geometrynodedistributepointsinvolume*", "modeling/geometry_nodes/point/distribute_points_in_volume.html#bpy-types-geometrynodedistributepointsinvolume"),
    ("bpy.types.geometrynodeinputmeshvertexneighbors*", "modeling/geometry_nodes/mesh/read/vertex_neighbors.html#bpy-types-geometrynodeinputmeshvertexneighbors"),
    ("bpy.types.geometrynodeinputnamedlayerselection*", "modeling/geometry_nodes/grease_pencil/read/named_layer_selection.html#bpy-types-geometrynodeinputnamedlayerselection"),
    ("bpy.types.greasepencillayer.use_onion_skinning*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-use-onion-skinning"),
    ("bpy.types.greasepencillineartmodifier.overscan*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier-overscan"),
    ("bpy.types.imageformatsettings.color_management*", "render/output/properties/output.html#bpy-types-imageformatsettings-color-management"),
    ("bpy.types.linestylegeometrymodifier_tipremover*", "render/freestyle/view_layer/line_style/modifiers/geometry/tip_remover.html#bpy-types-linestylegeometrymodifier-tipremover"),
    ("bpy.types.movieclipuser.use_render_undistorted*", "editors/clip/display/clip_display.html#bpy-types-movieclipuser-use-render-undistorted"),
    ("bpy.types.moviesequence.animation_offset_start*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-moviesequence-animation-offset-start"),
    ("bpy.types.movietrackingcamera.distortion_model*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-distortion-model"),
    ("bpy.types.movietrackingtrack.use_alpha_preview*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-alpha-preview"),
    ("bpy.types.movietrackingtrack.use_green_channel*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-green-channel"),
    ("bpy.types.preferencesfilepaths.sound_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-sound-directory"),
    ("bpy.types.preferencesinput.ndof_zoom_direction*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-zoom-direction"),
    ("bpy.types.preferencessystem.anisotropic_filter*", "editors/preferences/viewport.html#bpy-types-preferencessystem-anisotropic-filter"),
    ("bpy.types.preferencessystem.gpu_shader_workers*", "editors/preferences/system.html#bpy-types-preferencessystem-gpu-shader-workers"),
    ("bpy.types.preferencessystem.memory_cache_limit*", "editors/preferences/system.html#bpy-types-preferencessystem-memory-cache-limit"),
    ("bpy.types.preferencessystem.register_all_users*", "editors/preferences/system.html#bpy-types-preferencessystem-register-all-users"),
    ("bpy.types.preferencessystem.use_region_overlap*", "editors/preferences/interface.html#bpy-types-preferencessystem-use-region-overlap"),
    ("bpy.types.preferencesview.mini_axis_brightness*", "editors/preferences/viewport.html#bpy-types-preferencesview-mini-axis-brightness"),
    ("bpy.types.preferencesview.playback_fps_samples*", "editors/preferences/viewport.html#bpy-types-preferencesview-playback-fps-samples"),
    ("bpy.types.preferencesview.show_tooltips_python*", "editors/preferences/interface.html#bpy-types-preferencesview-show-tooltips-python"),
    ("bpy.types.rendersettings.resolution_percentage*", "render/output/properties/format.html#bpy-types-rendersettings-resolution-percentage"),
    ("bpy.types.rendersettings_simplify_gpencil_tint*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-gpencil-tint"),
    ("bpy.types.sceneeevee.volumetric_shadow_samples*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-volumetric-shadow-samples"),
    ("bpy.types.sequencercacheoverlay.show_cache_raw*", "editors/video_sequencer/sequencer/sidebar/cache.html#bpy-types-sequencercacheoverlay-show-cache-raw"),
    ("bpy.types.sequencertoolsettings.snap_to_center*", "editors/video_sequencer/preview/controls/snapping.html#bpy-types-sequencertoolsettings-snap-to-center"),
    ("bpy.types.softbodysettings.use_estimate_matrix*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-use-estimate-matrix"),
    ("bpy.types.softbodysettings.vertex_group_spring*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-vertex-group-spring"),
    ("bpy.types.spaceimageeditor.show_gizmo_navigate*", "editors/image/introduction.html#bpy-types-spaceimageeditor-show-gizmo-navigate"),
    ("bpy.types.spaceimageoverlay.passepartout_alpha*", "editors/image/overlays.html#bpy-types-spaceimageoverlay-passepartout-alpha"),
    ("bpy.types.spaceoutliner.lib_override_view_mode*", "editors/outliner/interface.html#bpy-types-spaceoutliner-lib-override-view-mode"),
    ("bpy.types.spaceoutliner.use_filter_object_mesh*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object-mesh"),
    ("bpy.types.spaceoutliner.use_filter_view_layers*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-view-layers"),
    ("bpy.types.spacesequenceeditor.show_overexposed*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-show-overexposed"),
    ("bpy.types.spacespreadsheet.show_region_toolbar*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-show-region-toolbar"),
    ("bpy.types.toolsettings.gpencil_selectmode_edit*", "grease_pencil/modes/edit/selecting.html#bpy-types-toolsettings-gpencil-selectmode-edit"),
    ("bpy.types.toolsettings.snap_angle_increment_3d*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-angle-increment-3d"),
    ("bpy.types.toolsettings.snap_face_nearest_steps*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-face-nearest-steps"),
    ("bpy.types.toolsettings.use_gpencil_draw_onback*", "grease_pencil/modes/draw/introduction.html#bpy-types-toolsettings-use-gpencil-draw-onback"),
    ("bpy.types.toolsettings.use_snap_align_rotation*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-align-rotation"),
    ("bpy.types.toolsettings.use_snap_to_same_target*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-to-same-target"),
    ("bpy.types.unifiedpaintsettings.secondary_color*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-secondary-color"),
    ("bpy.types.view3doverlay.use_gpencil_edit_lines*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-edit-lines"),
    ("bpy.types.view3doverlay.use_gpencil_onion_skin*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-onion-skin"),
    ("bpy.types.viewlayer.use_pass_ambient_occlusion*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-ambient-occlusion"),
    ("bpy.types.viewlayer.use_pass_cryptomatte_asset*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-cryptomatte-asset"),
    ("bpy.ops.anim.slot_channels_move_to_new_action*", "editors/dope_sheet/modes/action.html#bpy-ops-anim-slot-channels-move-to-new-action"),
    ("bpy.ops.armature.collection_create_and_assign*", "animation/armatures/bones/editing/parenting.html#bpy-ops-armature-collection-create-and-assign"),
    ("bpy.ops.grease_pencil.material_copy_to_object*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-copy-to-object"),
    ("bpy.ops.grease_pencil.relative_layer_mask_add*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-relative-layer-mask-add"),
    ("bpy.ops.grease_pencil.snap_cursor_to_selected*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-snap-cursor-to-selected"),
    ("bpy.ops.grease_pencil.stroke_subdivide_smooth*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-subdivide-smooth"),
    ("bpy.ops.grease_pencil.stroke_switch_direction*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-switch-direction"),
    ("bpy.ops.node.interface_item_make_panel_toggle*", "interface/controls/nodes/groups.html#bpy-ops-node-interface-item-make-panel-toggle"),
    ("bpy.ops.outliner.collection_indirect_only_set*", "render/layers/introduction.html#bpy-ops-outliner-collection-indirect-only-set"),
    ("bpy.ops.scene.freestyle_geometry_modifier_add*", "render/freestyle/view_layer/line_style/geometry.html#bpy-ops-scene-freestyle-geometry-modifier-add"),
    ("bpy.ops.scene.view_layer_add_used_lightgroups*", "render/layers/passes.html#bpy-ops-scene-view-layer-add-used-lightgroups"),
    ("bpy.ops.sequencer.deinterlace_selected_movies*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-deinterlace-selected-movies"),
    ("bpy.types.bakesettings.use_selected_to_active*", "render/cycles/baking.html#bpy-types-bakesettings-use-selected-to-active"),
    ("bpy.types.brush.surface_smooth_current_vertex*", "sculpt_paint/sculpting/brushes/smooth.html#bpy-types-brush-surface-smooth-current-vertex"),
    ("bpy.types.brush.use_multiplane_scrape_dynamic*", "sculpt_paint/sculpting/brushes/multiplane_scrape.html#bpy-types-brush-use-multiplane-scrape-dynamic"),
    ("bpy.types.brushgpencilsettings.fill_direction*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-direction"),
    ("bpy.types.brushgpencilsettings.fill_draw_mode*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-draw-mode"),
    ("bpy.types.brushgpencilsettings.fill_threshold*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-threshold"),
    ("bpy.types.brushgpencilsettings.use_fill_limit*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-use-fill-limit"),
    ("bpy.types.clothcollisionsettings.distance_min*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-distance-min"),
    ("bpy.types.clothsettings.bending_stiffness_max*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-bending-stiffness-max"),
    ("bpy.types.clothsettings.compression_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-compression-stiffness"),
    ("bpy.types.clothsettings.tension_stiffness_max*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-tension-stiffness-max"),
    ("bpy.types.clothsettings.vertex_group_pressure*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-vertex-group-pressure"),
    ("bpy.types.cyclesrendersettings.fast_gi_method*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-fast-gi-method"),
    ("bpy.types.cyclesrendersettings.glossy_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-glossy-bounces"),
    ("bpy.types.cyclesrendersettings.use_light_tree*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-light-tree"),
    ("bpy.types.cyclesrendersettings.volume_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-volume-bounces"),
    ("bpy.types.cyclesworldsettings.sampling_method*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-sampling-method"),
    ("bpy.types.cyclesworldsettings.volume_sampling*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-volume-sampling"),
    ("bpy.types.dynamicpaintbrushsettings.wave_type*", "physics/dynamic_paint/brush.html#bpy-types-dynamicpaintbrushsettings-wave-type"),
    ("bpy.types.editbone.bbone_handle_use_scale_end*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-use-scale-end"),
    ("bpy.types.ffmpegsettings.constant_rate_factor*", "render/output/properties/output.html#bpy-types-ffmpegsettings-constant-rate-factor"),
    ("bpy.types.fieldsettings.use_guide_path_weight*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-use-guide-path-weight"),
    ("bpy.types.fileassetselectparams.import_method*", "editors/asset_browser.html#bpy-types-fileassetselectparams-import-method"),
    ("bpy.types.fluiddomainsettings.adapt_threshold*", "physics/fluid/type/domain/gas/adaptive_domain.html#bpy-types-fluiddomainsettings-adapt-threshold"),
    ("bpy.types.fluiddomainsettings.cache_directory*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-directory"),
    ("bpy.types.fluiddomainsettings.cache_frame_end*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-frame-end"),
    ("bpy.types.fluiddomainsettings.flame_vorticity*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flame-vorticity"),
    ("bpy.types.fluiddomainsettings.noise_pos_scale*", "physics/fluid/type/domain/gas/noise.html#bpy-types-fluiddomainsettings-noise-pos-scale"),
    ("bpy.types.fluiddomainsettings.noise_time_anim*", "physics/fluid/type/domain/gas/noise.html#bpy-types-fluiddomainsettings-noise-time-anim"),
    ("bpy.types.fluiddomainsettings.particle_number*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-number"),
    ("bpy.types.fluiddomainsettings.particle_radius*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-radius"),
    ("bpy.types.fluiddomainsettings.slice_per_voxel*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-slice-per-voxel"),
    ("bpy.types.fluiddomainsettings.surface_tension*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-surface-tension"),
    ("bpy.types.fluiddomainsettings.viscosity_value*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-viscosity-value"),
    ("bpy.types.fluideffectorsettings.effector_type*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-effector-type"),
    ("bpy.types.fluidflowsettings.use_particle_size*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-particle-size"),
    ("bpy.types.freestylelineset.face_mark_negation*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-face-mark-negation"),
    ("bpy.types.freestylelinestyle.integration_type*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-integration-type"),
    ("bpy.types.freestylelinestyle.use_split_length*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-split-length"),
    ("bpy.types.geometrynodedistributepointsonfaces*", "modeling/geometry_nodes/point/distribute_points_on_faces.html#bpy-types-geometrynodedistributepointsonfaces"),
    ("bpy.types.geometrynodesetcurvehandlepositions*", "modeling/geometry_nodes/curve/write/set_handle_positions.html#bpy-types-geometrynodesetcurvehandlepositions"),
    ("bpy.types.geometrynodesetgreasepencilsoftness*", "modeling/geometry_nodes/grease_pencil/write/set_grease_pencil_softness.html#bpy-types-geometrynodesetgreasepencilsoftness"),
    ("bpy.types.geometrynodetree.is_type_pointcloud*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-type-pointcloud"),
    ("bpy.types.geometrynodetree.use_wait_for_click*", "editors/geometry_node.html#bpy-types-geometrynodetree-use-wait-for-click"),
    ("bpy.types.greasepencilweightproximitymodifier*", "grease_pencil/modifiers/edit/weight_proximity.html#bpy-types-greasepencilweightproximitymodifier"),
    ("bpy.types.linestylegeometrymodifier_blueprint*", "render/freestyle/view_layer/line_style/modifiers/geometry/blueprint.html#bpy-types-linestylegeometrymodifier-blueprint"),
    ("bpy.types.material.volume_intersection_method*", "render/eevee/material_settings.html#bpy-types-material-volume-intersection-method"),
    ("bpy.types.materialgpencilstyle.alignment_mode*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-alignment-mode"),
    ("bpy.types.materialgpencilstyle.texture_offset*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-texture-offset"),
    ("bpy.types.movietrackingtrack.use_blue_channel*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-blue-channel"),
    ("bpy.types.movietrackingtrack.use_custom_color*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-custom-color"),
    ("bpy.types.nodetreeinterfacesocket.description*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-description"),
    ("bpy.types.objectlineart.intersection_priority*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart-intersection-priority"),
    ("bpy.types.particlesettings.use_modifier_stack*", "physics/particles/emitter/emission.html#bpy-types-particlesettings-use-modifier-stack"),
    ("bpy.types.posebone.use_custom_shape_bone_size*", "animation/armatures/bones/properties/display.html#bpy-types-posebone-use-custom-shape-bone-size"),
    ("bpy.types.preferencesedit.key_insert_channels*", "editors/preferences/animation.html#bpy-types-preferencesedit-key-insert-channels"),
    ("bpy.types.preferencesedit.use_enter_edit_mode*", "editors/preferences/editing.html#bpy-types-preferencesedit-use-enter-edit-mode"),
    ("bpy.types.preferencesedit.use_negative_frames*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-negative-frames"),
    ("bpy.types.preferencesfilepaths.auto_save_time*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-auto-save-time"),
    ("bpy.types.preferencesfilepaths.font_directory*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-font-directory"),
    ("bpy.types.preferencesinput.use_emulate_numpad*", "editors/preferences/input.html#bpy-types-preferencesinput-use-emulate-numpad"),
    ("bpy.types.preferencesinput.view_rotate_method*", "editors/preferences/navigation.html#bpy-types-preferencesinput-view-rotate-method"),
    ("bpy.types.preferencessystem.audio_sample_rate*", "editors/preferences/system.html#bpy-types-preferencessystem-audio-sample-rate"),
    ("bpy.types.preferencessystem.image_draw_method*", "editors/preferences/viewport.html#bpy-types-preferencessystem-image-draw-method"),
    ("bpy.types.preferencessystem.use_online_access*", "editors/preferences/system.html#bpy-types-preferencessystem-use-online-access"),
    ("bpy.types.preferencesview.factor_display_type*", "editors/preferences/interface.html#bpy-types-preferencesview-factor-display-type"),
    ("bpy.types.preferencesview.lookdev_sphere_size*", "editors/preferences/viewport.html#bpy-types-preferencesview-lookdev-sphere-size"),
    ("bpy.types.preferencesview.open_sublevel_delay*", "editors/preferences/interface.html#bpy-types-preferencesview-open-sublevel-delay"),
    ("bpy.types.preferencesview.open_toplevel_delay*", "editors/preferences/interface.html#bpy-types-preferencesview-open-toplevel-delay"),
    ("bpy.types.preferencesview.pie_initial_timeout*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-initial-timeout"),
    ("bpy.types.preferencesview.render_display_type*", "editors/preferences/interface.html#bpy-types-preferencesview-render-display-type"),
    ("bpy.types.preferencesview.use_mouse_over_open*", "editors/preferences/interface.html#bpy-types-preferencesview-use-mouse-over-open"),
    ("bpy.types.rendersettings.compositor_precision*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-compositor-precision"),
    ("bpy.types.rendersettings.motion_blur_position*", "render/cycles/render_settings/motion_blur.html#bpy-types-rendersettings-motion-blur-position"),
    ("bpy.types.rendersettings.sequencer_gl_preview*", "editors/video_sequencer/preview/sidebar.html#bpy-types-rendersettings-sequencer-gl-preview"),
    ("bpy.types.rendersettings.simplify_subdivision*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-subdivision"),
    ("bpy.types.rendersettings.use_simplify_normals*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-use-simplify-normals"),
    ("bpy.types.sculpt.use_automasking_start_normal*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-start-normal"),
    ("bpy.types.sequencerpreviewoverlay.show_cursor*", "editors/video_sequencer/preview/display/overlays.html#bpy-types-sequencerpreviewoverlay-show-cursor"),
    ("bpy.types.softbodysettings.use_edge_collision*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-use-edge-collision"),
    ("bpy.types.softbodysettings.use_face_collision*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-use-face-collision"),
    ("bpy.types.softbodysettings.use_self_collision*", "physics/soft_body/settings/self_collision.html#bpy-types-softbodysettings-use-self-collision"),
    ("bpy.types.spaceclipeditor.show_gizmo_navigate*", "editors/clip/display/gizmo.html#bpy-types-spaceclipeditor-show-gizmo-navigate"),
    ("bpy.types.spacegrapheditor.show_extrapolation*", "editors/graph_editor/introduction.html#bpy-types-spacegrapheditor-show-extrapolation"),
    ("bpy.types.spacenodeeditor.geometry_nodes_type*", "editors/geometry_node.html#bpy-types-spacenodeeditor-geometry-nodes-type"),
    ("bpy.types.spaceoutliner.use_filter_collection*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-collection"),
    ("bpy.types.spacesequenceeditor.cursor_location*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-cursor-location"),
    ("bpy.types.spacesequenceeditor.display_channel*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-display-channel"),
    ("bpy.types.spacesequenceeditor.show_gizmo_tool*", "editors/video_sequencer/preview/display/gizmos.html#bpy-types-spacesequenceeditor-show-gizmo-tool"),
    ("bpy.types.spacesequenceeditor.show_region_hud*", "editors/video_sequencer/sequencer/navigating.html#bpy-types-spacesequenceeditor-show-region-hud"),
    ("bpy.types.spacesequenceeditor.use_zoom_to_fit*", "editors/video_sequencer/preview/header.html#bpy-types-spacesequenceeditor-use-zoom-to-fit"),
    ("bpy.types.spacespreadsheet.show_only_selected*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-show-only-selected"),
    ("bpy.types.spacespreadsheetrowfilter.operation*", "editors/spreadsheet.html#bpy-types-spacespreadsheetrowfilter-operation"),
    ("bpy.types.spacespreadsheetrowfilter.threshold*", "editors/spreadsheet.html#bpy-types-spacespreadsheetrowfilter-threshold"),
    ("bpy.types.toolsettings.gpencil_surface_offset*", "grease_pencil/modes/draw/stroke_placement.html#bpy-types-toolsettings-gpencil-surface-offset"),
    ("bpy.types.toolsettings.playhead_snap_distance*", "editors/timeline.html#bpy-types-toolsettings-playhead-snap-distance"),
    ("bpy.types.toolsettings.use_snap_time_absolute*", "editors/graph_editor/introduction.html#bpy-types-toolsettings-use-snap-time-absolute"),
    ("bpy.types.transformsequence.use_uniform_scale*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-use-uniform-scale"),
    ("bpy.types.view3doverlay.show_face_orientation*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-face-orientation"),
    ("bpy.types.view3doverlay.show_viewer_attribute*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-viewer-attribute"),
    ("bpy.types.viewlayer.use_pass_diffuse_indirect*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-diffuse-indirect"),
    ("bpy.types.viewlayereevee.use_pass_transparent*", "render/layers/passes.html#bpy-types-viewlayereevee-use-pass-transparent"),
    ("bpy.ops.armature.copy_bone_color_to_selected*", "animation/armatures/bones/properties/display.html#bpy-ops-armature-copy-bone-color-to-selected"),
    ("bpy.ops.grease_pencil.vertex_group_normalize*", "grease_pencil/modes/weight_paint/weights_menu.html#bpy-ops-grease-pencil-vertex-group-normalize"),
    ("bpy.ops.object.multires_higher_levels_delete*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-higher-levels-delete"),
    ("bpy.ops.object.simulation_nodes_cache_delete*", "physics/simulation_nodes.html#bpy-ops-object-simulation-nodes-cache-delete"),
    ("bpy.ops.object.vertex_group_copy_to_selected*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-copy-to-selected"),
    ("bpy.ops.outliner.collection_duplicate_linked*", "editors/outliner/editing.html#bpy-ops-outliner-collection-duplicate-linked"),
    ("bpy.ops.sequencer.retiming_segment_speed_set*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-segment-speed-set"),
    ("bpy.ops.view3d.edit_mesh_extrude_move_normal*", "modeling/meshes/editing/face/extrude_faces.html#bpy-ops-view3d-edit-mesh-extrude-move-normal"),
    ("bpy.types.bakesettings.use_pass_transmission*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-transmission"),
    ("bpy.types.brush.use_automasking_start_normal*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-start-normal"),
    ("bpy.types.clothsettings.use_internal_springs*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-use-internal-springs"),
    ("bpy.types.clothsettings.vertex_group_bending*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-vertex-group-bending"),
    ("bpy.types.cyclesrendersettings.dicing_camera*", "render/cycles/render_settings/subdivision.html#bpy-types-cyclesrendersettings-dicing-camera"),
    ("bpy.types.cyclesrendersettings.film_exposure*", "render/cycles/render_settings/film.html#bpy-types-cyclesrendersettings-film-exposure"),
    ("bpy.types.cyclesrendersettings.sample_offset*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-sample-offset"),
    ("bpy.types.cyclesrendersettings.texture_limit*", "render/cycles/render_settings/simplify.html#bpy-types-cyclesrendersettings-texture-limit"),
    ("bpy.types.cyclesrendersettings.use_denoising*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-denoising"),
    ("bpy.types.editbone.bbone_custom_handle_start*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-custom-handle-start"),
    ("bpy.types.editbone.bbone_handle_use_ease_end*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-use-ease-end"),
    ("bpy.types.fieldsettings.guide_kink_amplitude*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-kink-amplitude"),
    ("bpy.types.fieldsettings.guide_kink_frequency*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-kink-frequency"),
    ("bpy.types.fileselectparams.list_display_size*", "editors/asset_browser.html#bpy-types-fileselectparams-list-display-size"),
    ("bpy.types.fluiddomainsettings.additional_res*", "physics/fluid/type/domain/gas/adaptive_domain.html#bpy-types-fluiddomainsettings-additional-res"),
    ("bpy.types.fluiddomainsettings.dissolve_speed*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-dissolve-speed"),
    ("bpy.types.fluiddomainsettings.effector_group*", "physics/fluid/type/domain/collections.html#bpy-types-fluiddomainsettings-effector-group"),
    ("bpy.types.fluiddomainsettings.flame_ignition*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flame-ignition"),
    ("bpy.types.fluiddomainsettings.flame_max_temp*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flame-max-temp"),
    ("bpy.types.fluiddomainsettings.mesh_generator*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-generator"),
    ("bpy.types.fluiddomainsettings.noise_strength*", "physics/fluid/type/domain/gas/noise.html#bpy-types-fluiddomainsettings-noise-strength"),
    ("bpy.types.fluiddomainsettings.particle_scale*", "physics/fluid/type/domain/liquid/particles.html#bpy-types-fluiddomainsettings-particle-scale"),
    ("bpy.types.fluiddomainsettings.resolution_max*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-resolution-max"),
    ("bpy.types.fluiddomainsettings.show_gridlines*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-show-gridlines"),
    ("bpy.types.fluiddomainsettings.use_color_ramp*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-use-color-ramp"),
    ("bpy.types.fluiddomainsettings.viscosity_base*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-viscosity-base"),
    ("bpy.types.fluideffectorsettings.use_effector*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-use-effector"),
    ("bpy.types.fluidflowsettings.surface_distance*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-surface-distance"),
    ("bpy.types.fluidflowsettings.texture_map_type*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-texture-map-type"),
    ("bpy.types.freestylelineset.select_silhouette*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-silhouette"),
    ("bpy.types.freestylelinestyle.texture_spacing*", "render/freestyle/view_layer/line_style/texture.html#bpy-types-freestylelinestyle-texture-spacing"),
    ("bpy.types.freestylelinestyle.use_chain_count*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-chain-count"),
    ("bpy.types.freestylelinestyle.use_dashed_line*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-dashed-line"),
    ("bpy.types.freestylelinestyle.use_same_object*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-same-object"),
    ("bpy.types.functionnodeinputspecialcharacters*", "modeling/geometry_nodes/utilities/text/special_characters.html#bpy-types-functionnodeinputspecialcharacters"),
    ("bpy.types.geometrynodecurveendpointselection*", "modeling/geometry_nodes/curve/read/endpoint_selection.html#bpy-types-geometrynodecurveendpointselection"),
    ("bpy.types.geometrynodeinputmeshedgeneighbors*", "modeling/geometry_nodes/mesh/read/edge_neighbors.html#bpy-types-geometrynodeinputmeshedgeneighbors"),
    ("bpy.types.geometrynodeinputmeshfaceneighbors*", "modeling/geometry_nodes/mesh/read/face_neighbors.html#bpy-types-geometrynodeinputmeshfaceneighbors"),
    ("bpy.types.geometrynodeinputshortestedgepaths*", "modeling/geometry_nodes/mesh/read/shortest_edge_paths.html#bpy-types-geometrynodeinputshortestedgepaths"),
    ("bpy.types.greasepencillayer.viewlayer_render*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-viewlayer-render"),
    ("bpy.types.greasepencilv3.onion_keyframe_type*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-onion-keyframe-type"),
    ("bpy.types.greasepencilv3.use_autolock_layers*", "grease_pencil/properties/layers.html#bpy-types-greasepencilv3-use-autolock-layers"),
    ("bpy.types.linestylegeometrymodifier_2doffset*", "render/freestyle/view_layer/line_style/modifiers/geometry/2d_offset.html#bpy-types-linestylegeometrymodifier-2doffset"),
    ("bpy.types.linestylegeometrymodifier_sampling*", "render/freestyle/view_layer/line_style/modifiers/geometry/sampling.html#bpy-types-linestylegeometrymodifier-sampling"),
    ("bpy.types.material.use_thickness_from_shadow*", "render/eevee/material_settings.html#bpy-types-material-use-thickness-from-shadow"),
    ("bpy.types.materialgpencilstyle.gradient_type*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-gradient-type"),
    ("bpy.types.materialgpencilstyle.texture_angle*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-texture-angle"),
    ("bpy.types.materialgpencilstyle.texture_clamp*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-texture-clamp"),
    ("bpy.types.materialgpencilstyle.texture_scale*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-texture-scale"),
    ("bpy.types.moviesequence.animation_offset_end*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-moviesequence-animation-offset-end"),
    ("bpy.types.movietrackingdopesheet.sort_method*", "movie_clip/tracking/dope_sheet.html#bpy-types-movietrackingdopesheet-sort-method"),
    ("bpy.types.movietrackingtrack.use_red_channel*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-use-red-channel"),
    ("bpy.types.nodetreeinterfacesocket*.max_value*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-max-value"),
    ("bpy.types.nodetreeinterfacesocket*.min_value*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-min-value"),
    ("bpy.types.nodetreeinterfacesocket.hide_value*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-hide-value"),
    ("bpy.types.object.add_rest_position_attribute*", "animation/shape_keys/shape_keys_panel.html#bpy-types-object-add-rest-position-attribute"),
    ("bpy.types.preferencesinput.invert_mouse_zoom*", "editors/preferences/navigation.html#bpy-types-preferencesinput-invert-mouse-zoom"),
    ("bpy.types.preferencesinput.invert_zoom_wheel*", "editors/preferences/navigation.html#bpy-types-preferencesinput-invert-zoom-wheel"),
    ("bpy.types.preferencesinput.ndof_lock_horizon*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-lock-horizon"),
    ("bpy.types.preferencesinput.pressure_softness*", "editors/preferences/input.html#bpy-types-preferencesinput-pressure-softness"),
    ("bpy.types.preferencesinput.use_zoom_to_mouse*", "editors/preferences/navigation.html#bpy-types-preferencesinput-use-zoom-to-mouse"),
    ("bpy.types.preferencessystem.gl_texture_limit*", "editors/preferences/viewport.html#bpy-types-preferencessystem-gl-texture-limit"),
    ("bpy.types.preferencessystem.texture_time_out*", "editors/preferences/system.html#bpy-types-preferencessystem-texture-time-out"),
    ("bpy.types.preferencesview.pie_menu_threshold*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-menu-threshold"),
    ("bpy.types.rendersettings.line_thickness_mode*", "render/freestyle/render.html#bpy-types-rendersettings-line-thickness-mode"),
    ("bpy.types.rendersettings.motion_blur_shutter*", "render/cycles/render_settings/motion_blur.html#bpy-types-rendersettings-motion-blur-shutter"),
    ("bpy.types.rendersettings.use_persistent_data*", "render/cycles/render_settings/performance.html#bpy-types-rendersettings-use-persistent-data"),
    ("bpy.types.sceneeevee.gi_irradiance_pool_size*", "render/eevee/render_settings/performance.html#bpy-types-sceneeevee-gi-irradiance-pool-size"),
    ("bpy.types.sceneeevee.shadow_resolution_scale*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-shadow-resolution-scale"),
    ("bpy.types.sculpt.use_automasking_view_normal*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-view-normal"),
    ("bpy.types.sequencertimelineoverlay.show_grid*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay-show-grid"),
    ("bpy.types.sequencertoolsettings.overlap_mode*", "video_editing/edit/montage/editing.html#bpy-types-sequencertoolsettings-overlap-mode"),
    ("bpy.types.softbodysettings.aerodynamics_type*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-aerodynamics-type"),
    ("bpy.types.softbodysettings.vertex_group_goal*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-vertex-group-goal"),
    ("bpy.types.softbodysettings.vertex_group_mass*", "physics/soft_body/settings/object.html#bpy-types-softbodysettings-vertex-group-mass"),
    ("bpy.types.spaceclipeditor.show_green_channel*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-show-green-channel"),
    ("bpy.types.spaceimageoverlay.show_render_size*", "editors/image/overlays.html#bpy-types-spaceimageoverlay-show-render-size"),
    ("bpy.types.spacenodeoverlay.show_context_path*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-context-path"),
    ("bpy.types.spaceoutliner.show_restrict_column*", "editors/outliner/interface.html#bpy-types-spaceoutliner-show-restrict-column"),
    ("bpy.types.spacesequenceeditor.use_clamp_view*", "editors/video_sequencer/sequencer/navigating.html#bpy-types-spacesequenceeditor-use-clamp-view"),
    ("bpy.types.spacespreadsheet.object_eval_state*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-object-eval-state"),
    ("bpy.types.spaceuveditor.display_stretch_type*", "editors/uv/overlays.html#bpy-types-spaceuveditor-display-stretch-type"),
    ("bpy.types.spaceuveditor.show_grid_over_image*", "editors/uv/overlays.html#bpy-types-spaceuveditor-show-grid-over-image"),
    ("bpy.types.textcharacterformat.use_small_caps*", "modeling/texts/properties.html#bpy-types-textcharacterformat-use-small-caps"),
    ("bpy.types.toolsettings.proportional_distance*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-proportional-distance"),
    ("bpy.types.toolsettings.snap_playhead_element*", "editors/timeline.html#bpy-types-toolsettings-snap-playhead-element"),
    ("bpy.types.toolsettings.transform_pivot_point*", "editors/3dview/controls/pivot_point/index.html#bpy-types-toolsettings-transform-pivot-point"),
    ("bpy.types.toolsettings.use_proportional_edit*", "editors/3dview/controls/proportional_editing.html#bpy-types-toolsettings-use-proportional-edit"),
    ("bpy.types.toolsettings.uv_sticky_select_mode*", "editors/uv/selecting.html#bpy-types-toolsettings-uv-sticky-select-mode"),
    ("bpy.types.transformsequence.translation_unit*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-translation-unit"),
    ("bpy.types.userassetlibrary.use_relative_path*", "editors/preferences/file_paths.html#bpy-types-userassetlibrary-use-relative-path"),
    ("bpy.types.view3doverlay.gpencil_fade_objects*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-fade-objects"),
    ("bpy.types.view3doverlay.gpencil_grid_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-grid-opacity"),
    ("bpy.types.view3dshading.wireframe_color_type*", "editors/3dview/display/shading.html#bpy-types-view3dshading-wireframe-color-type"),
    ("bpy.types.viewlayer.use_pass_glossy_indirect*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-glossy-indirect"),
    ("bpy.types.volumedisplay.interpolation_method*", "modeling/volumes/properties.html#bpy-types-volumedisplay-interpolation-method"),
    ("bpy.ops.geometry.color_attribute_render_set*", "modeling/meshes/properties/object_data.html#bpy-ops-geometry-color-attribute-render-set"),
    ("bpy.ops.grease_pencil.frame_clean_duplicate*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-frame-clean-duplicate"),
    ("bpy.ops.grease_pencil.set_uniform_thickness*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-uniform-thickness"),
    ("bpy.ops.object.geometry_nodes_move_to_nodes*", "modeling/modifiers/generate/geometry_nodes.html#bpy-ops-object-geometry-nodes-move-to-nodes"),
    ("bpy.ops.object.light_linking_receivers_link*", "scene_layout/object/editing/link_transfer/link_data.html#bpy-ops-object-light-linking-receivers-link"),
    ("bpy.ops.preferences.script_directory_remove*", "editors/preferences/file_paths.html#bpy-ops-preferences-script-directory-remove"),
    ("bpy.ops.sculpt.mask_from_face_sets_boundary*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-from-face-sets-boundary"),
    ("bpy.types.brush.use_automasking_view_normal*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-view-normal"),
    ("bpy.types.brushgpencilsettings.angle_factor*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-angle-factor"),
    ("bpy.types.clothcollisionsettings.collection*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings-collection"),
    ("bpy.types.clothsettings.compression_damping*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-compression-damping"),
    ("bpy.types.clothsettings.shear_stiffness_max*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-shear-stiffness-max"),
    ("bpy.types.clothsettings.use_pressure_volume*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-use-pressure-volume"),
    ("bpy.types.clothsettings.vertex_group_intern*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-vertex-group-intern"),
    ("bpy.types.clothsettings.vertex_group_shrink*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-vertex-group-shrink"),
    ("bpy.types.colormanagedviewsettings.exposure*", "render/color_management.html#bpy-types-colormanagedviewsettings-exposure"),
    ("bpy.types.cyclesobjectsettings.motion_steps*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-motion-steps"),
    ("bpy.types.cyclesrendersettings.filter_width*", "render/cycles/render_settings/film.html#bpy-types-cyclesrendersettings-filter-width"),
    ("bpy.types.fileselectparams.list_column_size*", "editors/asset_browser.html#bpy-types-fileselectparams-list-column-size"),
    ("bpy.types.fluiddomainsettings.cfl_condition*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-cfl-condition"),
    ("bpy.types.fluiddomainsettings.show_velocity*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-show-velocity"),
    ("bpy.types.fluiddomainsettings.timesteps_max*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-timesteps-max"),
    ("bpy.types.fluiddomainsettings.timesteps_min*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-timesteps-min"),
    ("bpy.types.fluiddomainsettings.use_diffusion*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-use-diffusion"),
    ("bpy.types.fluiddomainsettings.use_fractions*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-use-fractions"),
    ("bpy.types.fluiddomainsettings.use_viscosity*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-types-fluiddomainsettings-use-viscosity"),
    ("bpy.types.fluidflowsettings.particle_system*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-particle-system"),
    ("bpy.types.fluidflowsettings.velocity_factor*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-velocity-factor"),
    ("bpy.types.fluidflowsettings.velocity_normal*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-velocity-normal"),
    ("bpy.types.freestylelineset.select_edge_mark*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-edge-mark"),
    ("bpy.types.freestylelinestyle.use_length_max*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-length-max"),
    ("bpy.types.freestylelinestyle.use_length_min*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-length-min"),
    ("bpy.types.functionnodealignrotationtovector*", "modeling/geometry_nodes/utilities/rotation/align_rotation_to_vector.html#bpy-types-functionnodealignrotationtovector"),
    ("bpy.types.geometrynodedeformcurvesonsurface*", "modeling/geometry_nodes/curve/operations/deform_curves_on_surface.html#bpy-types-geometrynodedeformcurvesonsurface"),
    ("bpy.types.geometrynodeinputinstancerotation*", "modeling/geometry_nodes/instances/instance_rotation.html#bpy-types-geometrynodeinputinstancerotation"),
    ("bpy.types.geometrynodeinputmeshedgevertices*", "modeling/geometry_nodes/mesh/read/edge_vertices.html#bpy-types-geometrynodeinputmeshedgevertices"),
    ("bpy.types.geometrynodeinputmeshfaceisplanar*", "modeling/geometry_nodes/mesh/read/face_is_planar.html#bpy-types-geometrynodeinputmeshfaceisplanar"),
    ("bpy.types.geometrynodeinputsplineresolution*", "modeling/geometry_nodes/curve/read/spline_resolution.html#bpy-types-geometrynodeinputsplineresolution"),
    ("bpy.types.geometrynodemeshfacesetboundaries*", "modeling/geometry_nodes/mesh/read/face_group_boundaries.html#bpy-types-geometrynodemeshfacesetboundaries"),
    ("bpy.types.greasepencilv3.ghost_before_range*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-ghost-before-range"),
    ("bpy.types.greasepencilv3.stroke_depth_order*", "grease_pencil/properties/strokes.html#bpy-types-greasepencilv3-stroke-depth-order"),
    ("bpy.types.material.use_transparency_overlap*", "render/eevee/material_settings.html#bpy-types-material-use-transparency-overlap"),
    ("bpy.types.materialgpencilstyle.stroke_style*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-stroke-style"),
    ("bpy.types.materiallineart.use_material_mask*", "render/materials/line_art.html#bpy-types-materiallineart-use-material-mask"),
    ("bpy.types.movietracking.active_object_index*", "movie_clip/tracking/clip/sidebar/track/objects.html#bpy-types-movietracking-active-object-index"),
    ("bpy.types.nodetree.default_group_node_width*", "interface/controls/nodes/groups.html#bpy-types-nodetree-default-group-node-width"),
    ("bpy.types.objectlineart.use_crease_override*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart-use-crease-override"),
    ("bpy.types.preferencesedit.undo_memory_limit*", "editors/preferences/system.html#bpy-types-preferencesedit-undo-memory-limit"),
    ("bpy.types.preferencesedit.use_visual_keying*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-visual-keying"),
    ("bpy.types.preferencesfilepaths.image_editor*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-image-editor"),
    ("bpy.types.preferencesfilepaths.recent_files*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-recent-files"),
    ("bpy.types.preferencesfilepaths.save_version*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-save-version"),
    ("bpy.types.preferencesinput.ndof_sensitivity*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-sensitivity"),
    ("bpy.types.preferencesinput.view_zoom_method*", "editors/preferences/navigation.html#bpy-types-preferencesinput-view-zoom-method"),
    ("bpy.types.preferencessystem.network_timeout*", "editors/preferences/system.html#bpy-types-preferencessystem-network-timeout"),
    ("bpy.types.preferencesview.color_picker_type*", "editors/preferences/interface.html#bpy-types-preferencesview-color-picker-type"),
    ("bpy.types.preferencesview.font_path_ui_mono*", "editors/preferences/interface.html#bpy-types-preferencesview-font-path-ui-mono"),
    ("bpy.types.preferencesview.show_developer_ui*", "editors/preferences/interface.html#bpy-types-preferencesview-show-developer-ui"),
    ("bpy.types.preferencesview.show_playback_fps*", "editors/preferences/viewport.html#bpy-types-preferencesview-show-playback-fps"),
    ("bpy.types.raytraceeevee.trace_max_roughness*", "render/eevee/render_settings/raytracing.html#bpy-types-raytraceeevee-trace-max-roughness"),
    ("bpy.types.rendersettings.preview_pixel_size*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-preview-pixel-size"),
    ("bpy.types.rendersettings.use_crop_to_border*", "render/output/properties/format.html#bpy-types-rendersettings-use-crop-to-border"),
    ("bpy.types.rendersettings.use_file_extension*", "render/output/properties/output.html#bpy-types-rendersettings-use-file-extension"),
    ("bpy.types.scene.use_custom_simulation_range*", "scene_layout/scene/properties.html#bpy-types-scene-use-custom-simulation-range"),
    ("bpy.types.sceneeevee.clamp_surface_indirect*", "render/eevee/render_settings/clamping.html#bpy-types-sceneeevee-clamp-surface-indirect"),
    ("bpy.types.sceneeevee.use_volumetric_shadows*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-use-volumetric-shadows"),
    ("bpy.types.sculpt.constant_detail_resolution*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-types-sculpt-constant-detail-resolution"),
    ("bpy.types.sequenceeditor.show_missing_media*", "editors/video_sequencer/preview/sidebar.html#bpy-types-sequenceeditor-show-missing-media"),
    ("bpy.types.sequencertoolsettings.pivot_point*", "editors/video_sequencer/preview/controls/pivot_point.html#bpy-types-sequencertoolsettings-pivot-point"),
    ("bpy.types.spaceclipeditor.annotation_source*", "movie_clip/tracking/clip/sidebar/view.html#bpy-types-spaceclipeditor-annotation-source"),
    ("bpy.types.spaceclipeditor.mask_display_type*", "editors/clip/display/mask_display.html#bpy-types-spaceclipeditor-mask-display-type"),
    ("bpy.types.spaceclipeditor.mask_overlay_mode*", "editors/clip/display/mask_display.html#bpy-types-spaceclipeditor-mask-overlay-mode"),
    ("bpy.types.spaceclipeditor.show_blue_channel*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-show-blue-channel"),
    ("bpy.types.spaceclipeditor.show_mask_overlay*", "editors/clip/display/mask_display.html#bpy-types-spaceclipeditor-show-mask-overlay"),
    ("bpy.types.spacefilebrowser.system_bookmarks*", "editors/file_browser.html#bpy-types-spacefilebrowser-system-bookmarks"),
    ("bpy.types.spaceimageeditor.display_channels*", "editors/image/introduction.html#bpy-types-spaceimageeditor-display-channels"),
    ("bpy.types.spaceoutliner.use_filter_children*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-children"),
    ("bpy.types.spaceoutliner.use_filter_complete*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-complete"),
    ("bpy.types.spacespreadsheet.attribute_domain*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-attribute-domain"),
    ("bpy.types.spacespreadsheetrowfilter.enabled*", "editors/spreadsheet.html#bpy-types-spacespreadsheetrowfilter-enabled"),
    ("bpy.types.spaceuveditor.show_modified_edges*", "editors/uv/overlays.html#bpy-types-spaceuveditor-show-modified-edges"),
    ("bpy.types.spaceview3d.transform_orientation*", "editors/3dview/controls/orientation.html#bpy-types-spaceview3d-transform-orientation"),
    ("bpy.types.spaceview3d.use_local_collections*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-use-local-collections"),
    ("bpy.types.textcharacterformat.use_underline*", "modeling/texts/properties.html#bpy-types-textcharacterformat-use-underline"),
    ("bpy.types.toolsettings.use_snap_peel_object*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-peel-object"),
    ("bpy.types.toolsettings.use_uv_select_island*", "editors/uv/selecting.html#bpy-types-toolsettings-use-uv-select-island"),
    ("bpy.types.transformsequence.translate_start*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-translate-start"),
    ("bpy.types.unifiedpaintsettings.value_jitter*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-value-jitter"),
    ("bpy.types.view3doverlay.fade_inactive_alpha*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-fade-inactive-alpha"),
    ("bpy.types.view3doverlay.gpencil_grid_offset*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-grid-offset"),
    ("bpy.types.view3doverlay.wireframe_threshold*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-wireframe-threshold"),
    ("bpy.types.viewlayer.active_lightgroup_index*", "render/layers/passes.html#bpy-types-viewlayer-active-lightgroup-index"),
    ("bpy.types.viewlayer.use_pass_diffuse_direct*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-diffuse-direct"),
    ("bpy.types.viewlayer.use_pass_material_index*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-material-index"),
    ("bpy.ops.brush.sculpt_curves_falloff_preset*", "sculpt_paint/curves_sculpting/brushes/comb_curves.html#bpy-ops-brush-sculpt-curves-falloff-preset"),
    ("bpy.ops.constraint.followpath_path_animate*", "animation/constraints/relationship/follow_path.html#bpy-ops-constraint-followpath-path-animate"),
    ("bpy.ops.ed.lib_id_override_editable_toggle*", "editors/outliner/interface.html#bpy-ops-ed-lib-id-override-editable-toggle"),
    ("bpy.ops.geometry.color_attribute_duplicate*", "modeling/meshes/properties/object_data.html#bpy-ops-geometry-color-attribute-duplicate"),
    ("bpy.ops.grease_pencil.interpolate_sequence*", "grease_pencil/animation/tools.html#bpy-ops-grease-pencil-interpolate-sequence"),
    ("bpy.ops.grease_pencil.material_lock_unused*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-lock-unused"),
    ("bpy.ops.grease_pencil.set_curve_resolution*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-curve-resolution"),
    ("bpy.ops.object.constraint_add_with_targets*", "scene_layout/object/editing/constraints.html#bpy-ops-object-constraint-add-with-targets"),
    ("bpy.ops.object.light_linking_blockers_link*", "scene_layout/object/editing/link_transfer/link_data.html#bpy-ops-object-light-linking-blockers-link"),
    ("bpy.ops.object.material_slot_remove_unused*", "scene_layout/object/editing/cleanup.html#bpy-ops-object-material-slot-remove-unused"),
    ("bpy.ops.object.simulation_nodes_cache_bake*", "physics/simulation_nodes.html#bpy-ops-object-simulation-nodes-cache-bake"),
    ("bpy.ops.outliner.collection_disable_render*", "editors/outliner/editing.html#bpy-ops-outliner-collection-disable-render"),
    ("bpy.ops.scene.freestyle_alpha_modifier_add*", "render/freestyle/view_layer/line_style/alpha.html#bpy-ops-scene-freestyle-alpha-modifier-add"),
    ("bpy.ops.scene.freestyle_color_modifier_add*", "render/freestyle/view_layer/line_style/color.html#bpy-ops-scene-freestyle-color-modifier-add"),
    ("bpy.ops.scene.view_layer_remove_lightgroup*", "render/layers/passes.html#bpy-ops-scene-view-layer-remove-lightgroup"),
    ("bpy.ops.sequencer.scene_frame_range_update*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-scene-frame-range-update"),
    ("bpy.types.brush.cloth_simulation_area_type*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-simulation-area-type"),
    ("bpy.types.brushgpencilsettings.eraser_mode*", "grease_pencil/modes/draw/tools/erase.html#bpy-types-brushgpencilsettings-eraser-mode"),
    ("bpy.types.brushgpencilsettings.fill_factor*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-fill-factor"),
    ("bpy.types.brushgpencilsettings.vertex_mode*", "grease_pencil/modes/draw/tool_settings/color.html#bpy-types-brushgpencilsettings-vertex-mode"),
    ("bpy.types.clothsettings.use_sewing_springs*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-use-sewing-springs"),
    ("bpy.types.collection.active_exporter_index*", "scene_layout/collections/collections.html#bpy-types-collection-active-exporter-index"),
    ("bpy.types.curve.bevel_factor_mapping_start*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-factor-mapping-start"),
    ("bpy.types.cyclesobjectsettings.ao_distance*", "render/cycles/object_settings/object_data.html#bpy-types-cyclesobjectsettings-ao-distance"),
    ("bpy.types.cyclesobjectsettings.dicing_rate*", "render/cycles/object_settings/adaptive_subdiv.html#bpy-types-cyclesobjectsettings-dicing-rate"),
    ("bpy.types.cyclesrendersettings.blur_glossy*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-blur-glossy"),
    ("bpy.types.cyclesrendersettings.dicing_rate*", "render/cycles/render_settings/subdivision.html#bpy-types-cyclesrendersettings-dicing-rate"),
    ("bpy.types.cyclesrendersettings.max_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-max-bounces"),
    ("bpy.types.cyclesrendersettings.use_fast_gi*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-use-fast-gi"),
    ("bpy.types.cyclesrendersettings.use_guiding*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-use-guiding"),
    ("bpy.types.drivervariable.type.context_prop*", "animation/drivers/drivers_panel.html#bpy-types-drivervariable-type-context-prop"),
    ("bpy.types.editbone.bbone_custom_handle_end*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-custom-handle-end"),
    ("bpy.types.editbone.bbone_handle_type_start*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-type-start"),
    ("bpy.types.fieldsettings.guide_clump_amount*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-clump-amount"),
    ("bpy.types.fieldsettings.use_guide_path_add*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-use-guide-path-add"),
    ("bpy.types.fileselectparams.recursion_level*", "editors/file_browser.html#bpy-types-fileselectparams-recursion-level"),
    ("bpy.types.fluiddomainsettings.adapt_margin*", "physics/fluid/type/domain/gas/adaptive_domain.html#bpy-types-fluiddomainsettings-adapt-margin"),
    ("bpy.types.fluiddomainsettings.burning_rate*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-burning-rate"),
    ("bpy.types.fluiddomainsettings.guide_parent*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-guide-parent"),
    ("bpy.types.fluiddomainsettings.guide_source*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-guide-source"),
    ("bpy.types.fluiddomainsettings.particle_max*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-max"),
    ("bpy.types.fluiddomainsettings.particle_min*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-particle-min"),
    ("bpy.types.fluiddomainsettings.vector_field*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-field"),
    ("bpy.types.fluiddomainsettings.vector_scale*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-vector-scale"),
    ("bpy.types.fluideffectorsettings.guide_mode*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-guide-mode"),
    ("bpy.types.fluidflowsettings.texture_offset*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-texture-offset"),
    ("bpy.types.fluidflowsettings.use_plane_init*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-plane-init"),
    ("bpy.types.fluidflowsettings.velocity_coord*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-velocity-coord"),
    ("bpy.types.fluidflowsettings.volume_density*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-volume-density"),
    ("bpy.types.freestylelinestyle.use_angle_max*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-angle-max"),
    ("bpy.types.freestylelinestyle.use_angle_min*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-angle-min"),
    ("bpy.types.freestylesettings.as_render_pass*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-as-render-pass"),
    ("bpy.types.freestylesettings.use_smoothness*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-use-smoothness"),
    ("bpy.types.functionnodequaterniontorotation*", "modeling/geometry_nodes/utilities/rotation/quaternion_to_rotation.html#bpy-types-functionnodequaterniontorotation"),
    ("bpy.types.functionnoderotationtoquaternion*", "modeling/geometry_nodes/utilities/rotation/rotation_to_quaternion.html#bpy-types-functionnoderotationtoquaternion"),
    ("bpy.types.geometrynodecurveprimitivecircle*", "modeling/geometry_nodes/curve/primitives/curve_circle.html#bpy-types-geometrynodecurveprimitivecircle"),
    ("bpy.types.geometrynodecurvequadraticbezier*", "modeling/geometry_nodes/curve/primitives/quadratic_bezier.html#bpy-types-geometrynodecurvequadraticbezier"),
    ("bpy.types.geometrynodecurvestogreasepencil*", "modeling/geometry_nodes/curve/operations/curves_to_grease_pencil.html#bpy-types-geometrynodecurvestogreasepencil"),
    ("bpy.types.geometrynodeedgepathstoselection*", "modeling/geometry_nodes/mesh/operations/edge_paths_to_selection.html#bpy-types-geometrynodeedgepathstoselection"),
    ("bpy.types.geometrynodegreasepenciltocurves*", "modeling/geometry_nodes/grease_pencil/operations/grease_pencil_to_curves.html#bpy-types-geometrynodegreasepenciltocurves"),
    ("bpy.types.geometrynodesamplenearestsurface*", "modeling/geometry_nodes/mesh/sample/sample_nearest_surface.html#bpy-types-geometrynodesamplenearestsurface"),
    ("bpy.types.geometrynodesetgreasepencilcolor*", "modeling/geometry_nodes/grease_pencil/write/set_grease_pencil_color.html#bpy-types-geometrynodesetgreasepencilcolor"),
    ("bpy.types.geometrynodesetgreasepencildepth*", "modeling/geometry_nodes/grease_pencil/write/set_grease_pencil_depth.html#bpy-types-geometrynodesetgreasepencildepth"),
    ("bpy.types.geometrynodesetinstancetransform*", "modeling/geometry_nodes/instances/set_instance_transform.html#bpy-types-geometrynodesetinstancetransform"),
    ("bpy.types.greasepencilv3.ghost_after_range*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-ghost-after-range"),
    ("bpy.types.imagetexture.use_calculate_alpha*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-calculate-alpha"),
    ("bpy.types.imagetexture.use_filter_size_min*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-filter-size-min"),
    ("bpy.types.material.max_vertex_displacement*", "render/eevee/material_settings.html#bpy-types-material-max-vertex-displacement"),
    ("bpy.types.material.use_raytrace_refraction*", "render/eevee/material_settings.html#bpy-types-material-use-raytrace-refraction"),
    ("bpy.types.materialgpencilstyle.show_stroke*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-show-stroke"),
    ("bpy.types.movietrackingcamera.focal_length*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-focal-length"),
    ("bpy.types.movietrackingcamera.pixel_aspect*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-pixel-aspect"),
    ("bpy.types.movietrackingcamera.sensor_width*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-sensor-width"),
    ("bpy.types.posebone.custom_shape_wire_width*", "animation/armatures/bones/properties/display.html#bpy-types-posebone-custom-shape-wire-width"),
    ("bpy.types.posebone.use_ik_rotation_control*", "animation/armatures/posing/bone_constraints/inverse_kinematics/introduction.html#bpy-types-posebone-use-ik-rotation-control"),
    ("bpy.types.preferencesfilepaths.text_editor*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths-text-editor"),
    ("bpy.types.preferencesfilepaths.use_load_ui*", "editors/preferences/save_load.html#bpy-types-preferencesfilepaths-use-load-ui"),
    ("bpy.types.preferencesinput.navigation_mode*", "editors/preferences/navigation.html#bpy-types-preferencesinput-navigation-mode"),
    ("bpy.types.preferencessystem.audio_channels*", "editors/preferences/system.html#bpy-types-preferencessystem-audio-channels"),
    ("bpy.types.preferencesview.pie_menu_confirm*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-menu-confirm"),
    ("bpy.types.preferencesview.show_navigate_ui*", "editors/preferences/interface.html#bpy-types-preferencesview-show-navigate-ui"),
    ("bpy.types.preferencesview.show_object_info*", "editors/preferences/viewport.html#bpy-types-preferencesview-show-object-info"),
    ("bpy.types.preferencesview.use_fresnel_edit*", "editors/preferences/viewport.html#bpy-types-preferencesview-use-fresnel-edit"),
    ("bpy.types.rendersettings.compositor_device*", "render/eevee/render_settings/performance.html#bpy-types-rendersettings-compositor-device"),
    ("bpy.types.rendersettings.use_bake_multires*", "render/cycles/baking.html#bpy-types-rendersettings-use-bake-multires"),
    ("bpy.types.rigidbodyobject.collision_margin*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-collision-margin"),
    ("bpy.types.sceneeevee.clamp_volume_indirect*", "render/eevee/render_settings/clamping.html#bpy-types-sceneeevee-clamp-volume-indirect"),
    ("bpy.types.scenegpencil.antialias_threshold*", "render/cycles/render_settings/grease_pencil.html#bpy-types-scenegpencil-antialias-threshold"),
    ("bpy.types.sculpt.use_automasking_face_sets*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-face-sets"),
    ("bpy.types.sequencercacheoverlay.show_cache*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencercacheoverlay-show-cache"),
    ("bpy.types.softbodysettings.error_threshold*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-error-threshold"),
    ("bpy.types.softbodysettings.use_stiff_quads*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-use-stiff-quads"),
    ("bpy.types.spaceclipeditor.show_mask_spline*", "editors/clip/display/mask_display.html#bpy-types-spaceclipeditor-show-mask-spline"),
    ("bpy.types.spaceclipeditor.show_red_channel*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-show-red-channel"),
    ("bpy.types.spaceclipeditor.use_mute_footage*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-use-mute-footage"),
    ("bpy.types.spaceimageoverlay.show_text_info*", "editors/image/overlays.html#bpy-types-spaceimageoverlay-show-text-info"),
    ("bpy.types.spacenodeoverlay.show_wire_color*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-wire-color"),
    ("bpy.types.spacesequenceeditor.display_mode*", "editors/video_sequencer/preview/display/display_mode.html#bpy-types-spacesequenceeditor-display-mode"),
    ("bpy.types.spaceview3d.show_object_viewport*", "editors/3dview/display/visibility.html#bpy-types-spaceview3d-show-object-viewport"),
    ("bpy.types.toolsettings.use_record_with_nla*", "editors/timeline.html#bpy-types-toolsettings-use-record-with-nla"),
    ("bpy.types.toolsettings.use_snap_selectable*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-selectable"),
    ("bpy.types.toolsettings.vertex_group_subset*", "modeling/meshes/properties/vertex_groups/vertex_weights.html#bpy-types-toolsettings-vertex-group-subset"),
    ("bpy.types.toolsettings.vertex_group_weight*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-types-toolsettings-vertex-group-weight"),
    ("bpy.types.transformsequence.rotation_start*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-rotation-start"),
    ("bpy.types.view3doverlay.gpencil_fade_layer*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-fade-layer"),
    ("bpy.types.view3doverlay.gpencil_grid_color*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-grid-color"),
    ("bpy.types.view3doverlay.gpencil_grid_scale*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-gpencil-grid-scale"),
    ("bpy.types.view3doverlay.show_camera_guides*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-camera-guides"),
    ("bpy.types.view3doverlay.show_curve_normals*", "modeling/curves/curve_display.html#bpy-types-view3doverlay-show-curve-normals"),
    ("bpy.types.view3doverlay.show_extra_indices*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-extra-indices"),
    ("bpy.types.view3doverlay.show_fade_inactive*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-fade-inactive"),
    ("bpy.types.viewlayer.pass_cryptomatte_depth*", "render/layers/passes.html#bpy-types-viewlayer-pass-cryptomatte-depth"),
    ("bpy.types.viewlayer.use_pass_diffuse_color*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-diffuse-color"),
    ("bpy.types.viewlayer.use_pass_glossy_direct*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-glossy-direct"),
    ("bpy.types.viewlayer.use_pass_grease_pencil*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-grease-pencil"),
    ("bpy.types.walknavigation.use_mouse_reverse*", "editors/preferences/navigation.html#bpy-types-walknavigation-use-mouse-reverse"),
    ("bpy.types.walknavigation.walk_speed_factor*", "editors/preferences/navigation.html#bpy-types-walknavigation-walk-speed-factor"),
    ("bpy.ops.armature.collection_remove_unused*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-remove-unused"),
    ("bpy.ops.clip.stabilize_2d_rotation_select*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-stabilize-2d-rotation-select"),
    ("bpy.ops.constraint.disable_keep_transform*", "animation/constraints/interface/common.html#bpy-ops-constraint-disable-keep-transform"),
    ("bpy.ops.curves.convert_to_particle_system*", "sculpt_paint/curves_sculpting/introduction.html#bpy-ops-curves-convert-to-particle-system"),
    ("bpy.ops.grease_pencil.material_unlock_all*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-unlock-all"),
    ("bpy.ops.grease_pencil.set_active_material*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-active-material"),
    ("bpy.ops.grease_pencil.set_uniform_opacity*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-uniform-opacity"),
    ("bpy.ops.grease_pencil.stroke_material_set*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-material-set"),
    ("bpy.ops.grease_pencil.vertex_color_invert*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-vertex-color-invert"),
    ("bpy.ops.grease_pencil.vertex_color_levels*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-vertex-color-levels"),
    ("bpy.ops.grease_pencil.vertex_group_smooth*", "grease_pencil/modes/weight_paint/weights_menu.html#bpy-ops-grease-pencil-vertex-group-smooth"),
    ("bpy.ops.object.modifier_apply_as_shapekey*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-apply-as-shapekey"),
    ("bpy.ops.object.modifiers_copy_to_selected*", "scene_layout/object/editing/modifiers.html#bpy-ops-object-modifiers-copy-to-selected"),
    ("bpy.ops.object.vertex_group_normalize_all*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-normalize-all"),
    ("bpy.ops.object.visual_geometry_to_objects*", "scene_layout/object/editing/apply.html#bpy-ops-object-visual-geometry-to-objects"),
    ("bpy.ops.outliner.collection_color_tag_set*", "editors/outliner/editing.html#bpy-ops-outliner-collection-color-tag-set"),
    ("bpy.ops.outliner.collection_enable_render*", "editors/outliner/editing.html#bpy-ops-outliner-collection-enable-render"),
    ("bpy.ops.outliner.collection_exclude_clear*", "editors/outliner/editing.html#bpy-ops-outliner-collection-exclude-clear"),
    ("bpy.ops.outliner.collection_holdout_clear*", "render/layers/introduction.html#bpy-ops-outliner-collection-holdout-clear"),
    ("bpy.ops.sculpt.face_set_change_visibility*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-face-set-change-visibility"),
    ("bpy.ops.sculpt.face_sets_randomize_colors*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-ops-sculpt-face-sets-randomize-colors"),
    ("bpy.ops.ui.copy_driver_to_selected_button*", "animation/drivers/usage.html#bpy-ops-ui-copy-driver-to-selected-button"),
    ("bpy.ops.ui.eyedropper_grease_pencil_color*", "grease_pencil/modes/draw/tools/eyedropper.html#bpy-ops-ui-eyedropper-grease-pencil-color"),
    ("bpy.types.animvizmotionpaths.frame_before*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-frame-before"),
    ("bpy.types.armature.relation_line_position*", "animation/armatures/properties/display.html#bpy-types-armature-relation-line-position"),
    ("bpy.types.brush.disconnected_distance_max*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-disconnected-distance-max"),
    ("bpy.types.brush.surface_smooth_iterations*", "sculpt_paint/sculpting/brushes/smooth.html#bpy-types-brush-surface-smooth-iterations"),
    ("bpy.types.brush.use_color_as_displacement*", "sculpt_paint/brush/texture.html#bpy-types-brush-use-color-as-displacement"),
    ("bpy.types.brushgpencilsettings.pen_jitter*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-pen-jitter"),
    ("bpy.types.brushgpencilsettings.show_lasso*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-show-lasso"),
    ("bpy.types.clothsettings.bending_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-bending-stiffness"),
    ("bpy.types.clothsettings.tension_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-tension-stiffness"),
    ("bpy.types.clothsettings.vertex_group_mass*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-vertex-group-mass"),
    ("bpy.types.compositornodeconvertcolorspace*", "compositing/types/color/convert_colorspace.html#bpy-types-compositornodeconvertcolorspace"),
    ("bpy.types.cyclescurverendersettings.shape*", "render/cycles/render_settings/hair.html#bpy-types-cyclescurverendersettings-shape"),
    ("bpy.types.cycleslightsettings.cast_shadow*", "render/cycles/light_settings.html#bpy-types-cycleslightsettings-cast-shadow"),
    ("bpy.types.cycleslightsettings.max_bounces*", "render/cycles/light_settings.html#bpy-types-cycleslightsettings-max-bounces"),
    ("bpy.types.cyclesrendersettings.ao_bounces*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-ao-bounces"),
    ("bpy.types.cyclesrendersettings.time_limit*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-time-limit"),
    ("bpy.types.cyclesvisibilitysettings.camera*", "render/cycles/world_settings.html#bpy-types-cyclesvisibilitysettings-camera"),
    ("bpy.types.cyclesworldsettings.max_bounces*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings-max-bounces"),
    ("bpy.types.ffmpegsettings.use_max_b_frames*", "render/output/properties/output.html#bpy-types-ffmpegsettings-use-max-b-frames"),
    ("bpy.types.fieldsettings.apply_to_location*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-apply-to-location"),
    ("bpy.types.fieldsettings.apply_to_rotation*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-apply-to-rotation"),
    ("bpy.types.fieldsettings.guide_clump_shape*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-clump-shape"),
    ("bpy.types.fluiddomainsettings.domain_type*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-domain-type"),
    ("bpy.types.fluiddomainsettings.flame_smoke*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flame-smoke"),
    ("bpy.types.fluiddomainsettings.fluid_group*", "physics/fluid/type/domain/collections.html#bpy-types-fluiddomainsettings-fluid-group"),
    ("bpy.types.fluiddomainsettings.guide_alpha*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-guide-alpha"),
    ("bpy.types.fluiddomainsettings.noise_scale*", "physics/fluid/type/domain/gas/noise.html#bpy-types-fluiddomainsettings-noise-scale"),
    ("bpy.types.fluiddomainsettings.slice_depth*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-slice-depth"),
    ("bpy.types.fluideffectorsettings.subframes*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings-subframes"),
    ("bpy.types.fluidflowsettings.flow_behavior*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-flow-behavior"),
    ("bpy.types.fluidflowsettings.noise_texture*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-noise-texture"),
    ("bpy.types.freestylelineset.select_contour*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-contour"),
    ("bpy.types.freestylelinestyle.split_length*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-split-length"),
    ("bpy.types.freestylelinestyle.use_chaining*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-chaining"),
    ("bpy.types.freestylesettings.sphere_radius*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-sphere-radius"),
    ("bpy.types.functionnodeaxisangletorotation*", "modeling/geometry_nodes/utilities/rotation/axis_angle_to_rotation.html#bpy-types-functionnodeaxisangletorotation"),
    ("bpy.types.geometrynodeattributedomainsize*", "modeling/geometry_nodes/attribute/domain_size.html#bpy-types-geometrynodeattributedomainsize"),
    ("bpy.types.geometrynodeinputinstancebounds*", "modeling/geometry_nodes/instances/instance_bounds.html#bpy-types-geometrynodeinputinstancebounds"),
    ("bpy.types.geometrynodeinputnamedattribute*", "modeling/geometry_nodes/geometry/read/named_attribute.html#bpy-types-geometrynodeinputnamedattribute"),
    ("bpy.types.geometrynodesetsplineresolution*", "modeling/geometry_nodes/curve/write/set_spline_resolution.html#bpy-types-geometrynodesetsplineresolution"),
    ("bpy.types.geometrynodestorenamedattribute*", "modeling/geometry_nodes/attribute/store_named_attribute.html#bpy-types-geometrynodestorenamedattribute"),
    ("bpy.types.geometrynodetree.is_mode_object*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-mode-object"),
    ("bpy.types.geometrynodetree.is_mode_sculpt*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-mode-sculpt"),
    ("bpy.types.gpencillayer.annotation_opacity*", "interface/annotate_tool.html#bpy-types-gpencillayer-annotation-opacity"),
    ("bpy.types.gpencilsculptsettings.lock_axis*", "grease_pencil/modes/draw/drawing_planes.html#bpy-types-gpencilsculptsettings-lock-axis"),
    ("bpy.types.greasepencillayer.channel_color*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-channel-color"),
    ("bpy.types.greasepencillayer.radius_offset*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-radius-offset"),
    ("bpy.types.greasepencilweightanglemodifier*", "grease_pencil/modifiers/edit/weight_angle.html#bpy-types-greasepencilweightanglemodifier"),
    ("bpy.types.imageformatsettings.color_depth*", "files/media/image_formats.html#bpy-types-imageformatsettings-color-depth"),
    ("bpy.types.imageformatsettings.file_format*", "render/output/properties/output.html#bpy-types-imageformatsettings-file-format"),
    ("bpy.types.imageformatsettings.use_preview*", "files/media/image_formats.html#bpy-types-imageformatsettings-use-preview"),
    ("bpy.types.imagepaint.use_backface_culling*", "sculpt_paint/texture_paint/tool_settings/options.html#bpy-types-imagepaint-use-backface-culling"),
    ("bpy.types.light.shadow_maximum_resolution*", "render/eevee/light_settings.html#bpy-types-light-shadow-maximum-resolution"),
    ("bpy.types.linestyle*modifier_curvature_3d*", "render/freestyle/view_layer/line_style/modifiers/color/curvature_3d.html#bpy-types-linestyle-modifier-curvature-3d"),
    ("bpy.types.material.use_transparent_shadow*", "render/eevee/material_settings.html#bpy-types-material-use-transparent-shadow"),
    ("bpy.types.materialgpencilstyle.fill_color*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-fill-color"),
    ("bpy.types.materialgpencilstyle.fill_image*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-fill-image"),
    ("bpy.types.materialgpencilstyle.fill_style*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-fill-style"),
    ("bpy.types.materialgpencilstyle.mix_factor*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-mix-factor"),
    ("bpy.types.materialgpencilstyle.pass_index*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-pass-index"),
    ("bpy.types.object.use_grease_pencil_lights*", "grease_pencil/properties/object.html#bpy-types-object-use-grease-pencil-lights"),
    ("bpy.types.object.use_parent_final_indices*", "scene_layout/object/properties/relations.html#bpy-types-object-use-parent-final-indices"),
    ("bpy.types.preferences.use_recent_searches*", "editors/preferences/interface.html#bpy-types-preferences-use-recent-searches"),
    ("bpy.types.preferencesedit.use_auto_keying*", "editors/preferences/animation.html#bpy-types-preferencesedit-use-auto-keying"),
    ("bpy.types.preferencesedit.use_global_undo*", "editors/preferences/system.html#bpy-types-preferencesedit-use-global-undo"),
    ("bpy.types.preferencesinput.drag_threshold*", "editors/preferences/input.html#bpy-types-preferencesinput-drag-threshold"),
    ("bpy.types.preferencesinput.move_threshold*", "editors/preferences/input.html#bpy-types-preferencesinput-move-threshold"),
    ("bpy.types.preferencesinput.view_zoom_axis*", "editors/preferences/navigation.html#bpy-types-preferencesinput-view-zoom-axis"),
    ("bpy.types.preferencessystem.gl_clip_alpha*", "editors/preferences/viewport.html#bpy-types-preferencessystem-gl-clip-alpha"),
    ("bpy.types.preferencessystem.light_ambient*", "editors/preferences/lights.html#bpy-types-preferencessystem-light-ambient"),
    ("bpy.types.preferencesview.pie_menu_radius*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-menu-radius"),
    ("bpy.types.preferencesview.pie_tap_timeout*", "editors/preferences/interface.html#bpy-types-preferencesview-pie-tap-timeout"),
    ("bpy.types.preferencesview.use_save_prompt*", "editors/preferences/save_load.html#bpy-types-preferencesview-use-save-prompt"),
    ("bpy.types.preferencesview.view_frame_type*", "editors/preferences/animation.html#bpy-types-preferencesview-view-frame-type"),
    ("bpy.types.rendersettings.dither_intensity*", "render/output/properties/post_processing.html#bpy-types-rendersettings-dither-intensity"),
    ("bpy.types.rendersettings.film_transparent*", "render/cycles/render_settings/film.html#bpy-types-rendersettings-film-transparent"),
    ("bpy.types.rendersettings.simplify_volumes*", "render/cycles/render_settings/simplify.html#bpy-types-rendersettings-simplify-volumes"),
    ("bpy.types.rendersettings.use_render_cache*", "render/output/properties/output.html#bpy-types-rendersettings-use-render-cache"),
    ("bpy.types.rendersettings.use_single_layer*", "render/layers/view_layer.html#bpy-types-rendersettings-use-single-layer"),
    ("bpy.types.rigidbodyobject.collision_shape*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-collision-shape"),
    ("bpy.types.sceneeevee.clamp_surface_direct*", "render/eevee/render_settings/clamping.html#bpy-types-sceneeevee-clamp-surface-direct"),
    ("bpy.types.sceneeevee.use_taa_reprojection*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-use-taa-reprojection"),
    ("bpy.types.sculpt.use_automasking_topology*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-topology"),
    ("bpy.types.shadernodevolumescatter.density*", "render/shader_nodes/shader/volume_scatter.html#bpy-types-shadernodevolumescatter-density"),
    ("bpy.types.softbodysettings.collision_type*", "physics/soft_body/settings/self_collision.html#bpy-types-softbodysettings-collision-type"),
    ("bpy.types.spaceclipeditor.cursor_location*", "editors/clip/sidebar.html#bpy-types-spaceclipeditor-cursor-location"),
    ("bpy.types.spacefilebrowser.recent_folders*", "editors/file_browser.html#bpy-types-spacefilebrowser-recent-folders"),
    ("bpy.types.spacefilebrowser.system_folders*", "editors/file_browser.html#bpy-types-spacefilebrowser-system-folders"),
    ("bpy.types.spacenodeeditor.show_annotation*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeeditor-show-annotation"),
    ("bpy.types.spaceoutliner.use_filter_object*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-filter-object"),
    ("bpy.types.spaceproperties.show_properties*", "editors/properties_editor.html#bpy-types-spaceproperties-show-properties"),
    ("bpy.types.spacesequenceeditor.use_proxies*", "editors/video_sequencer/preview/sidebar.html#bpy-types-spacesequenceeditor-use-proxies"),
    ("bpy.types.spacespreadsheet.show_region_ui*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-show-region-ui"),
    ("bpy.types.spaceuveditor.edge_display_type*", "editors/uv/overlays.html#bpy-types-spaceuveditor-edge-display-type"),
    ("bpy.types.spaceuveditor.grid_shape_source*", "editors/uv/overlays.html#bpy-types-spaceuveditor-grid-shape-source"),
    ("bpy.types.spaceuveditor.show_pixel_coords*", "editors/uv/sidebar.html#bpy-types-spaceuveditor-show-pixel-coords"),
    ("bpy.types.spaceview3d.show_reconstruction*", "editors/3dview/display/overlays.html#bpy-types-spaceview3d-show-reconstruction"),
    ("bpy.types.toolsettings.snap_elements_base*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-elements-base"),
    ("bpy.types.toolsettings.use_auto_normalize*", "sculpt_paint/weight_paint/tool_settings/options.html#bpy-types-toolsettings-use-auto-normalize"),
    ("bpy.types.toolsettings.use_mesh_automerge*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-use-mesh-automerge"),
    ("bpy.types.toolsettings.use_snap_sequencer*", "video_editing/edit/montage/editing.html#bpy-types-toolsettings-use-snap-sequencer"),
    ("bpy.types.toolsettings.use_snap_translate*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-translate"),
    ("bpy.types.toolsettings.use_uv_select_sync*", "editors/uv/selecting.html#bpy-types-toolsettings-use-uv-select-sync"),
    ("bpy.types.transformsequence.interpolation*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-interpolation"),
    ("bpy.types.unifiedpaintsettings.hue_jitter*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-hue-jitter"),
    ("bpy.types.view3doverlay.retopology_offset*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-retopology-offset"),
    ("bpy.types.view3doverlay.show_light_colors*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-light-colors"),
    ("bpy.types.view3doverlay.wireframe_opacity*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-wireframe-opacity"),
    ("bpy.types.viewlayer.use_pass_glossy_color*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-glossy-color"),
    ("bpy.types.viewlayer.use_pass_object_index*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-object-index"),
    ("bpy.ops.asset.open_containing_blend_file*", "editors/asset_browser.html#bpy-ops-asset-open-containing-blend-file"),
    ("bpy.ops.extensions.package_install_files*", "editors/preferences/extensions.html#bpy-ops-extensions-package-install-files"),
    ("bpy.ops.geometry.color_attribute_convert*", "modeling/meshes/properties/object_data.html#bpy-ops-geometry-color-attribute-convert"),
    ("bpy.ops.grease_pencil.insert_blank_frame*", "grease_pencil/animation/tools.html#bpy-ops-grease-pencil-insert-blank-frame"),
    ("bpy.ops.grease_pencil.layer_group_remove*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-group-remove"),
    ("bpy.ops.grease_pencil.set_selection_mode*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-set-selection-mode"),
    ("bpy.ops.node.collapse_hide_unused_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-collapse-hide-unused-toggle"),
    ("bpy.ops.object.anim_transforms_to_deltas*", "scene_layout/object/editing/apply.html#bpy-ops-object-anim-transforms-to-deltas"),
    ("bpy.ops.object.geometry_node_bake_single*", "modeling/geometry_nodes/baking.html#bpy-ops-object-geometry-node-bake-single"),
    ("bpy.ops.object.modifier_copy_to_selected*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-copy-to-selected"),
    ("bpy.ops.preferences.app_template_install*", "advanced/app_templates.html#bpy-ops-preferences-app-template-install"),
    ("bpy.ops.preferences.script_directory_add*", "editors/preferences/file_paths.html#bpy-ops-preferences-script-directory-add"),
    ("bpy.ops.sequencer.preview_duplicate_move*", "editors/video_sequencer/preview/header.html#bpy-ops-sequencer-preview-duplicate-move"),
    ("bpy.ops.wm.interface_theme_preset_remove*", "editors/preferences/themes.html#bpy-ops-wm-interface-theme-preset-remove"),
    ("bpy.types.animvizmotionpaths.frame_after*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-frame-after"),
    ("bpy.types.animvizmotionpaths.frame_start*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-frame-start"),
    ("bpy.types.bakesettings.use_pass_indirect*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-indirect"),
    ("bpy.types.brush.cloth_force_falloff_type*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-force-falloff-type"),
    ("bpy.types.brush.use_automasking_topology*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-topology"),
    ("bpy.types.brushgpencilsettings.caps_type*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-caps-type"),
    ("bpy.types.brushgpencilsettings.show_fill*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-show-fill"),
    ("bpy.types.brushgpencilsettings.uv_random*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-uv-random"),
    ("bpy.types.brushtextureslot.mask_map_mode*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-mask-map-mode"),
    ("bpy.types.clothsettings.sewing_force_max*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-sewing-force-max"),
    ("bpy.types.clothsettings.use_dynamic_mesh*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-use-dynamic-mesh"),
    ("bpy.types.colormanagedviewsettings.gamma*", "render/color_management.html#bpy-types-colormanagedviewsettings-gamma"),
    ("bpy.types.compositornodeimagecoordinates*", "compositing/types/input/image_coordinates.html#bpy-types-compositornodeimagecoordinates"),
    ("bpy.types.compositornodeplanetrackdeform*", "compositing/types/tracking/plane_track_deform.html#bpy-types-compositornodeplanetrackdeform"),
    ("bpy.types.curve.bevel_factor_mapping_end*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-factor-mapping-end"),
    ("bpy.types.editbone.bbone_handle_type_end*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-handle-type-end"),
    ("bpy.types.editbone.use_endroll_as_inroll*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-use-endroll-as-inroll"),
    ("bpy.types.fieldsettings.guide_kink_shape*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-kink-shape"),
    ("bpy.types.fieldsettings.use_max_distance*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-use-max-distance"),
    ("bpy.types.fieldsettings.use_min_distance*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-use-min-distance"),
    ("bpy.types.fileselectparams.filter_search*", "editors/file_browser.html#bpy-types-fileselectparams-filter-search"),
    ("bpy.types.fluiddomainsettings.cache_type*", "physics/fluid/type/domain/cache.html#bpy-types-fluiddomainsettings-cache-type"),
    ("bpy.types.fluiddomainsettings.flip_ratio*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-flip-ratio"),
    ("bpy.types.fluiddomainsettings.guide_beta*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-guide-beta"),
    ("bpy.types.fluiddomainsettings.mesh_scale*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-mesh-scale"),
    ("bpy.types.fluiddomainsettings.slice_axis*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-slice-axis"),
    ("bpy.types.fluiddomainsettings.time_scale*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-time-scale"),
    ("bpy.types.fluidflowsettings.texture_size*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-texture-size"),
    ("bpy.types.fluidflowsettings.use_absolute*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-absolute"),
    ("bpy.types.freestylelineset.select_border*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-border"),
    ("bpy.types.freestylelineset.select_crease*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-select-crease"),
    ("bpy.types.freestylelinestyle.chain_count*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-chain-count"),
    ("bpy.types.freestylelinestyle.use_sorting*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-use-sorting"),
    ("bpy.types.freestylesettings.crease_angle*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-crease-angle"),
    ("bpy.types.functionnodealigneulertovector*", "modeling/geometry_nodes/utilities/deprecated/align_euler_to_vector.html#bpy-types-functionnodealigneulertovector"),
    ("bpy.types.functionnodetransformdirection*", "modeling/geometry_nodes/utilities/matrix/transform_direction.html#bpy-types-functionnodetransformdirection"),
    ("bpy.types.geometrynodeattributestatistic*", "modeling/geometry_nodes/attribute/attribute_statistic.html#bpy-types-geometrynodeattributestatistic"),
    ("bpy.types.geometrynodecurveprimitiveline*", "modeling/geometry_nodes/curve/primitives/curve_line.html#bpy-types-geometrynodecurveprimitiveline"),
    ("bpy.types.geometrynodegeometrytoinstance*", "modeling/geometry_nodes/geometry/geometry_to_instance.html#bpy-types-geometrynodegeometrytoinstance"),
    ("bpy.types.geometrynodeinputinstancescale*", "modeling/geometry_nodes/instances/instance_scale.html#bpy-types-geometrynodeinputinstancescale"),
    ("bpy.types.geometrynodeinputmaterialindex*", "modeling/geometry_nodes/material/material_index.html#bpy-types-geometrynodeinputmaterialindex"),
    ("bpy.types.geometrynodeinputmeshedgeangle*", "modeling/geometry_nodes/mesh/read/edge_angle.html#bpy-types-geometrynodeinputmeshedgeangle"),
    ("bpy.types.geometrynodeoffsetcornerinface*", "modeling/geometry_nodes/mesh/topology/offset_corner_in_face.html#bpy-types-geometrynodeoffsetcornerinface"),
    ("bpy.types.geometrynodeoffsetpointincurve*", "modeling/geometry_nodes/curve/topology/offset_point_in_curve.html#bpy-types-geometrynodeoffsetpointincurve"),
    ("bpy.types.geometrynodeseparatecomponents*", "modeling/geometry_nodes/geometry/operations/separate_components.html#bpy-types-geometrynodeseparatecomponents"),
    ("bpy.types.geometrynodesubdivisionsurface*", "modeling/geometry_nodes/mesh/operations/subdivision_surface.html#bpy-types-geometrynodesubdivisionsurface"),
    ("bpy.types.geometrynodetranslateinstances*", "modeling/geometry_nodes/instances/translate_instances.html#bpy-types-geometrynodetranslateinstances"),
    ("bpy.types.geometrynodetree.is_mode_paint*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-mode-paint"),
    ("bpy.types.geometrynodetree.is_type_curve*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-type-curve"),
    ("bpy.types.greasepencilshrinkwrapmodifier*", "grease_pencil/modifiers/deform/shrinkwrap.html#bpy-types-greasepencilshrinkwrapmodifier"),
    ("bpy.types.imageformatsettings.color_mode*", "render/output/properties/output.html#bpy-types-imageformatsettings-color-mode"),
    ("bpy.types.imagetexture.use_interpolation*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-interpolation"),
    ("bpy.types.keyingsetpath.use_entire_array*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-use-entire-array"),
    ("bpy.types.linestyle*modifier_alongstroke*", "render/freestyle/view_layer/line_style/modifiers/color/along_stroke.html#bpy-types-linestyle-modifier-alongstroke"),
    ("bpy.types.linestyle*modifier_creaseangle*", "render/freestyle/view_layer/line_style/modifiers/color/crease_angle.html#bpy-types-linestyle-modifier-creaseangle"),
    ("bpy.types.linestylecolormodifier_tangent*", "render/freestyle/view_layer/line_style/modifiers/color/tangent.html#bpy-types-linestylecolormodifier-tangent"),
    ("bpy.types.material.surface_render_method*", "render/eevee/material_settings.html#bpy-types-material-surface-render-method"),
    ("bpy.types.materialgpencilstyle.mix_color*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-mix-color"),
    ("bpy.types.materialgpencilstyle.show_fill*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-show-fill"),
    ("bpy.types.movietrackingcamera.division_k*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-division-k"),
    ("bpy.types.movietrackingobject.keyframe_a*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingobject-keyframe-a"),
    ("bpy.types.movietrackingobject.keyframe_b*", "movie_clip/tracking/clip/toolbar/solve.html#bpy-types-movietrackingobject-keyframe-b"),
    ("bpy.types.movietrackingtrack.weight_stab*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-weight-stab"),
    ("bpy.types.object.use_shape_key_edit_mode*", "animation/shape_keys/shape_keys_panel.html#bpy-types-object-use-shape-key-edit-mode"),
    ("bpy.types.objectlineart.crease_threshold*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart-crease-threshold"),
    ("bpy.types.preferencesinput.ndof_deadzone*", "editors/preferences/input.html#bpy-types-preferencesinput-ndof-deadzone"),
    ("bpy.types.preferencessystem.audio_device*", "editors/preferences/system.html#bpy-types-preferencessystem-audio-device"),
    ("bpy.types.preferencessystem.vbo_time_out*", "editors/preferences/system.html#bpy-types-preferencessystem-vbo-time-out"),
    ("bpy.types.preferencesview.mini_axis_type*", "editors/preferences/viewport.html#bpy-types-preferencesview-mini-axis-type"),
    ("bpy.types.preferencesview.rotation_angle*", "editors/preferences/navigation.html#bpy-types-preferencesview-rotation-angle"),
    ("bpy.types.preferencesview.show_statusbar*", "editors/preferences/interface.html#bpy-types-preferencesview-show-statusbar"),
    ("bpy.types.preferencesview.show_view_name*", "editors/preferences/viewport.html#bpy-types-preferencesview-show-view-name"),
    ("bpy.types.preferencesview.timecode_style*", "editors/preferences/animation.html#bpy-types-preferencesview-timecode-style"),
    ("bpy.types.rendersettings.use_compositing*", "render/output/properties/post_processing.html#bpy-types-rendersettings-use-compositing"),
    ("bpy.types.rendersettings.use_motion_blur*", "render/cycles/render_settings/motion_blur.html#bpy-types-rendersettings-use-motion-blur"),
    ("bpy.types.rendersettings.use_placeholder*", "render/output/properties/output.html#bpy-types-rendersettings-use-placeholder"),
    ("bpy.types.sceneeevee.clamp_volume_direct*", "render/eevee/render_settings/clamping.html#bpy-types-sceneeevee-clamp-volume-direct"),
    ("bpy.types.sequenceeditor.use_cache_final*", "editors/video_sequencer/sequencer/sidebar/cache.html#bpy-types-sequenceeditor-use-cache-final"),
    ("bpy.types.shadernodesubsurfacescattering*", "render/shader_nodes/shader/sss.html#bpy-types-shadernodesubsurfacescattering"),
    ("bpy.types.softbodysettings.goal_friction*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-goal-friction"),
    ("bpy.types.softbodysettings.spring_length*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-spring-length"),
    ("bpy.types.softbodysettings.use_auto_step*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-use-auto-step"),
    ("bpy.types.spaceclipeditor.lock_selection*", "editors/clip/introduction.html#bpy-types-spaceclipeditor-lock-selection"),
    ("bpy.types.spacenodeoverlay.show_overlays*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-overlays"),
    ("bpy.types.spacenodeoverlay.show_previews*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-previews"),
    ("bpy.types.spaceoutliner.show_mode_column*", "editors/outliner/interface.html#bpy-types-spaceoutliner-show-mode-column"),
    ("bpy.types.spacesequenceeditor.show_gizmo*", "editors/video_sequencer/preview/display/gizmos.html#bpy-types-spacesequenceeditor-show-gizmo"),
    ("bpy.types.spacetexteditor.use_match_case*", "editors/text_editor.html#bpy-types-spacetexteditor-use-match-case"),
    ("bpy.types.spaceuveditor.pixel_round_mode*", "modeling/meshes/uv/editing.html#bpy-types-spaceuveditor-pixel-round-mode"),
    ("bpy.types.spaceview3d.show_object_select*", "editors/3dview/display/visibility.html#bpy-types-spaceview3d-show-object-select"),
    ("bpy.types.stripmodifier.input_mask_strip*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-stripmodifier-input-mask-strip"),
    ("bpy.types.textcharacterformat.use_italic*", "modeling/texts/properties.html#bpy-types-textcharacterformat-use-italic"),
    ("bpy.types.toolsettings.snap_anim_element*", "editors/graph_editor/introduction.html#bpy-types-toolsettings-snap-anim-element"),
    ("bpy.types.toolsettings.use_lock_relative*", "sculpt_paint/weight_paint/tool_settings/options.html#bpy-types-toolsettings-use-lock-relative"),
    ("bpy.types.toolsettings.use_snap_playhead*", "editors/timeline.html#bpy-types-toolsettings-use-snap-playhead"),
    ("bpy.types.toolsettings.vertex_group_user*", "editors/3dview/display/overlays.html#bpy-types-toolsettings-vertex-group-user"),
    ("bpy.types.userassetlibrary.import_method*", "editors/preferences/file_paths.html#bpy-types-userassetlibrary-import-method"),
    ("bpy.types.vertexpaint.use_group_restrict*", "sculpt_paint/weight_paint/tool_settings/options.html#bpy-types-vertexpaint-use-group-restrict"),
    ("bpy.types.view3doverlay.show_viewer_text*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-viewer-text"),
    ("bpy.types.view3doverlay.use_gpencil_grid*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-use-gpencil-grid"),
    ("bpy.types.viewlayer.pass_alpha_threshold*", "render/layers/passes.html#bpy-types-viewlayer-pass-alpha-threshold"),
    ("bpy.types.viewlayer.use_pass_environment*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-environment"),
    ("bpy.types.volumedisplay.wireframe_detail*", "modeling/volumes/properties.html#bpy-types-volumedisplay-wireframe-detail"),
    ("bpy.types.windowmanager.asset_path_dummy*", "editors/asset_browser.html#bpy-types-windowmanager-asset-path-dummy"),
    ("bpy.ops.armature.rigify_add_bone_groups*", "addons/rigging/rigify/metarigs.html#bpy-ops-armature-rigify-add-bone-groups"),
    ("bpy.ops.geometry.color_attribute_remove*", "modeling/meshes/properties/object_data.html#bpy-ops-geometry-color-attribute-remove"),
    ("bpy.ops.grease_pencil.material_lock_all*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-lock-all"),
    ("bpy.ops.object.assign_property_defaults*", "animation/armatures/posing/editing/apply.html#bpy-ops-object-assign-property-defaults"),
    ("bpy.ops.object.material_slot_remove_all*", "render/materials/assignment.html#bpy-ops-object-material-slot-remove-all"),
    ("bpy.ops.object.vertex_group_limit_total*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-limit-total"),
    ("bpy.ops.object.vertex_group_remove_from*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-remove-from"),
    ("bpy.ops.outliner.collection_exclude_set*", "editors/outliner/editing.html#bpy-ops-outliner-collection-exclude-set"),
    ("bpy.ops.outliner.collection_hide_inside*", "editors/outliner/editing.html#bpy-ops-outliner-collection-hide-inside"),
    ("bpy.ops.outliner.collection_holdout_set*", "render/layers/introduction.html#bpy-ops-outliner-collection-holdout-set"),
    ("bpy.ops.outliner.collection_show_inside*", "editors/outliner/editing.html#bpy-ops-outliner-collection-show-inside"),
    ("bpy.ops.pose.selection_set_remove_bones*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-remove-bones"),
    ("bpy.ops.poselib.pose_asset_select_bones*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-pose-asset-select-bones"),
    ("bpy.ops.preferences.reset_default_theme*", "editors/preferences/themes.html#bpy-ops-preferences-reset-default-theme"),
    ("bpy.ops.scene.view_layer_add_lightgroup*", "render/layers/passes.html#bpy-ops-scene-view-layer-add-lightgroup"),
    ("bpy.ops.sculpt.dyntopo_detail_size_edit*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-ops-sculpt-dyntopo-detail-size-edit"),
    ("bpy.ops.sculpt_curves.min_distance_edit*", "sculpt_paint/curves_sculpting/brushes/density_curves.html#bpy-ops-sculpt-curves-min-distance-edit"),
    ("bpy.ops.sequencer.strip_transform_clear*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-strip-transform-clear"),
    ("bpy.ops.sequencer.text_edit_mode_toggle*", "video_editing/edit/montage/strips/text.html#bpy-ops-sequencer-text-edit-mode-toggle"),
    ("bpy.ops.spreadsheet.add_row_filter_rule*", "editors/spreadsheet.html#bpy-ops-spreadsheet-add-row-filter-rule"),
    ("bpy.types.animdata.action_extrapolation*", "editors/nla/sidebar.html#bpy-types-animdata-action-extrapolation"),
    ("bpy.types.animvizmotionpaths.frame_step*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-frame-step"),
    ("bpy.types.bakesettings.max_ray_distance*", "render/cycles/baking.html#bpy-types-bakesettings-max-ray-distance"),
    ("bpy.types.brush.multiplane_scrape_angle*", "sculpt_paint/sculpting/brushes/multiplane_scrape.html#bpy-types-brush-multiplane-scrape-angle"),
    ("bpy.types.brush.wet_paint_radius_factor*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-wet-paint-radius-factor"),
    ("bpy.types.brushgpencilsettings.hardness*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-hardness"),
    ("bpy.types.brushgpencilsettings.use_trim*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-use-trim"),
    ("bpy.types.brushtextureslot.random_angle*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-random-angle"),
    ("bpy.types.clothsettings.bending_damping*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-bending-damping"),
    ("bpy.types.clothsettings.pressure_factor*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-pressure-factor"),
    ("bpy.types.clothsettings.shear_stiffness*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-shear-stiffness"),
    ("bpy.types.clothsettings.tension_damping*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-tension-damping"),
    ("bpy.types.colormanagedviewsettings.look*", "render/color_management.html#bpy-types-colormanagedviewsettings-look"),
    ("bpy.types.compositornodecolorcorrection*", "compositing/types/color/adjust/color_correction.html#bpy-types-compositornodecolorcorrection"),
    ("bpy.types.compositornodemoviedistortion*", "compositing/types/transform/movie_distortion.html#bpy-types-compositornodemoviedistortion"),
    ("bpy.types.compositornoderelativetopixel*", "compositing/types/utilities/relative_to_pixel.html#bpy-types-compositornoderelativetopixel"),
    ("bpy.types.cyclesrendersettings.caustics*", "render/cycles/render_settings/light_paths.html#bpy-types-cyclesrendersettings-caustics"),
    ("bpy.types.cyclesrendersettings.denoiser*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-denoiser"),
    ("bpy.types.editbone.use_inherit_rotation*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-use-inherit-rotation"),
    ("bpy.types.ffmpegsettings.audio_channels*", "render/output/properties/output.html#bpy-types-ffmpegsettings-audio-channels"),
    ("bpy.types.fieldsettings.guide_kink_axis*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-kink-axis"),
    ("bpy.types.fieldsettings.guide_kink_type*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-kink-type"),
    ("bpy.types.fileselectparams.display_size*", "editors/file_browser.html#bpy-types-fileselectparams-display-size"),
    ("bpy.types.fileselectparams.display_type*", "editors/file_browser.html#bpy-types-fileselectparams-display-type"),
    ("bpy.types.fluiddomainsettings.use_guide*", "physics/fluid/type/domain/guides.html#bpy-types-fluiddomainsettings-use-guide"),
    ("bpy.types.fluiddomainsettings.use_noise*", "physics/fluid/type/domain/gas/noise.html#bpy-types-fluiddomainsettings-use-noise"),
    ("bpy.types.fluiddomainsettings.use_slice*", "physics/fluid/type/domain/gas/viewport_display.html#bpy-types-fluiddomainsettings-use-slice"),
    ("bpy.types.fluiddomainsettings.vorticity*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-vorticity"),
    ("bpy.types.fluidflowsettings.flow_source*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-flow-source"),
    ("bpy.types.fluidflowsettings.fuel_amount*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-fuel-amount"),
    ("bpy.types.fluidflowsettings.smoke_color*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-smoke-color"),
    ("bpy.types.fluidflowsettings.temperature*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-temperature"),
    ("bpy.types.fluidflowsettings.use_texture*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-texture"),
    ("bpy.types.fmodifierenvelopecontrolpoint*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifierenvelopecontrolpoint"),
    ("bpy.types.freestylelinestyle.length_max*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-length-max"),
    ("bpy.types.freestylelinestyle.length_min*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-length-min"),
    ("bpy.types.freestylelinestyle.sort_order*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-sort-order"),
    ("bpy.types.freestylelinestyle.split_dash*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-split-dash"),
    ("bpy.types.freestylesettings.use_culling*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-use-culling"),
    ("bpy.types.functionnodematrixdeterminant*", "modeling/geometry_nodes/utilities/matrix/matrix_determinant.html#bpy-types-functionnodematrixdeterminant"),
    ("bpy.types.functionnodeseparatetransform*", "modeling/geometry_nodes/utilities/matrix/separate_transform.html#bpy-types-functionnodeseparatetransform"),
    ("bpy.types.geometrynodebake.active_index*", "modeling/geometry_nodes/geometry/operations/bake.html#bpy-types-geometrynodebake-active-index"),
    ("bpy.types.geometrynodeduplicateelements*", "modeling/geometry_nodes/geometry/operations/duplicate_elements.html#bpy-types-geometrynodeduplicateelements"),
    ("bpy.types.geometrynodeedgepathstocurves*", "modeling/geometry_nodes/mesh/operations/edge_paths_to_curves.html#bpy-types-geometrynodeedgepathstocurves"),
    ("bpy.types.geometrynodeedgestofacegroups*", "modeling/geometry_nodes/mesh/read/edges_to_face_groups.html#bpy-types-geometrynodeedgestofacegroups"),
    ("bpy.types.geometrynodeinputactivecamera*", "modeling/geometry_nodes/input/scene/active_camera.html#bpy-types-geometrynodeinputactivecamera"),
    ("bpy.types.geometrynodeinputmeshfacearea*", "modeling/geometry_nodes/mesh/read/face_area.html#bpy-types-geometrynodeinputmeshfacearea"),
    ("bpy.types.geometrynodeinputsplinecyclic*", "modeling/geometry_nodes/curve/read/is_spline_cyclic.html#bpy-types-geometrynodeinputsplinecyclic"),
    ("bpy.types.geometrynodeinstancestopoints*", "modeling/geometry_nodes/instances/instances_to_points.html#bpy-types-geometrynodeinstancestopoints"),
    ("bpy.types.geometrynodeinstancetransform*", "modeling/geometry_nodes/instances/instance_transform.html#bpy-types-geometrynodeinstancetransform"),
    ("bpy.types.geometrynodeinterpolatecurves*", "modeling/geometry_nodes/curve/operations/interpolate_curves.html#bpy-types-geometrynodeinterpolatecurves"),
    ("bpy.types.geometrynodematerialselection*", "modeling/geometry_nodes/material/material_selection.html#bpy-types-geometrynodematerialselection"),
    ("bpy.types.geometrynodetoolactiveelement*", "modeling/geometry_nodes/geometry/read/active_element.html#bpy-types-geometrynodetoolactiveelement"),
    ("bpy.types.geometrynodetoolmouseposition*", "modeling/geometry_nodes/input/scene/mouse_position.html#bpy-types-geometrynodetoolmouseposition"),
    ("bpy.types.geometrynodetree.is_mode_edit*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-mode-edit"),
    ("bpy.types.geometrynodetree.is_type_mesh*", "editors/geometry_node.html#bpy-types-geometrynodetree-is-type-mesh"),
    ("bpy.types.geometrynodeviewporttransform*", "modeling/geometry_nodes/input/scene/viewport_transform.html#bpy-types-geometrynodeviewporttransform"),
    ("bpy.types.greasepencillayer.tint_factor*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-tint-factor"),
    ("bpy.types.greasepencilthickmodifierdata*", "grease_pencil/modifiers/deform/thickness.html#bpy-types-greasepencilthickmodifierdata"),
    ("bpy.types.greasepencilv3.use_onion_fade*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-use-onion-fade"),
    ("bpy.types.greasepencilv3.use_onion_loop*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-use-onion-loop"),
    ("bpy.types.imageformatsettings.exr_codec*", "files/media/image_formats.html#bpy-types-imageformatsettings-exr-codec"),
    ("bpy.types.imagepaint.use_normal_falloff*", "sculpt_paint/brush/falloff.html#bpy-types-imagepaint-use-normal-falloff"),
    ("bpy.types.imagetexture.use_mipmap_gauss*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-mipmap-gauss"),
    ("bpy.types.layercollection.hide_viewport*", "editors/outliner/interface.html#bpy-types-layercollection-hide-viewport"),
    ("bpy.types.layercollection.indirect_only*", "editors/outliner/interface.html#bpy-types-layercollection-indirect-only"),
    ("bpy.types.light.use_absolute_resolution*", "render/eevee/light_settings.html#bpy-types-light-use-absolute-resolution"),
    ("bpy.types.lightprobe.influence_distance*", "render/eevee/light_probes/plane.html#bpy-types-lightprobe-influence-distance"),
    ("bpy.types.material.use_backface_culling*", "render/eevee/material_settings.html#bpy-types-material-use-backface-culling"),
    ("bpy.types.materiallineart.mat_occlusion*", "render/materials/line_art.html#bpy-types-materiallineart-mat-occlusion"),
    ("bpy.types.mesh.use_mirror_vertex_groups*", "sculpt_paint/weight_paint/tool_settings/symmetry.html#bpy-types-mesh-use-mirror-vertex-groups"),
    ("bpy.types.meshuvlooplayer.active_render*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-types-meshuvlooplayer-active-render"),
    ("bpy.types.movieclipsequence.stabilize2d*", "video_editing/edit/montage/strips/clip.html#bpy-types-movieclipsequence-stabilize2d"),
    ("bpy.types.movietrackingcamera.principal*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-principal"),
    ("bpy.types.object.active_shape_key_index*", "animation/shape_keys/shape_keys_panel.html#bpy-types-object-active-shape-key-index"),
    ("bpy.types.object.use_camera_lock_parent*", "scene_layout/object/properties/relations.html#bpy-types-object-use-camera-lock-parent"),
    ("bpy.types.object.visible_volume_scatter*", "render/cycles/object_settings/object_data.html#bpy-types-object-visible-volume-scatter"),
    ("bpy.types.preferencesedit.material_link*", "editors/preferences/editing.html#bpy-types-preferencesedit-material-link"),
    ("bpy.types.preferencesedit.use_duplicate*", "editors/preferences/editing.html#bpy-types-preferencesedit-use-duplicate"),
    ("bpy.types.preferencessystem.gpu_backend*", "editors/preferences/system.html#bpy-types-preferencessystem-gpu-backend"),
    ("bpy.types.preferencessystem.viewport_aa*", "editors/preferences/viewport.html#bpy-types-preferencessystem-viewport-aa"),
    ("bpy.types.preferencesview.show_tooltips*", "editors/preferences/interface.html#bpy-types-preferencesview-show-tooltips"),
    ("bpy.types.preferencesview.ui_line_width*", "editors/preferences/interface.html#bpy-types-preferencesview-ui-line-width"),
    ("bpy.types.preferencesview.use_translate*", "editors/preferences/interface.html#bpy-types-preferencesview-use-translate"),
    ("bpy.types.rendersettings.line_thickness*", "render/freestyle/render.html#bpy-types-rendersettings-line-thickness"),
    ("bpy.types.rendersettings.pixel_aspect_x*", "render/output/properties/format.html#bpy-types-rendersettings-pixel-aspect-x"),
    ("bpy.types.rendersettings.pixel_aspect_y*", "render/output/properties/format.html#bpy-types-rendersettings-pixel-aspect-y"),
    ("bpy.types.rigidbodyconstraint.use_limit*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-use-limit"),
    ("bpy.types.sceneeevee.taa_render_samples*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-taa-render-samples"),
    ("bpy.types.sceneeevee.use_bokeh_jittered*", "render/eevee/render_settings/depth_of_field.html#bpy-types-sceneeevee-use-bokeh-jittered"),
    ("bpy.types.sculpt.use_automasking_cavity*", "sculpt_paint/sculpting/controls.html#bpy-types-sculpt-use-automasking-cavity"),
    ("bpy.types.shadernodevolumescatter.phase*", "render/shader_nodes/shader/volume_scatter.html#bpy-types-shadernodevolumescatter-phase"),
    ("bpy.types.softbodysettings.goal_default*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-goal-default"),
    ("bpy.types.softbodysettings.use_diagnose*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-use-diagnose"),
    ("bpy.types.spaceoutliner.use_sync_select*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-sync-select"),
    ("bpy.types.spaceproperties.outliner_sync*", "editors/properties_editor.html#bpy-types-spaceproperties-outliner-sync"),
    ("bpy.types.spaceproperties.search_filter*", "editors/properties_editor.html#bpy-types-spaceproperties-search-filter"),
    ("bpy.types.spacesequenceeditor.view_type*", "editors/video_sequencer/introduction.html#bpy-types-spacesequenceeditor-view-type"),
    ("bpy.types.spacetexteditor.margin_column*", "editors/text_editor.html#bpy-types-spacetexteditor-margin-column"),
    ("bpy.types.spacetexteditor.use_find_wrap*", "editors/text_editor.html#bpy-types-spacetexteditor-use-find-wrap"),
    ("bpy.types.spacetexteditor.use_live_edit*", "editors/text_editor.html#bpy-types-spacetexteditor-use-live-edit"),
    ("bpy.types.spaceuveditor.tile_grid_shape*", "editors/uv/overlays.html#bpy-types-spaceuveditor-tile-grid-shape"),
    ("bpy.types.spaceuveditor.use_live_unwrap*", "modeling/meshes/uv/editing.html#bpy-types-spaceuveditor-use-live-unwrap"),
    ("bpy.types.spaceuveditor.uv_face_opacity*", "editors/uv/overlays.html#bpy-types-spaceuveditor-uv-face-opacity"),
    ("bpy.types.spaceview3d.use_render_border*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-use-render-border"),
    ("bpy.types.stripmodifier.input_mask_type*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-stripmodifier-input-mask-type"),
    ("bpy.types.toolsettings.auto_keying_mode*", "editors/timeline.html#bpy-types-toolsettings-auto-keying-mode"),
    ("bpy.types.toolsettings.double_threshold*", "modeling/meshes/tools/tool_settings.html#bpy-types-toolsettings-double-threshold"),
    ("bpy.types.toolsettings.lock_object_mode*", "interface/window_system/topbar.html#bpy-types-toolsettings-lock-object-mode"),
    ("bpy.types.toolsettings.mesh_select_mode*", "modeling/meshes/selecting/introduction.html#bpy-types-toolsettings-mesh-select-mode"),
    ("bpy.types.toolsettings.use_snap_nonedit*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-nonedit"),
    ("bpy.types.transformorientationslot.type*", "editors/3dview/controls/orientation.html#bpy-types-transformorientationslot-type"),
    ("bpy.types.transformsequence.scale_start*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence-scale-start"),
    ("bpy.types.unitsettings.temperature_unit*", "scene_layout/scene/properties.html#bpy-types-unitsettings-temperature-unit"),
    ("bpy.types.usersolidlight.specular_color*", "editors/preferences/lights.html#bpy-types-usersolidlight-specular-color"),
    ("bpy.types.vertexweightproximitymodifier*", "modeling/modifiers/modify/weight_proximity.html#bpy-types-vertexweightproximitymodifier"),
    ("bpy.types.view3doverlay.bone_wire_alpha*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-bone-wire-alpha"),
    ("bpy.types.view3doverlay.show_retopology*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-retopology"),
    ("bpy.types.view3doverlay.show_wireframes*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-wireframes"),
    ("bpy.types.view3dshading.background_type*", "editors/3dview/display/shading.html#bpy-types-view3dshading-background-type"),
    ("bpy.types.view3dshading.use_scene_world*", "editors/3dview/display/shading.html#bpy-types-view3dshading-use-scene-world"),
    ("bpy.types.workspace.use_filter_by_owner*", "interface/window_system/workspaces.html#bpy-types-workspace-use-filter-by-owner"),
    ("bpy.ops.armature.collection_unsolo_all*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-unsolo-all"),
    ("bpy.ops.brush.stencil_fit_image_aspect*", "sculpt_paint/brush/texture.html#bpy-ops-brush-stencil-fit-image-aspect"),
    ("bpy.ops.extensions.package_upgrade_all*", "editors/preferences/extensions.html#bpy-ops-extensions-package-upgrade-all"),
    ("bpy.ops.grease_pencil.material_isolate*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-isolate"),
    ("bpy.ops.grease_pencil.select_alternate*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-alternate"),
    ("bpy.ops.grease_pencil.stroke_subdivide*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-subdivide"),
    ("bpy.ops.grease_pencil.vertex_color_hsv*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-vertex-color-hsv"),
    ("bpy.ops.grease_pencil.vertex_color_set*", "grease_pencil/modes/vertex_paint/editing.html#bpy-ops-grease-pencil-vertex-color-set"),
    ("bpy.ops.mesh.vertices_smooth_laplacian*", "modeling/meshes/editing/vertex/laplacian_smooth.html#bpy-ops-mesh-vertices-smooth-laplacian"),
    ("bpy.ops.node.format_string_item_remove*", "modeling/geometry_nodes/utilities/text/format_string.html#bpy-ops-node-format-string-item-remove"),
    ("bpy.ops.object.multires_rebuild_subdiv*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-rebuild-subdiv"),
    ("bpy.ops.object.vertex_group_set_active*", "modeling/meshes/editing/vertex/vertex_groups.html#bpy-ops-object-vertex-group-set-active"),
    ("bpy.ops.outliner.liboverride_operation*", "files/linked_libraries/library_overrides.html#bpy-ops-outliner-liboverride-operation"),
    ("bpy.ops.paint.vertex_color_from_weight*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-from-weight"),
    ("bpy.ops.screen.space_type_set_or_cycle*", "editors/index.html#bpy-ops-screen-space-type-set-or-cycle"),
    ("bpy.ops.sculpt.dynamic_topology_toggle*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-ops-sculpt-dynamic-topology-toggle"),
    ("bpy.ops.sculpt.mask_from_mesh_boundary*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-from-mesh-boundary"),
    ("bpy.ops.sequencer.select_side_of_frame*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-side-of-frame"),
    ("bpy.ops.wm.interface_theme_preset_save*", "editors/preferences/themes.html#bpy-ops-wm-interface-theme-preset-save"),
    ("bpy.types.animvizmotionpaths.frame_end*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-frame-end"),
    ("bpy.types.armature.rigify_colors_index*", "addons/rigging/rigify/metarigs.html#bpy-types-armature-rigify-colors-index"),
    ("bpy.types.armature.rigify_theme_to_add*", "addons/rigging/rigify/metarigs.html#bpy-types-armature-rigify-theme-to-add"),
    ("bpy.types.bakesettings.use_pass_direct*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-direct"),
    ("bpy.types.bakesettings.use_pass_glossy*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-glossy"),
    ("bpy.types.bonecollection.rigify_ui_row*", "addons/rigging/rigify/metarigs.html#bpy-types-bonecollection-rigify-ui-row"),
    ("bpy.types.brush.pose_smooth_iterations*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-pose-smooth-iterations"),
    ("bpy.types.brush.snake_hook_deform_type*", "sculpt_paint/sculpting/brushes/snake_hook.html#bpy-types-brush-snake-hook-deform-type"),
    ("bpy.types.brush.use_automasking_cavity*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-automasking-cavity"),
    ("bpy.types.brush.use_grab_active_vertex*", "sculpt_paint/sculpting/brushes/grab.html#bpy-types-brush-use-grab-active-vertex"),
    ("bpy.types.brush.use_pose_lock_rotation*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-use-pose-lock-rotation"),
    ("bpy.types.clothsettings.rest_shape_key*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-rest-shape-key"),
    ("bpy.types.compositornodebrightcontrast*", "compositing/types/color/adjust/bright_contrast.html#bpy-types-compositornodebrightcontrast"),
    ("bpy.types.compositornodedoubleedgemask*", "compositing/types/mask/double_edge_mask.html#bpy-types-compositornodedoubleedgemask"),
    ("bpy.types.cyclesrendersettings.samples*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-samples"),
    ("bpy.types.dopesheet.filter_fcurve_name*", "editors/dope_sheet/navigating.html#bpy-types-dopesheet-filter-fcurve-name"),
    ("bpy.types.dopesheet.show_only_selected*", "editors/dope_sheet/navigating.html#bpy-types-dopesheet-show-only-selected"),
    ("bpy.types.ffmpegsettings.audio_bitrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-audio-bitrate"),
    ("bpy.types.ffmpegsettings.audio_mixrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-audio-mixrate"),
    ("bpy.types.ffmpegsettings.ffmpeg_preset*", "render/output/properties/output.html#bpy-types-ffmpegsettings-ffmpeg-preset"),
    ("bpy.types.ffmpegsettings.use_autosplit*", "render/output/properties/output.html#bpy-types-ffmpegsettings-use-autosplit"),
    ("bpy.types.ffmpegsettings.video_bitrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-video-bitrate"),
    ("bpy.types.fieldsettings.use_absorption*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-use-absorption"),
    ("bpy.types.fileselectparams.show_hidden*", "editors/file_browser.html#bpy-types-fileselectparams-show-hidden"),
    ("bpy.types.fileselectparams.sort_method*", "editors/file_browser.html#bpy-types-fileselectparams-sort-method"),
    ("bpy.types.fluiddomainsettings.clipping*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-clipping"),
    ("bpy.types.fluiddomainsettings.use_mesh*", "physics/fluid/type/domain/liquid/mesh.html#bpy-types-fluiddomainsettings-use-mesh"),
    ("bpy.types.freestylelinestyle.angle_max*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-angle-max"),
    ("bpy.types.freestylelinestyle.angle_min*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-angle-min"),
    ("bpy.types.freestylelinestyle.split_gap*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-split-gap"),
    ("bpy.types.freestylelinestyle.use_nodes*", "render/freestyle/view_layer/line_style/texture.html#bpy-types-freestylelinestyle-use-nodes"),
    ("bpy.types.functionnodecombinetransform*", "modeling/geometry_nodes/utilities/matrix/combine_transform.html#bpy-types-functionnodecombinetransform"),
    ("bpy.types.geometrynodecaptureattribute*", "modeling/geometry_nodes/attribute/capture_attribute.html#bpy-types-geometrynodecaptureattribute"),
    ("bpy.types.geometrynodeinputshadesmooth*", "modeling/geometry_nodes/mesh/read/is_face_smooth.html#bpy-types-geometrynodeinputshadesmooth"),
    ("bpy.types.geometrynodeinstanceonpoints*", "modeling/geometry_nodes/instances/instance_on_points.html#bpy-types-geometrynodeinstanceonpoints"),
    ("bpy.types.geometrynodepointstovertices*", "modeling/geometry_nodes/point/points_to_vertices.html#bpy-types-geometrynodepointstovertices"),
    ("bpy.types.geometrynoderealizeinstances*", "modeling/geometry_nodes/instances/realize_instances.html#bpy-types-geometrynoderealizeinstances"),
    ("bpy.types.geometrynodeseparategeometry*", "modeling/geometry_nodes/geometry/operations/separate_geometry.html#bpy-types-geometrynodeseparategeometry"),
    ("bpy.types.geometrynodesetmaterialindex*", "modeling/geometry_nodes/material/set_material_index.html#bpy-types-geometrynodesetmaterialindex"),
    ("bpy.types.geometrynodesimulationoutput*", "modeling/geometry_nodes/simulation/simulation_zone.html#bpy-types-geometrynodesimulationoutput"),
    ("bpy.types.geometrynodesplittoinstances*", "modeling/geometry_nodes/geometry/operations/split_to_instances.html#bpy-types-geometrynodesplittoinstances"),
    ("bpy.types.geometrynodetoolsetselection*", "modeling/geometry_nodes/geometry/write/set_selection.html#bpy-types-geometrynodetoolsetselection"),
    ("bpy.types.geometrynodetree.is_modifier*", "interface/controls/nodes/groups.html#bpy-types-geometrynodetree-is-modifier"),
    ("bpy.types.greasepencil.edit_line_color*", "grease_pencil/properties/display.html#bpy-types-greasepencil-edit-line-color"),
    ("bpy.types.greasepencilarmaturemodifier*", "grease_pencil/modifiers/deform/armature.html#bpy-types-greasepencilarmaturemodifier"),
    ("bpy.types.greasepencildashmodifierdata*", "grease_pencil/modifiers/generate/dash.html#bpy-types-greasepencildashmodifierdata"),
    ("bpy.types.greasepencilenvelopemodifier*", "grease_pencil/modifiers/generate/envelope.html#bpy-types-greasepencilenvelopemodifier"),
    ("bpy.types.greasepencillayer.blend_mode*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-blend-mode"),
    ("bpy.types.greasepencillayer.pass_index*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-pass-index"),
    ("bpy.types.greasepencillayer.tint_color*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-tint-color"),
    ("bpy.types.greasepencillayer.use_lights*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-use-lights"),
    ("bpy.types.greasepencilmultiplymodifier*", "grease_pencil/modifiers/generate/multiple_strokes.html#bpy-types-greasepencilmultiplymodifier"),
    ("bpy.types.greasepencilsimplifymodifier*", "grease_pencil/modifiers/generate/simplify.html#bpy-types-greasepencilsimplifymodifier"),
    ("bpy.types.light.shadow_jitter_overblur*", "render/eevee/light_settings.html#bpy-types-light-shadow-jitter-overblur"),
    ("bpy.types.material.displacement_method*", "render/cycles/material_settings.html#bpy-types-material-displacement-method"),
    ("bpy.types.material.preview_render_type*", "render/materials/preview.html#bpy-types-material-preview-render-type"),
    ("bpy.types.modifier.use_apply_on_spline*", "modeling/modifiers/introduction.html#bpy-types-modifier-use-apply-on-spline"),
    ("bpy.types.movietrackingplanetrack.name*", "movie_clip/tracking/clip/sidebar/track/plane_track.html#bpy-types-movietrackingplanetrack-name"),
    ("bpy.types.nodefunctionformatstringitem*", "modeling/geometry_nodes/utilities/text/format_string.html#bpy-types-nodefunctionformatstringitem"),
    ("bpy.types.nodetreeinterfacesocket.name*", "interface/controls/nodes/groups.html#bpy-types-nodetreeinterfacesocket-name"),
    ("bpy.types.object.use_empty_image_alpha*", "modeling/empties.html#bpy-types-object-use-empty-image-alpha"),
    ("bpy.types.preferencesedit.object_align*", "editors/preferences/editing.html#bpy-types-preferencesedit-object-align"),
    ("bpy.types.preferencessystem.scrollback*", "editors/preferences/system.html#bpy-types-preferencessystem-scrollback"),
    ("bpy.types.preferencesview.border_width*", "editors/preferences/interface.html#bpy-types-preferencesview-border-width"),
    ("bpy.types.preferencesview.font_path_ui*", "editors/preferences/interface.html#bpy-types-preferencesview-font-path-ui"),
    ("bpy.types.preferencesview.header_align*", "editors/preferences/interface.html#bpy-types-preferencesview-header-align"),
    ("bpy.types.preferencesview.text_hinting*", "editors/preferences/interface.html#bpy-types-preferencesview-text-hinting"),
    ("bpy.types.rendersettings.frame_map_new*", "render/output/properties/frame_range.html#bpy-types-rendersettings-frame-map-new"),
    ("bpy.types.rendersettings.frame_map_old*", "render/output/properties/frame_range.html#bpy-types-rendersettings-frame-map-old"),
    ("bpy.types.rendersettings.use_auto_tile*", "render/cycles/render_settings/performance.html#bpy-types-rendersettings-use-auto-tile"),
    ("bpy.types.rendersettings.use_overwrite*", "render/output/properties/output.html#bpy-types-rendersettings-use-overwrite"),
    ("bpy.types.rendersettings.use_sequencer*", "render/output/properties/post_processing.html#bpy-types-rendersettings-use-sequencer"),
    ("bpy.types.sceneeevee.shadow_step_count*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-shadow-step-count"),
    ("bpy.types.sequenceeditor.overlay_frame*", "editors/video_sequencer/preview/sidebar.html#bpy-types-sequenceeditor-overlay-frame"),
    ("bpy.types.sequenceeditor.proxy_storage*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-sequenceeditor-proxy-storage"),
    ("bpy.types.sequenceeditor.use_cache_raw*", "editors/video_sequencer/sequencer/sidebar/cache.html#bpy-types-sequenceeditor-use-cache-raw"),
    ("bpy.types.sequencetimelinechannel.lock*", "editors/video_sequencer/sequencer/channels.html#bpy-types-sequencetimelinechannel-lock"),
    ("bpy.types.sequencetimelinechannel.mute*", "editors/video_sequencer/sequencer/channels.html#bpy-types-sequencetimelinechannel-mute"),
    ("bpy.types.sequencetimelinechannel.name*", "editors/video_sequencer/sequencer/channels.html#bpy-types-sequencetimelinechannel-name"),
    ("bpy.types.shadernodebsdfhairprincipled*", "render/shader_nodes/shader/hair_principled.html#bpy-types-shadernodebsdfhairprincipled"),
    ("bpy.types.shadernodevectordisplacement*", "render/shader_nodes/vector/vector_displacement.html#bpy-types-shadernodevectordisplacement"),
    ("bpy.types.shadernodevolumecoefficients*", "render/shader_nodes/shader/volume_coefficients.html#bpy-types-shadernodevolumecoefficients"),
    ("bpy.types.softbodysettings.goal_spring*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-goal-spring"),
    ("bpy.types.spaceclipeditor.blend_factor*", "editors/clip/display/mask_display.html#bpy-types-spaceclipeditor-blend-factor"),
    ("bpy.types.spacegrapheditor.show_cursor*", "editors/graph_editor/introduction.html#bpy-types-spacegrapheditor-show-cursor"),
    ("bpy.types.spaceimageeditor.show_region*", "editors/image/navigating.html#bpy-types-spaceimageeditor-show-region"),
    ("bpy.types.spaceimageeditor.show_repeat*", "editors/image/sidebar.html#bpy-types-spaceimageeditor-show-repeat"),
    ("bpy.types.spacenodeoverlay.show_timing*", "interface/controls/nodes/introduction.html#bpy-types-spacenodeoverlay-show-timing"),
    ("bpy.types.spaceoutliner.use_sort_alpha*", "editors/outliner/interface.html#bpy-types-spaceoutliner-use-sort-alpha"),
    ("bpy.types.spacepreferences.filter_text*", "editors/preferences/keymap.html#bpy-types-spacepreferences-filter-text"),
    ("bpy.types.spacepreferences.filter_type*", "editors/preferences/keymap.html#bpy-types-spacepreferences-filter-type"),
    ("bpy.types.spacetexteditor.replace_text*", "editors/text_editor.html#bpy-types-spacetexteditor-replace-text"),
    ("bpy.types.spacetexteditor.use_find_all*", "editors/text_editor.html#bpy-types-spacetexteditor-use-find-all"),
    ("bpy.types.spaceview3d.use_local_camera*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-use-local-camera"),
    ("bpy.types.textcharacterformat.use_bold*", "modeling/texts/properties.html#bpy-types-textcharacterformat-use-bold"),
    ("bpy.types.textcurve.underline_position*", "modeling/texts/properties.html#bpy-types-textcurve-underline-position"),
    ("bpy.types.toolsettings.snap_uv_element*", "editors/uv/controls/snapping.html#bpy-types-toolsettings-snap-uv-element"),
    ("bpy.types.toolsettings.use_snap_rotate*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-rotate"),
    ("bpy.types.toolsettings.uv_relax_method*", "modeling/meshes/uv/tools/relax.html#bpy-types-toolsettings-uv-relax-method"),
    ("bpy.types.unitsettings.system_rotation*", "scene_layout/scene/properties.html#bpy-types-unitsettings-system-rotation"),
    ("bpy.types.usersolidlight.diffuse_color*", "editors/preferences/lights.html#bpy-types-usersolidlight-diffuse-color"),
    ("bpy.types.view3doverlay.display_handle*", "modeling/curves_new/curve_display.html#bpy-types-view3doverlay-display-handle"),
    ("bpy.types.view3doverlay.normals_length*", "modeling/curves/curve_display.html#bpy-types-view3doverlay-normals-length"),
    ("bpy.types.view3dshading.use_compositor*", "editors/3dview/display/shading.html#bpy-types-view3dshading-use-compositor"),
    ("bpy.types.volumedisplay.wireframe_type*", "modeling/volumes/properties.html#bpy-types-volumedisplay-wireframe-type"),
    ("bpy.types.walknavigation.teleport_time*", "editors/preferences/navigation.html#bpy-types-walknavigation-teleport-time"),
    ("bpy.types.windowmanager.extension_type*", "editors/preferences/extensions.html#bpy-types-windowmanager-extension-type"),
    ("bpy.ops.anim.channels_editable_toggle*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-editable-toggle"),
    ("bpy.ops.anim.channels_setting_disable*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-setting-disable"),
    ("bpy.ops.armature.assign_to_collection*", "animation/armatures/bones/editing/parenting.html#bpy-ops-armature-assign-to-collection"),
    ("bpy.ops.brush.stencil_reset_transform*", "sculpt_paint/brush/texture.html#bpy-ops-brush-stencil-reset-transform"),
    ("bpy.ops.curve.normals_make_consistent*", "modeling/curves/editing/control_points.html#bpy-ops-curve-normals-make-consistent"),
    ("bpy.ops.curves.snap_curves_to_surface*", "sculpt_paint/curves_sculpting/introduction.html#bpy-ops-curves-snap-curves-to-surface"),
    ("bpy.ops.ed.lib_id_load_custom_preview*", "editors/asset_browser.html#bpy-ops-ed-lib-id-load-custom-preview"),
    ("bpy.ops.grease_pencil.frame_duplicate*", "grease_pencil/animation/tools.html#bpy-ops-grease-pencil-frame-duplicate"),
    ("bpy.ops.grease_pencil.layer_duplicate*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-duplicate"),
    ("bpy.ops.grease_pencil.layer_group_add*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-group-add"),
    ("bpy.ops.grease_pencil.material_reveal*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-reveal"),
    ("bpy.ops.grease_pencil.set_handle_type*", "grease_pencil/modes/edit/point_menu.html#bpy-ops-grease-pencil-set-handle-type"),
    ("bpy.ops.grease_pencil.set_start_point*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-start-point"),
    ("bpy.ops.grease_pencil.stroke_simplify*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-simplify"),
    ("bpy.ops.mask.add_feather_vertex_slide*", "movie_clip/masking/editing.html#bpy-ops-mask-add-feather-vertex-slide"),
    ("bpy.ops.mesh.faces_select_linked_flat*", "modeling/meshes/selecting/linked.html#bpy-ops-mesh-faces-select-linked-flat"),
    ("bpy.ops.mesh.primitive_ico_sphere_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-ico-sphere-add"),
    ("bpy.ops.node.interface_item_duplicate*", "interface/controls/nodes/groups.html#bpy-ops-node-interface-item-duplicate"),
    ("bpy.ops.object.clear_override_library*", "files/linked_libraries/library_overrides.html#bpy-ops-object-clear-override-library"),
    ("bpy.ops.object.material_slot_deselect*", "render/materials/assignment.html#bpy-ops-object-material-slot-deselect"),
    ("bpy.ops.object.modifier_move_to_index*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-move-to-index"),
    ("bpy.ops.object.multires_external_save*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-external-save"),
    ("bpy.ops.object.reset_override_library*", "files/linked_libraries/library_overrides.html#bpy-ops-object-reset-override-library"),
    ("bpy.ops.object.vertex_group_normalize*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-normalize"),
    ("bpy.ops.object.visual_transform_apply*", "scene_layout/object/editing/apply.html#bpy-ops-object-visual-transform-apply"),
    ("bpy.ops.outliner.collection_duplicate*", "editors/outliner/editing.html#bpy-ops-outliner-collection-duplicate"),
    ("bpy.ops.pose.select_constraint_target*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-constraint-target"),
    ("bpy.ops.pose.selection_set_delete_all*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-delete-all"),
    ("bpy.ops.preferences.unassociate_blend*", "editors/preferences/system.html#bpy-ops-preferences-unassociate-blend"),
    ("bpy.ops.sequencer.strip_color_tag_set*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-ops-sequencer-strip-color-tag-set"),
    ("bpy.ops.sequencer.strip_modifier_copy*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-ops-sequencer-strip-modifier-copy"),
    ("bpy.ops.sequencer.strip_transform_fit*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-strip-transform-fit"),
    ("bpy.ops.wm.doc_view_manual_ui_context*", "getting_started/help.html#bpy-ops-wm-doc-view-manual-ui-context"),
    ("bpy.ops.wm.interface_theme_preset_add*", "editors/preferences/themes.html#bpy-ops-wm-interface-theme-preset-add"),
    ("bpy.types.armature.rigify_colors_lock*", "addons/rigging/rigify/metarigs.html#bpy-types-armature-rigify-colors-lock"),
    ("bpy.types.bakesettings.cage_extrusion*", "render/cycles/baking.html#bpy-types-bakesettings-cage-extrusion"),
    ("bpy.types.bakesettings.use_pass_color*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-color"),
    ("bpy.types.brush.boundary_falloff_type*", "sculpt_paint/sculpting/brushes/boundary.html#bpy-types-brush-boundary-falloff-type"),
    ("bpy.types.brush.cursor_color_subtract*", "sculpt_paint/brush/cursor.html#bpy-types-brush-cursor-color-subtract"),
    ("bpy.types.brush.texture_overlay_alpha*", "sculpt_paint/brush/cursor.html#bpy-types-brush-texture-overlay-alpha"),
    ("bpy.types.brush.use_frontface_falloff*", "sculpt_paint/brush/falloff.html#bpy-types-brush-use-frontface-falloff"),
    ("bpy.types.brush.use_space_attenuation*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-space-attenuation"),
    ("bpy.types.brushgpencilsettings.aspect*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-aspect"),
    ("bpy.types.brushgpencilsettings.dilate*", "grease_pencil/modes/draw/tools/fill.html#bpy-types-brushgpencilsettings-dilate"),
    ("bpy.types.brushgpencilsettings.random*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-random"),
    ("bpy.types.brushtextureslot.use_random*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-use-random"),
    ("bpy.types.clothsettings.bending_model*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-bending-model"),
    ("bpy.types.clothsettings.fluid_density*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-fluid-density"),
    ("bpy.types.clothsettings.pin_stiffness*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-pin-stiffness"),
    ("bpy.types.clothsettings.shear_damping*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-shear-damping"),
    ("bpy.types.clothsettings.target_volume*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-target-volume"),
    ("bpy.types.colormanageddisplaysettings*", "render/color_management.html#bpy-types-colormanageddisplaysettings"),
    ("bpy.types.colorramp.hue_interpolation*", "interface/controls/templates/color_ramp.html#bpy-types-colorramp-hue-interpolation"),
    ("bpy.types.compositornodebilateralblur*", "compositing/types/filter/blur/bilateral_blur.html#bpy-types-compositornodebilateralblur"),
    ("bpy.types.compositornodecryptomattev2*", "compositing/types/mask/cryptomatte.html#bpy-types-compositornodecryptomattev2"),
    ("bpy.types.compositornodedistancematte*", "compositing/types/keying/distance_key.html#bpy-types-compositornodedistancematte"),
    ("bpy.types.compositornodeseparatecolor*", "compositing/types/color/mix/separate_color.html#bpy-types-compositornodeseparatecolor"),
    ("bpy.types.compositornodesetalpha.mode*", "compositing/types/color/set_alpha.html#bpy-types-compositornodesetalpha-mode"),
    ("bpy.types.curves.use_sculpt_collision*", "sculpt_paint/curves_sculpting/introduction.html#bpy-types-curves-use-sculpt-collision"),
    ("bpy.types.dopesheet.use_filter_invert*", "editors/graph_editor/channels/introduction.html#bpy-types-dopesheet-use-filter-invert"),
    ("bpy.types.editbone.bbone_mapping_mode*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-mapping-mode"),
    ("bpy.types.editbone.use_local_location*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-use-local-location"),
    ("bpy.types.ffmpegsettings.audio_volume*", "render/output/properties/output.html#bpy-types-ffmpegsettings-audio-volume"),
    ("bpy.types.ffmpegsettings.max_b_frames*", "render/output/properties/output.html#bpy-types-ffmpegsettings-max-b-frames"),
    ("bpy.types.fieldsettings.falloff_power*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-falloff-power"),
    ("bpy.types.fieldsettings.guide_minimum*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-minimum"),
    ("bpy.types.fileselectparams.use_filter*", "editors/file_browser.html#bpy-types-fileselectparams-use-filter"),
    ("bpy.types.fluiddomainsettings.gravity*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-gravity"),
    ("bpy.types.fluidflowsettings.flow_type*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-flow-type"),
    ("bpy.types.fluidflowsettings.subframes*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-subframes"),
    ("bpy.types.freestylelineset.collection*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-collection"),
    ("bpy.types.freestylelineset.visibility*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-visibility"),
    ("bpy.types.freestylelinestyle.chaining*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-chaining"),
    ("bpy.types.freestylelinestyle.sort_key*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-sort-key"),
    ("bpy.types.functionnodeeulertorotation*", "modeling/geometry_nodes/utilities/rotation/euler_to_rotation.html#bpy-types-functionnodeeulertorotation"),
    ("bpy.types.functionnoderotationtoeuler*", "modeling/geometry_nodes/utilities/rotation/rotation_to_euler.html#bpy-types-functionnoderotationtoeuler"),
    ("bpy.types.functionnodetransposematrix*", "modeling/geometry_nodes/utilities/matrix/transpose_matrix.html#bpy-types-functionnodetransposematrix"),
    ("bpy.types.geometrynodeaccumulatefield*", "modeling/geometry_nodes/utilities/field/accumulate_field.html#bpy-types-geometrynodeaccumulatefield"),
    ("bpy.types.geometrynodecornersofvertex*", "modeling/geometry_nodes/mesh/topology/corners_of_vertex.html#bpy-types-geometrynodecornersofvertex"),
    ("bpy.types.geometrynodecurvesethandles*", "modeling/geometry_nodes/curve/write/set_handle_type.html#bpy-types-geometrynodecurvesethandles"),
    ("bpy.types.geometrynodecurvesplinetype*", "modeling/geometry_nodes/curve/write/set_spline_type.html#bpy-types-geometrynodecurvesplinetype"),
    ("bpy.types.geometrynodeinputcollection*", "modeling/geometry_nodes/input/constant/collection.html#bpy-types-geometrynodeinputcollection"),
    ("bpy.types.geometrynodeinputedgesmooth*", "modeling/geometry_nodes/mesh/read/is_edge_smooth.html#bpy-types-geometrynodeinputedgesmooth"),
    ("bpy.types.geometrynodeinputmeshisland*", "modeling/geometry_nodes/mesh/read/mesh_island.html#bpy-types-geometrynodeinputmeshisland"),
    ("bpy.types.geometrynodemergebydistance*", "modeling/geometry_nodes/geometry/operations/merge_by_distance.html#bpy-types-geometrynodemergebydistance"),
    ("bpy.types.geometrynoderemoveattribute*", "modeling/geometry_nodes/attribute/remove_named_attribute.html#bpy-types-geometrynoderemoveattribute"),
    ("bpy.types.geometrynodereplacematerial*", "modeling/geometry_nodes/material/replace_material.html#bpy-types-geometrynodereplacematerial"),
    ("bpy.types.geometrynoderotateinstances*", "modeling/geometry_nodes/instances/rotate_instances.html#bpy-types-geometrynoderotateinstances"),
    ("bpy.types.geometrynodesampleuvsurface*", "modeling/geometry_nodes/mesh/sample/sample_uv_surface.html#bpy-types-geometrynodesampleuvsurface"),
    ("bpy.types.geometrynodesetgeometryname*", "modeling/geometry_nodes/geometry/write/set_geometry_name.html#bpy-types-geometrynodesetgeometryname"),
    ("bpy.types.geometrynodesetsplinecyclic*", "modeling/geometry_nodes/curve/write/set_spline_cyclic.html#bpy-types-geometrynodesetsplinecyclic"),
    ("bpy.types.geometrynodesimulationinput*", "modeling/geometry_nodes/simulation/simulation_zone.html#bpy-types-geometrynodesimulationinput"),
    ("bpy.types.geometrynodesplineparameter*", "modeling/geometry_nodes/curve/read/spline_parameter.html#bpy-types-geometrynodesplineparameter"),
    ("bpy.types.greasepencillatticemodifier*", "grease_pencil/modifiers/deform/lattice.html#bpy-types-greasepencillatticemodifier"),
    ("bpy.types.greasepencillayer.use_masks*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-use-masks"),
    ("bpy.types.greasepencillineartmodifier*", "grease_pencil/modifiers/generate/line_art.html#bpy-types-greasepencillineartmodifier"),
    ("bpy.types.greasepencilopacitymodifier*", "grease_pencil/modifiers/color/opacity.html#bpy-types-greasepencilopacitymodifier"),
    ("bpy.types.greasepenciloutlinemodifier*", "grease_pencil/modifiers/generate/outline.html#bpy-types-greasepenciloutlinemodifier"),
    ("bpy.types.greasepenciltexturemodifier*", "grease_pencil/modifiers/edit/texture_mapping.html#bpy-types-greasepenciltexturemodifier"),
    ("bpy.types.greasepencilv3.before_color*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-before-color"),
    ("bpy.types.greasepencilv3.onion_factor*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-onion-factor"),
    ("bpy.types.imageformatsettings.quality*", "files/media/image_formats.html#bpy-types-imageformatsettings-quality"),
    ("bpy.types.imagepaint.screen_grab_size*", "sculpt_paint/texture_paint/tool_settings/options.html#bpy-types-imagepaint-screen-grab-size"),
    ("bpy.types.keyingsetpaths.active_index*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpaths-active-index"),
    ("bpy.types.linestyle*modifier_material*", "render/freestyle/view_layer/line_style/modifiers/color/material.html#bpy-types-linestyle-modifier-material"),
    ("bpy.types.motionpath.use_custom_color*", "animation/motion_paths.html#bpy-types-motionpath-use-custom-color"),
    ("bpy.types.movieclipsequence.undistort*", "video_editing/edit/montage/strips/clip.html#bpy-types-movieclipsequence-undistort"),
    ("bpy.types.movietrackingcamera.brown_k*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-brown-k"),
    ("bpy.types.movietrackingcamera.brown_p*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-brown-p"),
    ("bpy.types.object.active_selection_set*", "animation/armatures/properties/selection_sets.html#bpy-types-object-active-selection-set"),
    ("bpy.types.object.use_simulation_cache*", "physics/simulation_nodes.html#bpy-types-object-use-simulation-cache"),
    ("bpy.types.object.visible_transmission*", "render/cycles/object_settings/object_data.html#bpy-types-object-visible-transmission"),
    ("bpy.types.particlesettingstextureslot*", "physics/particles/texture_influence.html#bpy-types-particlesettingstextureslot"),
    ("bpy.types.pointlight.shadow_soft_size*", "render/lights/light_object.html#bpy-types-pointlight-shadow-soft-size"),
    ("bpy.types.pointlight.use_soft_falloff*", "render/lights/light_object.html#bpy-types-pointlight-use-soft-falloff"),
    ("bpy.types.posebone.ik_rotation_weight*", "animation/armatures/posing/bone_constraints/inverse_kinematics/introduction.html#bpy-types-posebone-ik-rotation-weight"),
    ("bpy.types.preferencesedit.node_margin*", "editors/preferences/editing.html#bpy-types-preferencesedit-node-margin"),
    ("bpy.types.preferencesinput.tablet_api*", "editors/preferences/input.html#bpy-types-preferencesinput-tablet-api"),
    ("bpy.types.preferencesview.show_splash*", "editors/preferences/interface.html#bpy-types-preferencesview-show-splash"),
    ("bpy.types.preferencesview.smooth_view*", "editors/preferences/navigation.html#bpy-types-preferencesview-smooth-view"),
    ("bpy.types.regionview3d.show_sync_view*", "editors/3dview/navigate/views.html#bpy-types-regionview3d-show-sync-view"),
    ("bpy.types.rendersettings.resolution_x*", "render/output/properties/format.html#bpy-types-rendersettings-resolution-x"),
    ("bpy.types.rendersettings.resolution_y*", "render/output/properties/format.html#bpy-types-rendersettings-resolution-y"),
    ("bpy.types.rendersettings.threads_mode*", "render/cycles/render_settings/performance.html#bpy-types-rendersettings-threads-mode"),
    ("bpy.types.rigidbodyconstraint.enabled*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-enabled"),
    ("bpy.types.rigidbodyconstraint.object1*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-object1"),
    ("bpy.types.rigidbodyconstraint.object2*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-object2"),
    ("bpy.types.rigidbodyobject.mesh_source*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-mesh-source"),
    ("bpy.types.rigidbodyobject.restitution*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-restitution"),
    ("bpy.types.sceneeevee.shadow_pool_size*", "render/eevee/render_settings/performance.html#bpy-types-sceneeevee-shadow-pool-size"),
    ("bpy.types.sceneeevee.shadow_ray_count*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-shadow-ray-count"),
    ("bpy.types.sculpt.detail_refine_method*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-types-sculpt-detail-refine-method"),
    ("bpy.types.sculpt.symmetrize_direction*", "sculpt_paint/sculpting/tool_settings/symmetry.html#bpy-types-sculpt-symmetrize-direction"),
    ("bpy.types.sequenceeditor.show_overlay*", "editors/video_sequencer/preview/sidebar.html#bpy-types-sequenceeditor-show-overlay"),
    ("bpy.types.sequenceeditor.use_prefetch*", "editors/video_sequencer/sequencer/sidebar/cache.html#bpy-types-sequenceeditor-use-prefetch"),
    ("bpy.types.softbodysettings.ball_stiff*", "physics/soft_body/settings/self_collision.html#bpy-types-softbodysettings-ball-stiff"),
    ("bpy.types.soundsequence.show_waveform*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-soundsequence-show-waveform"),
    ("bpy.types.spaceclipeditor.show_stable*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-show-stable"),
    ("bpy.types.spaceimageeditor.show_gizmo*", "editors/image/introduction.html#bpy-types-spaceimageeditor-show-gizmo"),
    ("bpy.types.spaceoutliner.filter_invert*", "editors/outliner/interface.html#bpy-types-spaceoutliner-filter-invert"),
    ("bpy.types.spacespreadsheet.use_filter*", "editors/spreadsheet.html#bpy-types-spacespreadsheet-use-filter"),
    ("bpy.types.spacetexteditor.show_margin*", "editors/text_editor.html#bpy-types-spacetexteditor-show-margin"),
    ("bpy.types.spline.radius_interpolation*", "modeling/curves/properties/active_spline.html#bpy-types-spline-radius-interpolation"),
    ("bpy.types.stripmodifier.input_mask_id*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-stripmodifier-input-mask-id"),
    ("bpy.types.toolsettings.use_multipaint*", "sculpt_paint/weight_paint/tool_settings/options.html#bpy-types-toolsettings-use-multipaint"),
    ("bpy.types.toolsettings.use_snap_scale*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-scale"),
    ("bpy.types.toolsettings.uv_select_mode*", "editors/uv/selecting.html#bpy-types-toolsettings-uv-select-mode"),
    ("bpy.types.viewlayer.material_override*", "render/layers/introduction.html#bpy-types-viewlayer-material-override"),
    ("bpy.types.viewlayer.use_grease_pencil*", "render/layers/filter.html#bpy-types-viewlayer-use-grease-pencil"),
    ("bpy.types.viewlayer.use_pass_combined*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-combined"),
    ("bpy.types.viewlayer.use_pass_position*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-position"),
    ("bpy.types.volumerender.velocity_scale*", "modeling/volumes/properties.html#bpy-types-volumerender-velocity-scale"),
    ("bpy.ops.anim.channels_fcurves_enable*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-fcurves-enable"),
    ("bpy.ops.anim.channels_setting_enable*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-setting-enable"),
    ("bpy.ops.anim.channels_setting_toggle*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-setting-toggle"),
    ("bpy.ops.armature.collection_deselect*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-deselect"),
    ("bpy.ops.armature.collection_show_all*", "animation/armatures/bones/editing/parenting.html#bpy-ops-armature-collection-show-all"),
    ("bpy.ops.armature.collection_unassign*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-unassign"),
    ("bpy.ops.clip.set_viewport_background*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-set-viewport-background"),
    ("bpy.ops.geometry.color_attribute_add*", "modeling/meshes/properties/object_data.html#bpy-ops-geometry-color-attribute-add"),
    ("bpy.ops.graph.driver_variables_paste*", "animation/drivers/drivers_panel.html#bpy-ops-graph-driver-variables-paste"),
    ("bpy.ops.grease_pencil.duplicate_move*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-duplicate-move"),
    ("bpy.ops.grease_pencil.join_selection*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-join-selection"),
    ("bpy.ops.grease_pencil.layer_lock_all*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-lock-all"),
    ("bpy.ops.grease_pencil.select_similar*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-similar"),
    ("bpy.ops.grease_pencil.set_curve_type*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-set-curve-type"),
    ("bpy.ops.grease_pencil.snap_to_cursor*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-snap-to-cursor"),
    ("bpy.ops.mask.normals_make_consistent*", "movie_clip/masking/editing.html#bpy-ops-mask-normals-make-consistent"),
    ("bpy.ops.mesh.normals_make_consistent*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-normals-make-consistent"),
    ("bpy.ops.mesh.offset_edge_loops_slide*", "modeling/meshes/editing/edge/offset_edge_slide.html#bpy-ops-mesh-offset-edge-loops-slide"),
    ("bpy.ops.mesh.primitive_uv_sphere_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-uv-sphere-add"),
    ("bpy.ops.node.default_group_width_set*", "interface/controls/nodes/groups.html#bpy-ops-node-default-group-width-set"),
    ("bpy.ops.node.format_string_item_move*", "modeling/geometry_nodes/utilities/text/format_string.html#bpy-ops-node-format-string-item-move"),
    ("bpy.ops.object.curves_empty_hair_add*", "modeling/curves_new/primitives.html#bpy-ops-object-curves-empty-hair-add"),
    ("bpy.ops.object.duplicate_move_linked*", "scene_layout/object/editing/duplicate_linked.html#bpy-ops-object-duplicate-move-linked"),
    ("bpy.ops.object.make_override_library*", "files/linked_libraries/library_overrides.html#bpy-ops-object-make-override-library"),
    ("bpy.ops.object.parent_no_inverse_set*", "scene_layout/object/editing/parent.html#bpy-ops-object-parent-no-inverse-set"),
    ("bpy.ops.object.vertex_group_quantize*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-quantize"),
    ("bpy.ops.outliner.collection_instance*", "editors/outliner/editing.html#bpy-ops-outliner-collection-instance"),
    ("bpy.ops.preferences.keyconfig_export*", "editors/preferences/keymap.html#bpy-ops-preferences-keyconfig-export"),
    ("bpy.ops.preferences.keyconfig_import*", "editors/preferences/keymap.html#bpy-ops-preferences-keyconfig-import"),
    ("bpy.ops.sequencer.change_effect_type*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-change-effect-type"),
    ("bpy.ops.transform.create_orientation*", "editors/3dview/controls/orientation.html#bpy-ops-transform-create-orientation"),
    ("bpy.ops.transform.delete_orientation*", "editors/3dview/controls/orientation.html#bpy-ops-transform-delete-orientation"),
    ("bpy.ops.ui.override_idtemplate_clear*", "files/linked_libraries/library_overrides.html#bpy-ops-ui-override-idtemplate-clear"),
    ("bpy.ops.ui.override_idtemplate_reset*", "files/linked_libraries/library_overrides.html#bpy-ops-ui-override-idtemplate-reset"),
    ("bpy.ops.view3d.localview_remove_from*", "editors/3dview/navigate/local_view.html#bpy-ops-view3d-localview-remove-from"),
    ("bpy.ops.view3d.snap_cursor_to_center*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-view3d-snap-cursor-to-center"),
    ("bpy.types.animdata.action_blend_type*", "editors/nla/sidebar.html#bpy-types-animdata-action-blend-type"),
    ("bpy.types.attributegroupgreasepencil*", "grease_pencil/properties/data.html#bpy-types-attributegroupgreasepencil"),
    ("bpy.types.bakesettings.use_pass_emit*", "render/cycles/baking.html#bpy-types-bakesettings-use-pass-emit"),
    ("bpy.types.bone.use_envelope_multiply*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-use-envelope-multiply"),
    ("bpy.types.brush.boundary_deform_type*", "sculpt_paint/sculpting/brushes/boundary.html#bpy-types-brush-boundary-deform-type"),
    ("bpy.types.brush.cursor_overlay_alpha*", "sculpt_paint/brush/cursor.html#bpy-types-brush-cursor-overlay-alpha"),
    ("bpy.types.brush.normal_radius_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-normal-radius-factor"),
    ("bpy.types.brush.smooth_stroke_factor*", "sculpt_paint/brush/stroke.html#bpy-types-brush-smooth-stroke-factor"),
    ("bpy.types.brush.smooth_stroke_radius*", "sculpt_paint/brush/stroke.html#bpy-types-brush-smooth-stroke-radius"),
    ("bpy.types.brush.tilt_strength_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-tilt-strength-factor"),
    ("bpy.types.brush.topology_rake_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-topology-rake-factor"),
    ("bpy.types.brush.use_pose_ik_anchored*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-use-pose-ik-anchored"),
    ("bpy.types.brush.use_pressure_masking*", "sculpt_paint/brush/texture.html#bpy-types-brush-use-pressure-masking"),
    ("bpy.types.brush.use_pressure_spacing*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-pressure-spacing"),
    ("bpy.types.brushgpencilsettings.angle*", "grease_pencil/modes/draw/brushes/draw.html#bpy-types-brushgpencilsettings-angle"),
    ("bpy.types.clothsettings.use_pressure*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-use-pressure"),
    ("bpy.types.compositornodeantialiasing*", "compositing/types/filter/anti_aliasing.html#bpy-types-compositornodeantialiasing"),
    ("bpy.types.compositornodechannelmatte*", "compositing/types/keying/channel_key.html#bpy-types-compositornodechannelmatte"),
    ("bpy.types.compositornodecolorbalance*", "compositing/types/color/adjust/color_balance.html#bpy-types-compositornodecolorbalance"),
    ("bpy.types.compositornodecombinecolor*", "compositing/types/color/mix/combine_color.html#bpy-types-compositornodecombinecolor"),
    ("bpy.types.compositornodekeyingscreen*", "compositing/types/keying/keying_screen.html#bpy-types-compositornodekeyingscreen"),
    ("bpy.types.dynamicpaintcanvassettings*", "physics/dynamic_paint/canvas.html#bpy-types-dynamicpaintcanvassettings"),
    ("bpy.types.ffmpegsettings.audio_codec*", "render/output/properties/output.html#bpy-types-ffmpegsettings-audio-codec"),
    ("bpy.types.fieldsettings.distance_max*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-distance-max"),
    ("bpy.types.fieldsettings.distance_min*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-distance-min"),
    ("bpy.types.fieldsettings.falloff_type*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-falloff-type"),
    ("bpy.types.fileselectparams.directory*", "editors/file_browser.html#bpy-types-fileselectparams-directory"),
    ("bpy.types.fluidflowsettings.use_flow*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-use-flow"),
    ("bpy.types.fmodifierfunctiongenerator*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifierfunctiongenerator"),
    ("bpy.types.functionnodeaxestorotation*", "modeling/geometry_nodes/utilities/rotation/axes_to_rotation.html#bpy-types-functionnodeaxestorotation"),
    ("bpy.types.functionnodeinvertrotation*", "modeling/geometry_nodes/utilities/rotation/invert_rotation.html#bpy-types-functionnodeinvertrotation"),
    ("bpy.types.functionnodematrixmultiply*", "modeling/geometry_nodes/utilities/matrix/multiply_matrices.html#bpy-types-functionnodematrixmultiply"),
    ("bpy.types.functionnoderotaterotation*", "modeling/geometry_nodes/utilities/rotation/rotate_rotation.html#bpy-types-functionnoderotaterotation"),
    ("bpy.types.functionnodeseparatematrix*", "modeling/geometry_nodes/utilities/matrix/separate_matrix.html#bpy-types-functionnodeseparatematrix"),
    ("bpy.types.functionnodetransformpoint*", "modeling/geometry_nodes/utilities/matrix/transform_point.html#bpy-types-functionnodetransformpoint"),
    ("bpy.types.geometrynodecollectioninfo*", "modeling/geometry_nodes/input/scene/collection_info.html#bpy-types-geometrynodecollectioninfo"),
    ("bpy.types.geometrynodedeletegeometry*", "modeling/geometry_nodes/geometry/operations/delete_geometry.html#bpy-types-geometrynodedeletegeometry"),
    ("bpy.types.geometrynodefieldminandmax*", "modeling/geometry_nodes/utilities/field/field_min_max.html#bpy-types-geometrynodefieldminandmax"),
    ("bpy.types.geometrynodegizmotransform*", "modeling/geometry_nodes/input/gizmo/transform_gizmo.html#bpy-types-geometrynodegizmotransform"),
    ("bpy.types.geometrynodeindexofnearest*", "modeling/geometry_nodes/geometry/sample/index_of_nearest.html#bpy-types-geometrynodeindexofnearest"),
    ("bpy.types.geometrynodeinputcurvetilt*", "modeling/geometry_nodes/curve/read/curve_tilt.html#bpy-types-geometrynodeinputcurvetilt"),
    ("bpy.types.geometrynodeinputscenetime*", "modeling/geometry_nodes/input/scene/scene_time.html#bpy-types-geometrynodeinputscenetime"),
    ("bpy.types.geometrynodepointstocurves*", "modeling/geometry_nodes/point/points_to_curves.html#bpy-types-geometrynodepointstocurves"),
    ("bpy.types.geometrynodepointstovolume*", "modeling/geometry_nodes/point/points_to_volume.html#bpy-types-geometrynodepointstovolume"),
    ("bpy.types.geometrynodescaleinstances*", "modeling/geometry_nodes/instances/scale_instances.html#bpy-types-geometrynodescaleinstances"),
    ("bpy.types.geometrynodesetcurvenormal*", "modeling/geometry_nodes/curve/write/set_curve_normal.html#bpy-types-geometrynodesetcurvenormal"),
    ("bpy.types.geometrynodesetcurveradius*", "modeling/geometry_nodes/curve/write/set_curve_radius.html#bpy-types-geometrynodesetcurveradius"),
    ("bpy.types.geometrynodesetpointradius*", "modeling/geometry_nodes/point/set_point_radius.html#bpy-types-geometrynodesetpointradius"),
    ("bpy.types.geometrynodesetshadesmooth*", "modeling/geometry_nodes/mesh/write/set_shade_smooth.html#bpy-types-geometrynodesetshadesmooth"),
    ("bpy.types.geometrynodestringtocurves*", "modeling/geometry_nodes/utilities/text/string_to_curves.html#bpy-types-geometrynodestringtocurves"),
    ("bpy.types.geometrynodesubdividecurve*", "modeling/geometry_nodes/curve/operations/subdivide_curve.html#bpy-types-geometrynodesubdividecurve"),
    ("bpy.types.geometrynodetoolsetfaceset*", "modeling/geometry_nodes/mesh/write/set_face_set.html#bpy-types-geometrynodetoolsetfaceset"),
    ("bpy.types.geometrynodevertexofcorner*", "modeling/geometry_nodes/mesh/topology/vertex_of_corner.html#bpy-types-geometrynodevertexofcorner"),
    ("bpy.types.greasepencillayer.location*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-location"),
    ("bpy.types.greasepencillayer.rotation*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-rotation"),
    ("bpy.types.greasepencillengthmodifier*", "grease_pencil/modifiers/generate/length.html#bpy-types-greasepencillengthmodifier"),
    ("bpy.types.greasepencilmirrormodifier*", "grease_pencil/modifiers/generate/mirror.html#bpy-types-greasepencilmirrormodifier"),
    ("bpy.types.greasepenciloffsetmodifier*", "grease_pencil/modifiers/deform/offset.html#bpy-types-greasepenciloffsetmodifier"),
    ("bpy.types.greasepencilsmoothmodifier*", "grease_pencil/modifiers/deform/smooth.html#bpy-types-greasepencilsmoothmodifier"),
    ("bpy.types.greasepencilsubdivmodifier*", "grease_pencil/modifiers/generate/subdivide.html#bpy-types-greasepencilsubdivmodifier"),
    ("bpy.types.greasepencilv3.after_color*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-after-color"),
    ("bpy.types.imagetexture.use_flip_axis*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-flip-axis"),
    ("bpy.types.keyframe.handle_right_type*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-handle-right-type"),
    ("bpy.types.keyingsetpath.group_method*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-group-method"),
    ("bpy.types.light.shadow_filter_radius*", "render/eevee/light_settings.html#bpy-types-light-shadow-filter-radius"),
    ("bpy.types.materialgpencilstyle.color*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-color"),
    ("bpy.types.materialgpencilstyle.ghost*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-ghost"),
    ("bpy.types.movietrackingcamera.nuke_k*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-nuke-k"),
    ("bpy.types.movietrackingstabilization*", "movie_clip/tracking/clip/sidebar/stabilization/index.html#bpy-types-movietrackingstabilization"),
    ("bpy.types.object.display_bounds_type*", "scene_layout/object/properties/display.html#bpy-types-object-display-bounds-type"),
    ("bpy.types.object.show_only_shape_key*", "animation/shape_keys/shape_keys_panel.html#bpy-types-object-show-only-shape-key"),
    ("bpy.types.preferencesedit.undo_steps*", "editors/preferences/system.html#bpy-types-preferencesedit-undo-steps"),
    ("bpy.types.preferencesview.gizmo_size*", "editors/preferences/viewport.html#bpy-types-preferencesview-gizmo-size"),
    ("bpy.types.regionview3d.lock_rotation*", "editors/3dview/sidebar.html#bpy-types-regionview3d-lock-rotation"),
    ("bpy.types.rendersettings.hair_subdiv*", "render/eevee/render_settings/curves.html#bpy-types-rendersettings-hair-subdiv"),
    ("bpy.types.rigidbodyobject.use_deform*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-use-deform"),
    ("bpy.types.scene.audio_distance_model*", "scene_layout/scene/properties.html#bpy-types-scene-audio-distance-model"),
    ("bpy.types.scene.audio_doppler_factor*", "scene_layout/scene/properties.html#bpy-types-scene-audio-doppler-factor"),
    ("bpy.types.sceneeevee.light_threshold*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-light-threshold"),
    ("bpy.types.shadernodeambientocclusion*", "render/shader_nodes/input/ao.html#bpy-types-shadernodeambientocclusion"),
    ("bpy.types.shadernodevolumeabsorption*", "render/shader_nodes/shader/volume_absorption.html#bpy-types-shadernodevolumeabsorption"),
    ("bpy.types.shadernodevolumeprincipled*", "render/shader_nodes/shader/volume_principled.html#bpy-types-shadernodevolumeprincipled"),
    ("bpy.types.softbodysettings.ball_damp*", "physics/soft_body/settings/self_collision.html#bpy-types-softbodysettings-ball-damp"),
    ("bpy.types.softbodysettings.ball_size*", "physics/soft_body/settings/self_collision.html#bpy-types-softbodysettings-ball-size"),
    ("bpy.types.softbodysettings.use_edges*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-use-edges"),
    ("bpy.types.soundsequence.sound_offset*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-soundsequence-sound-offset"),
    ("bpy.types.spaceclipeditor.show_gizmo*", "editors/clip/display/gizmo.html#bpy-types-spaceclipeditor-show-gizmo"),
    ("bpy.types.spacefilebrowser.bookmarks*", "editors/file_browser.html#bpy-types-spacefilebrowser-bookmarks"),
    ("bpy.types.spacenodeeditor.show_gizmo*", "editors/compositor.html#bpy-types-spacenodeeditor-show-gizmo"),
    ("bpy.types.spaceoutliner.display_mode*", "editors/outliner/interface.html#bpy-types-spaceoutliner-display-mode"),
    ("bpy.types.spaceoutliner.filter_state*", "editors/outliner/interface.html#bpy-types-spaceoutliner-filter-state"),
    ("bpy.types.spaceuveditor.show_stretch*", "editors/uv/overlays.html#bpy-types-spaceuveditor-show-stretch"),
    ("bpy.types.spotlight.shadow_soft_size*", "render/lights/light_object.html#bpy-types-spotlight-shadow-soft-size"),
    ("bpy.types.spotlight.use_soft_falloff*", "render/lights/light_object.html#bpy-types-spotlight-use-soft-falloff"),
    ("bpy.types.strip.frame_final_duration*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-frame-final-duration"),
    ("bpy.types.strip.use_linear_modifiers*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-strip-use-linear-modifiers"),
    ("bpy.types.textcurve.small_caps_scale*", "modeling/texts/properties.html#bpy-types-textcurve-small-caps-scale"),
    ("bpy.types.toolsettings.keyframe_type*", "editors/timeline.html#bpy-types-toolsettings-keyframe-type"),
    ("bpy.types.toolsettings.use_snap_anim*", "editors/graph_editor/introduction.html#bpy-types-toolsettings-use-snap-anim"),
    ("bpy.types.toolsettings.use_snap_edit*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-edit"),
    ("bpy.types.toolsettings.use_snap_node*", "interface/controls/nodes/arranging.html#bpy-types-toolsettings-use-snap-node"),
    ("bpy.types.toolsettings.use_snap_self*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap-self"),
    ("bpy.types.unifiedpaintsettings.color*", "sculpt_paint/brush/brush_settings.html#bpy-types-unifiedpaintsettings-color"),
    ("bpy.types.viewlayer.active_aov_index*", "render/layers/passes.html#bpy-types-viewlayer-active-aov-index"),
    ("bpy.types.volumerender.velocity_grid*", "modeling/volumes/properties.html#bpy-types-volumerender-velocity-grid"),
    ("bpy.types.volumerender.velocity_unit*", "modeling/volumes/properties.html#bpy-types-volumerender-velocity-unit"),
    ("bpy.types.walknavigation.jump_height*", "editors/preferences/navigation.html#bpy-types-walknavigation-jump-height"),
    ("bpy.types.walknavigation.mouse_speed*", "editors/preferences/navigation.html#bpy-types-walknavigation-mouse-speed"),
    ("bpy.types.walknavigation.use_gravity*", "editors/preferences/navigation.html#bpy-types-walknavigation-use-gravity"),
    ("bpy.types.walknavigation.view_height*", "editors/preferences/navigation.html#bpy-types-walknavigation-view-height"),
    ("bpy.types.windowmanager.addon_search*", "editors/preferences/addons.html#bpy-types-windowmanager-addon-search"),
    ("bpy.ops.anim.channels_view_selected*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-view-selected"),
    ("bpy.ops.anim.keying_set_path_remove*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keying-set-path-remove"),
    ("bpy.ops.armature.move_to_collection*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-move-to-collection"),
    ("bpy.ops.clip.tracking_object_remove*", "movie_clip/tracking/clip/sidebar/track/objects.html#bpy-ops-clip-tracking-object-remove"),
    ("bpy.ops.constraint.copy_to_selected*", "animation/constraints/interface/header.html#bpy-ops-constraint-copy-to-selected"),
    ("bpy.ops.curves.set_selection_domain*", "modeling/curves_new/selecting.html#bpy-ops-curves-set-selection-domain"),
    ("bpy.ops.extensions.repo_refresh_all*", "editors/preferences/extensions.html#bpy-ops-extensions-repo-refresh-all"),
    ("bpy.ops.graph.driver_variables_copy*", "animation/drivers/drivers_panel.html#bpy-ops-graph-driver-variables-copy"),
    ("bpy.ops.grease_pencil.layer_isolate*", "editors/dope_sheet/modes/grease_pencil.html#bpy-ops-grease-pencil-layer-isolate"),
    ("bpy.ops.grease_pencil.material_hide*", "grease_pencil/materials/properties.html#bpy-ops-grease-pencil-material-hide"),
    ("bpy.ops.grease_pencil.move_to_layer*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-move-to-layer"),
    ("bpy.ops.grease_pencil.select_linked*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-linked"),
    ("bpy.ops.grease_pencil.select_random*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-random"),
    ("bpy.ops.grease_pencil.stroke_smooth*", "grease_pencil/modes/edit/point_menu.html#bpy-ops-grease-pencil-stroke-smooth"),
    ("bpy.ops.grease_pencil.weight_invert*", "grease_pencil/modes/weight_paint/weights_menu.html#bpy-ops-grease-pencil-weight-invert"),
    ("bpy.ops.grease_pencil.weight_sample*", "grease_pencil/modes/weight_paint/weights_menu.html#bpy-ops-grease-pencil-weight-sample"),
    ("bpy.ops.image.convert_to_mesh_plane*", "scene_layout/object/editing/convert.html#bpy-ops-image-convert-to-mesh-plane"),
    ("bpy.ops.image.import_as_mesh_planes*", "modeling/meshes/import_images_as_planes.html#bpy-ops-image-import-as-mesh-planes"),
    ("bpy.ops.mask.slide_spline_curvature*", "movie_clip/masking/editing.html#bpy-ops-mask-slide-spline-curvature"),
    ("bpy.ops.mesh.flip_quad_tessellation*", "modeling/meshes/editing/face/face_data.html#bpy-ops-mesh-flip-quad-tessellation"),
    ("bpy.ops.mesh.primitive_cylinder_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-cylinder-add"),
    ("bpy.ops.mesh.set_normals_from_faces*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-set-normals-from-faces"),
    ("bpy.ops.mesh.set_sharpness_by_angle*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-mesh-set-sharpness-by-angle"),
    ("bpy.ops.mesh.shape_propagate_to_all*", "modeling/meshes/editing/vertex/propagate_shapes.html#bpy-ops-mesh-shape-propagate-to-all"),
    ("bpy.ops.mesh.vert_connect_nonplanar*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-vert-connect-nonplanar"),
    ("bpy.ops.node.format_string_item_add*", "modeling/geometry_nodes/utilities/text/format_string.html#bpy-ops-node-format-string-item-add"),
    ("bpy.ops.object.duplicates_make_real*", "scene_layout/object/editing/apply.html#bpy-ops-object-duplicates-make-real"),
    ("bpy.ops.object.material_slot_assign*", "render/materials/assignment.html#bpy-ops-object-material-slot-assign"),
    ("bpy.ops.object.material_slot_remove*", "render/materials/assignment.html#bpy-ops-object-material-slot-remove"),
    ("bpy.ops.object.material_slot_select*", "render/materials/assignment.html#bpy-ops-object-material-slot-select"),
    ("bpy.ops.object.multires_unsubdivide*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-unsubdivide"),
    ("bpy.ops.object.parent_inverse_apply*", "scene_layout/object/editing/apply.html#bpy-ops-object-parent-inverse-apply"),
    ("bpy.ops.object.paths_update_visible*", "animation/motion_paths.html#bpy-ops-object-paths-update-visible"),
    ("bpy.ops.object.transforms_to_deltas*", "scene_layout/object/editing/apply.html#bpy-ops-object-transforms-to-deltas"),
    ("bpy.ops.object.vertex_weight_delete*", "modeling/meshes/properties/vertex_groups/vertex_weights.html#bpy-ops-object-vertex-weight-delete"),
    ("bpy.ops.outliner.collection_disable*", "editors/outliner/editing.html#bpy-ops-outliner-collection-disable"),
    ("bpy.ops.outliner.collection_isolate*", "editors/outliner/editing.html#bpy-ops-outliner-collection-isolate"),
    ("bpy.ops.pose.selection_set_deselect*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-deselect"),
    ("bpy.ops.pose.selection_set_unassign*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-unassign"),
    ("bpy.ops.pose.visual_transform_apply*", "animation/armatures/posing/editing/apply.html#bpy-ops-pose-visual-transform-apply"),
    ("bpy.ops.poselib.convert_old_poselib*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-convert-old-poselib"),
    ("bpy.ops.preferences.associate_blend*", "editors/preferences/system.html#bpy-ops-preferences-associate-blend"),
    ("bpy.ops.preferences.studiolight_new*", "editors/preferences/lights.html#bpy-ops-preferences-studiolight-new"),
    ("bpy.ops.render.shutter_curve_preset*", "render/cycles/render_settings/motion_blur.html#bpy-ops-render-shutter-curve-preset"),
    ("bpy.ops.scene.view_layer_remove_aov*", "render/layers/passes.html#bpy-ops-scene-view-layer-remove-aov"),
    ("bpy.ops.sculpt.project_line_gesture*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-project-line-gesture"),
    ("bpy.ops.sculpt_curves.select_random*", "sculpt_paint/curves_sculpting/introduction.html#bpy-ops-sculpt-curves-select-random"),
    ("bpy.ops.sequencer.view_ghost_border*", "editors/video_sequencer/preview/sidebar.html#bpy-ops-sequencer-view-ghost-border"),
    ("bpy.ops.spreadsheet.reorder_columns*", "editors/spreadsheet.html#bpy-ops-spreadsheet-reorder-columns"),
    ("bpy.ops.ui.override_idtemplate_make*", "files/linked_libraries/library_overrides.html#bpy-ops-ui-override-idtemplate-make"),
    ("bpy.ops.ui.override_type_set_button*", "files/linked_libraries/library_overrides.html#bpy-ops-ui-override-type-set-button"),
    ("bpy.ops.wm.grease_pencil_export_pdf*", "files/import_export/grease_pencil_pdf.html#bpy-ops-wm-grease-pencil-export-pdf"),
    ("bpy.ops.wm.grease_pencil_export_svg*", "files/import_export/grease_pencil_svg.html#bpy-ops-wm-grease-pencil-export-svg"),
    ("bpy.ops.wm.grease_pencil_import_svg*", "files/import_export/grease_pencil_svg.html#bpy-ops-wm-grease-pencil-import-svg"),
    ("bpy.types.animdata.action_influence*", "editors/nla/sidebar.html#bpy-types-animdata-action-influence"),
    ("bpy.types.assetmetadata.description*", "editors/asset_browser.html#bpy-types-assetmetadata-description"),
    ("bpy.types.bakesettings.normal_space*", "render/cycles/baking.html#bpy-types-bakesettings-normal-space"),
    ("bpy.types.bonecollection.is_visible*", "animation/armatures/properties/bone_collections.html#bpy-types-bonecollection-is-visible"),
    ("bpy.types.brush.crease_pinch_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-crease-pinch-factor"),
    ("bpy.types.brush.elastic_deform_type*", "sculpt_paint/sculpting/brushes/elastic_deform.html#bpy-types-brush-elastic-deform-type"),
    ("bpy.types.brush.texture_sample_bias*", "sculpt_paint/brush/texture.html#bpy-types-brush-texture-sample-bias"),
    ("bpy.types.brush.use_cloth_collision*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-use-cloth-collision"),
    ("bpy.types.brush.use_grab_silhouette*", "sculpt_paint/sculpting/brushes/grab.html#bpy-types-brush-use-grab-silhouette"),
    ("bpy.types.brush.use_original_normal*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-original-normal"),
    ("bpy.types.brush.use_pressure_jitter*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-pressure-jitter"),
    ("bpy.types.brush.use_primary_overlay*", "sculpt_paint/brush/cursor.html#bpy-types-brush-use-primary-overlay"),
    ("bpy.types.brushcurvessculptsettings*", "sculpt_paint/curves_sculpting/index.html#bpy-types-brushcurvessculptsettings"),
    ("bpy.types.brushtextureslot.map_mode*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-map-mode"),
    ("bpy.types.brushtextureslot.use_rake*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-use-rake"),
    ("bpy.types.camera.passepartout_alpha*", "render/cameras.html#bpy-types-camera-passepartout-alpha"),
    ("bpy.types.clothsettings.air_damping*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-air-damping"),
    ("bpy.types.colorrampelement.position*", "interface/controls/templates/color_ramp.html#bpy-types-colorrampelement-position"),
    ("bpy.types.compositornodechromamatte*", "compositing/types/keying/chroma_key.html#bpy-types-compositornodechromamatte"),
    ("bpy.types.compositornodecryptomatte*", "compositing/types/mask/cryptomatte_legacy.html#bpy-types-compositornodecryptomatte"),
    ("bpy.types.compositornodedilateerode*", "compositing/types/filter/dilate_erode.html#bpy-types-compositornodedilateerode"),
    ("bpy.types.compositornodeellipsemask*", "compositing/types/mask/ellipse_mask.html#bpy-types-compositornodeellipsemask"),
    ("bpy.types.curve.render_resolution_u*", "modeling/curves/properties/shape.html#bpy-types-curve-render-resolution-u"),
    ("bpy.types.cyclesrendersettings.seed*", "render/cycles/render_settings/sampling.html#bpy-types-cyclesrendersettings-seed"),
    ("bpy.types.dynamicpaintbrushsettings*", "physics/dynamic_paint/brush.html#bpy-types-dynamicpaintbrushsettings"),
    ("bpy.types.editbone.use_scale_easing*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-use-scale-easing"),
    ("bpy.types.ffmpegsettings.buffersize*", "render/output/properties/output.html#bpy-types-ffmpegsettings-buffersize"),
    ("bpy.types.ffmpegsettings.packetsize*", "render/output/properties/output.html#bpy-types-ffmpegsettings-packetsize"),
    ("bpy.types.fieldsettings.wind_factor*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-wind-factor"),
    ("bpy.types.fieldsettings.z_direction*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-z-direction"),
    ("bpy.types.fileselectparams.filename*", "editors/file_browser.html#bpy-types-fileselectparams-filename"),
    ("bpy.types.fluiddomainsettings.alpha*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-alpha"),
    ("bpy.types.fluidflowsettings.density*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings-density"),
    ("bpy.types.freestylelineset.qi_start*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-qi-start"),
    ("bpy.types.freestylelinestyle.rounds*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-rounds"),
    ("bpy.types.functionnodecombinematrix*", "modeling/geometry_nodes/utilities/matrix/combine_matrix.html#bpy-types-functionnodecombinematrix"),
    ("bpy.types.functionnodeinputrotation*", "modeling/geometry_nodes/input/constant/rotation.html#bpy-types-functionnodeinputrotation"),
    ("bpy.types.functionnodereplacestring*", "modeling/geometry_nodes/utilities/text/replace_string.html#bpy-types-functionnodereplacestring"),
    ("bpy.types.functionnodeseparatecolor*", "modeling/geometry_nodes/utilities/color/separate_color.html#bpy-types-functionnodeseparatecolor"),
    ("bpy.types.functionnodevaluetostring*", "modeling/geometry_nodes/utilities/text/value_to_string.html#bpy-types-functionnodevaluetostring"),
    ("bpy.types.geometrynodeblurattribute*", "modeling/geometry_nodes/attribute/blur_attribute.html#bpy-types-geometrynodeblurattribute"),
    ("bpy.types.geometrynodecornersofedge*", "modeling/geometry_nodes/mesh/topology/corners_of_edge.html#bpy-types-geometrynodecornersofedge"),
    ("bpy.types.geometrynodecornersofface*", "modeling/geometry_nodes/mesh/topology/corners_of_face.html#bpy-types-geometrynodecornersofface"),
    ("bpy.types.geometrynodecurvetopoints*", "modeling/geometry_nodes/curve/operations/curve_to_points.html#bpy-types-geometrynodecurvetopoints"),
    ("bpy.types.geometrynodeedgesofcorner*", "modeling/geometry_nodes/mesh/topology/edges_of_corner.html#bpy-types-geometrynodeedgesofcorner"),
    ("bpy.types.geometrynodeedgesofvertex*", "modeling/geometry_nodes/mesh/topology/edges_of_vertex.html#bpy-types-geometrynodeedgesofvertex"),
    ("bpy.types.geometrynodefieldondomain*", "modeling/geometry_nodes/utilities/field/evaluate_on_domain.html#bpy-types-geometrynodefieldondomain"),
    ("bpy.types.geometrynodefieldvariance*", "modeling/geometry_nodes/utilities/field/field_variance.html#bpy-types-geometrynodefieldvariance"),
    ("bpy.types.geometrynodeinputmaterial*", "modeling/geometry_nodes/input/constant/material.html#bpy-types-geometrynodeinputmaterial"),
    ("bpy.types.geometrynodeinputposition*", "modeling/geometry_nodes/geometry/read/position.html#bpy-types-geometrynodeinputposition"),
    ("bpy.types.geometrynodemeshicosphere*", "modeling/geometry_nodes/mesh/primitives/icosphere.html#bpy-types-geometrynodemeshicosphere"),
    ("bpy.types.geometrynodepointsofcurve*", "modeling/geometry_nodes/curve/topology/points_of_curve.html#bpy-types-geometrynodepointsofcurve"),
    ("bpy.types.geometrynoderesamplecurve*", "modeling/geometry_nodes/curve/operations/resample_curve.html#bpy-types-geometrynoderesamplecurve"),
    ("bpy.types.geometrynodesamplenearest*", "modeling/geometry_nodes/geometry/sample/sample_nearest.html#bpy-types-geometrynodesamplenearest"),
    ("bpy.types.geometrynodescaleelements*", "modeling/geometry_nodes/mesh/operations/scale_elements.html#bpy-types-geometrynodescaleelements"),
    ("bpy.types.geometrynodesetmeshnormal*", "modeling/geometry_nodes/mesh/write/set_mesh_normal.html#bpy-types-geometrynodesetmeshnormal"),
    ("bpy.types.geometrynodesubdividemesh*", "modeling/geometry_nodes/mesh/operations/subdivide_mesh.html#bpy-types-geometrynodesubdividemesh"),
    ("bpy.types.geometrynodetoolselection*", "modeling/geometry_nodes/geometry/read/selection.html#bpy-types-geometrynodetoolselection"),
    ("bpy.types.geometrynodeuvpackislands*", "modeling/geometry_nodes/mesh/uv/pack_uv_islands.html#bpy-types-geometrynodeuvpackislands"),
    ("bpy.types.greasepencilarraymodifier*", "grease_pencil/modifiers/generate/array.html#bpy-types-greasepencilarraymodifier"),
    ("bpy.types.greasepencilbuildmodifier*", "grease_pencil/modifiers/generate/build.html#bpy-types-greasepencilbuildmodifier"),
    ("bpy.types.greasepencilcolormodifier*", "grease_pencil/modifiers/color/hue_saturation.html#bpy-types-greasepencilcolormodifier"),
    ("bpy.types.greasepencillayer.opacity*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-opacity"),
    ("bpy.types.greasepencilnoisemodifier*", "grease_pencil/modifiers/deform/noise.html#bpy-types-greasepencilnoisemodifier"),
    ("bpy.types.greasepencilv3.onion_mode*", "grease_pencil/properties/onion_skinning.html#bpy-types-greasepencilv3-onion-mode"),
    ("bpy.types.imagetexture.invert_alpha*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-invert-alpha"),
    ("bpy.types.keyframe.handle_left_type*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-handle-left-type"),
    ("bpy.types.light.transmission_factor*", "render/eevee/light_settings.html#bpy-types-light-transmission-factor"),
    ("bpy.types.light.use_custom_distance*", "render/eevee/light_settings.html#bpy-types-light-use-custom-distance"),
    ("bpy.types.materialgpencilstyle.flip*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-flip"),
    ("bpy.types.materialgpencilstyle.hide*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-hide"),
    ("bpy.types.materialgpencilstyle.lock*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-lock"),
    ("bpy.types.materialgpencilstyle.mode*", "grease_pencil/materials/properties.html#bpy-types-materialgpencilstyle-mode"),
    ("bpy.types.meshsequencecachemodifier*", "modeling/modifiers/modify/mesh_sequence_cache.html#bpy-types-meshsequencecachemodifier"),
    ("bpy.types.modifier.show_in_editmode*", "modeling/modifiers/introduction.html#bpy-types-modifier-show-in-editmode"),
    ("bpy.types.motionpath.line_thickness*", "animation/motion_paths.html#bpy-types-motionpath-line-thickness"),
    ("bpy.types.movietrackingtrack.weight*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-weight"),
    ("bpy.types.object.empty_display_size*", "modeling/empties.html#bpy-types-object-empty-display-size"),
    ("bpy.types.object.empty_display_type*", "modeling/empties.html#bpy-types-object-empty-display-type"),
    ("bpy.types.object.show_texture_space*", "scene_layout/object/properties/display.html#bpy-types-object-show-texture-space"),
    ("bpy.types.raytraceeevee.use_denoise*", "render/eevee/render_settings/raytracing.html#bpy-types-raytraceeevee-use-denoise"),
    ("bpy.types.regionview3d.use_box_clip*", "editors/3dview/navigate/views.html#bpy-types-regionview3d-use-box-clip"),
    ("bpy.types.rendersettings.ppm_factor*", "render/output/properties/output.html#bpy-types-rendersettings-ppm-factor"),
    ("bpy.types.rendersettings.use_border*", "render/output/properties/format.html#bpy-types-rendersettings-use-border"),
    ("bpy.types.rigidbodyconstraint.limit*", "physics/rigid_body/constraints/introduction.html#bpy-types-rigidbodyconstraint-limit"),
    ("bpy.types.rigidbodyobject.kinematic*", "physics/rigid_body/properties/settings.html#bpy-types-rigidbodyobject-kinematic"),
    ("bpy.types.scene.audio_doppler_speed*", "scene_layout/scene/properties.html#bpy-types-scene-audio-doppler-speed"),
    ("bpy.types.sceneeevee.use_raytracing*", "render/eevee/render_settings/raytracing.html#bpy-types-sceneeevee-use-raytracing"),
    ("bpy.types.sculpt.detail_type_method*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-types-sculpt-detail-type-method"),
    ("bpy.types.shadernodebsdfanisotropic*", "render/shader_nodes/shader/glossy.html#bpy-types-shadernodebsdfanisotropic"),
    ("bpy.types.shadernodebsdftranslucent*", "render/shader_nodes/shader/translucent.html#bpy-types-shadernodebsdftranslucent"),
    ("bpy.types.shadernodebsdftransparent*", "render/shader_nodes/shader/transparent.html#bpy-types-shadernodebsdftransparent"),
    ("bpy.types.shadernodetexpointdensity*", "render/shader_nodes/textures/point_density.html#bpy-types-shadernodetexpointdensity"),
    ("bpy.types.shadernodevectortransform*", "render/shader_nodes/vector/transform.html#bpy-types-shadernodevectortransform"),
    ("bpy.types.softbodysettings.friction*", "physics/soft_body/settings/object.html#bpy-types-softbodysettings-friction"),
    ("bpy.types.softbodysettings.goal_max*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-goal-max"),
    ("bpy.types.softbodysettings.goal_min*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-goal-min"),
    ("bpy.types.softbodysettings.step_max*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-step-max"),
    ("bpy.types.softbodysettings.step_min*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-step-min"),
    ("bpy.types.softbodysettings.use_goal*", "physics/soft_body/settings/goal.html#bpy-types-softbodysettings-use-goal"),
    ("bpy.types.spaceclipeditor.show_grid*", "editors/clip/display/clip_display.html#bpy-types-spaceclipeditor-show-grid"),
    ("bpy.types.spaceoutliner.filter_text*", "editors/outliner/interface.html#bpy-types-spaceoutliner-filter-text"),
    ("bpy.types.spacetexteditor.find_text*", "editors/text_editor.html#bpy-types-spacetexteditor-find-text"),
    ("bpy.types.spacetexteditor.font_size*", "editors/text_editor.html#bpy-types-spacetexteditor-font-size"),
    ("bpy.types.spacetexteditor.tab_width*", "editors/text_editor.html#bpy-types-spacetexteditor-tab-width"),
    ("bpy.types.spaceuveditor.lock_bounds*", "modeling/meshes/uv/editing.html#bpy-types-spaceuveditor-lock-bounds"),
    ("bpy.types.spline.tilt_interpolation*", "modeling/curves/properties/active_spline.html#bpy-types-spline-tilt-interpolation"),
    ("bpy.types.textcurve.space_character*", "modeling/texts/properties.html#bpy-types-textcurve-space-character"),
    ("bpy.types.transformorientation.name*", "editors/3dview/controls/orientation.html#bpy-types-transformorientation-name"),
    ("bpy.types.unitsettings.scale_length*", "scene_layout/scene/properties.html#bpy-types-unitsettings-scale-length"),
    ("bpy.types.unitsettings.use_separate*", "scene_layout/scene/properties.html#bpy-types-unitsettings-use-separate"),
    ("bpy.types.vertexgroups.active_index*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-types-vertexgroups-active-index"),
    ("bpy.types.view3doverlay.show_extras*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-extras"),
    ("bpy.types.view3doverlay.show_weight*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay-show-weight"),
    ("bpy.types.viewlayer.use_motion_blur*", "render/layers/filter.html#bpy-types-viewlayer-use-motion-blur"),
    ("bpy.types.viewlayer.use_pass_normal*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-normal"),
    ("bpy.types.viewlayer.use_pass_vector*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-vector"),
    ("bpy.types.volumedisplay.slice_depth*", "modeling/volumes/properties.html#bpy-types-volumedisplay-slice-depth"),
    ("bpy.types.walknavigation.walk_speed*", "editors/preferences/navigation.html#bpy-types-walknavigation-walk-speed"),
    ("bpy.types.worldmistsettings.falloff*", "render/eevee/world_settings.html#bpy-types-worldmistsettings-falloff"),
    ("bpy.ops.anim.keying_set_active_set*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keying-set-active-set"),
    ("bpy.ops.anim.slot_unassign_from_id*", "editors/dope_sheet/modes/action.html#bpy-ops-anim-slot-unassign-from-id"),
    ("bpy.ops.armature.collection_assign*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-assign"),
    ("bpy.ops.armature.collection_select*", "animation/armatures/properties/bone_collections.html#bpy-ops-armature-collection-select"),
    ("bpy.ops.clip.lock_selection_toggle*", "editors/clip/introduction.html#bpy-ops-clip-lock-selection-toggle"),
    ("bpy.ops.collection.exporter_remove*", "scene_layout/collections/collections.html#bpy-ops-collection-exporter-remove"),
    ("bpy.ops.ed.lib_id_generate_preview*", "editors/asset_browser.html#bpy-ops-ed-lib-id-generate-preview"),
    ("bpy.ops.geometry.attribute_convert*", "modeling/geometry_nodes/attributes_reference.html#bpy-ops-geometry-attribute-convert"),
    ("bpy.ops.grease_pencil.cyclical_set*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-cyclical-set"),
    ("bpy.ops.grease_pencil.delete_frame*", "grease_pencil/animation/tools.html#bpy-ops-grease-pencil-delete-frame"),
    ("bpy.ops.grease_pencil.extrude_move*", "grease_pencil/modes/edit/point_menu.html#bpy-ops-grease-pencil-extrude-move"),
    ("bpy.ops.grease_pencil.layer_remove*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-remove"),
    ("bpy.ops.grease_pencil.layer_reveal*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-reveal"),
    ("bpy.ops.grease_pencil.snap_to_grid*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-snap-to-grid"),
    ("bpy.ops.grease_pencil.stroke_split*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-stroke-split"),
    ("bpy.ops.mesh.customdata_mask_clear*", "modeling/meshes/properties/custom_data.html#bpy-ops-mesh-customdata-mask-clear"),
    ("bpy.ops.mesh.customdata_skin_clear*", "modeling/meshes/properties/custom_data.html#bpy-ops-mesh-customdata-skin-clear"),
    ("bpy.ops.mesh.extrude_vertices_move*", "modeling/meshes/editing/vertex/extrude_vertices.html#bpy-ops-mesh-extrude-vertices-move"),
    ("bpy.ops.mesh.mod_weighted_strength*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-mod-weighted-strength"),
    ("bpy.ops.mesh.quads_convert_to_tris*", "modeling/meshes/editing/face/triangulate_faces.html#bpy-ops-mesh-quads-convert-to-tris"),
    ("bpy.ops.mesh.select_interior_faces*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-interior-faces"),
    ("bpy.ops.mesh.select_similar_region*", "modeling/meshes/selecting/similar.html#bpy-ops-mesh-select-similar-region"),
    ("bpy.ops.mesh.tris_convert_to_quads*", "modeling/meshes/editing/face/triangles_quads.html#bpy-ops-mesh-tris-convert-to-quads"),
    ("bpy.ops.node.duplicate_move_linked*", "interface/controls/nodes/editing.html#bpy-ops-node-duplicate-move-linked"),
    ("bpy.ops.node.interface_item_remove*", "interface/controls/nodes/groups.html#bpy-ops-node-interface-item-remove"),
    ("bpy.ops.object.datalayout_transfer*", "scene_layout/object/editing/link_transfer/transfer_mesh_data_layout.html#bpy-ops-object-datalayout-transfer"),
    ("bpy.ops.object.multires_base_apply*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-base-apply"),
    ("bpy.ops.object.randomize_transform*", "scene_layout/object/editing/transform/randomize.html#bpy-ops-object-randomize-transform"),
    ("bpy.ops.object.vertex_group_invert*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-invert"),
    ("bpy.ops.object.vertex_group_levels*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-levels"),
    ("bpy.ops.object.vertex_group_mirror*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-mirror"),
    ("bpy.ops.object.vertex_group_remove*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-remove"),
    ("bpy.ops.object.vertex_group_smooth*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-smooth"),
    ("bpy.ops.object.vertex_weight_paste*", "modeling/meshes/properties/vertex_groups/vertex_weights.html#bpy-ops-object-vertex-weight-paste"),
    ("bpy.ops.outliner.collection_enable*", "editors/outliner/editing.html#bpy-ops-outliner-collection-enable"),
    ("bpy.ops.palette.extract_from_image*", "editors/image/editing.html#bpy-ops-palette-extract-from-image"),
    ("bpy.ops.pose.user_transforms_clear*", "animation/armatures/posing/editing/clear.html#bpy-ops-pose-user-transforms-clear"),
    ("bpy.ops.screen.drivers_editor_show*", "animation/drivers/usage.html#bpy-ops-screen-drivers-editor-show"),
    ("bpy.ops.screen.region_context_menu*", "interface/controls/buttons/menus.html#bpy-ops-screen-region-context-menu"),
    ("bpy.ops.sculpt.set_persistent_base*", "sculpt_paint/sculpting/brushes/layer.html#bpy-ops-sculpt-set-persistent-base"),
    ("bpy.ops.sequencer.crossfade_sounds*", "video_editing/edit/montage/strips/transitions/sound_crossfade.html#bpy-ops-sequencer-crossfade-sounds"),
    ("bpy.ops.sequencer.export_subtitles*", "editors/video_sequencer/preview/header.html#bpy-ops-sequencer-export-subtitles"),
    ("bpy.ops.sequencer.previewrange_set*", "editors/video_sequencer/sequencer/navigating.html#bpy-ops-sequencer-previewrange-set"),
    ("bpy.ops.sequencer.retiming_key_add*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-key-add"),
    ("bpy.ops.text.jump_to_file_at_point*", "editors/text_editor.html#bpy-ops-text-jump-to-file-at-point"),
    ("bpy.ops.transform.edge_bevelweight*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-transform-edge-bevelweight"),
    ("bpy.ops.view3d.clear_render_border*", "editors/3dview/navigate/regions.html#bpy-ops-view3d-clear-render-border"),
    ("bpy.ops.wm.keyconfig_preset_remove*", "editors/preferences/keymap.html#bpy-ops-wm-keyconfig-preset-remove"),
    ("bpy.ops.wm.previews_batch_generate*", "files/blend/previews.html#bpy-ops-wm-previews-batch-generate"),
    ("bpy.types.animvizmotionpaths.range*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-range"),
    ("bpy.types.assetmetadata.active_tag*", "editors/asset_browser.html#bpy-types-assetmetadata-active-tag"),
    ("bpy.types.bakesettings.cage_object*", "render/cycles/baking.html#bpy-types-bakesettings-cage-object"),
    ("bpy.types.bakesettings.margin_type*", "render/cycles/baking.html#bpy-types-bakesettings-margin-type"),
    ("bpy.types.bone.use_relative_parent*", "animation/armatures/bones/properties/relations.html#bpy-types-bone-use-relative-parent"),
    ("bpy.types.brush.area_radius_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-area-radius-factor"),
    ("bpy.types.brush.auto_smooth_factor*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-auto-smooth-factor"),
    ("bpy.types.brush.smooth_deform_type*", "sculpt_paint/sculpting/brushes/smooth.html#bpy-types-brush-smooth-deform-type"),
    ("bpy.types.brush.use_connected_only*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-use-connected-only"),
    ("bpy.types.brush.use_cursor_overlay*", "sculpt_paint/brush/cursor.html#bpy-types-brush-use-cursor-overlay"),
    ("bpy.types.brush.use_original_plane*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-use-original-plane"),
    ("bpy.types.camera.show_passepartout*", "render/cameras.html#bpy-types-camera-show-passepartout"),
    ("bpy.types.clothsettings.shrink_max*", "physics/cloth/settings/property_weights.html#bpy-types-clothsettings-shrink-max"),
    ("bpy.types.clothsettings.shrink_min*", "physics/cloth/settings/shape.html#bpy-types-clothsettings-shrink-min"),
    ("bpy.types.clothsettings.time_scale*", "physics/cloth/settings/index.html#bpy-types-clothsettings-time-scale"),
    ("bpy.types.collection.hide_viewport*", "editors/outliner/interface.html#bpy-types-collection-hide-viewport"),
    ("bpy.types.collection.lineart_usage*", "scene_layout/collections/collections.html#bpy-types-collection-lineart-usage"),
    ("bpy.types.colormanagedviewsettings*", "render/color_management.html#bpy-types-colormanagedviewsettings"),
    ("bpy.types.compositornodebokehimage*", "compositing/types/input/bokeh_image.html#bpy-types-compositornodebokehimage"),
    ("bpy.types.compositornodecolormatte*", "compositing/types/keying/color_key.html#bpy-types-compositornodecolormatte"),
    ("bpy.types.compositornodecolorspill*", "compositing/types/keying/color_spill.html#bpy-types-compositornodecolorspill"),
    ("bpy.types.compositornodehuecorrect*", "compositing/types/color/adjust/hue_correct.html#bpy-types-compositornodehuecorrect"),
    ("bpy.types.compositornodeoutputfile*", "compositing/types/output/file_output.html#bpy-types-compositornodeoutputfile"),
    ("bpy.types.compositornodeswitchview*", "compositing/types/utilities/switch_stereo_view.html#bpy-types-compositornodeswitchview"),
    ("bpy.types.copytransformsconstraint*", "animation/constraints/transform/copy_transforms.html#bpy-types-copytransformsconstraint"),
    ("bpy.types.correctivesmoothmodifier*", "modeling/modifiers/deform/corrective_smooth.html#bpy-types-correctivesmoothmodifier"),
    ("bpy.types.curve.bevel_factor_start*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-factor-start"),
    ("bpy.types.fieldsettings.guide_free*", "physics/forces/force_fields/types/curve_guide.html#bpy-types-fieldsettings-guide-free"),
    ("bpy.types.fluiddomainsettings.beta*", "physics/fluid/type/domain/settings.html#bpy-types-fluiddomainsettings-beta"),
    ("bpy.types.fluidmodifier.fluid_type*", "physics/fluid/type/index.html#bpy-types-fluidmodifier-fluid-type"),
    ("bpy.types.freestylelineset.exclude*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-exclude"),
    ("bpy.types.freestylelinestyle.alpha*", "render/freestyle/view_layer/line_style/alpha.html#bpy-types-freestylelinestyle-alpha"),
    ("bpy.types.freestylelinestyle.color*", "render/freestyle/view_layer/line_style/color.html#bpy-types-freestylelinestyle-color"),
    ("bpy.types.functionnodecombinecolor*", "modeling/geometry_nodes/utilities/color/combine_color.html#bpy-types-functionnodecombinecolor"),
    ("bpy.types.functionnodefindinstring*", "modeling/geometry_nodes/utilities/text/find_in_string.html#bpy-types-functionnodefindinstring"),
    ("bpy.types.functionnodeformatstring*", "modeling/geometry_nodes/utilities/text/format_string.html#bpy-types-functionnodeformatstring"),
    ("bpy.types.functionnodeinvertmatrix*", "modeling/geometry_nodes/utilities/matrix/invert_matrix.html#bpy-types-functionnodeinvertmatrix"),
    ("bpy.types.functionnodeprojectpoint*", "modeling/geometry_nodes/utilities/matrix/project_point.html#bpy-types-functionnodeprojectpoint"),
    ("bpy.types.functionnoderotatevector*", "modeling/geometry_nodes/utilities/rotation/rotate_vector.html#bpy-types-functionnoderotatevector"),
    ("bpy.types.functionnodestringlength*", "modeling/geometry_nodes/utilities/text/string_length.html#bpy-types-functionnodestringlength"),
    ("bpy.types.geometrynodecurveofpoint*", "modeling/geometry_nodes/curve/topology/curve_of_point.html#bpy-types-geometrynodecurveofpoint"),
    ("bpy.types.geometrynodefaceofcorner*", "modeling/geometry_nodes/mesh/topology/face_of_corner.html#bpy-types-geometrynodefaceofcorner"),
    ("bpy.types.geometrynodefieldatindex*", "modeling/geometry_nodes/utilities/field/evaluate_at_index.html#bpy-types-geometrynodefieldatindex"),
    ("bpy.types.geometrynodefieldaverage*", "modeling/geometry_nodes/utilities/field/field_average.html#bpy-types-geometrynodefieldaverage"),
    ("bpy.types.geometrynodeimagetexture*", "modeling/geometry_nodes/texture/image.html#bpy-types-geometrynodeimagetexture"),
    ("bpy.types.geometrynodeinputtangent*", "modeling/geometry_nodes/curve/read/curve_tangent.html#bpy-types-geometrynodeinputtangent"),
    ("bpy.types.geometrynodejoingeometry*", "modeling/geometry_nodes/geometry/join_geometry.html#bpy-types-geometrynodejoingeometry"),
    ("bpy.types.geometrynodemeshcylinder*", "modeling/geometry_nodes/mesh/primitives/cylinder.html#bpy-types-geometrynodemeshcylinder"),
    ("bpy.types.geometrynodemeshtopoints*", "modeling/geometry_nodes/mesh/operations/mesh_to_points.html#bpy-types-geometrynodemeshtopoints"),
    ("bpy.types.geometrynodemeshtovolume*", "modeling/geometry_nodes/mesh/operations/mesh_to_volume.html#bpy-types-geometrynodemeshtovolume"),
    ("bpy.types.geometrynodemeshuvsphere*", "modeling/geometry_nodes/mesh/primitives/uv_sphere.html#bpy-types-geometrynodemeshuvsphere"),
    ("bpy.types.geometrynoderepeatoutput*", "modeling/geometry_nodes/utilities/repeat_zone.html#bpy-types-geometrynoderepeatoutput"),
    ("bpy.types.geometrynodereversecurve*", "modeling/geometry_nodes/curve/operations/reverse_curve.html#bpy-types-geometrynodereversecurve"),
    ("bpy.types.geometrynodesetcurvetilt*", "modeling/geometry_nodes/curve/write/set_curve_tilt.html#bpy-types-geometrynodesetcurvetilt"),
    ("bpy.types.geometrynodesortelements*", "modeling/geometry_nodes/geometry/operations/sort_elements.html#bpy-types-geometrynodesortelements"),
    ("bpy.types.geometrynodesplinelength*", "modeling/geometry_nodes/curve/read/spline_length.html#bpy-types-geometrynodesplinelength"),
    ("bpy.types.geometrynodetool3dcursor*", "modeling/geometry_nodes/input/scene/3d_cursor.html#bpy-types-geometrynodetool3dcursor"),
    ("bpy.types.geometrynodetree.is_tool*", "interface/controls/nodes/groups.html#bpy-types-geometrynodetree-is-tool"),
    ("bpy.types.geometrynodevolumetomesh*", "modeling/geometry_nodes/volume/operations/volume_to_mesh.html#bpy-types-geometrynodevolumetomesh"),
    ("bpy.types.greasepencilhookmodifier*", "grease_pencil/modifiers/deform/hook.html#bpy-types-greasepencilhookmodifier"),
    ("bpy.types.greasepencillayer.parent*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-parent"),
    ("bpy.types.greasepenciltimemodifier*", "grease_pencil/modifiers/edit/time_offset.html#bpy-types-greasepenciltimemodifier"),
    ("bpy.types.greasepenciltintmodifier*", "grease_pencil/modifiers/color/tint.html#bpy-types-greasepenciltintmodifier"),
    ("bpy.types.image.use_half_precision*", "editors/image/image_settings.html#bpy-types-image-use-half-precision"),
    ("bpy.types.image.use_view_as_render*", "editors/image/image_settings.html#bpy-types-image-use-view-as-render"),
    ("bpy.types.imagepaint.interpolation*", "sculpt_paint/texture_paint/tool_settings/texture_slots.html#bpy-types-imagepaint-interpolation"),
    ("bpy.types.imagetexture.filter_size*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-filter-size"),
    ("bpy.types.imagetexture.filter_type*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-filter-type"),
    ("bpy.types.keyingset.bl_description*", "animation/keyframes/keying_sets.html#bpy-types-keyingset-bl-description"),
    ("bpy.types.linestyle*modifier_noise*", "render/freestyle/view_layer/line_style/modifiers/color/noise.html#bpy-types-linestyle-modifier-noise"),
    ("bpy.types.maintainvolumeconstraint*", "animation/constraints/transform/maintain_volume.html#bpy-types-maintainvolumeconstraint"),
    ("bpy.types.mesh.use_mirror_topology*", "modeling/meshes/tools/tool_settings.html#bpy-types-mesh-use-mirror-topology"),
    ("bpy.types.modifier.use_pin_to_last*", "modeling/modifiers/introduction.html#bpy-types-modifier-use-pin-to-last"),
    ("bpy.types.movieclip.display_aspect*", "editors/clip/display/clip_display.html#bpy-types-movieclip-display-aspect"),
    ("bpy.types.movietrackingcamera.lens*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-lens"),
    ("bpy.types.movietrackingtrack.color*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-color"),
    ("bpy.types.node.warning_propagation*", "interface/controls/nodes/sidebar.html#bpy-types-node-warning-propagation"),
    ("bpy.types.object.hide_probe_sphere*", "render/eevee/object_settings/object_data.html#bpy-types-object-hide-probe-sphere"),
    ("bpy.types.object.hide_probe_volume*", "render/eevee/object_settings/object_data.html#bpy-types-object-hide-probe-volume"),
    ("bpy.types.object.is_shadow_catcher*", "render/cycles/object_settings/object_data.html#bpy-types-object-is-shadow-catcher"),
    ("bpy.types.particleedit.select_mode*", "physics/particles/mode.html#bpy-types-particleedit-select-mode"),
    ("bpy.types.particleinstancemodifier*", "modeling/modifiers/physics/particle_instance.html#bpy-types-particleinstancemodifier"),
    ("bpy.types.preferencesview.language*", "editors/preferences/interface.html#bpy-types-preferencesview-language"),
    ("bpy.types.preferencesview.ui_scale*", "editors/preferences/interface.html#bpy-types-preferencesview-ui-scale"),
    ("bpy.types.rendersettings.hair_type*", "render/eevee/render_settings/curves.html#bpy-types-rendersettings-hair-type"),
    ("bpy.types.rendersettings.tile_size*", "render/cycles/render_settings/performance.html#bpy-types-rendersettings-tile-size"),
    ("bpy.types.rigidbodyobject.friction*", "physics/rigid_body/properties/collisions.html#bpy-types-rigidbodyobject-friction"),
    ("bpy.types.scenedisplay.viewport_aa*", "render/workbench/sampling.html#bpy-types-scenedisplay-viewport-aa"),
    ("bpy.types.sceneeevee.gtao_distance*", "render/layers/passes.html#bpy-types-sceneeevee-gtao-distance"),
    ("bpy.types.sceneeevee.overscan_size*", "render/eevee/render_settings/film.html#bpy-types-sceneeevee-overscan-size"),
    ("bpy.types.selectionset.is_selected*", "animation/armatures/properties/selection_sets.html#bpy-types-selectionset-is-selected"),
    ("bpy.types.sequenceeditor.proxy_dir*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-sequenceeditor-proxy-dir"),
    ("bpy.types.sequencertimelineoverlay*", "editors/video_sequencer/sequencer/display.html#bpy-types-sequencertimelineoverlay"),
    ("bpy.types.shadernodebrightcontrast*", "render/shader_nodes/color/bright_contrast.html#bpy-types-shadernodebrightcontrast"),
    ("bpy.types.shadernodebsdfprincipled*", "render/shader_nodes/shader/principled.html#bpy-types-shadernodebsdfprincipled"),
    ("bpy.types.shadernodebsdfrefraction*", "render/shader_nodes/shader/refraction.html#bpy-types-shadernodebsdfrefraction"),
    ("bpy.types.shadernodeoutputmaterial*", "render/shader_nodes/output/material.html#bpy-types-shadernodeoutputmaterial"),
    ("bpy.types.shadernodetexenvironment*", "render/shader_nodes/textures/environment.html#bpy-types-shadernodetexenvironment"),
    ("bpy.types.softbodysettings.damping*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-damping"),
    ("bpy.types.softbodysettings.plastic*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-plastic"),
    ("bpy.types.spacesequenceeditor.show*", "editors/video_sequencer/preview/header.html#bpy-types-spacesequenceeditor-show"),
    ("bpy.types.spaceuveditor.show_faces*", "editors/uv/overlays.html#bpy-types-spaceuveditor-show-faces"),
    ("bpy.types.spaceuveditor.uv_opacity*", "editors/uv/overlays.html#bpy-types-spaceuveditor-uv-opacity"),
    ("bpy.types.strip.frame_offset_start*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-frame-offset-start"),
    ("bpy.types.strip.show_retiming_keys*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-show-retiming-keys"),
    ("bpy.types.stripproxy.use_overwrite*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-stripproxy-use-overwrite"),
    ("bpy.types.texturenodehuesaturation*", "editors/texture_node/types/color/hue_saturation.html#bpy-types-texturenodehuesaturation"),
    ("bpy.types.texturenodeseparatecolor*", "editors/texture_node/types/color/separate_color.html#bpy-types-texturenodeseparatecolor"),
    ("bpy.types.toolsettings.snap_target*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-snap-target"),
    ("bpy.types.transformcacheconstraint*", "animation/constraints/transform/transform_cache.html#bpy-types-transformcacheconstraint"),
    ("bpy.types.unitsettings.length_unit*", "scene_layout/scene/properties.html#bpy-types-unitsettings-length-unit"),
    ("bpy.types.usersolidlight.direction*", "editors/preferences/lights.html#bpy-types-usersolidlight-direction"),
    ("bpy.types.vertexweighteditmodifier*", "modeling/modifiers/modify/weight_edit.html#bpy-types-vertexweighteditmodifier"),
    ("bpy.types.view3dshading.color_type*", "render/workbench/color.html#bpy-types-view3dshading-color-type"),
    ("bpy.types.viewlayer.world_override*", "render/layers/introduction.html#bpy-types-viewlayer-world-override"),
    ("bpy.types.volumedisplay.slice_axis*", "modeling/volumes/properties.html#bpy-types-volumedisplay-slice-axis"),
    ("bpy.ops.action.markers_make_local*", "animation/markers.html#bpy-ops-action-markers-make-local"),
    ("bpy.ops.anim.channels_clean_empty*", "editors/nla/editing/track.html#bpy-ops-anim-channels-clean-empty"),
    ("bpy.ops.anim.driver_button_remove*", "animation/drivers/usage.html#bpy-ops-anim-driver-button-remove"),
    ("bpy.ops.anim.keyframe_insert_menu*", "animation/keyframes/editing.html#bpy-ops-anim-keyframe-insert-menu"),
    ("bpy.ops.anim.keyingset_button_add*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keyingset-button-add"),
    ("bpy.ops.armature.select_hierarchy*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-hierarchy"),
    ("bpy.ops.armature.switch_direction*", "animation/armatures/bones/editing/switch_direction.html#bpy-ops-armature-switch-direction"),
    ("bpy.ops.clip.apply_solution_scale*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-apply-solution-scale"),
    ("bpy.ops.clip.set_center_principal*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-set-center-principal"),
    ("bpy.ops.clip.setup_tracking_scene*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-setup-tracking-scene"),
    ("bpy.ops.curve.match_texture_space*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-ops-curve-match-texture-space"),
    ("bpy.ops.curves.select_linked_pick*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-linked-pick"),
    ("bpy.ops.font.text_paste_from_file*", "modeling/texts/editing.html#bpy-ops-font-text-paste-from-file"),
    ("bpy.ops.geometry.attribute_remove*", "modeling/geometry_nodes/attributes_reference.html#bpy-ops-geometry-attribute-remove"),
    ("bpy.ops.graph.scale_from_neighbor*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-scale-from-neighbor"),
    ("bpy.ops.grease_pencil.clean_loose*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-clean-loose"),
    ("bpy.ops.grease_pencil.erase_lasso*", "grease_pencil/modes/draw/editing.html#bpy-ops-grease-pencil-erase-lasso"),
    ("bpy.ops.grease_pencil.layer_merge*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-merge"),
    ("bpy.ops.grease_pencil.select_ends*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-ends"),
    ("bpy.ops.grease_pencil.select_less*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-less"),
    ("bpy.ops.grease_pencil.select_more*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-more"),
    ("bpy.ops.grease_pencil.stroke_trim*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-stroke-trim"),
    ("bpy.ops.grease_pencil.trace_image*", "grease_pencil/modes/object/trace_image.html#bpy-ops-grease-pencil-trace-image"),
    ("bpy.ops.mask.feather_weight_clear*", "movie_clip/masking/editing.html#bpy-ops-mask-feather-weight-clear"),
    ("bpy.ops.mask.primitive_circle_add*", "movie_clip/masking/scurve.html#bpy-ops-mask-primitive-circle-add"),
    ("bpy.ops.mask.primitive_square_add*", "movie_clip/masking/scurve.html#bpy-ops-mask-primitive-square-add"),
    ("bpy.ops.mesh.dupli_extrude_cursor*", "modeling/meshes/editing/vertex/extrude_cursor.html#bpy-ops-mesh-dupli-extrude-cursor"),
    ("bpy.ops.mesh.primitive_circle_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-circle-add"),
    ("bpy.ops.mesh.primitive_monkey_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-monkey-add"),
    ("bpy.ops.mesh.select_by_pole_count*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-by-pole-count"),
    ("bpy.ops.mesh.select_face_by_sides*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-face-by-sides"),
    ("bpy.ops.mesh.shortest_path_select*", "modeling/meshes/selecting/linked.html#bpy-ops-mesh-shortest-path-select"),
    ("bpy.ops.mesh.vert_connect_concave*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-vert-connect-concave"),
    ("bpy.ops.nla.duplicate_linked_move*", "editors/nla/editing/strip.html#bpy-ops-nla-duplicate-linked-move"),
    ("bpy.ops.node.shader_script_update*", "render/shader_nodes/script.html#bpy-ops-node-shader-script-update"),
    ("bpy.ops.object.multires_subdivide*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-subdivide"),
    ("bpy.ops.object.shape_key_transfer*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-transfer"),
    ("bpy.ops.object.vertex_group_clean*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-clean"),
    ("bpy.ops.object.vertex_weight_copy*", "modeling/meshes/properties/vertex_groups/vertex_weights.html#bpy-ops-object-vertex-weight-copy"),
    ("bpy.ops.paint.vertex_color_invert*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-invert"),
    ("bpy.ops.paint.vertex_color_levels*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-levels"),
    ("bpy.ops.paint.vertex_color_smooth*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-smooth"),
    ("bpy.ops.paint.weight_sample_group*", "sculpt_paint/weight_paint/editing.html#bpy-ops-paint-weight-sample-group"),
    ("bpy.ops.pointcloud.duplicate_move*", "modeling/point_cloud/editing.html#bpy-ops-pointcloud-duplicate-move"),
    ("bpy.ops.pose.selection_set_assign*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-assign"),
    ("bpy.ops.pose.selection_set_select*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-select"),
    ("bpy.ops.poselib.create_pose_asset*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-create-pose-asset"),
    ("bpy.ops.poselib.modify_pose_asset*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-modify-pose-asset"),
    ("bpy.ops.preferences.addon_disable*", "editors/preferences/addons.html#bpy-ops-preferences-addon-disable"),
    ("bpy.ops.preferences.addon_install*", "advanced/extensions/addons.html#bpy-ops-preferences-addon-install"),
    ("bpy.ops.preferences.autoexec_path*", "editors/preferences/save_load.html#bpy-ops-preferences-autoexec-path"),
    ("bpy.ops.preferences.theme_install*", "editors/preferences/themes.html#bpy-ops-preferences-theme-install"),
    ("bpy.ops.render.play_rendered_anim*", "render/output/animation_player.html#bpy-ops-render-play-rendered-anim"),
    ("bpy.ops.sculpt.set_pivot_position*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-set-pivot-position"),
    ("bpy.ops.sculpt.trim_lasso_gesture*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-trim-lasso-gesture"),
    ("bpy.ops.sculpt_curves.select_ends*", "sculpt_paint/curves_sculpting/introduction.html#bpy-ops-sculpt-curves-select-ends"),
    ("bpy.ops.sculpt_curves.select_grow*", "sculpt_paint/curves_sculpting/introduction.html#bpy-ops-sculpt-curves-select-grow"),
    ("bpy.ops.sequencer.image_strip_add*", "video_editing/edit/montage/strips/image.html#bpy-ops-sequencer-image-strip-add"),
    ("bpy.ops.sequencer.images_separate*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-images-separate"),
    ("bpy.ops.sequencer.movie_strip_add*", "video_editing/edit/montage/strips/movie.html#bpy-ops-sequencer-movie-strip-add"),
    ("bpy.ops.sequencer.reassign_inputs*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-reassign-inputs"),
    ("bpy.ops.sequencer.sound_strip_add*", "video_editing/edit/montage/strips/sound.html#bpy-ops-sequencer-sound-strip-add"),
    ("bpy.ops.spreadsheet.resize_column*", "editors/spreadsheet.html#bpy-ops-spreadsheet-resize-column"),
    ("bpy.ops.ui.remove_override_button*", "files/linked_libraries/library_overrides.html#bpy-ops-ui-remove-override-button"),
    ("bpy.ops.view3d.view_center_camera*", "editors/3dview/navigate/camera_view.html#bpy-ops-view3d-view-center-camera"),
    ("bpy.ops.view3d.zoom_camera_1_to_1*", "editors/3dview/navigate/camera_view.html#bpy-ops-view3d-zoom-camera-1-to-1"),
    ("bpy.types.actionslot.name_display*", "editors/dope_sheet/modes/action.html#bpy-types-actionslot-name-display"),
    ("bpy.types.animvizmotionpaths.type*", "animation/motion_paths.html#bpy-types-animvizmotionpaths-type"),
    ("bpy.types.assetmetadata.copyright*", "editors/asset_browser.html#bpy-types-assetmetadata-copyright"),
    ("bpy.types.brush.cloth_deform_type*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-deform-type"),
    ("bpy.types.brush.cloth_sim_falloff*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-sim-falloff"),
    ("bpy.types.brush.slide_deform_type*", "sculpt_paint/sculpting/brushes/slide_relax.html#bpy-types-brush-slide-deform-type"),
    ("bpy.types.brush.smear_deform_type*", "sculpt_paint/sculpting/brushes/multires_displacement_smear.html#bpy-types-brush-smear-deform-type"),
    ("bpy.types.brush.use_scene_spacing*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-scene-spacing"),
    ("bpy.types.brush.use_smooth_stroke*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-smooth-stroke"),
    ("bpy.types.camera.show_composition*", "render/cameras.html#bpy-types-camera-show-composition"),
    ("bpy.types.colorramp.interpolation*", "interface/controls/templates/color_ramp.html#bpy-types-colorramp-interpolation"),
    ("bpy.types.compositornodealphaover*", "compositing/types/color/mix/alpha_over.html#bpy-types-compositornodealphaover"),
    ("bpy.types.compositornodebokehblur*", "compositing/types/filter/blur/bokeh_blur.html#bpy-types-compositornodebokehblur"),
    ("bpy.types.compositornodecomposite*", "compositing/types/output/composite.html#bpy-types-compositornodecomposite"),
    ("bpy.types.compositornodecornerpin*", "compositing/types/transform/corner_pin.html#bpy-types-compositornodecornerpin"),
    ("bpy.types.compositornodedespeckle*", "compositing/types/filter/despeckle.html#bpy-types-compositornodedespeckle"),
    ("bpy.types.compositornodediffmatte*", "compositing/types/keying/difference_key.html#bpy-types-compositornodediffmatte"),
    ("bpy.types.compositornodeimageinfo*", "compositing/types/input/image_info.html#bpy-types-compositornodeimageinfo"),
    ("bpy.types.compositornodelumamatte*", "compositing/types/keying/luminance_key.html#bpy-types-compositornodelumamatte"),
    ("bpy.types.compositornodemovieclip*", "compositing/types/input/movie_clip.html#bpy-types-compositornodemovieclip"),
    ("bpy.types.compositornodenormalize*", "compositing/types/utilities/normalize.html#bpy-types-compositornodenormalize"),
    ("bpy.types.compositornodeposterize*", "compositing/types/filter/posterize.html#bpy-types-compositornodeposterize"),
    ("bpy.types.compositornodepremulkey*", "compositing/types/color/alpha_convert.html#bpy-types-compositornodepremulkey"),
    ("bpy.types.compositornodescenetime*", "compositing/types/input/scene/scene_time.html#bpy-types-compositornodescenetime"),
    ("bpy.types.compositornodestabilize*", "compositing/types/tracking/stabilize_2d.html#bpy-types-compositornodestabilize"),
    ("bpy.types.compositornodetransform*", "compositing/types/transform/transform.html#bpy-types-compositornodetransform"),
    ("bpy.types.compositornodetranslate*", "compositing/types/transform/translate.html#bpy-types-compositornodetranslate"),
    ("bpy.types.constraint.target_space*", "animation/constraints/interface/common.html#bpy-types-constraint-target-space"),
    ("bpy.types.curve.use_deform_bounds*", "modeling/curves/properties/shape.html#bpy-types-curve-use-deform-bounds"),
    ("bpy.types.editbone.bbone_curveinx*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-curveinx"),
    ("bpy.types.editbone.bbone_curveinz*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-curveinz"),
    ("bpy.types.editbone.bbone_scaleout*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-scaleout"),
    ("bpy.types.editbone.bbone_segments*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-segments"),
    ("bpy.types.freestylelineset.qi_end*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset-qi-end"),
    ("bpy.types.freestylelinestyle.caps*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-caps"),
    ("bpy.types.freestylelinestyle.dash*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-dash"),
    ("bpy.types.freestylemodulesettings*", "render/freestyle/python.html#bpy-types-freestylemodulesettings"),
    ("bpy.types.functionnodebooleanmath*", "modeling/geometry_nodes/utilities/math/boolean_math.html#bpy-types-functionnodebooleanmath"),
    ("bpy.types.functionnodeinputstring*", "modeling/geometry_nodes/input/constant/string.html#bpy-types-functionnodeinputstring"),
    ("bpy.types.functionnodeinputvector*", "modeling/geometry_nodes/input/constant/vector.html#bpy-types-functionnodeinputvector"),
    ("bpy.types.functionnodeintegermath*", "modeling/geometry_nodes/utilities/math/integer_math.html#bpy-types-functionnodeintegermath"),
    ("bpy.types.functionnoderandomvalue*", "modeling/geometry_nodes/utilities/random_value.html#bpy-types-functionnoderandomvalue"),
    ("bpy.types.functionnoderotateeuler*", "modeling/geometry_nodes/utilities/deprecated/rotate_euler.html#bpy-types-functionnoderotateeuler"),
    ("bpy.types.functionnodeslicestring*", "modeling/geometry_nodes/utilities/text/slice_string.html#bpy-types-functionnodeslicestring"),
    ("bpy.types.geometrynodecurvelength*", "modeling/geometry_nodes/curve/read/curve_length.html#bpy-types-geometrynodecurvelength"),
    ("bpy.types.geometrynodecurvespiral*", "modeling/geometry_nodes/curve/primitives/curve_spiral.html#bpy-types-geometrynodecurvespiral"),
    ("bpy.types.geometrynodecurvetomesh*", "modeling/geometry_nodes/curve/operations/curve_to_mesh.html#bpy-types-geometrynodecurvetomesh"),
    ("bpy.types.geometrynodeextrudemesh*", "modeling/geometry_nodes/mesh/operations/extrude_mesh.html#bpy-types-geometrynodeextrudemesh"),
    ("bpy.types.geometrynodefilletcurve*", "modeling/geometry_nodes/curve/operations/fillet_curve.html#bpy-types-geometrynodefilletcurve"),
    ("bpy.types.geometrynodegizmolinear*", "modeling/geometry_nodes/input/gizmo/linear_gizmo.html#bpy-types-geometrynodegizmolinear"),
    ("bpy.types.geometrynodeindexswitch*", "modeling/geometry_nodes/utilities/index_switch.html#bpy-types-geometrynodeindexswitch"),
    ("bpy.types.geometrynodeinputnormal*", "modeling/geometry_nodes/geometry/read/normal.html#bpy-types-geometrynodeinputnormal"),
    ("bpy.types.geometrynodeinputobject*", "modeling/geometry_nodes/input/constant/object.html#bpy-types-geometrynodeinputobject"),
    ("bpy.types.geometrynodeinputradius*", "modeling/geometry_nodes/geometry/read/radius.html#bpy-types-geometrynodeinputradius"),
    ("bpy.types.geometrynodematchstring*", "modeling/geometry_nodes/utilities/text/match_string.html#bpy-types-geometrynodematchstring"),
    ("bpy.types.geometrynodemergelayers*", "modeling/geometry_nodes/grease_pencil/operations/merge_layers.html#bpy-types-geometrynodemergelayers"),
    ("bpy.types.geometrynodemeshboolean*", "modeling/geometry_nodes/mesh/operations/mesh_boolean.html#bpy-types-geometrynodemeshboolean"),
    ("bpy.types.geometrynodemeshtocurve*", "modeling/geometry_nodes/mesh/operations/mesh_to_curve.html#bpy-types-geometrynodemeshtocurve"),
    ("bpy.types.geometrynoderepeatinput*", "modeling/geometry_nodes/utilities/repeat_zone.html#bpy-types-geometrynoderepeatinput"),
    ("bpy.types.geometrynodesamplecurve*", "modeling/geometry_nodes/curve/sample/sample_curve.html#bpy-types-geometrynodesamplecurve"),
    ("bpy.types.geometrynodesampleindex*", "modeling/geometry_nodes/geometry/sample/sample_index.html#bpy-types-geometrynodesampleindex"),
    ("bpy.types.geometrynodesetmaterial*", "modeling/geometry_nodes/material/set_material.html#bpy-types-geometrynodesetmaterial"),
    ("bpy.types.geometrynodesetposition*", "modeling/geometry_nodes/geometry/write/set_position.html#bpy-types-geometrynodesetposition"),
    ("bpy.types.geometrynodetoolfaceset*", "modeling/geometry_nodes/mesh/read/face_set.html#bpy-types-geometrynodetoolfaceset"),
    ("bpy.types.geometrynodetriangulate*", "modeling/geometry_nodes/mesh/operations/triangulate.html#bpy-types-geometrynodetriangulate"),
    ("bpy.types.greasepencilgrid.offset*", "grease_pencil/properties/display.html#bpy-types-greasepencilgrid-offset"),
    ("bpy.types.greasepencillayer.scale*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-scale"),
    ("bpy.types.imagepaint.normal_angle*", "sculpt_paint/brush/falloff.html#bpy-types-imagepaint-normal-angle"),
    ("bpy.types.imagetexture.use_mipmap*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-mipmap"),
    ("bpy.types.keyingsetpath.data_path*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-data-path"),
    ("bpy.types.keyingsets.active_index*", "animation/keyframes/keying_sets.html#bpy-types-keyingsets-active-index"),
    ("bpy.types.keymapitem.key_modifier*", "editors/preferences/keymap.html#bpy-types-keymapitem-key-modifier"),
    ("bpy.types.laplaciandeformmodifier*", "modeling/modifiers/deform/laplacian_deform.html#bpy-types-laplaciandeformmodifier"),
    ("bpy.types.laplaciansmoothmodifier*", "modeling/modifiers/deform/laplacian_smooth.html#bpy-types-laplaciansmoothmodifier"),
    ("bpy.types.layercollection.exclude*", "editors/outliner/interface.html#bpy-types-layercollection-exclude"),
    ("bpy.types.layercollection.holdout*", "editors/outliner/interface.html#bpy-types-layercollection-holdout"),
    ("bpy.types.light.use_shadow_jitter*", "render/eevee/light_settings.html#bpy-types-light-use-shadow-jitter"),
    ("bpy.types.limitdistanceconstraint*", "animation/constraints/transform/limit_distance.html#bpy-types-limitdistanceconstraint"),
    ("bpy.types.limitlocationconstraint*", "animation/constraints/transform/limit_location.html#bpy-types-limitlocationconstraint"),
    ("bpy.types.limitrotationconstraint*", "animation/constraints/transform/limit_rotation.html#bpy-types-limitrotationconstraint"),
    ("bpy.types.material.thickness_mode*", "render/eevee/material_settings.html#bpy-types-material-thickness-mode"),
    ("bpy.types.movieclipproxy.timecode*", "editors/clip/sidebar.html#bpy-types-movieclipproxy-timecode"),
    ("bpy.types.moviesequence.use_proxy*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-moviesequence-use-proxy"),
    ("bpy.types.movietrackingtrack.lock*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-lock"),
    ("bpy.types.movietrackingtrack.name*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-types-movietrackingtrack-name"),
    ("bpy.types.object.hide_probe_plane*", "render/eevee/object_settings/object_data.html#bpy-types-object-hide-probe-plane"),
    ("bpy.types.preferencesexperimental*", "editors/preferences/experimental.html#bpy-types-preferencesexperimental"),
    ("bpy.types.rendersettings.filepath*", "render/output/properties/output.html#bpy-types-rendersettings-filepath"),
    ("bpy.types.rendersettings.fps_base*", "render/output/properties/format.html#bpy-types-rendersettings-fps-base"),
    ("bpy.types.rendersettings.ppm_base*", "render/output/properties/output.html#bpy-types-rendersettings-ppm-base"),
    ("bpy.types.rigidbodyobject.enabled*", "physics/rigid_body/properties/settings.html#bpy-types-rigidbodyobject-enabled"),
    ("bpy.types.scene.use_preview_range*", "editors/timeline.html#bpy-types-scene-use-preview-range"),
    ("bpy.types.sceneeevee.use_overscan*", "render/eevee/render_settings/film.html#bpy-types-sceneeevee-use-overscan"),
    ("bpy.types.scenegpencil.aa_samples*", "render/cycles/render_settings/grease_pencil.html#bpy-types-scenegpencil-aa-samples"),
    ("bpy.types.sequencerpreviewoverlay*", "editors/video_sequencer/preview/display/overlays.html#bpy-types-sequencerpreviewoverlay"),
    ("bpy.types.sequencetimelinechannel*", "editors/video_sequencer/sequencer/channels.html#bpy-types-sequencetimelinechannel"),
    ("bpy.types.shadernodebsdfrayportal*", "render/shader_nodes/shader/ray_portal.html#bpy-types-shadernodebsdfrayportal"),
    ("bpy.types.shadernodeeeveespecular*", "render/shader_nodes/shader/specular_bsdf.html#bpy-types-shadernodeeeveespecular"),
    ("bpy.types.shadernodehuesaturation*", "render/shader_nodes/color/hue_saturation.html#bpy-types-shadernodehuesaturation"),
    ("bpy.types.shadernodeseparatecolor*", "render/shader_nodes/converter/separate_color.html#bpy-types-shadernodeseparatecolor"),
    ("bpy.types.shadernodetexwhitenoise*", "render/shader_nodes/textures/white_noise.html#bpy-types-shadernodetexwhitenoise"),
    ("bpy.types.shadernodevolumescatter*", "render/shader_nodes/shader/volume_scatter.html#bpy-types-shadernodevolumescatter"),
    ("bpy.types.spacegrapheditor.cursor*", "editors/graph_editor/introduction.html#bpy-types-spacegrapheditor-cursor"),
    ("bpy.types.spaceview3d.lock_camera*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-lock-camera"),
    ("bpy.types.spaceview3d.lock_cursor*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-lock-cursor"),
    ("bpy.types.spaceview3d.lock_object*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-lock-object"),
    ("bpy.types.spaceview3d.show_viewer*", "modeling/geometry_nodes/output/viewer.html#bpy-types-spaceview3d-show-viewer"),
    ("bpy.types.stripmodifier.mask_time*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-stripmodifier-mask-time"),
    ("bpy.types.striptransform.rotation*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-striptransform-rotation"),
    ("bpy.types.textcurve.use_fast_edit*", "modeling/texts/properties.html#bpy-types-textcurve-use-fast-edit"),
    ("bpy.types.texturenodecombinecolor*", "editors/texture_node/types/color/combine_color.html#bpy-types-texturenodecombinecolor"),
    ("bpy.types.texturenodetexdistnoise*", "editors/texture_node/types/textures/distorted_noise.html#bpy-types-texturenodetexdistnoise"),
    ("bpy.types.vertexweightmixmodifier*", "modeling/modifiers/modify/weight_mix.html#bpy-types-vertexweightmixmodifier"),
    ("bpy.types.view3dshading.show_xray*", "editors/3dview/display/shading.html#bpy-types-view3dshading-show-xray"),
    ("bpy.types.viewlayer.use_freestyle*", "render/freestyle/view_layer/freestyle.html#bpy-types-viewlayer-use-freestyle"),
    ("bpy.types.viewlayer.use_pass_emit*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-emit"),
    ("bpy.types.viewlayer.use_pass_mist*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-mist"),
    ("bpy.types.volumedisplay.use_slice*", "modeling/volumes/properties.html#bpy-types-volumedisplay-use-slice"),
    ("bpy.types.worldlighting.ao_factor*", "render/cycles/render_settings/light_paths.html#bpy-types-worldlighting-ao-factor"),
    ("bpy.types.worldmistsettings.depth*", "render/eevee/world_settings.html#bpy-types-worldmistsettings-depth"),
    ("bpy.types.worldmistsettings.start*", "render/eevee/world_settings.html#bpy-types-worldmistsettings-start"),
    ("bpy.ops.anim.keying_set_path_add*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keying-set-path-add"),
    ("bpy.ops.anim.paste_driver_button*", "animation/drivers/usage.html#bpy-ops-anim-paste-driver-button"),
    ("bpy.ops.asset.screenshot_preview*", "editors/asset_browser.html#bpy-ops-asset-screenshot-preview"),
    ("bpy.ops.clip.stabilize_2d_select*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-stabilize-2d-select"),
    ("bpy.ops.clip.tracking_object_new*", "movie_clip/tracking/clip/sidebar/track/objects.html#bpy-ops-clip-tracking-object-new"),
    ("bpy.ops.constraint.move_to_index*", "animation/constraints/interface/header.html#bpy-ops-constraint-move-to-index"),
    ("bpy.ops.constraint.remove_target*", "animation/constraints/relationship/armature.html#bpy-ops-constraint-remove-target"),
    ("bpy.ops.curve.select_linked_pick*", "modeling/curves/selecting.html#bpy-ops-curve-select-linked-pick"),
    ("bpy.ops.ed.lib_id_remove_preview*", "editors/asset_browser.html#bpy-ops-ed-lib-id-remove-preview"),
    ("bpy.ops.extensions.repo_sync_all*", "editors/preferences/extensions.html#bpy-ops-extensions-repo-sync-all"),
    ("bpy.ops.font.text_insert_unicode*", "modeling/texts/editing.html#bpy-ops-font-text-insert-unicode"),
    ("bpy.ops.graph.extrapolation_type*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-extrapolation-type"),
    ("bpy.ops.graph.interpolation_type*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-interpolation-type"),
    ("bpy.ops.graph.select_key_handles*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-key-handles"),
    ("bpy.ops.grease_pencil.layer_hide*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-layer-hide"),
    ("bpy.ops.grease_pencil.layer_move*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-move"),
    ("bpy.ops.grease_pencil.select_all*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select-all"),
    ("bpy.ops.image.match_movie_length*", "editors/image/image_settings.html#bpy-ops-image-match-movie-length"),
    ("bpy.ops.mesh.customdata_skin_add*", "modeling/meshes/properties/custom_data.html#bpy-ops-mesh-customdata-skin-add"),
    ("bpy.ops.mesh.dissolve_degenerate*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-dissolve-degenerate"),
    ("bpy.ops.mesh.face_split_by_edges*", "modeling/meshes/editing/face/weld_edges_faces.html#bpy-ops-mesh-face-split-by-edges"),
    ("bpy.ops.mesh.mark_freestyle_edge*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-mesh-mark-freestyle-edge"),
    ("bpy.ops.mesh.mark_freestyle_face*", "modeling/meshes/editing/face/face_data.html#bpy-ops-mesh-mark-freestyle-face"),
    ("bpy.ops.mesh.primitive_plane_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-plane-add"),
    ("bpy.ops.mesh.primitive_torus_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-torus-add"),
    ("bpy.ops.mesh.select_by_attribute*", "modeling/meshes/selecting/by_attribute.html#bpy-ops-mesh-select-by-attribute"),
    ("bpy.ops.mesh.select_non_manifold*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-non-manifold"),
    ("bpy.ops.nla.selected_objects_add*", "editors/nla/introduction.html#bpy-ops-nla-selected-objects-add"),
    ("bpy.ops.node.viewer_shortcut_get*", "compositing/types/output/viewer.html#bpy-ops-node-viewer-shortcut-get"),
    ("bpy.ops.node.viewer_shortcut_set*", "compositing/types/output/viewer.html#bpy-ops-node-viewer-shortcut-set"),
    ("bpy.ops.object.constraints_clear*", "scene_layout/object/editing/constraints.html#bpy-ops-object-constraints-clear"),
    ("bpy.ops.object.grease_pencil_add*", "grease_pencil/primitives.html#bpy-ops-object-grease-pencil-add"),
    ("bpy.ops.object.material_slot_add*", "render/materials/assignment.html#bpy-ops-object-material-slot-add"),
    ("bpy.ops.object.quadriflow_remesh*", "modeling/meshes/retopology.html#bpy-ops-object-quadriflow-remesh"),
    ("bpy.ops.object.shade_auto_smooth*", "scene_layout/object/editing/shading.html#bpy-ops-object-shade-auto-smooth"),
    ("bpy.ops.object.vertex_group_copy*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-copy"),
    ("bpy.ops.object.vertex_group_lock*", "sculpt_paint/weight_paint/editing.html#bpy-ops-object-vertex-group-lock"),
    ("bpy.ops.object.vertex_group_move*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-move"),
    ("bpy.ops.object.vertex_group_sort*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-sort"),
    ("bpy.ops.object.vertex_parent_set*", "modeling/meshes/editing/vertex/make_vertex_parent.html#bpy-ops-object-vertex-parent-set"),
    ("bpy.ops.outliner.collection_hide*", "editors/outliner/editing.html#bpy-ops-outliner-collection-hide"),
    ("bpy.ops.outliner.collection_show*", "editors/outliner/editing.html#bpy-ops-outliner-collection-show"),
    ("bpy.ops.outliner.expanded_toggle*", "editors/outliner/editing.html#bpy-ops-outliner-expanded-toggle"),
    ("bpy.ops.paint.mask_lasso_gesture*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-paint-mask-lasso-gesture"),
    ("bpy.ops.paint.vert_select_linked*", "sculpt_paint/selection_visibility.html#bpy-ops-paint-vert-select-linked"),
    ("bpy.ops.pointcloud.attribute_set*", "modeling/point_cloud/editing.html#bpy-ops-pointcloud-attribute-set"),
    ("bpy.ops.pointcloud.select_random*", "modeling/point_cloud/selecting.html#bpy-ops-pointcloud-select-random"),
    ("bpy.ops.pose.selection_set_paste*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-paste"),
    ("bpy.ops.poselib.apply_pose_asset*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-apply-pose-asset"),
    ("bpy.ops.poselib.blend_pose_asset*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-blend-pose-asset"),
    ("bpy.ops.preferences.addon_enable*", "editors/preferences/addons.html#bpy-ops-preferences-addon-enable"),
    ("bpy.ops.rigidbody.mass_calculate*", "scene_layout/object/editing/rigid_body.html#bpy-ops-rigidbody-mass-calculate"),
    ("bpy.ops.scene.view_layer_add_aov*", "render/layers/passes.html#bpy-ops-scene-view-layer-add-aov"),
    ("bpy.ops.screen.spacedata_cleanup*", "advanced/operators.html#bpy-ops-screen-spacedata-cleanup"),
    ("bpy.ops.sculpt.detail_flood_fill*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-ops-sculpt-detail-flood-fill"),
    ("bpy.ops.sequencer.duplicate_move*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-duplicate-move"),
    ("bpy.ops.sequencer.enable_proxies*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-ops-sequencer-enable-proxies"),
    ("bpy.ops.sequencer.retiming_reset*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-reset"),
    ("bpy.ops.sequencer.select_grouped*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-grouped"),
    ("bpy.ops.sequencer.select_handles*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-handles"),
    ("bpy.ops.ui.copy_as_driver_button*", "animation/drivers/usage.html#bpy-ops-ui-copy-as-driver-button"),
    ("bpy.ops.ui.copy_data_path_button*", "interface/controls/buttons/menus.html#bpy-ops-ui-copy-data-path-button"),
    ("bpy.ops.uv.average_islands_scale*", "modeling/meshes/uv/editing.html#bpy-ops-uv-average-islands-scale"),
    ("bpy.ops.wm.collection_export_all*", "interface/window_system/topbar.html#bpy-ops-wm-collection-export-all"),
    ("bpy.ops.wm.read_factory_settings*", "interface/window_system/topbar.html#bpy-ops-wm-read-factory-settings"),
    ("bpy.ops.wm.read_factory_userpref*", "editors/preferences/introduction.html#bpy-ops-wm-read-factory-userpref"),
    ("bpy.types.action.use_frame_range*", "animation/actions.html#bpy-types-action-use-frame-range"),
    ("bpy.types.armature.axes_position*", "animation/armatures/properties/display.html#bpy-types-armature-axes-position"),
    ("bpy.types.armature.pose_position*", "animation/armatures/properties/introduction.html#bpy-types-armature-pose-position"),
    ("bpy.types.armature.rigify_colors*", "addons/rigging/rigify/metarigs.html#bpy-types-armature-rigify-colors"),
    ("bpy.types.bakesettings.use_clear*", "render/cycles/baking.html#bpy-types-bakesettings-use-clear"),
    ("bpy.types.bakesettings.view_from*", "render/cycles/baking.html#bpy-types-bakesettings-view-from"),
    ("bpy.types.bone.envelope_distance*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-envelope-distance"),
    ("bpy.types.bonecollection.is_solo*", "animation/armatures/properties/bone_collections.html#bpy-types-bonecollection-is-solo"),
    ("bpy.types.brightcontrastmodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-brightcontrastmodifier"),
    ("bpy.types.brush.cursor_color_add*", "sculpt_paint/brush/cursor.html#bpy-types-brush-cursor-color-add"),
    ("bpy.types.brush.pose_deform_type*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-pose-deform-type"),
    ("bpy.types.brush.pose_ik_segments*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-pose-ik-segments"),
    ("bpy.types.brush.pose_origin_type*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-pose-origin-type"),
    ("bpy.types.brush.use_edge_to_edge*", "sculpt_paint/brush/stroke.html#bpy-types-brush-use-edge-to-edge"),
    ("bpy.types.brushtextureslot.angle*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot-angle"),
    ("bpy.types.camerasolverconstraint*", "animation/constraints/motion_tracking/camera_solver.html#bpy-types-camerasolverconstraint"),
    ("bpy.types.clothcollisionsettings*", "physics/cloth/settings/collisions.html#bpy-types-clothcollisionsettings"),
    ("bpy.types.collection.hide_render*", "editors/outliner/interface.html#bpy-types-collection-hide-render"),
    ("bpy.types.collection.hide_select*", "editors/outliner/interface.html#bpy-types-collection-hide-select"),
    ("bpy.types.colorrampelement.color*", "interface/controls/templates/color_ramp.html#bpy-types-colorrampelement-color"),
    ("bpy.types.compositornodecurvergb*", "compositing/types/color/adjust/rgb_curves.html#bpy-types-compositornodecurvergb"),
    ("bpy.types.compositornodedisplace*", "compositing/types/transform/displace.html#bpy-types-compositornodedisplace"),
    ("bpy.types.compositornodeexposure*", "compositing/types/color/adjust/exposure.html#bpy-types-compositornodeexposure"),
    ("bpy.types.compositornodekuwahara*", "compositing/types/filter/kuwahara.html#bpy-types-compositornodekuwahara"),
    ("bpy.types.compositornodelensdist*", "compositing/types/transform/lens_distortion.html#bpy-types-compositornodelensdist"),
    ("bpy.types.compositornodepixelate*", "compositing/types/filter/pixelate.html#bpy-types-compositornodepixelate"),
    ("bpy.types.compositornodesetalpha*", "compositing/types/color/set_alpha.html#bpy-types-compositornodesetalpha"),
    ("bpy.types.compositornodesunbeams*", "compositing/types/filter/sun_beams.html#bpy-types-compositornodesunbeams"),
    ("bpy.types.compositornodetrackpos*", "compositing/types/tracking/track_position.html#bpy-types-compositornodetrackpos"),
    ("bpy.types.compositornodezcombine*", "compositing/types/color/mix/z_combine.html#bpy-types-compositornodezcombine"),
    ("bpy.types.constraint.owner_space*", "animation/constraints/interface/common.html#bpy-types-constraint-owner-space"),
    ("bpy.types.copylocationconstraint*", "animation/constraints/transform/copy_location.html#bpy-types-copylocationconstraint"),
    ("bpy.types.copyrotationconstraint*", "animation/constraints/transform/copy_rotation.html#bpy-types-copyrotationconstraint"),
    ("bpy.types.curve.bevel_factor_end*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-factor-end"),
    ("bpy.types.curve.bevel_resolution*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-resolution"),
    ("bpy.types.cyclesmaterialsettings*", "render/cycles/material_settings.html#bpy-types-cyclesmaterialsettings"),
    ("bpy.types.dopesheet.show_summary*", "editors/dope_sheet/navigating.html#bpy-types-dopesheet-show-summary"),
    ("bpy.types.editbone.bbone_easeout*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-easeout"),
    ("bpy.types.editbone.bbone_rollout*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-rollout"),
    ("bpy.types.editbone.bbone_scalein*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-scalein"),
    ("bpy.types.editbone.inherit_scale*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-inherit-scale"),
    ("bpy.types.ffmpegsettings.gopsize*", "render/output/properties/output.html#bpy-types-ffmpegsettings-gopsize"),
    ("bpy.types.ffmpegsettings.maxrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-maxrate"),
    ("bpy.types.ffmpegsettings.minrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-minrate"),
    ("bpy.types.ffmpegsettings.muxrate*", "render/output/properties/output.html#bpy-types-ffmpegsettings-muxrate"),
    ("bpy.types.fieldsettings.strength*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-strength"),
    ("bpy.types.freestylelinestyle.gap*", "render/freestyle/view_layer/line_style/strokes.html#bpy-types-freestylelinestyle-gap"),
    ("bpy.types.freestylesettings.mode*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings-mode"),
    ("bpy.types.functionnodefloattoint*", "modeling/geometry_nodes/utilities/math/float_to_integer.html#bpy-types-functionnodefloattoint"),
    ("bpy.types.functionnodeinputcolor*", "modeling/geometry_nodes/input/constant/color.html#bpy-types-functionnodeinputcolor"),
    ("bpy.types.geometrynodeconvexhull*", "modeling/geometry_nodes/geometry/operations/convex_hull.html#bpy-types-geometrynodeconvexhull"),
    ("bpy.types.geometrynodeimporttext*", "modeling/geometry_nodes/input/import/txt.html#bpy-types-geometrynodeimporttext"),
    ("bpy.types.geometrynodeinputimage*", "modeling/geometry_nodes/input/constant/image.html#bpy-types-geometrynodeinputimage"),
    ("bpy.types.geometrynodeinputindex*", "modeling/geometry_nodes/geometry/read/input_index.html#bpy-types-geometrynodeinputindex"),
    ("bpy.types.geometrynodeisviewport*", "modeling/geometry_nodes/input/scene/is_viewport.html#bpy-types-geometrynodeisviewport"),
    ("bpy.types.geometrynodemenuswitch*", "modeling/geometry_nodes/utilities/menu_switch.html#bpy-types-geometrynodemenuswitch"),
    ("bpy.types.geometrynodemeshcircle*", "modeling/geometry_nodes/mesh/primitives/mesh_circle.html#bpy-types-geometrynodemeshcircle"),
    ("bpy.types.geometrynodeobjectinfo*", "modeling/geometry_nodes/input/scene/object_info.html#bpy-types-geometrynodeobjectinfo"),
    ("bpy.types.geometrynodeselfobject*", "modeling/geometry_nodes/input/scene/self_object.html#bpy-types-geometrynodeselfobject"),
    ("bpy.types.geometrynodesplitedges*", "modeling/geometry_nodes/mesh/operations/split_edges.html#bpy-types-geometrynodesplitedges"),
    ("bpy.types.geometrynodestringjoin*", "modeling/geometry_nodes/utilities/text/join_strings.html#bpy-types-geometrynodestringjoin"),
    ("bpy.types.geometrynodevolumecube*", "modeling/geometry_nodes/volume/primitives/volume_cube.html#bpy-types-geometrynodevolumecube"),
    ("bpy.types.greasepencilgrid.color*", "grease_pencil/properties/display.html#bpy-types-greasepencilgrid-color"),
    ("bpy.types.greasepencilgrid.lines*", "grease_pencil/properties/display.html#bpy-types-greasepencilgrid-lines"),
    ("bpy.types.greasepencilgrid.scale*", "grease_pencil/properties/display.html#bpy-types-greasepencilgrid-scale"),
    ("bpy.types.greasepencillayer.hide*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-hide"),
    ("bpy.types.greasepencillayer.lock*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer-lock"),
    ("bpy.types.imagepaint.use_occlude*", "sculpt_paint/texture_paint/tool_settings/options.html#bpy-types-imagepaint-use-occlude"),
    ("bpy.types.imagesequence.use_flip*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-imagesequence-use-flip"),
    ("bpy.types.imagetexture.extension*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-extension"),
    ("bpy.types.imagetexture.use_alpha*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture-use-alpha"),
    ("bpy.types.keyframe.interpolation*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-interpolation"),
    ("bpy.types.material.diffuse_color*", "render/materials/settings.html#bpy-types-material-diffuse-color"),
    ("bpy.types.material.line_priority*", "render/freestyle/material.html#bpy-types-material-line-priority"),
    ("bpy.types.mesh.texspace_location*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-types-mesh-texspace-location"),
    ("bpy.types.mesh.use_auto_texspace*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-types-mesh-use-auto-texspace"),
    ("bpy.types.modifier.show_viewport*", "modeling/modifiers/introduction.html#bpy-types-modifier-show-viewport"),
    ("bpy.types.motionpath.frame_start*", "animation/motion_paths.html#bpy-types-motionpath-frame-start"),
    ("bpy.types.movieclipsequence.clip*", "video_editing/edit/montage/strips/clip.html#bpy-types-movieclipsequence-clip"),
    ("bpy.types.nodesmodifierdatablock*", "modeling/geometry_nodes/baking.html#bpy-types-nodesmodifierdatablock"),
    ("bpy.types.object.parent_vertices*", "scene_layout/object/properties/relations.html#bpy-types-object-parent-vertices"),
    ("bpy.types.object.visible_diffuse*", "render/cycles/object_settings/object_data.html#bpy-types-object-visible-diffuse"),
    ("bpy.types.objectsolverconstraint*", "animation/constraints/motion_tracking/object_solver.html#bpy-types-objectsolverconstraint"),
    ("bpy.types.particlesystemmodifier*", "physics/particles/index.html#bpy-types-particlesystemmodifier"),
    ("bpy.types.rendersettings.threads*", "render/cycles/render_settings/performance.html#bpy-types-rendersettings-threads"),
    ("bpy.types.scenedisplay.render_aa*", "render/workbench/sampling.html#bpy-types-scenedisplay-render-aa"),
    ("bpy.types.sceneeevee.motion_blur*", "render/eevee/render_settings/motion_blur.html#bpy-types-sceneeevee-motion-blur"),
    ("bpy.types.sceneeevee.taa_samples*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-taa-samples"),
    ("bpy.types.sceneeevee.use_fast_gi*", "render/eevee/render_settings/raytracing.html#bpy-types-sceneeevee-use-fast-gi"),
    ("bpy.types.sceneeevee.use_shadows*", "render/eevee/render_settings/sampling.html#bpy-types-sceneeevee-use-shadows"),
    ("bpy.types.sculpt.radial_symmetry*", "sculpt_paint/sculpting/tool_settings/symmetry.html#bpy-types-sculpt-radial-symmetry"),
    ("bpy.types.shadernodebsdfmetallic*", "render/shader_nodes/shader/metallic.html#bpy-types-shadernodebsdfmetallic"),
    ("bpy.types.shadernodecombinecolor*", "render/shader_nodes/converter/combine_color.html#bpy-types-shadernodecombinecolor"),
    ("bpy.types.shadernodedisplacement*", "render/shader_nodes/vector/displacement.html#bpy-types-shadernodedisplacement"),
    ("bpy.types.shadernodelightfalloff*", "render/shader_nodes/color/light_falloff.html#bpy-types-shadernodelightfalloff"),
    ("bpy.types.shadernodeparticleinfo*", "render/shader_nodes/input/particle_info.html#bpy-types-shadernodeparticleinfo"),
    ("bpy.types.shadernodevectorrotate*", "render/shader_nodes/vector/vector_rotate.html#bpy-types-shadernodevectorrotate"),
    ("bpy.types.shapekey.interpolation*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-interpolation"),
    ("bpy.types.softbodysettings.choke*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-choke"),
    ("bpy.types.softbodysettings.fuzzy*", "physics/soft_body/settings/solver.html#bpy-types-softbodysettings-fuzzy"),
    ("bpy.types.softbodysettings.shear*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-shear"),
    ("bpy.types.softbodysettings.speed*", "physics/soft_body/settings/simulation.html#bpy-types-softbodysettings-speed"),
    ("bpy.types.sound.use_memory_cache*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-sound-use-memory-cache"),
    ("bpy.types.soundequalizermodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-soundequalizermodifier"),
    ("bpy.types.spaceclipeditor.cursor*", "editors/clip/sidebar.html#bpy-types-spaceclipeditor-cursor"),
    ("bpy.types.spaceview3d.clip_start*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-clip-start"),
    ("bpy.types.spaceview3d.show_gizmo*", "editors/3dview/display/gizmo.html#bpy-types-spaceview3d-show-gizmo"),
    ("bpy.types.strip.frame_offset_end*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-frame-offset-end"),
    ("bpy.types.textcurve.follow_curve*", "modeling/texts/properties.html#bpy-types-textcurve-follow-curve"),
    ("bpy.types.texturenodecoordinates*", "editors/texture_node/types/input/coordinates.html#bpy-types-texturenodecoordinates"),
    ("bpy.types.texturenodetexmusgrave*", "editors/texture_node/types/textures/musgrave.html#bpy-types-texturenodetexmusgrave"),
    ("bpy.types.unitsettings.mass_unit*", "scene_layout/scene/properties.html#bpy-types-unitsettings-mass-unit"),
    ("bpy.types.unitsettings.time_unit*", "scene_layout/scene/properties.html#bpy-types-unitsettings-time-unit"),
    ("bpy.types.volumedisplacemodifier*", "modeling/modifiers/deform/volume_displace.html#bpy-types-volumedisplacemodifier"),
    ("bpy.types.volumerender.precision*", "modeling/volumes/properties.html#bpy-types-volumerender-precision"),
    ("bpy.types.volumerender.step_size*", "modeling/volumes/properties.html#bpy-types-volumerender-step-size"),
    ("bpy.types.weightednormalmodifier*", "modeling/modifiers/normals/weighted_normal.html#bpy-types-weightednormalmodifier"),
    ("bpy.types.world.probe_resolution*", "render/eevee/world_settings.html#bpy-types-world-probe-resolution"),
    ("bpy.types.worldlighting.distance*", "render/cycles/render_settings/light_paths.html#bpy-types-worldlighting-distance"),
    ("bpy.ops.anim.copy_driver_button*", "animation/drivers/usage.html#bpy-ops-anim-copy-driver-button"),
    ("bpy.ops.anim.driver_button_edit*", "animation/drivers/usage.html#bpy-ops-anim-driver-button-edit"),
    ("bpy.ops.armature.autoside_names*", "animation/armatures/bones/editing/naming.html#bpy-ops-armature-autoside-names"),
    ("bpy.ops.armature.calculate_roll*", "animation/armatures/bones/editing/bone_roll.html#bpy-ops-armature-calculate-roll"),
    ("bpy.ops.armature.duplicate_move*", "animation/armatures/bones/editing/duplicate.html#bpy-ops-armature-duplicate-move"),
    ("bpy.ops.armature.extrude_forked*", "animation/armatures/bones/editing/extrude.html#bpy-ops-armature-extrude-forked"),
    ("bpy.ops.armature.select_similar*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-similar"),
    ("bpy.ops.clip.create_plane_track*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-create-plane-track"),
    ("bpy.ops.collection.exporter_add*", "scene_layout/collections/collections.html#bpy-ops-collection-exporter-add"),
    ("bpy.ops.curve.spline_weight_set*", "modeling/curves/editing/other.html#bpy-ops-curve-spline-weight-set"),
    ("bpy.ops.curves.switch_direction*", "modeling/curves_new/editing/segments.html#bpy-ops-curves-switch-direction"),
    ("bpy.ops.file.external_operation*", "editors/file_browser.html#bpy-ops-file-external-operation"),
    ("bpy.ops.graph.blend_to_neighbor*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-blend-to-neighbor"),
    ("bpy.ops.graph.snap_cursor_value*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-snap-cursor-value"),
    ("bpy.ops.grease_pencil.erase_box*", "grease_pencil/modes/draw/editing.html#bpy-ops-grease-pencil-erase-box"),
    ("bpy.ops.grease_pencil.layer_add*", "grease_pencil/properties/layers.html#bpy-ops-grease-pencil-layer-add"),
    ("bpy.ops.grease_pencil.reproject*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-reproject"),
    ("bpy.ops.grease_pencil.reset_uvs*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-reset-uvs"),
    ("bpy.ops.image.rotate_orthogonal*", "editors/image/editing.html#bpy-ops-image-rotate-orthogonal"),
    ("bpy.ops.image.save_all_modified*", "editors/image/editing.html#bpy-ops-image-save-all-modified"),
    ("bpy.ops.marker.make_links_scene*", "animation/markers.html#bpy-ops-marker-make-links-scene"),
    ("bpy.ops.marker.select_leftright*", "animation/markers.html#bpy-ops-marker-select-leftright"),
    ("bpy.ops.mesh.extrude_edges_move*", "modeling/meshes/editing/edge/extrude_edges.html#bpy-ops-mesh-extrude-edges-move"),
    ("bpy.ops.mesh.extrude_faces_move*", "modeling/meshes/editing/face/extrude_individual_faces.html#bpy-ops-mesh-extrude-faces-move"),
    ("bpy.ops.mesh.faces_shade_smooth*", "modeling/meshes/editing/face/shading.html#bpy-ops-mesh-faces-shade-smooth"),
    ("bpy.ops.mesh.paint_mask_extract*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-mesh-paint-mask-extract"),
    ("bpy.ops.mesh.primitive_cone_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-cone-add"),
    ("bpy.ops.mesh.primitive_cube_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-cube-add"),
    ("bpy.ops.mesh.primitive_grid_add*", "modeling/meshes/primitives.html#bpy-ops-mesh-primitive-grid-add"),
    ("bpy.ops.mesh.shortest_path_pick*", "modeling/meshes/selecting/linked.html#bpy-ops-mesh-shortest-path-pick"),
    ("bpy.ops.mesh.subdivide_edgering*", "modeling/meshes/editing/edge/subdivide_edge_ring.html#bpy-ops-mesh-subdivide-edgering"),
    ("bpy.ops.node.hide_socket_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-hide-socket-toggle"),
    ("bpy.ops.node.interface_item_new*", "interface/controls/nodes/groups.html#bpy-ops-node-interface-item-new"),
    ("bpy.ops.object.attribute_remove*", "modeling/geometry_nodes/attributes_reference.html#bpy-ops-object-attribute-remove"),
    ("bpy.ops.object.constraints_copy*", "scene_layout/object/editing/constraints.html#bpy-ops-object-constraints-copy"),
    ("bpy.ops.object.make_links_scene*", "scene_layout/object/editing/link_transfer/link_data.html#bpy-ops-object-make-links-scene"),
    ("bpy.ops.object.make_single_user*", "scene_layout/object/editing/relations/make_single_user.html#bpy-ops-object-make-single-user"),
    ("bpy.ops.object.multires_reshape*", "modeling/modifiers/generate/multiresolution.html#bpy-ops-object-multires-reshape"),
    ("bpy.ops.object.select_hierarchy*", "scene_layout/object/selecting.html#bpy-ops-object-select-hierarchy"),
    ("bpy.ops.object.shape_key_mirror*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-mirror"),
    ("bpy.ops.object.shape_key_remove*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-remove"),
    ("bpy.ops.object.shape_key_retime*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-retime"),
    ("bpy.ops.object.vertex_group_add*", "modeling/meshes/properties/vertex_groups/vertex_groups.html#bpy-ops-object-vertex-group-add"),
    ("bpy.ops.outliner.collection_new*", "editors/outliner/editing.html#bpy-ops-outliner-collection-new"),
    ("bpy.ops.outliner.orphans_manage*", "interface/window_system/topbar.html#bpy-ops-outliner-orphans-manage"),
    ("bpy.ops.outliner.show_hierarchy*", "editors/outliner/editing.html#bpy-ops-outliner-show-hierarchy"),
    ("bpy.ops.outliner.show_one_level*", "editors/outliner/editing.html#bpy-ops-outliner-show-one-level"),
    ("bpy.ops.paint.brush_colors_flip*", "sculpt_paint/brush/brush_settings.html#bpy-ops-paint-brush-colors-flip"),
    ("bpy.ops.paint.vertex_color_dirt*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-dirt"),
    ("bpy.ops.paint.visibility_filter*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-paint-visibility-filter"),
    ("bpy.ops.paint.visibility_invert*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-paint-visibility-invert"),
    ("bpy.ops.paint.weight_from_bones*", "sculpt_paint/weight_paint/editing.html#bpy-ops-paint-weight-from-bones"),
    ("bpy.ops.paintcurve.delete_point*", "sculpt_paint/brush/stroke.html#bpy-ops-paintcurve-delete-point"),
    ("bpy.ops.pose.paths_range_update*", "animation/motion_paths.html#bpy-ops-pose-paths-range-update"),
    ("bpy.ops.pose.selection_set_copy*", "animation/armatures/properties/selection_sets.html#bpy-ops-pose-selection-set-copy"),
    ("bpy.ops.preferences.studiolight*", "editors/preferences/lights.html#bpy-ops-preferences-studiolight"),
    ("bpy.ops.scene.view_layer_remove*", "render/layers/introduction.html#bpy-ops-scene-view-layer-remove"),
    ("bpy.ops.screen.screen_full_area*", "interface/window_system/areas.html#bpy-ops-screen-screen-full-area"),
    ("bpy.ops.sculpt.face_sets_create*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-ops-sculpt-face-sets-create"),
    ("bpy.ops.sculpt.mask_from_cavity*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-from-cavity"),
    ("bpy.ops.sculpt.trim_box_gesture*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-trim-box-gesture"),
    ("bpy.ops.sequencer.rebuild_proxy*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-ops-sequencer-rebuild-proxy"),
    ("bpy.ops.sequencer.retiming_show*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-retiming-show"),
    ("bpy.ops.sequencer.select_linked*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-linked"),
    ("bpy.ops.transform.rotate_normal*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-transform-rotate-normal"),
    ("bpy.ops.transform.shrink_fatten*", "modeling/meshes/editing/mesh/transform/shrink-fatten.html#bpy-ops-transform-shrink-fatten"),
    ("bpy.ops.transform.vertex_random*", "modeling/meshes/editing/mesh/transform/randomize.html#bpy-ops-transform-vertex-random"),
    ("bpy.ops.ui.reset_default_button*", "interface/controls/buttons/menus.html#bpy-ops-ui-reset-default-button"),
    ("bpy.ops.uv.shortest_path_select*", "editors/uv/selecting.html#bpy-ops-uv-shortest-path-select"),
    ("bpy.ops.view3d.object_as_camera*", "editors/3dview/navigate/camera_view.html#bpy-ops-view3d-object-as-camera"),
    ("bpy.ops.wm.keyconfig_preset_add*", "editors/preferences/keymap.html#bpy-ops-wm-keyconfig-preset-add"),
    ("bpy.ops.wm.operator_cheat_sheet*", "advanced/operators.html#bpy-ops-wm-operator-cheat-sheet"),
    ("bpy.ops.wm.previews_batch_clear*", "files/blend/previews.html#bpy-ops-wm-previews-batch-clear"),
    ("bpy.ops.wm.recover_last_session*", "files/blend/open_save.html#bpy-ops-wm-recover-last-session"),
    ("bpy.ops.wm.toolbar_fallback_pie*", "interface/tool_system.html#bpy-ops-wm-toolbar-fallback-pie"),
    ("bpy.types.armature.display_type*", "animation/armatures/properties/display.html#bpy-types-armature-display-type"),
    ("bpy.types.armature.use_mirror_x*", "animation/armatures/bones/tools/tool_settings.html#bpy-types-armature-use-mirror-x"),
    ("bpy.types.assetmetadata.license*", "editors/asset_browser.html#bpy-types-assetmetadata-license"),
    ("bpy.types.bakesettings.normal_b*", "render/cycles/baking.html#bpy-types-bakesettings-normal-b"),
    ("bpy.types.bakesettings.normal_g*", "render/cycles/baking.html#bpy-types-bakesettings-normal-g"),
    ("bpy.types.bakesettings.normal_r*", "render/cycles/baking.html#bpy-types-bakesettings-normal-r"),
    ("bpy.types.bakesettings.use_cage*", "render/cycles/baking.html#bpy-types-bakesettings-use-cage"),
    ("bpy.types.brush.boundary_offset*", "sculpt_paint/sculpting/brushes/boundary.html#bpy-types-brush-boundary-offset"),
    ("bpy.types.brush.cloth_sim_limit*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-sim-limit"),
    ("bpy.types.brush.secondary_color*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-secondary-color"),
    ("bpy.types.brush.wet_persistence*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-wet-persistence"),
    ("bpy.types.camerabackgroundimage*", "render/cameras.html#bpy-types-camerabackgroundimage"),
    ("bpy.types.clothsettings.quality*", "physics/cloth/settings/index.html#bpy-types-clothsettings-quality"),
    ("bpy.types.compositornodeboxmask*", "compositing/types/mask/box_mask.html#bpy-types-compositornodeboxmask"),
    ("bpy.types.compositornodedefocus*", "compositing/types/filter/blur/defocus.html#bpy-types-compositornodedefocus"),
    ("bpy.types.compositornodedenoise*", "compositing/types/filter/denoise.html#bpy-types-compositornodedenoise"),
    ("bpy.types.compositornodeinpaint*", "compositing/types/filter/inpaint.html#bpy-types-compositornodeinpaint"),
    ("bpy.types.compositornodergbtobw*", "compositing/types/color/rgb_to_bw.html#bpy-types-compositornodergbtobw"),
    ("bpy.types.compositornoderlayers*", "compositing/types/input/scene/render_layers.html#bpy-types-compositornoderlayers"),
    ("bpy.types.compositornodetexture*", "compositing/types/input/texture.html#bpy-types-compositornodetexture"),
    ("bpy.types.compositornodetonemap*", "compositing/types/color/adjust/tone_map.html#bpy-types-compositornodetonemap"),
    ("bpy.types.compositornodevecblur*", "compositing/types/filter/blur/vector_blur.html#bpy-types-compositornodevecblur"),
    ("bpy.types.curve.use_fill_deform*", "modeling/curves/properties/shape.html#bpy-types-curve-use-fill-deform"),
    ("bpy.types.curve.use_path_follow*", "modeling/curves/properties/path_animation.html#bpy-types-curve-use-path-follow"),
    ("bpy.types.curves.surface_uv_map*", "modeling/curves_new/properties.html#bpy-types-curves-surface-uv-map"),
    ("bpy.types.dampedtrackconstraint*", "animation/constraints/tracking/damped_track.html#bpy-types-dampedtrackconstraint"),
    ("bpy.types.distortednoisetexture*", "render/materials/legacy_textures/types/distorted_noise.html#bpy-types-distortednoisetexture"),
    ("bpy.types.dopesheet.filter_text*", "editors/graph_editor/channels/introduction.html#bpy-types-dopesheet-filter-text"),
    ("bpy.types.editbone.bbone_easein*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-easein"),
    ("bpy.types.editbone.bbone_rollin*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-rollin"),
    ("bpy.types.fcurve.auto_smoothing*", "editors/graph_editor/fcurves/properties.html#bpy-types-fcurve-auto-smoothing"),
    ("bpy.types.ffmpegsettings.format*", "render/output/properties/output.html#bpy-types-ffmpegsettings-format"),
    ("bpy.types.fluideffectorsettings*", "physics/fluid/type/effector.html#bpy-types-fluideffectorsettings"),
    ("bpy.types.followtrackconstraint*", "animation/constraints/motion_tracking/follow_track.html#bpy-types-followtrackconstraint"),
    ("bpy.types.functionnodehashvalue*", "modeling/geometry_nodes/utilities/math/hash_value.html#bpy-types-functionnodehashvalue"),
    ("bpy.types.functionnodeinputbool*", "modeling/geometry_nodes/input/constant/boolean.html#bpy-types-functionnodeinputbool"),
    ("bpy.types.geometrynodecurvestar*", "modeling/geometry_nodes/curve/primitives/star.html#bpy-types-geometrynodecurvestar"),
    ("bpy.types.geometrynodefillcurve*", "modeling/geometry_nodes/curve/operations/fill_curve.html#bpy-types-geometrynodefillcurve"),
    ("bpy.types.geometrynodeflipfaces*", "modeling/geometry_nodes/mesh/operations/flip_faces.html#bpy-types-geometrynodeflipfaces"),
    ("bpy.types.geometrynodegizmodial*", "modeling/geometry_nodes/input/gizmo/dial_gizmo.html#bpy-types-geometrynodegizmodial"),
    ("bpy.types.geometrynodeimageinfo*", "modeling/geometry_nodes/input/scene/image_info.html#bpy-types-geometrynodeimageinfo"),
    ("bpy.types.geometrynodeimportcsv*", "modeling/geometry_nodes/input/import/csv.html#bpy-types-geometrynodeimportcsv"),
    ("bpy.types.geometrynodeimportobj*", "modeling/geometry_nodes/input/import/obj.html#bpy-types-geometrynodeimportobj"),
    ("bpy.types.geometrynodeimportply*", "modeling/geometry_nodes/input/import/ply.html#bpy-types-geometrynodeimportply"),
    ("bpy.types.geometrynodeimportstl*", "modeling/geometry_nodes/input/import/stl.html#bpy-types-geometrynodeimportstl"),
    ("bpy.types.geometrynodeimportvdb*", "modeling/geometry_nodes/input/import/vdb.html#bpy-types-geometrynodeimportvdb"),
    ("bpy.types.geometrynodeproximity*", "modeling/geometry_nodes/geometry/sample/geometry_proximity.html#bpy-types-geometrynodeproximity"),
    ("bpy.types.geometrynodetransform*", "modeling/geometry_nodes/geometry/operations/transform_geometry.html#bpy-types-geometrynodetransform"),
    ("bpy.types.geometrynodetrimcurve*", "modeling/geometry_nodes/curve/operations/trim_curve.html#bpy-types-geometrynodetrimcurve"),
    ("bpy.types.gpencilsculptsettings*", "grease_pencil/properties/index.html#bpy-types-gpencilsculptsettings"),
    ("bpy.types.keyframe.handle_right*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-handle-right"),
    ("bpy.types.keyingsetpath.id_type*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-id-type"),
    ("bpy.types.light.cutoff_distance*", "render/eevee/light_settings.html#bpy-types-light-cutoff-distance"),
    ("bpy.types.light.specular_factor*", "render/eevee/light_settings.html#bpy-types-light-specular-factor"),
    ("bpy.types.light.use_temperature*", "render/lights/light_object.html#bpy-types-light-use-temperature"),
    ("bpy.types.lightprobe.clip_start*", "render/eevee/light_probes/plane.html#bpy-types-lightprobe-clip-start"),
    ("bpy.types.lockedtrackconstraint*", "animation/constraints/tracking/locked_track.html#bpy-types-lockedtrackconstraint"),
    ("bpy.types.material.displacement*", "render/eevee/material_settings.html#bpy-types-material-displacement"),
    ("bpy.types.modifier.show_on_cage*", "modeling/modifiers/introduction.html#bpy-types-modifier-show-on-cage"),
    ("bpy.types.movietrackingcamera.k*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera-k"),
    ("bpy.types.node.use_custom_color*", "interface/controls/nodes/sidebar.html#bpy-types-node-use-custom-color"),
    ("bpy.types.object.visible_camera*", "render/eevee/object_settings/object_data.html#bpy-types-object-visible-camera"),
    ("bpy.types.object.visible_glossy*", "render/cycles/object_settings/object_data.html#bpy-types-object-visible-glossy"),
    ("bpy.types.object.visible_shadow*", "render/eevee/object_settings/object_data.html#bpy-types-object-visible-shadow"),
    ("bpy.types.posebone.custom_shape*", "animation/armatures/bones/properties/display.html#bpy-types-posebone-custom-shape"),
    ("bpy.types.rigifyselectioncolors*", "addons/rigging/rigify/metarigs.html#bpy-types-rigifyselectioncolors"),
    ("bpy.types.sceneeevee.raytracing*", "render/eevee/render_settings/raytracing.html#bpy-types-sceneeevee-raytracing"),
    ("bpy.types.sceneeevee.volumetric*", "render/eevee/render_settings/volumes.html#bpy-types-sceneeevee-volumetric"),
    ("bpy.types.screen.show_statusbar*", "interface/window_system/topbar.html#bpy-types-screen-show-statusbar"),
    ("bpy.types.sculpt.detail_percent*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-types-sculpt-detail-percent"),
    ("bpy.types.sculpt.gravity_object*", "sculpt_paint/sculpting/tool_settings/options.html#bpy-types-sculpt-gravity-object"),
    ("bpy.types.sculpt.show_face_sets*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-types-sculpt-show-face-sets"),
    ("bpy.types.shadernodebsdfdiffuse*", "render/shader_nodes/shader/diffuse.html#bpy-types-shadernodebsdfdiffuse"),
    ("bpy.types.shadernodelayerweight*", "render/shader_nodes/input/layer_weight.html#bpy-types-shadernodelayerweight"),
    ("bpy.types.shadernodenewgeometry*", "render/shader_nodes/input/geometry.html#bpy-types-shadernodenewgeometry"),
    ("bpy.types.shadernodeoutputlight*", "render/shader_nodes/output/light.html#bpy-types-shadernodeoutputlight"),
    ("bpy.types.shadernodeoutputworld*", "render/shader_nodes/output/world.html#bpy-types-shadernodeoutputworld"),
    ("bpy.types.shadernodeseparatexyz*", "render/shader_nodes/converter/separate_xyz.html#bpy-types-shadernodeseparatexyz"),
    ("bpy.types.shadernodeshadertorgb*", "render/shader_nodes/converter/shader_to_rgb.html#bpy-types-shadernodeshadertorgb"),
    ("bpy.types.shadernodetexgradient*", "render/shader_nodes/textures/gradient.html#bpy-types-shadernodetexgradient"),
    ("bpy.types.shadernodetexmusgrave*", "render/shader_nodes/textures/musgrave.html#bpy-types-shadernodetexmusgrave"),
    ("bpy.types.shadernodevectorcurve*", "render/shader_nodes/vector/curves.html#bpy-types-shadernodevectorcurve"),
    ("bpy.types.shadernodevertexcolor*", "render/shader_nodes/input/vertex_color.html#bpy-types-shadernodevertexcolor"),
    ("bpy.types.shapekey.relative_key*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-relative-key"),
    ("bpy.types.shapekey.vertex_group*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-vertex-group"),
    ("bpy.types.softbodysettings.aero*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-aero"),
    ("bpy.types.softbodysettings.bend*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-bend"),
    ("bpy.types.softbodysettings.mass*", "physics/soft_body/settings/object.html#bpy-types-softbodysettings-mass"),
    ("bpy.types.softbodysettings.pull*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-pull"),
    ("bpy.types.softbodysettings.push*", "physics/soft_body/settings/edges.html#bpy-types-softbodysettings-push"),
    ("bpy.types.spaceuveditor.show_uv*", "editors/uv/overlays.html#bpy-types-spaceuveditor-show-uv"),
    ("bpy.types.spline.use_endpoint_u*", "modeling/curves/properties/active_spline.html#bpy-types-spline-use-endpoint-u"),
    ("bpy.types.stripelement.filename*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-stripelement-filename"),
    ("bpy.types.striptransform.filter*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-striptransform-filter"),
    ("bpy.types.striptransform.offset*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-striptransform-offset"),
    ("bpy.types.surfacedeformmodifier*", "modeling/modifiers/deform/surface_deform.html#bpy-types-surfacedeformmodifier"),
    ("bpy.types.texturenodetexvoronoi*", "editors/texture_node/types/textures/voronoi.html#bpy-types-texturenodetexvoronoi"),
    ("bpy.types.toolsettings.use_snap*", "editors/3dview/controls/snapping.html#bpy-types-toolsettings-use-snap"),
    ("bpy.types.usersolidlight.smooth*", "editors/preferences/lights.html#bpy-types-usersolidlight-smooth"),
    ("bpy.types.viewlayer.use_pass_uv*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-uv"),
    ("bpy.types.viewlayer.use_volumes*", "render/layers/filter.html#bpy-types-viewlayer-use-volumes"),
    ("bpy.types.volume.frame_duration*", "modeling/volumes/properties.html#bpy-types-volume-frame-duration"),
    ("bpy.types.volumedisplay.density*", "modeling/volumes/properties.html#bpy-types-volumedisplay-density"),
    ("bpy.types.volumerender.clipping*", "modeling/volumes/properties.html#bpy-types-volumerender-clipping"),
    ("bpy.types.workspace.object_mode*", "interface/window_system/workspaces.html#bpy-types-workspace-object-mode"),
    ("bpy.ops.anim.channels_collapse*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-collapse"),
    ("bpy.ops.anim.driver_button_add*", "animation/drivers/usage.html#bpy-ops-anim-driver-button-add"),
    ("bpy.ops.anim.keying_set_export*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keying-set-export"),
    ("bpy.ops.anim.keying_set_remove*", "animation/keyframes/keying_sets.html#bpy-ops-anim-keying-set-remove"),
    ("bpy.ops.anim.scene_range_frame*", "editors/timeline.html#bpy-ops-anim-scene-range-frame"),
    ("bpy.ops.armature.click_extrude*", "animation/armatures/bones/editing/extrude.html#bpy-ops-armature-click-extrude"),
    ("bpy.ops.armature.select_linked*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-linked"),
    ("bpy.ops.armature.select_mirror*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-mirror"),
    ("bpy.ops.curve.switch_direction*", "modeling/curves/editing/segments.html#bpy-ops-curve-switch-direction"),
    ("bpy.ops.curves.handle_type_set*", "modeling/curves_new/editing/control_points.html#bpy-ops-curves-handle-type-set"),
    ("bpy.ops.curves.spline_type_set*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-spline-type-set"),
    ("bpy.ops.geometry.attribute_add*", "modeling/geometry_nodes/attributes_reference.html#bpy-ops-geometry-attribute-add"),
    ("bpy.ops.graph.blend_to_default*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-blend-to-default"),
    ("bpy.ops.graph.equalize_handles*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-equalize-handles"),
    ("bpy.ops.graph.select_leftright*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-leftright"),
    ("bpy.ops.graph.sound_to_samples*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-sound-to-samples"),
    ("bpy.ops.grease_pencil.caps_set*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-caps-set"),
    ("bpy.ops.grease_pencil.dissolve*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-dissolve"),
    ("bpy.ops.grease_pencil.separate*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-separate"),
    ("bpy.ops.mball.delete_metaelems*", "modeling/metas/editing.html#bpy-ops-mball-delete-metaelems"),
    ("bpy.ops.mball.reveal_metaelems*", "modeling/metas/properties.html#bpy-ops-mball-reveal-metaelems"),
    ("bpy.ops.mesh.bridge_edge_loops*", "modeling/meshes/editing/edge/bridge_edge_loops.html#bpy-ops-mesh-bridge-edge-loops"),
    ("bpy.ops.mesh.intersect_boolean*", "modeling/meshes/editing/face/intersect_boolean.html#bpy-ops-mesh-intersect-boolean"),
    ("bpy.ops.mesh.loop_multi_select*", "modeling/meshes/selecting/loops.html#bpy-ops-mesh-loop-multi-select"),
    ("bpy.ops.mesh.vert_connect_path*", "modeling/meshes/editing/vertex/connect_vertex_path.html#bpy-ops-mesh-vert-connect-path"),
    ("bpy.ops.nla.action_sync_length*", "editors/nla/editing/strip.html#bpy-ops-nla-action-sync-length"),
    ("bpy.ops.node.connect_to_output*", "interface/controls/nodes/editing.html#bpy-ops-node-connect-to-output"),
    ("bpy.ops.node.move_detach_links*", "interface/controls/nodes/editing.html#bpy-ops-node-move-detach-links"),
    ("bpy.ops.object.make_links_data*", "scene_layout/object/editing/link_transfer/link_data.html#bpy-ops-object-make-links-data"),
    ("bpy.ops.object.modifier_remove*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-remove"),
    ("bpy.ops.object.modifiers_clear*", "scene_layout/object/editing/modifiers.html#bpy-ops-object-modifiers-clear"),
    ("bpy.ops.object.paths_calculate*", "animation/motion_paths.html#bpy-ops-object-paths-calculate"),
    ("bpy.ops.object.shape_key_clear*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-clear"),
    ("bpy.ops.object.transform_apply*", "scene_layout/object/editing/apply.html#bpy-ops-object-transform-apply"),
    ("bpy.ops.outliner.lib_operation*", "files/linked_libraries/link_append.html#bpy-ops-outliner-lib-operation"),
    ("bpy.ops.outliner.orphans_purge*", "editors/outliner/interface.html#bpy-ops-outliner-orphans-purge"),
    ("bpy.ops.paint.hide_show_masked*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-paint-hide-show-masked"),
    ("bpy.ops.paint.mask_box_gesture*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-paint-mask-box-gesture"),
    ("bpy.ops.paint.vertex_color_hsv*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-hsv"),
    ("bpy.ops.paint.vertex_color_set*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-vertex-color-set"),
    ("bpy.ops.particle.select_random*", "physics/particles/mode.html#bpy-ops-particle-select-random"),
    ("bpy.ops.pose.blend_to_neighbor*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-blend-to-neighbor"),
    ("bpy.ops.screen.region_quadview*", "editors/3dview/navigate/views.html#bpy-ops-screen-region-quadview"),
    ("bpy.ops.screen.screenshot_area*", "interface/window_system/topbar.html#bpy-ops-screen-screenshot-area"),
    ("bpy.ops.screen.workspace_cycle*", "interface/window_system/workspaces.html#bpy-ops-screen-workspace-cycle"),
    ("bpy.ops.sequencer.offset_clear*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-offset-clear"),
    ("bpy.ops.spreadsheet.fit_column*", "editors/spreadsheet.html#bpy-ops-spreadsheet-fit-column"),
    ("bpy.ops.spreadsheet.toggle_pin*", "editors/spreadsheet.html#bpy-ops-spreadsheet-toggle-pin"),
    ("bpy.ops.uv.follow_active_quads*", "modeling/meshes/editing/uv.html#bpy-ops-uv-follow-active-quads"),
    ("bpy.ops.view3d.view_persportho*", "editors/3dview/navigate/projections.html#bpy-ops-view3d-view-persportho"),
    ("bpy.types.assetmetadata.author*", "editors/asset_browser.html#bpy-types-assetmetadata-author"),
    ("bpy.types.bone.envelope_weight*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-envelope-weight"),
    ("bpy.types.brush.use_persistent*", "sculpt_paint/sculpting/brushes/layer.html#bpy-types-brush-use-persistent"),
    ("bpy.types.brushgpencilsettings*", "grease_pencil/modes/draw/tools/index.html#bpy-types-brushgpencilsettings"),
    ("bpy.types.camera.panorama_type*", "render/cycles/object_settings/cameras.html#bpy-types-camera-panorama-type"),
    ("bpy.types.camera.sensor_height*", "render/cameras.html#bpy-types-camera-sensor-height"),
    ("bpy.types.colorbalancemodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-colorbalancemodifier"),
    ("bpy.types.colorramp.color_mode*", "interface/controls/templates/color_ramp.html#bpy-types-colorramp-color-mode"),
    ("bpy.types.compositornodefilter*", "compositing/types/filter/filter.html#bpy-types-compositornodefilter"),
    ("bpy.types.compositornodehuesat*", "compositing/types/color/adjust/hue_saturation.html#bpy-types-compositornodehuesat"),
    ("bpy.types.compositornodeidmask*", "compositing/types/mask/id_mask.html#bpy-types-compositornodeidmask"),
    ("bpy.types.compositornodeinvert*", "compositing/types/color/invert_color.html#bpy-types-compositornodeinvert"),
    ("bpy.types.compositornodekeying*", "compositing/types/keying/keying.html#bpy-types-compositornodekeying"),
    ("bpy.types.compositornodelevels*", "compositing/types/utilities/levels.html#bpy-types-compositornodelevels"),
    ("bpy.types.compositornodenormal*", "compositing/types/vector/normal.html#bpy-types-compositornodenormal"),
    ("bpy.types.compositornoderotate*", "compositing/types/transform/rotate.html#bpy-types-compositornoderotate"),
    ("bpy.types.compositornodeswitch*", "compositing/types/utilities/switch.html#bpy-types-compositornodeswitch"),
    ("bpy.types.compositornodeviewer*", "compositing/types/output/viewer.html#bpy-types-compositornodeviewer"),
    ("bpy.types.constraint.influence*", "animation/constraints/interface/common.html#bpy-types-constraint-influence"),
    ("bpy.types.curve.use_path_clamp*", "modeling/curves/properties/path_animation.html#bpy-types-curve-use-path-clamp"),
    ("bpy.types.datatransfermodifier*", "modeling/modifiers/modify/data_transfer.html#bpy-types-datatransfermodifier"),
    ("bpy.types.dynamicpaintmodifier*", "physics/dynamic_paint/index.html#bpy-types-dynamicpaintmodifier"),
    ("bpy.types.editbone.collections*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-collections"),
    ("bpy.types.editbone.use_connect*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-use-connect"),
    ("bpy.types.ffmpegsettings.codec*", "render/output/properties/output.html#bpy-types-ffmpegsettings-codec"),
    ("bpy.types.followpathconstraint*", "animation/constraints/relationship/follow_path.html#bpy-types-followpathconstraint"),
    ("bpy.types.functionnodeinputint*", "modeling/geometry_nodes/input/constant/integer.html#bpy-types-functionnodeinputint"),
    ("bpy.types.gaussianblursequence*", "video_editing/edit/montage/strips/effects/blur.html#bpy-types-gaussianblursequence"),
    ("bpy.types.geometrynodeboundbox*", "modeling/geometry_nodes/geometry/operations/bounding_box.html#bpy-types-geometrynodeboundbox"),
    ("bpy.types.geometrynodecurvearc*", "modeling/geometry_nodes/curve/primitives/arc.html#bpy-types-geometrynodecurvearc"),
    ("bpy.types.geometrynodedualmesh*", "modeling/geometry_nodes/mesh/operations/dual_mesh.html#bpy-types-geometrynodedualmesh"),
    ("bpy.types.geometrynodemeshcone*", "modeling/geometry_nodes/mesh/primitives/cone.html#bpy-types-geometrynodemeshcone"),
    ("bpy.types.geometrynodemeshcube*", "modeling/geometry_nodes/mesh/primitives/cube.html#bpy-types-geometrynodemeshcube"),
    ("bpy.types.geometrynodemeshgrid*", "modeling/geometry_nodes/mesh/primitives/grid.html#bpy-types-geometrynodemeshgrid"),
    ("bpy.types.geometrynodemeshline*", "modeling/geometry_nodes/mesh/primitives/mesh_line.html#bpy-types-geometrynodemeshline"),
    ("bpy.types.geometrynodeuvunwrap*", "modeling/geometry_nodes/mesh/uv/uv_unwrap.html#bpy-types-geometrynodeuvunwrap"),
    ("bpy.types.image.display_aspect*", "editors/image/sidebar.html#bpy-types-image-display-aspect"),
    ("bpy.types.keyframe.handle_left*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-handle-left"),
    ("bpy.types.keyingsetsall.active*", "editors/timeline.html#bpy-types-keyingsetsall-active"),
    ("bpy.types.light.diffuse_factor*", "render/eevee/light_settings.html#bpy-types-light-diffuse-factor"),
    ("bpy.types.limitscaleconstraint*", "animation/constraints/transform/limit_scale.html#bpy-types-limitscaleconstraint"),
    ("bpy.types.materialgpencilstyle*", "grease_pencil/materials/index.html#bpy-types-materialgpencilstyle"),
    ("bpy.types.meshtovolumemodifier*", "modeling/modifiers/generate/mesh_to_volume.html#bpy-types-meshtovolumemodifier"),
    ("bpy.types.modifier.show_render*", "modeling/modifiers/introduction.html#bpy-types-modifier-show-render"),
    ("bpy.types.motionpath.frame_end*", "animation/motion_paths.html#bpy-types-motionpath-frame-end"),
    ("bpy.types.nodetree.description*", "interface/controls/nodes/groups.html#bpy-types-nodetree-description"),
    ("bpy.types.object.hide_viewport*", "scene_layout/object/properties/visibility.html#bpy-types-object-hide-viewport"),
    ("bpy.types.object.light_linking*", "render/lights/light_linking.html#bpy-types-object-light-linking"),
    ("bpy.types.object.rotation_mode*", "scene_layout/object/properties/transforms.html#bpy-types-object-rotation-mode"),
    ("bpy.types.object.show_in_front*", "scene_layout/object/properties/display.html#bpy-types-object-show-in-front"),
    ("bpy.types.posebone.collections*", "animation/armatures/bones/properties/relations.html#bpy-types-posebone-collections"),
    ("bpy.types.posebone.rigify_type*", "addons/rigging/rigify/metarigs.html#bpy-types-posebone-rigify-type"),
    ("bpy.types.preferencesfilepaths*", "editors/preferences/file_paths.html#bpy-types-preferencesfilepaths"),
    ("bpy.types.rigidbodyobject.mass*", "physics/rigid_body/properties/settings.html#bpy-types-rigidbodyobject-mass"),
    ("bpy.types.scene.background_set*", "scene_layout/scene/properties.html#bpy-types-scene-background-set"),
    ("bpy.types.scenesequence.volume*", "video_editing/edit/montage/strips/scene.html#bpy-types-scenesequence-volume"),
    ("bpy.types.shadernodebackground*", "render/shader_nodes/shader/background.html#bpy-types-shadernodebackground"),
    ("bpy.types.shadernodebsdfvelvet*", "render/shader_nodes/shader/sheen.html#bpy-types-shadernodebsdfvelvet"),
    ("bpy.types.shadernodecameradata*", "render/shader_nodes/input/camera_data.html#bpy-types-shadernodecameradata"),
    ("bpy.types.shadernodecombinexyz*", "render/shader_nodes/converter/combine_xyz.html#bpy-types-shadernodecombinexyz"),
    ("bpy.types.shadernodefloatcurve*", "render/shader_nodes/converter/float_curve.html#bpy-types-shadernodefloatcurve"),
    ("bpy.types.shadernodeobjectinfo*", "render/shader_nodes/input/object_info.html#bpy-types-shadernodeobjectinfo"),
    ("bpy.types.shadernodetexchecker*", "render/shader_nodes/textures/checker.html#bpy-types-shadernodetexchecker"),
    ("bpy.types.shadernodetexvoronoi*", "render/shader_nodes/textures/voronoi.html#bpy-types-shadernodetexvoronoi"),
    ("bpy.types.shadernodevectormath*", "render/shader_nodes/converter/vector_math.html#bpy-types-shadernodevectormath"),
    ("bpy.types.shadernodevolumeinfo*", "render/shader_nodes/input/volume_info.html#bpy-types-shadernodevolumeinfo"),
    ("bpy.types.shadernodewavelength*", "render/shader_nodes/converter/wavelength.html#bpy-types-shadernodewavelength"),
    ("bpy.types.shrinkwrapconstraint*", "animation/constraints/relationship/shrinkwrap.html#bpy-types-shrinkwrapconstraint"),
    ("bpy.types.simpledeformmodifier*", "modeling/modifiers/deform/simple_deform.html#bpy-types-simpledeformmodifier"),
    ("bpy.types.soundsequence.volume*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-soundsequence-volume"),
    ("bpy.types.spaceclipeditor.show*", "editors/clip/display/index.html#bpy-types-spaceclipeditor-show"),
    ("bpy.types.spacedopesheeteditor*", "editors/dope_sheet/index.html#bpy-types-spacedopesheeteditor"),
    ("bpy.types.spaceview3d.clip_end*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-clip-end"),
    ("bpy.types.speedcontrolsequence*", "video_editing/edit/montage/strips/effects/speed_control.html#bpy-types-speedcontrolsequence"),
    ("bpy.types.spotlight.spot_blend*", "render/lights/light_object.html#bpy-types-spotlight-spot-blend"),
    ("bpy.types.striptransform.scale*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-striptransform-scale"),
    ("bpy.types.textcurve.space_line*", "modeling/texts/properties.html#bpy-types-textcurve-space-line"),
    ("bpy.types.textcurve.space_word*", "modeling/texts/properties.html#bpy-types-textcurve-space-word"),
    ("bpy.types.texturenodecurvetime*", "editors/texture_node/types/input/time.html#bpy-types-texturenodecurvetime"),
    ("bpy.types.texturenodetexclouds*", "editors/texture_node/types/textures/clouds.html#bpy-types-texturenodetexclouds"),
    ("bpy.types.texturenodetexmarble*", "editors/texture_node/types/textures/marble.html#bpy-types-texturenodetexmarble"),
    ("bpy.types.texturenodetexstucci*", "editors/texture_node/types/textures/stucci.html#bpy-types-texturenodetexstucci"),
    ("bpy.types.texturenodetranslate*", "editors/texture_node/types/distort/translate.html#bpy-types-texturenodetranslate"),
    ("bpy.types.transformorientation*", "editors/3dview/controls/orientation.html#bpy-types-transformorientation"),
    ("bpy.types.unifiedpaintsettings*", "sculpt_paint/brush/introduction.html#bpy-types-unifiedpaintsettings"),
    ("bpy.types.viewlayer.use_pass_z*", "render/layers/passes.html#bpy-types-viewlayer-use-pass-z"),
    ("bpy.types.viewlayer.use_strand*", "render/layers/filter.html#bpy-types-viewlayer-use-strand"),
    ("bpy.types.volume.sequence_mode*", "modeling/volumes/properties.html#bpy-types-volume-sequence-mode"),
    ("bpy.types.volumetomeshmodifier*", "modeling/modifiers/generate/volume_to_mesh.html#bpy-types-volumetomeshmodifier"),
    ("bpy.types.whitebalancemodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-whitebalancemodifier"),
    ("bpy.ops.anim.channels_ungroup*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-ungroup"),
    ("bpy.ops.armature.extrude_move*", "animation/armatures/bones/editing/extrude.html#bpy-ops-armature-extrude-move"),
    ("bpy.ops.armature.parent_clear*", "animation/armatures/bones/editing/parenting.html#bpy-ops-armature-parent-clear"),
    ("bpy.ops.clip.clear_track_path*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-clear-track-path"),
    ("bpy.ops.clip.set_scene_frames*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-set-scene-frames"),
    ("bpy.ops.clip.track_copy_color*", "movie_clip/tracking/clip/sidebar/track/track.html#bpy-ops-clip-track-copy-color"),
    ("bpy.ops.collection.export_all*", "scene_layout/collections/collections.html#bpy-ops-collection-export-all"),
    ("bpy.ops.curve.de_select_first*", "modeling/curves/selecting.html#bpy-ops-curve-de-select-first"),
    ("bpy.ops.curve.handle_type_set*", "modeling/curves/editing/control_points.html#bpy-ops-curve-handle-type-set"),
    ("bpy.ops.curve.select_previous*", "modeling/curves/selecting.html#bpy-ops-curve-select-previous"),
    ("bpy.ops.curve.spline_type_set*", "modeling/curves/editing/curve.html#bpy-ops-curve-spline-type-set"),
    ("bpy.ops.curves.duplicate_move*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-duplicate-move"),
    ("bpy.ops.file.unpack_libraries*", "files/blend/packed_data.html#bpy-ops-file-unpack-libraries"),
    ("bpy.ops.graph.gaussian_smooth*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-gaussian-smooth"),
    ("bpy.ops.graph.keyframe_insert*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-keyframe-insert"),
    ("bpy.ops.graph.keys_to_samples*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-keys-to-samples"),
    ("bpy.ops.graph.samples_to_keys*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-samples-to-keys"),
    ("bpy.ops.grease_pencil.outline*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-outline"),
    ("bpy.ops.grease_pencil.reorder*", "grease_pencil/modes/edit/stroke_menu.html#bpy-ops-grease-pencil-reorder"),
    ("bpy.ops.image.read_viewlayers*", "editors/image/editing.html#bpy-ops-image-read-viewlayers"),
    ("bpy.ops.mask.add_vertex_slide*", "movie_clip/masking/editing.html#bpy-ops-mask-add-vertex-slide"),
    ("bpy.ops.mask.shape_key_insert*", "movie_clip/masking/editing.html#bpy-ops-mask-shape-key-insert"),
    ("bpy.ops.mask.switch_direction*", "movie_clip/masking/editing.html#bpy-ops-mask-switch-direction"),
    ("bpy.ops.mesh.blend_from_shape*", "modeling/meshes/editing/vertex/blend_shape.html#bpy-ops-mesh-blend-from-shape"),
    ("bpy.ops.mesh.dissolve_limited*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-dissolve-limited"),
    ("bpy.ops.mesh.face_make_planar*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-face-make-planar"),
    ("bpy.ops.mesh.face_set_extract*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-ops-mesh-face-set-extract"),
    ("bpy.ops.mesh.faces_shade_flat*", "modeling/meshes/editing/face/shading.html#bpy-ops-mesh-faces-shade-flat"),
    ("bpy.ops.mesh.paint_mask_slice*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-mesh-paint-mask-slice"),
    ("bpy.ops.mesh.select_edge_ring*", "modeling/meshes/selecting/loops.html#bpy-ops-mesh-select-edge-ring"),
    ("bpy.ops.mesh.select_next_item*", "modeling/meshes/selecting/more_less.html#bpy-ops-mesh-select-next-item"),
    ("bpy.ops.mesh.select_prev_item*", "modeling/meshes/selecting/more_less.html#bpy-ops-mesh-select-prev-item"),
    ("bpy.ops.mesh.select_ungrouped*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-ungrouped"),
    ("bpy.ops.node.delete_reconnect*", "interface/controls/nodes/editing.html#bpy-ops-node-delete-reconnect"),
    ("bpy.ops.node.tree_path_parent*", "interface/controls/nodes/groups.html#bpy-ops-node-tree-path-parent"),
    ("bpy.ops.object.duplicate_move*", "scene_layout/object/editing/duplicate.html#bpy-ops-object-duplicate-move"),
    ("bpy.ops.object.hook_add_selob*", "modeling/meshes/editing/vertex/hooks.html#bpy-ops-object-hook-add-selob"),
    ("bpy.ops.object.modifier_apply*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-apply"),
    ("bpy.ops.object.select_by_type*", "scene_layout/object/selecting.html#bpy-ops-object-select-by-type"),
    ("bpy.ops.object.select_grouped*", "scene_layout/object/selecting.html#bpy-ops-object-select-grouped"),
    ("bpy.ops.object.select_pattern*", "scene_layout/object/selecting.html#bpy-ops-object-select-pattern"),
    ("bpy.ops.object.shape_key_copy*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-copy"),
    ("bpy.ops.outliner.id_operation*", "editors/outliner/editing.html#bpy-ops-outliner-id-operation"),
    ("bpy.ops.paint.mask_flood_fill*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-paint-mask-flood-fill"),
    ("bpy.ops.paint.weight_gradient*", "sculpt_paint/weight_paint/editing.html#bpy-ops-paint-weight-gradient"),
    ("bpy.ops.pointcloud.select_all*", "modeling/point_cloud/selecting.html#bpy-ops-pointcloud-select-all"),
    ("bpy.ops.pose.quaternions_flip*", "animation/armatures/posing/editing/flip_quats.html#bpy-ops-pose-quaternions-flip"),
    ("bpy.ops.pose.select_hierarchy*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-hierarchy"),
    ("bpy.ops.pose.transforms_clear*", "animation/armatures/posing/editing/clear.html#bpy-ops-pose-transforms-clear"),
    ("bpy.ops.poselib.copy_as_asset*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib-copy-as-asset"),
    ("bpy.ops.preferences.copy_prev*", "editors/preferences/introduction.html#bpy-ops-preferences-copy-prev"),
    ("bpy.ops.screen.animation_play*", "editors/timeline.html#bpy-ops-screen-animation-play"),
    ("bpy.ops.screen.repeat_history*", "interface/undo_redo.html#bpy-ops-screen-repeat-history"),
    ("bpy.ops.script.execute_preset*", "interface/window_system/tabs_panels.html#bpy-ops-script-execute-preset"),
    ("bpy.ops.sculpt.face_sets_init*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-ops-sculpt-face-sets-init"),
    ("bpy.ops.sequencer.change_path*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-change-path"),
    ("bpy.ops.sequencer.refresh_all*", "editors/video_sequencer/sequencer/navigating.html#bpy-ops-sequencer-refresh-all"),
    ("bpy.ops.sequencer.select_less*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-less"),
    ("bpy.ops.sequencer.select_more*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-more"),
    ("bpy.ops.sequencer.select_side*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-side"),
    ("bpy.ops.sequencer.swap_inputs*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-swap-inputs"),
    ("bpy.ops.surface.primitive*add*", "modeling/surfaces/primitives.html#bpy-ops-surface-primitive-add"),
    ("bpy.ops.text.resolve_conflict*", "editors/text_editor.html#bpy-ops-text-resolve-conflict"),
    ("bpy.ops.transform.edge_crease*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-transform-edge-crease"),
    ("bpy.ops.transform.skin_resize*", "modeling/meshes/editing/mesh/transform/skin_resize.html#bpy-ops-transform-skin-resize"),
    ("bpy.ops.uv.seams_from_islands*", "modeling/meshes/uv/editing.html#bpy-ops-uv-seams-from-islands"),
    ("bpy.ops.uv.shortest_path_pick*", "editors/uv/selecting.html#bpy-ops-uv-shortest-path-pick"),
    ("bpy.ops.view3d.camera_to_view*", "editors/3dview/navigate/camera_view.html#bpy-ops-view3d-camera-to-view"),
    ("bpy.ops.wm.clear_recent_files*", "files/blend/open_save.html#bpy-ops-wm-clear-recent-files"),
    ("bpy.ops.wm.search_single_menu*", "interface/controls/buttons/menus.html#bpy-ops-wm-search-single-menu"),
    ("bpy.types.bakesettings.margin*", "render/cycles/baking.html#bpy-types-bakesettings-margin"),
    ("bpy.types.bakesettings.target*", "render/cycles/baking.html#bpy-types-bakesettings-target"),
    ("bpy.types.brush.cloth_damping*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-damping"),
    ("bpy.types.brush.deform_target*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-deform-target"),
    ("bpy.types.brush.falloff_shape*", "sculpt_paint/brush/falloff.html#bpy-types-brush-falloff-shape"),
    ("bpy.types.brush.stroke_method*", "sculpt_paint/brush/stroke.html#bpy-types-brush-stroke-method"),
    ("bpy.types.brush.tip_roundness*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-tip-roundness"),
    ("bpy.types.camera.display_size*", "render/cameras.html#bpy-types-camera-display-size"),
    ("bpy.types.camera.fisheye_lens*", "render/cycles/object_settings/cameras.html#bpy-types-camera-fisheye-lens"),
    ("bpy.types.camera.sensor_width*", "render/cameras.html#bpy-types-camera-sensor-width"),
    ("bpy.types.compositornodedblur*", "compositing/types/filter/blur/directional_blur.html#bpy-types-compositornodedblur"),
    ("bpy.types.compositornodegamma*", "compositing/types/color/adjust/gamma.html#bpy-types-compositornodegamma"),
    ("bpy.types.compositornodeglare*", "compositing/types/filter/glare.html#bpy-types-compositornodeglare"),
    ("bpy.types.compositornodegroup*", "compositing/types/groups.html#bpy-types-compositornodegroup"),
    ("bpy.types.compositornodeimage*", "compositing/types/input/image.html#bpy-types-compositornodeimage"),
    ("bpy.types.compositornodemapuv*", "compositing/types/transform/map_uv.html#bpy-types-compositornodemapuv"),
    ("bpy.types.compositornodescale*", "compositing/types/transform/scale.html#bpy-types-compositornodescale"),
    ("bpy.types.compositornodesplit*", "compositing/types/utilities/split.html#bpy-types-compositornodesplit"),
    ("bpy.types.copyscaleconstraint*", "animation/constraints/transform/copy_scale.html#bpy-types-copyscaleconstraint"),
    ("bpy.types.curve.path_duration*", "modeling/curves/properties/path_animation.html#bpy-types-curve-path-duration"),
    ("bpy.types.curve.use_fill_caps*", "modeling/curves/properties/geometry.html#bpy-types-curve-use-fill-caps"),
    ("bpy.types.curve.use_map_taper*", "modeling/curves/properties/geometry.html#bpy-types-curve-use-map-taper"),
    ("bpy.types.curves.use_mirror_z*", "sculpt_paint/curves_sculpting/introduction.html#bpy-types-curves-use-mirror-z"),
    ("bpy.types.cyclesworldsettings*", "render/cycles/world_settings.html#bpy-types-cyclesworldsettings"),
    ("bpy.types.drivervariable.name*", "animation/drivers/drivers_panel.html#bpy-types-drivervariable-name"),
    ("bpy.types.drivervariable.type*", "animation/drivers/drivers_panel.html#bpy-types-drivervariable-type"),
    ("bpy.types.fieldsettings.noise*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-noise"),
    ("bpy.types.fieldsettings.shape*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-shape"),
    ("bpy.types.fluiddomainsettings*", "physics/fluid/type/domain/index.html#bpy-types-fluiddomainsettings"),
    ("bpy.types.functionnodebitmath*", "modeling/geometry_nodes/utilities/math/bit_math.html#bpy-types-functionnodebitmath"),
    ("bpy.types.functionnodecompare*", "modeling/geometry_nodes/utilities/math/compare.html#bpy-types-functionnodecompare"),
    ("bpy.types.geometrynodeinputid*", "modeling/geometry_nodes/geometry/read/id.html#bpy-types-geometrynodeinputid"),
    ("bpy.types.geometrynoderaycast*", "modeling/geometry_nodes/geometry/sample/raycast.html#bpy-types-geometrynoderaycast"),
    ("bpy.types.geometrynodewarning*", "modeling/geometry_nodes/output/warning.html#bpy-types-geometrynodewarning"),
    ("bpy.types.imageformatsettings*", "files/media/image_formats.html#bpy-types-imageformatsettings"),
    ("bpy.types.keymapitem.map_type*", "editors/preferences/keymap.html#bpy-types-keymapitem-map-type"),
    ("bpy.types.kinematicconstraint*", "animation/constraints/tracking/ik_solver.html#bpy-types-kinematicconstraint"),
    ("bpy.types.light.volume_factor*", "render/eevee/light_settings.html#bpy-types-light-volume-factor"),
    ("bpy.types.material.line_color*", "render/freestyle/material.html#bpy-types-material-line-color"),
    ("bpy.types.material.pass_index*", "render/materials/settings.html#bpy-types-material-pass-index"),
    ("bpy.types.mesh.use_paint_mask*", "sculpt_paint/selection_visibility.html#bpy-types-mesh-use-paint-mask"),
    ("bpy.types.movietrackingcamera*", "movie_clip/tracking/clip/sidebar/track/camera.html#bpy-types-movietrackingcamera"),
    ("bpy.types.object.display_type*", "scene_layout/object/properties/display.html#bpy-types-object-display-type"),
    ("bpy.types.objectlineart.usage*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart-usage"),
    ("bpy.types.paint.input_samples*", "sculpt_paint/brush/stroke.html#bpy-types-paint-input-samples"),
    ("bpy.types.particledupliweight*", "physics/particles/emitter/vertex_groups.html#bpy-types-particledupliweight"),
    ("bpy.types.poseboneconstraints*", "animation/armatures/posing/bone_constraints/index.html#bpy-types-poseboneconstraints"),
    ("bpy.types.rigidbodyconstraint*", "physics/rigid_body/constraints/index.html#bpy-types-rigidbodyconstraint"),
    ("bpy.types.rigifyarmaturelayer*", "addons/rigging/rigify/metarigs.html#bpy-types-rigifyarmaturelayer"),
    ("bpy.types.scene.frame_current*", "editors/timeline.html#bpy-types-scene-frame-current"),
    ("bpy.types.scene.show_subframe*", "editors/timeline.html#bpy-types-scene-show-subframe"),
    ("bpy.types.shadernodeaddshader*", "render/shader_nodes/shader/add.html#bpy-types-shadernodeaddshader"),
    ("bpy.types.shadernodeattribute*", "render/shader_nodes/input/attribute.html#bpy-types-shadernodeattribute"),
    ("bpy.types.shadernodeblackbody*", "render/shader_nodes/converter/blackbody.html#bpy-types-shadernodeblackbody"),
    ("bpy.types.shadernodebsdfglass*", "render/shader_nodes/shader/glass.html#bpy-types-shadernodebsdfglass"),
    ("bpy.types.shadernodebsdfsheen*", "render/shader_nodes/shader/sheen.html#bpy-types-shadernodebsdfsheen"),
    ("bpy.types.shadernodelightpath*", "render/shader_nodes/input/light_path.html#bpy-types-shadernodelightpath"),
    ("bpy.types.shadernodemixshader*", "render/shader_nodes/shader/mix.html#bpy-types-shadernodemixshader"),
    ("bpy.types.shadernodenormalmap*", "render/shader_nodes/vector/normal_map.html#bpy-types-shadernodenormalmap"),
    ("bpy.types.shadernodeoutputaov*", "render/shader_nodes/output/aov.html#bpy-types-shadernodeoutputaov"),
    ("bpy.types.shadernodepointinfo*", "render/shader_nodes/input/point_info.html#bpy-types-shadernodepointinfo"),
    ("bpy.types.shadernodewireframe*", "render/shader_nodes/input/wireframe.html#bpy-types-shadernodewireframe"),
    ("bpy.types.shapekey.lock_shape*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-lock-shape"),
    ("bpy.types.shapekey.slider_max*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-slider-max"),
    ("bpy.types.shapekey.slider_min*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-slider-min"),
    ("bpy.types.spacesequenceeditor*", "editors/video_sequencer/index.html#bpy-types-spacesequenceeditor"),
    ("bpy.types.spline.resolution_u*", "modeling/curves/properties/active_spline.html#bpy-types-spline-resolution-u"),
    ("bpy.types.spline.use_bezier_u*", "modeling/curves/properties/active_spline.html#bpy-types-spline-use-bezier-u"),
    ("bpy.types.spline.use_cyclic_u*", "modeling/curves/properties/active_spline.html#bpy-types-spline-use-cyclic-u"),
    ("bpy.types.spotlight.show_cone*", "render/lights/light_object.html#bpy-types-spotlight-show-cone"),
    ("bpy.types.spotlight.spot_size*", "render/lights/light_object.html#bpy-types-spotlight-spot-size"),
    ("bpy.types.stretchtoconstraint*", "animation/constraints/tracking/stretch_to.html#bpy-types-stretchtoconstraint"),
    ("bpy.types.stripproxy.timecode*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-stripproxy-timecode"),
    ("bpy.types.texturenodecurvergb*", "editors/texture_node/types/color/rgb_curves.html#bpy-types-texturenodecurvergb"),
    ("bpy.types.texturenodedistance*", "editors/texture_node/types/converter/distance.html#bpy-types-texturenodedistance"),
    ("bpy.types.texturenodetexblend*", "editors/texture_node/types/textures/blend.html#bpy-types-texturenodetexblend"),
    ("bpy.types.texturenodetexmagic*", "editors/texture_node/types/textures/magic.html#bpy-types-texturenodetexmagic"),
    ("bpy.types.texturenodetexnoise*", "editors/texture_node/types/textures/noise.html#bpy-types-texturenodetexnoise"),
    ("bpy.types.texturenodevaltonor*", "editors/texture_node/types/converter/value_to_normal.html#bpy-types-texturenodevaltonor"),
    ("bpy.types.texturenodevaltorgb*", "editors/texture_node/types/converter/color_ramp.html#bpy-types-texturenodevaltorgb"),
    ("bpy.types.transformconstraint*", "animation/constraints/transform/transformation.html#bpy-types-transformconstraint"),
    ("bpy.types.triangulatemodifier*", "modeling/modifiers/generate/triangulate.html#bpy-types-triangulatemodifier"),
    ("bpy.types.unitsettings.system*", "scene_layout/scene/properties.html#bpy-types-unitsettings-system"),
    ("bpy.types.view3dshading.light*", "render/workbench/lighting.html#bpy-types-view3dshading-light"),
    ("bpy.types.viewlayer.use_solid*", "render/layers/filter.html#bpy-types-viewlayer-use-solid"),
    ("bpy.types.volume.frame_offset*", "modeling/volumes/properties.html#bpy-types-volume-frame-offset"),
    ("bpy.types.windowmanager.addon*", "editors/preferences/extensions.html#bpy-types-windowmanager-addon"),
    ("bpy.ops.anim.channels_delete*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-delete"),
    ("bpy.ops.anim.channels_expand*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-expand"),
    ("bpy.ops.anim.keyframe_delete*", "animation/keyframes/editing.html#bpy-ops-anim-keyframe-delete"),
    ("bpy.ops.anim.keyframe_insert*", "animation/keyframes/editing.html#bpy-ops-anim-keyframe-insert"),
    ("bpy.ops.anim.merge_animation*", "editors/dope_sheet/modes/action.html#bpy-ops-anim-merge-animation"),
    ("bpy.ops.anim.slot_new_for_id*", "editors/dope_sheet/modes/action.html#bpy-ops-anim-slot-new-for-id"),
    ("bpy.ops.armature.select_less*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-less"),
    ("bpy.ops.armature.select_more*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-more"),
    ("bpy.ops.asset.bundle_install*", "editors/asset_browser.html#bpy-ops-asset-bundle-install"),
    ("bpy.ops.buttons.clear_filter*", "interface/controls/buttons/fields.html#bpy-ops-buttons-clear-filter"),
    ("bpy.ops.buttons.context_menu*", "interface/controls/buttons/menus.html#bpy-ops-buttons-context-menu"),
    ("bpy.ops.buttons.start_filter*", "interface/controls/buttons/fields.html#bpy-ops-buttons-start-filter"),
    ("bpy.ops.clip.add_marker_move*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-add-marker-move"),
    ("bpy.ops.clip.bundles_to_mesh*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-bundles-to-mesh"),
    ("bpy.ops.clip.detect_features*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-detect-features"),
    ("bpy.ops.console.autocomplete*", "editors/python_console.html#bpy-ops-console-autocomplete"),
    ("bpy.ops.curve.de_select_last*", "modeling/curves/selecting.html#bpy-ops-curve-de-select-last"),
    ("bpy.ops.curve.dissolve_verts*", "modeling/curves/editing/curve.html#bpy-ops-curve-dissolve-verts"),
    ("bpy.ops.curve.duplicate_move*", "modeling/curves/editing/curve.html#bpy-ops-curve-duplicate-move"),
    ("bpy.ops.curve.select_similar*", "modeling/curves/selecting.html#bpy-ops-curve-select-similar"),
    ("bpy.ops.curves.attribute_set*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-attribute-set"),
    ("bpy.ops.curves.select_linked*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-linked"),
    ("bpy.ops.curves.select_random*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-random"),
    ("bpy.ops.file.autopack_toggle*", "files/blend/packed_data.html#bpy-ops-file-autopack-toggle"),
    ("bpy.ops.fluid.bake_particles*", "physics/fluid/type/domain/liquid/particles.html#bpy-ops-fluid-bake-particles"),
    ("bpy.ops.fluid.free_particles*", "physics/fluid/type/domain/liquid/particles.html#bpy-ops-fluid-free-particles"),
    ("bpy.ops.graph.duplicate_move*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-duplicate-move"),
    ("bpy.ops.grease_pencil.delete*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-delete"),
    ("bpy.ops.grease_pencil.select*", "grease_pencil/modes/edit/selecting.html#bpy-ops-grease-pencil-select"),
    ("bpy.ops.mask.handle_type_set*", "movie_clip/masking/editing.html#bpy-ops-mask-handle-type-set"),
    ("bpy.ops.mask.hide_view_clear*", "movie_clip/masking/editing.html#bpy-ops-mask-hide-view-clear"),
    ("bpy.ops.mask.shape_key_clear*", "movie_clip/masking/editing.html#bpy-ops-mask-shape-key-clear"),
    ("bpy.ops.mball.hide_metaelems*", "modeling/metas/properties.html#bpy-ops-mball-hide-metaelems"),
    ("bpy.ops.mesh.average_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-average-normals"),
    ("bpy.ops.mesh.delete_edgeloop*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-delete-edgeloop"),
    ("bpy.ops.mesh.vertices_smooth*", "modeling/meshes/editing/vertex/smooth_vertices.html#bpy-ops-mesh-vertices-smooth"),
    ("bpy.ops.nla.make_single_user*", "editors/nla/editing/strip.html#bpy-ops-nla-make-single-user"),
    ("bpy.ops.node.add_import_node*", "modeling/geometry_nodes/input/import/index.html#bpy-ops-node-add-import-node"),
    ("bpy.ops.node.clipboard_paste*", "interface/controls/nodes/editing.html#bpy-ops-node-clipboard-paste"),
    ("bpy.ops.node.node_copy_color*", "interface/controls/nodes/sidebar.html#bpy-ops-node-node-copy-color"),
    ("bpy.ops.node.nw_add_sequence*", "addons/node/node_wrangler.html#bpy-ops-node-nw-add-sequence"),
    ("bpy.ops.node.read_viewlayers*", "interface/controls/nodes/editing.html#bpy-ops-node-read-viewlayers"),
    ("bpy.ops.object.attribute_add*", "modeling/geometry_nodes/attributes_reference.html#bpy-ops-object-attribute-add"),
    ("bpy.ops.object.data_transfer*", "scene_layout/object/editing/link_transfer/transfer_mesh_data.html#bpy-ops-object-data-transfer"),
    ("bpy.ops.object.modifier_copy*", "modeling/modifiers/introduction.html#bpy-ops-object-modifier-copy"),
    ("bpy.ops.object.select_camera*", "scene_layout/object/selecting.html#bpy-ops-object-select-camera"),
    ("bpy.ops.object.select_linked*", "scene_layout/object/selecting.html#bpy-ops-object-select-linked"),
    ("bpy.ops.object.select_mirror*", "scene_layout/object/selecting.html#bpy-ops-object-select-mirror"),
    ("bpy.ops.object.select_random*", "scene_layout/object/selecting.html#bpy-ops-object-select-random"),
    ("bpy.ops.object.shape_key_add*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-shape-key-add"),
    ("bpy.ops.object.transfer_mode*", "editors/3dview/modes.html#bpy-ops-object-transfer-mode"),
    ("bpy.ops.object.update_shapes*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-update-shapes"),
    ("bpy.ops.outliner.show_active*", "editors/outliner/editing.html#bpy-ops-outliner-show-active"),
    ("bpy.ops.paint.add_simple_uvs*", "sculpt_paint/texture_paint/tool_settings/texture_slots.html#bpy-ops-paint-add-simple-uvs"),
    ("bpy.ops.pose.paths_calculate*", "animation/motion_paths.html#bpy-ops-pose-paths-calculate"),
    ("bpy.ops.pose.rigify_generate*", "addons/rigging/rigify/basics.html#bpy-ops-pose-rigify-generate"),
    ("bpy.ops.preferences.autoexec*", "editors/preferences/save_load.html#bpy-ops-preferences-autoexec"),
    ("bpy.ops.scene.view_layer_add*", "render/layers/introduction.html#bpy-ops-scene-view-layer-add"),
    ("bpy.ops.screen.keyframe_jump*", "editors/timeline.html#bpy-ops-screen-keyframe-jump"),
    ("bpy.ops.sculpt.face_set_edit*", "sculpt_paint/sculpting/editing/face_sets.html#bpy-ops-sculpt-face-set-edit"),
    ("bpy.ops.sculpt.mask_by_color*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-by-color"),
    ("bpy.ops.sequencer.gap_insert*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-gap-insert"),
    ("bpy.ops.sequencer.gap_remove*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-gap-remove"),
    ("bpy.ops.sequencer.hold_split*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-hold-split"),
    ("bpy.ops.sequencer.rendersize*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-rendersize"),
    ("bpy.ops.sequencer.select_all*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-all"),
    ("bpy.ops.sequencer.select_box*", "video_editing/edit/montage/selecting.html#bpy-ops-sequencer-select-box"),
    ("bpy.ops.sound.bake_animation*", "scene_layout/scene/properties.html#bpy-ops-sound-bake-animation"),
    ("bpy.ops.transform.edge_slide*", "modeling/meshes/editing/edge/edge_slide.html#bpy-ops-transform-edge-slide"),
    ("bpy.ops.transform.vert_slide*", "modeling/meshes/editing/vertex/slide_vertices.html#bpy-ops-transform-vert-slide"),
    ("bpy.ops.ui.list_start_filter*", "interface/controls/templates/list_view.html#bpy-ops-ui-list-start-filter"),
    ("bpy.ops.uv.project_from_view*", "modeling/meshes/editing/uv.html#bpy-ops-uv-project-from-view"),
    ("bpy.ops.view3d.render_border*", "editors/3dview/navigate/regions.html#bpy-ops-view3d-render-border"),
    ("bpy.ops.view3d.view_selected*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-view-selected"),
    ("bpy.ops.wm.memory_statistics*", "advanced/operators.html#bpy-ops-wm-memory-statistics"),
    ("bpy.ops.wm.recover_auto_save*", "files/blend/open_save.html#bpy-ops-wm-recover-auto-save"),
    ("bpy.types.adjustmentsequence*", "video_editing/edit/montage/strips/adjustment.html#bpy-types-adjustmentsequence"),
    ("bpy.types.alphaundersequence*", "video_editing/edit/montage/strips/effects/alpha_over_under.html#bpy-types-alphaundersequence"),
    ("bpy.types.animvizmotionpaths*", "animation/motion_paths.html#bpy-types-animvizmotionpaths"),
    ("bpy.types.armature.show_axes*", "animation/armatures/properties/display.html#bpy-types-armature-show-axes"),
    ("bpy.types.armatureconstraint*", "animation/constraints/relationship/armature.html#bpy-types-armatureconstraint"),
    ("bpy.types.brush.dash_samples*", "sculpt_paint/brush/stroke.html#bpy-types-brush-dash-samples"),
    ("bpy.types.brush.sculpt_plane*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-sculpt-plane"),
    ("bpy.types.camera.fisheye_fov*", "render/cycles/object_settings/cameras.html#bpy-types-camera-fisheye-fov"),
    ("bpy.types.clothsettings.mass*", "physics/cloth/settings/physical_properties.html#bpy-types-clothsettings-mass"),
    ("bpy.types.compositornodeblur*", "compositing/types/filter/blur/blur.html#bpy-types-compositornodeblur"),
    ("bpy.types.compositornodecrop*", "compositing/types/transform/crop.html#bpy-types-compositornodecrop"),
    ("bpy.types.compositornodeflip*", "compositing/types/transform/flip.html#bpy-types-compositornodeflip"),
    ("bpy.types.compositornodemask*", "compositing/types/input/mask.html#bpy-types-compositornodemask"),
    ("bpy.types.compositornodemath*", "compositing/types/utilities/math.html#bpy-types-compositornodemath"),
    ("bpy.types.compositornodetime*", "compositing/types/input/scene/time_curve.html#bpy-types-compositornodetime"),
    ("bpy.types.constraint.enabled*", "animation/constraints/interface/header.html#bpy-types-constraint-enabled"),
    ("bpy.types.curve.bevel_object*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-object"),
    ("bpy.types.curve.resolution_u*", "modeling/curves/properties/shape.html#bpy-types-curve-resolution-u"),
    ("bpy.types.curve.resolution_v*", "modeling/surfaces/properties/shape.html#bpy-types-curve-resolution-v"),
    ("bpy.types.curve.taper_object*", "modeling/curves/properties/geometry.html#bpy-types-curve-taper-object"),
    ("bpy.types.curve.twist_smooth*", "modeling/curves/properties/shape.html#bpy-types-curve-twist-smooth"),
    ("bpy.types.curvepaintsettings*", "modeling/curves_new/tools/draw.html#bpy-types-curvepaintsettings"),
    ("bpy.types.fcurve.array_index*", "editors/graph_editor/fcurves/properties.html#bpy-types-fcurve-array-index"),
    ("bpy.types.fieldsettings.flow*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-flow"),
    ("bpy.types.fieldsettings.seed*", "physics/forces/force_fields/introduction.html#bpy-types-fieldsettings-seed"),
    ("bpy.types.fieldsettings.size*", "physics/forces/force_fields/types/turbulence.html#bpy-types-fieldsettings-size"),
    ("bpy.types.fileselectidfilter*", "editors/file_browser.html#bpy-types-fileselectidfilter"),
    ("bpy.types.fmodifiergenerator*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifiergenerator"),
    ("bpy.types.freestylelinestyle*", "render/freestyle/view_layer/line_style/index.html#bpy-types-freestylelinestyle"),
    ("bpy.types.gammacrosssequence*", "video_editing/edit/montage/strips/transitions/gamma_cross.html#bpy-types-gammacrosssequence"),
    ("bpy.types.geometrynodepoints*", "modeling/geometry_nodes/point/points.html#bpy-types-geometrynodepoints"),
    ("bpy.types.geometrynodeswitch*", "modeling/geometry_nodes/utilities/switch.html#bpy-types-geometrynodeswitch"),
    ("bpy.types.geometrynodeviewer*", "modeling/geometry_nodes/output/viewer.html#bpy-types-geometrynodeviewer"),
    ("bpy.types.gppaint.color_mode*", "grease_pencil/modes/draw/tool_settings/color.html#bpy-types-gppaint-color-mode"),
    ("bpy.types.huecorrectmodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-huecorrectmodifier"),
    ("bpy.types.image.is_multiview*", "editors/image/image_settings.html#bpy-types-image-is-multiview"),
    ("bpy.types.imagepaint.stencil*", "sculpt_paint/texture_paint/tool_settings/mask.html#bpy-types-imagepaint-stencil"),
    ("bpy.types.material.roughness*", "render/materials/settings.html#bpy-types-material-roughness"),
    ("bpy.types.mesh.texspace_size*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-types-mesh-texspace-size"),
    ("bpy.types.meshdeformmodifier*", "modeling/modifiers/deform/mesh_deform.html#bpy-types-meshdeformmodifier"),
    ("bpy.types.movietrackingtrack*", "movie_clip/tracking/clip/sidebar/track/index.html#bpy-types-movietrackingtrack"),
    ("bpy.types.nlastrip.influence*", "editors/nla/sidebar.html#bpy-types-nlastrip-influence"),
    ("bpy.types.nodeoutputfileslot*", "compositing/types/output/file_output.html#bpy-types-nodeoutputfileslot"),
    ("bpy.types.nodetree.color_tag*", "interface/controls/nodes/groups.html#bpy-types-nodetree-color-tag"),
    ("bpy.types.normaleditmodifier*", "modeling/modifiers/normals/normal_edit.html#bpy-types-normaleditmodifier"),
    ("bpy.types.object.empty_image*", "modeling/empties.html#bpy-types-object-empty-image"),
    ("bpy.types.object.hide_render*", "scene_layout/object/properties/visibility.html#bpy-types-object-hide-render"),
    ("bpy.types.object.hide_select*", "scene_layout/object/properties/visibility.html#bpy-types-object-hide-select"),
    ("bpy.types.object.parent_type*", "scene_layout/object/properties/relations.html#bpy-types-object-parent-type"),
    ("bpy.types.object.show_bounds*", "scene_layout/object/properties/display.html#bpy-types-object-show-bounds"),
    ("bpy.types.paint.eraser_brush*", "grease_pencil/modes/draw/tools/brush.html#bpy-types-paint-eraser-brush"),
    ("bpy.types.palettecolor.color*", "interface/controls/templates/color_palette.html#bpy-types-palettecolor-color"),
    ("bpy.types.rendersettings.fps*", "render/output/properties/format.html#bpy-types-rendersettings-fps"),
    ("bpy.types.scene.audio_volume*", "scene_layout/scene/properties.html#bpy-types-scene-audio-volume"),
    ("bpy.types.sceneeevee.fast_gi*", "render/eevee/render_settings/raytracing.html#bpy-types-sceneeevee-fast-gi"),
    ("bpy.types.sculpt.detail_size*", "sculpt_paint/sculpting/tool_settings/dyntopo.html#bpy-types-sculpt-detail-size"),
    ("bpy.types.shadernodebsdfhair*", "render/shader_nodes/shader/hair.html#bpy-types-shadernodebsdfhair"),
    ("bpy.types.shadernodebsdftoon*", "render/shader_nodes/shader/toon.html#bpy-types-shadernodebsdftoon"),
    ("bpy.types.shadernodeemission*", "render/shader_nodes/shader/emission.html#bpy-types-shadernodeemission"),
    ("bpy.types.shadernodehairinfo*", "render/shader_nodes/input/hair_info.html#bpy-types-shadernodehairinfo"),
    ("bpy.types.shadernodemaprange*", "render/shader_nodes/converter/map_range.html#bpy-types-shadernodemaprange"),
    ("bpy.types.shadernodergbcurve*", "modeling/geometry_nodes/utilities/color/rgb_curves.html#bpy-types-shadernodergbcurve"),
    ("bpy.types.shadernodetexbrick*", "render/shader_nodes/textures/brick.html#bpy-types-shadernodetexbrick"),
    ("bpy.types.shadernodetexcoord*", "render/shader_nodes/input/texture_coordinate.html#bpy-types-shadernodetexcoord"),
    ("bpy.types.shadernodetexgabor*", "render/shader_nodes/textures/gabor.html#bpy-types-shadernodetexgabor"),
    ("bpy.types.shadernodeteximage*", "render/shader_nodes/textures/image.html#bpy-types-shadernodeteximage"),
    ("bpy.types.shadernodetexmagic*", "render/shader_nodes/textures/magic.html#bpy-types-shadernodetexmagic"),
    ("bpy.types.shadernodetexnoise*", "render/shader_nodes/textures/noise.html#bpy-types-shadernodetexnoise"),
    ("bpy.types.shadernodevaltorgb*", "render/shader_nodes/converter/color_ramp.html#bpy-types-shadernodevaltorgb"),
    ("bpy.types.shrinkwrapmodifier*", "modeling/modifiers/deform/shrinkwrap.html#bpy-types-shrinkwrapmodifier"),
    ("bpy.types.spaceview3d.camera*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-camera"),
    ("bpy.types.splineikconstraint*", "animation/constraints/tracking/spline_ik.html#bpy-types-splineikconstraint"),
    ("bpy.types.stripproxy.quality*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-stripproxy-quality"),
    ("bpy.types.textcurve.offset_x*", "modeling/texts/properties.html#bpy-types-textcurve-offset-x"),
    ("bpy.types.textcurve.offset_y*", "modeling/texts/properties.html#bpy-types-textcurve-offset-y"),
    ("bpy.types.textcurve.overflow*", "modeling/texts/properties.html#bpy-types-textcurve-overflow"),
    ("bpy.types.texturenodechecker*", "editors/texture_node/types/patterns/checker.html#bpy-types-texturenodechecker"),
    ("bpy.types.texturenodergbtobw*", "editors/texture_node/types/converter/rgb_to_bw.html#bpy-types-texturenodergbtobw"),
    ("bpy.types.texturenodetexture*", "editors/texture_node/types/input/texture.html#bpy-types-texturenodetexture"),
    ("bpy.types.texturenodetexwood*", "editors/texture_node/types/textures/wood.html#bpy-types-texturenodetexwood"),
    ("bpy.types.textureslot.offset*", "sculpt_paint/brush/texture.html#bpy-types-textureslot-offset"),
    ("bpy.types.usersolidlight.use*", "editors/preferences/lights.html#bpy-types-usersolidlight-use"),
    ("bpy.types.view3dshading.type*", "editors/3dview/display/shading.html#bpy-types-view3dshading-type"),
    ("bpy.types.volume.frame_start*", "modeling/volumes/properties.html#bpy-types-volume-frame-start"),
    ("bpy.types.volume.is_sequence*", "modeling/volumes/properties.html#bpy-types-volume-is-sequence"),
    ("bpy.types.volumerender.space*", "modeling/volumes/properties.html#bpy-types-volumerender-space"),
    ("bpy.ops.anim.channels_group*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-group"),
    ("bpy.ops.anim.keyframe_clear*", "animation/keyframes/editing.html#bpy-ops-anim-keyframe-clear"),
    ("bpy.ops.anim.separate_slots*", "editors/dope_sheet/modes/action.html#bpy-ops-anim-separate-slots"),
    ("bpy.ops.armature.flip_names*", "animation/armatures/bones/editing/naming.html#bpy-ops-armature-flip-names"),
    ("bpy.ops.armature.parent_set*", "animation/armatures/bones/editing/parenting.html#bpy-ops-armature-parent-set"),
    ("bpy.ops.armature.select_all*", "animation/armatures/bones/selecting.html#bpy-ops-armature-select-all"),
    ("bpy.ops.armature.symmetrize*", "animation/armatures/bones/editing/symmetrize.html#bpy-ops-armature-symmetrize"),
    ("bpy.ops.asset.catalogs_save*", "files/asset_libraries/catalogs.html#bpy-ops-asset-catalogs-save"),
    ("bpy.ops.brush.asset_save_as*", "sculpt_paint/brush/brush_management.html#bpy-ops-brush-asset-save-as"),
    ("bpy.ops.clip.average_tracks*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-average-tracks"),
    ("bpy.ops.clip.refine_markers*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-refine-markers"),
    ("bpy.ops.clip.select_grouped*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-select-grouped"),
    ("bpy.ops.clip.track_to_empty*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-track-to-empty"),
    ("bpy.ops.curve.cyclic_toggle*", "modeling/curves/editing/curve.html#bpy-ops-curve-cyclic-toggle"),
    ("bpy.ops.curve.primitive*add*", "modeling/curves/primitives.html#bpy-ops-curve-primitive-add"),
    ("bpy.ops.curve.select_linked*", "modeling/curves/selecting.html#bpy-ops-curve-select-linked"),
    ("bpy.ops.curve.select_random*", "modeling/curves/selecting.html#bpy-ops-curve-select-random"),
    ("bpy.ops.curve.smooth_radius*", "modeling/curves/editing/control_points.html#bpy-ops-curve-smooth-radius"),
    ("bpy.ops.curve.smooth_weight*", "modeling/curves/editing/control_points.html#bpy-ops-curve-smooth-weight"),
    ("bpy.ops.curves.extrude_move*", "modeling/curves_new/editing/control_points.html#bpy-ops-curves-extrude-move"),
    ("bpy.ops.file.pack_libraries*", "files/blend/packed_data.html#bpy-ops-file-pack-libraries"),
    ("bpy.ops.font.change_spacing*", "modeling/texts/editing.html#bpy-ops-font-change-spacing"),
    ("bpy.ops.font.textbox_remove*", "modeling/texts/properties.html#bpy-ops-font-textbox-remove"),
    ("bpy.ops.graph.blend_to_ease*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-blend-to-ease"),
    ("bpy.ops.graph.channels_bake*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-channels-bake"),
    ("bpy.ops.graph.fmodifier_add*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-fmodifier-add"),
    ("bpy.ops.graph.scale_average*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-scale-average"),
    ("bpy.ops.graph.select_column*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-column"),
    ("bpy.ops.graph.select_linked*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-linked"),
    ("bpy.ops.grease_pencil.paste*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-paste"),
    ("bpy.ops.image.external_edit*", "editors/image/editing.html#bpy-ops-image-external-edit"),
    ("bpy.ops.mesh.colors_reverse*", "modeling/meshes/editing/face/face_data.html#bpy-ops-mesh-colors-reverse"),
    ("bpy.ops.mesh.dissolve_edges*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-dissolve-edges"),
    ("bpy.ops.mesh.dissolve_faces*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-dissolve-faces"),
    ("bpy.ops.mesh.dissolve_verts*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-dissolve-verts"),
    ("bpy.ops.mesh.duplicate_move*", "modeling/meshes/editing/mesh/duplicate.html#bpy-ops-mesh-duplicate-move"),
    ("bpy.ops.mesh.extrude_region*", "modeling/meshes/tools/extrude_region.html#bpy-ops-mesh-extrude-region"),
    ("bpy.ops.mesh.extrude_repeat*", "modeling/meshes/editing/mesh/extrude.html#bpy-ops-mesh-extrude-repeat"),
    ("bpy.ops.mesh.loop_to_region*", "modeling/meshes/selecting/loops.html#bpy-ops-mesh-loop-to-region"),
    ("bpy.ops.mesh.region_to_loop*", "modeling/meshes/selecting/loops.html#bpy-ops-mesh-region-to-loop"),
    ("bpy.ops.mesh.remove_doubles*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-remove-doubles"),
    ("bpy.ops.mesh.select_similar*", "modeling/meshes/selecting/similar.html#bpy-ops-mesh-select-similar"),
    ("bpy.ops.mesh.smooth_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-smooth-normals"),
    ("bpy.ops.nla.action_pushdown*", "editors/nla/tracks.html#bpy-ops-nla-action-pushdown"),
    ("bpy.ops.nla.tweakmode_enter*", "editors/nla/editing/strip.html#bpy-ops-nla-tweakmode-enter"),
    ("bpy.ops.node.clipboard_copy*", "interface/controls/nodes/editing.html#bpy-ops-node-clipboard-copy"),
    ("bpy.ops.node.duplicate_move*", "interface/controls/nodes/editing.html#bpy-ops-node-duplicate-move"),
    ("bpy.ops.node.options_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-options-toggle"),
    ("bpy.ops.node.preview_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-preview-toggle"),
    ("bpy.ops.object.modifier_add*", "scene_layout/object/editing/modifiers.html#bpy-ops-object-modifier-add"),
    ("bpy.ops.object.origin_clear*", "scene_layout/object/editing/clear.html#bpy-ops-object-origin-clear"),
    ("bpy.ops.object.parent_clear*", "scene_layout/object/editing/parent.html#bpy-ops-object-parent-clear"),
    ("bpy.ops.object.paths_update*", "animation/motion_paths.html#bpy-ops-object-paths-update"),
    ("bpy.ops.object.shade_smooth*", "scene_layout/object/editing/shading.html#bpy-ops-object-shade-smooth"),
    ("bpy.ops.object.voxel_remesh*", "modeling/meshes/retopology.html#bpy-ops-object-voxel-remesh"),
    ("bpy.ops.paint.hide_show_all*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-paint-hide-show-all"),
    ("bpy.ops.pointcloud.separate*", "modeling/point_cloud/editing.html#bpy-ops-pointcloud-separate"),
    ("bpy.ops.pose.armature_apply*", "animation/armatures/posing/editing/apply.html#bpy-ops-pose-armature-apply"),
    ("bpy.ops.pose.select_grouped*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-grouped"),
    ("bpy.ops.preferences.keyitem*", "editors/preferences/keymap.html#bpy-ops-preferences-keyitem"),
    ("bpy.ops.screen.area_options*", "interface/window_system/areas.html#bpy-ops-screen-area-options"),
    ("bpy.ops.screen.region_blend*", "interface/window_system/regions.html#bpy-ops-screen-region-blend"),
    ("bpy.ops.screen.region_scale*", "interface/window_system/regions.html#bpy-ops-screen-region-scale"),
    ("bpy.ops.sculpt.sample_color*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-sample-color"),
    ("bpy.ops.sequencer.swap_data*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-swap-data"),
    ("bpy.ops.transform.push_pull*", "modeling/meshes/editing/mesh/transform/push_pull.html#bpy-ops-transform-push-pull"),
    ("bpy.ops.transform.seq_slide*", "video_editing/edit/montage/editing.html#bpy-ops-transform-seq-slide"),
    ("bpy.ops.transform.trackball*", "scene_layout/object/editing/transform/rotate.html#bpy-ops-transform-trackball"),
    ("bpy.ops.transform.transform*", "scene_layout/object/editing/transform/align_transform_orientation.html#bpy-ops-transform-transform"),
    ("bpy.ops.transform.translate*", "scene_layout/object/editing/transform/move.html#bpy-ops-transform-translate"),
    ("bpy.ops.ui.eyedropper_color*", "interface/controls/templates/color_picker.html#bpy-ops-ui-eyedropper-color"),
    ("bpy.ops.uv.cylinder_project*", "modeling/meshes/editing/uv.html#bpy-ops-uv-cylinder-project"),
    ("bpy.ops.uv.minimize_stretch*", "modeling/meshes/uv/editing.html#bpy-ops-uv-minimize-stretch"),
    ("bpy.ops.uv.select_edge_loop*", "editors/uv/selecting.html#bpy-ops-uv-select-edge-loop"),
    ("bpy.ops.uv.select_edge_ring*", "editors/uv/selecting.html#bpy-ops-uv-select-edge-ring"),
    ("bpy.ops.wm.save_as_mainfile*", "files/blend/open_save.html#bpy-ops-wm-save-as-mainfile"),
    ("bpy.types.action.use_cyclic*", "animation/actions.html#bpy-types-action-use-cyclic"),
    ("bpy.types.alphaoversequence*", "video_editing/edit/montage/strips/effects/alpha_over_under.html#bpy-types-alphaoversequence"),
    ("bpy.types.armatureeditbones*", "animation/armatures/bones/editing/index.html#bpy-types-armatureeditbones"),
    ("bpy.types.bone.display_type*", "animation/armatures/bones/properties/display.html#bpy-types-bone-display-type"),
    ("bpy.types.bonecolor.palette*", "animation/armatures/bones/properties/display.html#bpy-types-bonecolor-palette"),
    ("bpy.types.brush.automasking*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-automasking"),
    ("bpy.types.brush.jitter_unit*", "sculpt_paint/brush/stroke.html#bpy-types-brush-jitter-unit"),
    ("bpy.types.brush.pose_offset*", "sculpt_paint/sculpting/brushes/pose.html#bpy-types-brush-pose-offset"),
    ("bpy.types.brush.rake_factor*", "sculpt_paint/sculpting/brushes/snake_hook.html#bpy-types-brush-rake-factor"),
    ("bpy.types.brush.tip_scale_x*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-tip-scale-x"),
    ("bpy.types.camera.sensor_fit*", "render/cameras.html#bpy-types-camera-sensor-fit"),
    ("bpy.types.cameradofsettings*", "render/cameras.html#bpy-types-cameradofsettings"),
    ("bpy.types.childofconstraint*", "animation/constraints/relationship/child_of.html#bpy-types-childofconstraint"),
    ("bpy.types.clamptoconstraint*", "animation/constraints/tracking/clamp_to.html#bpy-types-clamptoconstraint"),
    ("bpy.types.collisionmodifier*", "physics/collision.html#bpy-types-collisionmodifier"),
    ("bpy.types.collisionsettings*", "physics/collision.html#bpy-types-collisionsettings"),
    ("bpy.types.compositornodergb*", "compositing/types/input/constant/rgb.html#bpy-types-compositornodergb"),
    ("bpy.types.curve.bevel_depth*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel-depth"),
    ("bpy.types.curve.use_stretch*", "modeling/curves/properties/shape.html#bpy-types-curve-use-stretch"),
    ("bpy.types.driver.expression*", "animation/drivers/drivers_panel.html#bpy-types-driver-expression"),
    ("bpy.types.edgesplitmodifier*", "modeling/modifiers/generate/edge_split.html#bpy-types-edgesplitmodifier"),
    ("bpy.types.fcurve.color_mode*", "editors/graph_editor/fcurves/properties.html#bpy-types-fcurve-color-mode"),
    ("bpy.types.fluidflowsettings*", "physics/fluid/type/flow.html#bpy-types-fluidflowsettings"),
    ("bpy.types.fmodifierenvelope*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifierenvelope"),
    ("bpy.types.freestylesettings*", "render/freestyle/view_layer/freestyle.html#bpy-types-freestylesettings"),
    ("bpy.types.geometrynodegroup*", "modeling/geometry_nodes/group.html#bpy-types-geometrynodegroup"),
    ("bpy.types.geometrynodesetid*", "modeling/geometry_nodes/geometry/write/set_id.html#bpy-types-geometrynodesetid"),
    ("bpy.types.greasepencillayer*", "grease_pencil/properties/layers.html#bpy-types-greasepencillayer"),
    ("bpy.types.image.seam_margin*", "editors/image/image_settings.html#bpy-types-image-seam-margin"),
    ("bpy.types.imagepaint.dither*", "sculpt_paint/texture_paint/tool_settings/options.html#bpy-types-imagepaint-dither"),
    ("bpy.types.keymapitem.active*", "editors/preferences/keymap.html#bpy-types-keymapitem-active"),
    ("bpy.types.keymapitem.idname*", "editors/preferences/keymap.html#bpy-types-keymapitem-idname"),
    ("bpy.types.light.temperature*", "render/lights/light_object.html#bpy-types-light-temperature"),
    ("bpy.types.material.metallic*", "render/materials/settings.html#bpy-types-material-metallic"),
    ("bpy.types.materialslot.link*", "render/materials/assignment.html#bpy-types-materialslot-link"),
    ("bpy.types.mesh.texture_mesh*", "modeling/meshes/uv/uv_texture_spaces.html#bpy-types-mesh-texture-mesh"),
    ("bpy.types.mesh.use_mirror_x*", "modeling/meshes/tools/tool_settings.html#bpy-types-mesh-use-mirror-x"),
    ("bpy.types.mesh.use_mirror_y*", "modeling/meshes/tools/tool_settings.html#bpy-types-mesh-use-mirror-y"),
    ("bpy.types.mesh.use_mirror_z*", "modeling/meshes/tools/tool_settings.html#bpy-types-mesh-use-mirror-z"),
    ("bpy.types.meshcachemodifier*", "modeling/modifiers/modify/mesh_cache.html#bpy-types-meshcachemodifier"),
    ("bpy.types.movieclipsequence*", "video_editing/edit/montage/strips/clip.html#bpy-types-movieclipsequence"),
    ("bpy.types.object.dimensions*", "scene_layout/object/properties/transforms.html#bpy-types-object-dimensions"),
    ("bpy.types.object.is_holdout*", "scene_layout/object/properties/visibility.html#bpy-types-object-is-holdout"),
    ("bpy.types.object.lightgroup*", "render/cycles/object_settings/object_data.html#bpy-types-object-lightgroup"),
    ("bpy.types.object.pass_index*", "scene_layout/object/properties/relations.html#bpy-types-object-pass-index"),
    ("bpy.types.object.track_axis*", "scene_layout/object/properties/relations.html#bpy-types-object-track-axis"),
    ("bpy.types.pointlight.energy*", "render/lights/light_object.html#bpy-types-pointlight-energy"),
    ("bpy.types.pose.use_mirror_x*", "animation/armatures/posing/tool_settings.html#bpy-types-pose-use-mirror-x"),
    ("bpy.types.preferencessystem*", "editors/preferences/system.html#bpy-types-preferencessystem"),
    ("bpy.types.scene.active_clip*", "scene_layout/scene/properties.html#bpy-types-scene-active-clip"),
    ("bpy.types.scene.frame_start*", "render/output/properties/frame_range.html#bpy-types-scene-frame-start"),
    ("bpy.types.screen.use_follow*", "editors/timeline.html#bpy-types-screen-use-follow"),
    ("bpy.types.shadernodefresnel*", "render/shader_nodes/input/fresnel.html#bpy-types-shadernodefresnel"),
    ("bpy.types.shadernodeholdout*", "render/shader_nodes/shader/holdout.html#bpy-types-shadernodeholdout"),
    ("bpy.types.shadernodemapping*", "render/shader_nodes/vector/mapping.html#bpy-types-shadernodemapping"),
    ("bpy.types.shadernodergbtobw*", "render/shader_nodes/converter/rgb_to_bw.html#bpy-types-shadernodergbtobw"),
    ("bpy.types.shadernodetangent*", "render/shader_nodes/input/tangent.html#bpy-types-shadernodetangent"),
    ("bpy.types.shadernodetexwave*", "render/shader_nodes/textures/wave.html#bpy-types-shadernodetexwave"),
    ("bpy.types.soundsequence.pan*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-soundsequence-pan"),
    ("bpy.types.spaceimageoverlay*", "editors/image/overlays.html#bpy-types-spaceimageoverlay"),
    ("bpy.types.spline.use_smooth*", "modeling/curves/properties/active_spline.html#bpy-types-spline-use-smooth"),
    ("bpy.types.strip.frame_start*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-frame-start"),
    ("bpy.types.textcurve.align_x*", "modeling/texts/properties.html#bpy-types-textcurve-align-x"),
    ("bpy.types.textcurve.align_y*", "modeling/texts/properties.html#bpy-types-textcurve-align-y"),
    ("bpy.types.texturenodebricks*", "editors/texture_node/types/patterns/bricks.html#bpy-types-texturenodebricks"),
    ("bpy.types.texturenodeinvert*", "editors/texture_node/types/color/invert_color.html#bpy-types-texturenodeinvert"),
    ("bpy.types.texturenodemixrgb*", "editors/texture_node/types/color/mix_rgb.html#bpy-types-texturenodemixrgb"),
    ("bpy.types.texturenodeoutput*", "editors/texture_node/types/output/output.html#bpy-types-texturenodeoutput"),
    ("bpy.types.texturenoderotate*", "editors/texture_node/types/distort/rotate.html#bpy-types-texturenoderotate"),
    ("bpy.types.texturenodeviewer*", "editors/texture_node/types/output/viewer.html#bpy-types-texturenodeviewer"),
    ("bpy.types.textureslot.scale*", "sculpt_paint/brush/texture.html#bpy-types-textureslot-scale"),
    ("bpy.types.themebonecolorset*", "animation/armatures/bones/properties/display.html#bpy-types-themebonecolorset"),
    ("bpy.types.tracktoconstraint*", "animation/constraints/tracking/track_to.html#bpy-types-tracktoconstraint"),
    ("bpy.types.transformsequence*", "video_editing/edit/montage/strips/effects/transform.html#bpy-types-transformsequence"),
    ("bpy.types.uvprojectmodifier*", "modeling/modifiers/modify/uv_project.html#bpy-types-uvprojectmodifier"),
    ("bpy.types.viewlayer.samples*", "render/layers/introduction.html#bpy-types-viewlayer-samples"),
    ("bpy.types.viewlayer.use_sky*", "render/layers/filter.html#bpy-types-viewlayer-use-sky"),
    ("bpy.types.wireframemodifier*", "modeling/modifiers/generate/wireframe.html#bpy-types-wireframemodifier"),
    ("bpy.types.worldmistsettings*", "render/eevee/world_settings.html#bpy-types-worldmistsettings"),
    ("bpy.ops.anim.channels_move*", "editors/graph_editor/channels/editing.html#bpy-ops-anim-channels-move"),
    ("bpy.ops.armature.subdivide*", "animation/armatures/bones/editing/subdivide.html#bpy-ops-armature-subdivide"),
    ("bpy.ops.asset.catalog_redo*", "editors/asset_browser.html#bpy-ops-asset-catalog-redo"),
    ("bpy.ops.asset.catalog_undo*", "editors/asset_browser.html#bpy-ops-asset-catalog-undo"),
    ("bpy.ops.brush.curve_preset*", "sculpt_paint/brush/falloff.html#bpy-ops-brush-curve-preset"),
    ("bpy.ops.buttons.toggle_pin*", "editors/properties_editor.html#bpy-ops-buttons-toggle-pin"),
    ("bpy.ops.clip.delete_marker*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-delete-marker"),
    ("bpy.ops.clip.filter_tracks*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-filter-tracks"),
    ("bpy.ops.clip.select_circle*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-select-circle"),
    ("bpy.ops.clip.track_markers*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-track-markers"),
    ("bpy.ops.curve.extrude_move*", "modeling/curves/editing/control_points.html#bpy-ops-curve-extrude-move"),
    ("bpy.ops.curve.make_segment*", "modeling/curves/editing/control_points.html#bpy-ops-curve-make-segment"),
    ("bpy.ops.curves.select_less*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-less"),
    ("bpy.ops.curves.select_more*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-more"),
    ("bpy.ops.file.directory_new*", "editors/file_browser.html#bpy-ops-file-directory-new"),
    ("bpy.ops.graph.blend_offset*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-blend-offset"),
    ("bpy.ops.graph.euler_filter*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-euler-filter"),
    ("bpy.ops.grease_pencil.copy*", "grease_pencil/modes/edit/grease_pencil_menu.html#bpy-ops-grease-pencil-copy"),
    ("bpy.ops.marker.camera_bind*", "animation/markers.html#bpy-ops-marker-camera-bind"),
    ("bpy.ops.mask.cyclic_toggle*", "movie_clip/masking/editing.html#bpy-ops-mask-cyclic-toggle"),
    ("bpy.ops.mask.hide_view_set*", "movie_clip/masking/editing.html#bpy-ops-mask-hide-view-set"),
    ("bpy.ops.mask.paste_splines*", "movie_clip/masking/editing.html#bpy-ops-mask-paste-splines"),
    ("bpy.ops.mask.select_circle*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-circle"),
    ("bpy.ops.mask.select_linked*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-linked"),
    ("bpy.ops.mesh.attribute_set*", "modeling/meshes/editing/mesh/set_attribute.html#bpy-ops-mesh-attribute-set"),
    ("bpy.ops.mesh.beautify_fill*", "modeling/meshes/editing/face/beautify_faces.html#bpy-ops-mesh-beautify-fill"),
    ("bpy.ops.mesh.colors_rotate*", "modeling/meshes/editing/face/face_data.html#bpy-ops-mesh-colors-rotate"),
    ("bpy.ops.mesh.edge_collapse*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-edge-collapse"),
    ("bpy.ops.mesh.edge_face_add*", "modeling/meshes/editing/vertex/make_face_edge.html#bpy-ops-mesh-edge-face-add"),
    ("bpy.ops.mesh.extrude_indiv*", "modeling/meshes/editing/face/extrude_faces.html#bpy-ops-mesh-extrude-indiv"),
    ("bpy.ops.mesh.knife_project*", "modeling/meshes/editing/mesh/knife_project.html#bpy-ops-mesh-knife-project"),
    ("bpy.ops.mesh.loopcut_slide*", "modeling/meshes/editing/edge/loopcut_slide.html#bpy-ops-mesh-loopcut-slide"),
    ("bpy.ops.mesh.merge_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-merge-normals"),
    ("bpy.ops.mesh.normals_tools*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-normals-tools"),
    ("bpy.ops.mesh.point_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-point-normals"),
    ("bpy.ops.mesh.rip_edge_move*", "modeling/meshes/editing/vertex/rip_vertices_extend.html#bpy-ops-mesh-rip-edge-move"),
    ("bpy.ops.mesh.select_linked*", "modeling/meshes/selecting/linked.html#bpy-ops-mesh-select-linked"),
    ("bpy.ops.mesh.select_mirror*", "modeling/meshes/selecting/mirror.html#bpy-ops-mesh-select-mirror"),
    ("bpy.ops.mesh.select_random*", "modeling/meshes/selecting/random.html#bpy-ops-mesh-select-random"),
    ("bpy.ops.mesh.sort_elements*", "modeling/meshes/editing/mesh/sort_elements.html#bpy-ops-mesh-sort-elements"),
    ("bpy.ops.mesh.split_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-split-normals"),
    ("bpy.ops.mesh.symmetry_snap*", "modeling/meshes/editing/mesh/snap_symmetry.html#bpy-ops-mesh-symmetry-snap"),
    ("bpy.ops.nla.actionclip_add*", "editors/nla/strips.html#bpy-ops-nla-actionclip-add"),
    ("bpy.ops.nla.transition_add*", "editors/nla/strips.html#bpy-ops-nla-transition-add"),
    ("bpy.ops.node.group_ungroup*", "interface/controls/nodes/groups.html#bpy-ops-node-group-ungroup"),
    ("bpy.ops.object.join_shapes*", "animation/shape_keys/shape_keys_panel.html#bpy-ops-object-join-shapes"),
    ("bpy.ops.object.paths_clear*", "animation/motion_paths.html#bpy-ops-object-paths-clear"),
    ("bpy.ops.object.select_less*", "scene_layout/object/selecting.html#bpy-ops-object-select-less"),
    ("bpy.ops.object.select_more*", "scene_layout/object/selecting.html#bpy-ops-object-select-more"),
    ("bpy.ops.object.track_clear*", "scene_layout/object/editing/track.html#bpy-ops-object-track-clear"),
    ("bpy.ops.paint.sample_color*", "sculpt_paint/vertex_paint/editing.html#bpy-ops-paint-sample-color"),
    ("bpy.ops.pose.select_linked*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-linked"),
    ("bpy.ops.pose.select_mirror*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-mirror"),
    ("bpy.ops.screen.marker_jump*", "animation/markers.html#bpy-ops-screen-marker-jump"),
    ("bpy.ops.screen.repeat_last*", "interface/undo_redo.html#bpy-ops-screen-repeat-last"),
    ("bpy.ops.sculpt.mask_expand*", "sculpt_paint/sculpting/editing/expand.html#bpy-ops-sculpt-mask-expand"),
    ("bpy.ops.sculpt.mask_filter*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-filter"),
    ("bpy.ops.sculpt.mesh_filter*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-mesh-filter"),
    ("bpy.ops.transform.tosphere*", "modeling/meshes/editing/mesh/transform/to_sphere.html#bpy-ops-transform-tosphere"),
    ("bpy.ops.view3d.clip_border*", "editors/3dview/navigate/regions.html#bpy-ops-view3d-clip-border"),
    ("bpy.ops.view3d.toggle_xray*", "editors/3dview/introduction.html#bpy-ops-view3d-toggle-xray"),
    ("bpy.ops.view3d.view_camera*", "editors/3dview/navigate/camera_view.html#bpy-ops-view3d-view-camera"),
    ("bpy.ops.view3d.zoom_border*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-zoom-border"),
    ("bpy.ops.wm.previews_ensure*", "files/blend/previews.html#bpy-ops-wm-previews-ensure"),
    ("bpy.ops.wm.properties_edit*", "files/custom_properties.html#bpy-ops-wm-properties-edit"),
    ("bpy.ops.wm.search_operator*", "interface/operators.html#bpy-ops-wm-search-operator"),
    ("bpy.types.actionconstraint*", "animation/constraints/relationship/action.html#bpy-types-actionconstraint"),
    ("bpy.types.addonpreferences*", "editors/preferences/addons.html#bpy-types-addonpreferences"),
    ("bpy.types.arealight.energy*", "render/lights/light_object.html#bpy-types-arealight-energy"),
    ("bpy.types.arealight.spread*", "render/cycles/light_settings.html#bpy-types-arealight-spread"),
    ("bpy.types.armaturemodifier*", "modeling/modifiers/deform/armature.html#bpy-types-armaturemodifier"),
    ("bpy.types.bone.collections*", "animation/armatures/bones/properties/relations.html#bpy-types-bone-collections"),
    ("bpy.types.bone.head_radius*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-head-radius"),
    ("bpy.types.bone.tail_radius*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-tail-radius"),
    ("bpy.types.brush.cloth_mass*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth-mass"),
    ("bpy.types.brush.dash_ratio*", "sculpt_paint/brush/stroke.html#bpy-types-brush-dash-ratio"),
    ("bpy.types.brushtextureslot*", "sculpt_paint/brush/texture.html#bpy-types-brushtextureslot"),
    ("bpy.types.camera.longitude*", "render/cycles/object_settings/cameras.html#bpy-types-camera-longitude"),
    ("bpy.types.colormixsequence*", "video_editing/edit/montage/strips/effects/color_mix.html#bpy-types-colormixsequence"),
    ("bpy.types.curve.dimensions*", "modeling/curves/properties/shape.html#bpy-types-curve-dimensions"),
    ("bpy.types.curve.taper_mode*", "modeling/curves/properties/geometry.html#bpy-types-curve-taper-mode"),
    ("bpy.types.curve.twist_mode*", "modeling/curves/properties/shape.html#bpy-types-curve-twist-mode"),
    ("bpy.types.curve.use_radius*", "modeling/curves/properties/shape.html#bpy-types-curve-use-radius"),
    ("bpy.types.decimatemodifier*", "modeling/modifiers/generate/decimate.html#bpy-types-decimatemodifier"),
    ("bpy.types.displacemodifier*", "modeling/modifiers/deform/displace.html#bpy-types-displacemodifier"),
    ("bpy.types.displaysafeareas*", "render/cameras.html#bpy-types-displaysafeareas"),
    ("bpy.types.editbone.bbone_x*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-x"),
    ("bpy.types.editbone.bbone_z*", "animation/armatures/bones/properties/bendy_bones.html#bpy-types-editbone-bbone-z"),
    ("bpy.types.fcurve.data_path*", "editors/graph_editor/fcurves/properties.html#bpy-types-fcurve-data-path"),
    ("bpy.types.fileselectparams*", "editors/file_browser.html#bpy-types-fileselectparams"),
    ("bpy.types.fmodifierstepped*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifierstepped"),
    ("bpy.types.freestylelineset*", "render/freestyle/view_layer/line_set.html#bpy-types-freestylelineset"),
    ("bpy.types.geometrynodebake*", "modeling/geometry_nodes/geometry/operations/bake.html#bpy-types-geometrynodebake"),
    ("bpy.types.geometrynodetree*", "modeling/geometry_nodes/tools.html#bpy-types-geometrynodetree"),
    ("bpy.types.greasepencilgrid*", "grease_pencil/properties/display.html#bpy-types-greasepencilgrid"),
    ("bpy.types.image.alpha_mode*", "editors/image/image_settings.html#bpy-types-image-alpha-mode"),
    ("bpy.types.key.use_relative*", "animation/shape_keys/shape_keys_panel.html#bpy-types-key-use-relative"),
    ("bpy.types.keyingsetpath.id*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath-id"),
    ("bpy.types.keymapitem.value*", "editors/preferences/keymap.html#bpy-types-keymapitem-value"),
    ("bpy.types.mask.frame_start*", "movie_clip/masking/sidebar.html#bpy-types-mask-frame-start"),
    ("bpy.types.motionpath.color*", "animation/motion_paths.html#bpy-types-motionpath-color"),
    ("bpy.types.motionpath.lines*", "animation/motion_paths.html#bpy-types-motionpath-lines"),
    ("bpy.types.multicamsequence*", "video_editing/edit/montage/strips/effects/multicam.html#bpy-types-multicamsequence"),
    ("bpy.types.multiplysequence*", "video_editing/edit/montage/strips/effects/multiply.html#bpy-types-multiplysequence"),
    ("bpy.types.multiresmodifier*", "modeling/modifiers/generate/multiresolution.html#bpy-types-multiresmodifier"),
    ("bpy.types.paint.show_brush*", "sculpt_paint/brush/cursor.html#bpy-types-paint-show-brush"),
    ("bpy.types.paint.use_cavity*", "sculpt_paint/texture_paint/tool_settings/mask.html#bpy-types-paint-use-cavity"),
    ("bpy.types.particlesettings*", "physics/particles/index.html#bpy-types-particlesettings"),
    ("bpy.types.pose.use_auto_ik*", "animation/armatures/posing/tool_settings.html#bpy-types-pose-use-auto-ik"),
    ("bpy.types.preferencesinput*", "editors/preferences/input.html#bpy-types-preferencesinput"),
    ("bpy.types.rigifyparameters*", "addons/rigging/rigify/metarigs.html#bpy-types-rigifyparameters"),
    ("bpy.types.scene.frame_step*", "render/output/properties/frame_range.html#bpy-types-scene-frame-step"),
    ("bpy.types.sceneeevee.bokeh*", "render/eevee/render_settings/depth_of_field.html#bpy-types-sceneeevee-bokeh"),
    ("bpy.types.sculpt.show_mask*", "sculpt_paint/sculpting/editing/mask.html#bpy-types-sculpt-show-mask"),
    ("bpy.types.shaderfxcolorize*", "grease_pencil/visual_effects/colorize.html#bpy-types-shaderfxcolorize"),
    ("bpy.types.shaderfxpixelate*", "grease_pencil/visual_effects/pixelate.html#bpy-types-shaderfxpixelate"),
    ("bpy.types.shadernodeinvert*", "render/shader_nodes/color/invert_color.html#bpy-types-shadernodeinvert"),
    ("bpy.types.shadernodemixrgb*", "render/shader_nodes/color/mix.html#bpy-types-shadernodemixrgb"),
    ("bpy.types.shadernodenormal*", "render/shader_nodes/vector/normal.html#bpy-types-shadernodenormal"),
    ("bpy.types.shadernodescript*", "render/shader_nodes/script.html#bpy-types-shadernodescript"),
    ("bpy.types.shadernodetexies*", "render/shader_nodes/textures/ies.html#bpy-types-shadernodetexies"),
    ("bpy.types.shadernodetexsky*", "render/shader_nodes/textures/sky.html#bpy-types-shadernodetexsky"),
    ("bpy.types.softbodymodifier*", "physics/soft_body/index.html#bpy-types-softbodymodifier"),
    ("bpy.types.softbodysettings*", "physics/soft_body/settings/index.html#bpy-types-softbodysettings"),
    ("bpy.types.solidifymodifier*", "modeling/modifiers/generate/solidify.html#bpy-types-solidifymodifier"),
    ("bpy.types.spacefilebrowser*", "editors/file_browser.html#bpy-types-spacefilebrowser"),
    ("bpy.types.spacegrapheditor*", "editors/graph_editor/index.html#bpy-types-spacegrapheditor"),
    ("bpy.types.spaceimageeditor*", "editors/image/index.html#bpy-types-spaceimageeditor"),
    ("bpy.types.spacepreferences*", "editors/preferences/index.html#bpy-types-spacepreferences"),
    ("bpy.types.spacespreadsheet*", "editors/spreadsheet.html#bpy-types-spacespreadsheet"),
    ("bpy.types.spaceview3d.lens*", "editors/3dview/sidebar.html#bpy-types-spaceview3d-lens"),
    ("bpy.types.spaceview3d.show*", "editors/3dview/display/index.html#bpy-types-spaceview3d-show"),
    ("bpy.types.sphfluidsettings*", "physics/particles/emitter/physics/fluid.html#bpy-types-sphfluidsettings"),
    ("bpy.types.spotlight.energy*", "render/lights/light_object.html#bpy-types-spotlight-energy"),
    ("bpy.types.stripproxy.build*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-stripproxy-build"),
    ("bpy.types.subtractsequence*", "video_editing/edit/montage/strips/effects/subtract.html#bpy-types-subtractsequence"),
    ("bpy.types.text.indentation*", "editors/text_editor.html#bpy-types-text-indentation"),
    ("bpy.types.textcurve.family*", "modeling/texts/properties.html#bpy-types-textcurve-family"),
    ("bpy.types.texture.contrast*", "render/materials/legacy_textures/colors.html#bpy-types-texture-contrast"),
    ("bpy.types.texturenodegroup*", "editors/texture_node/types/groups.html#bpy-types-texturenodegroup"),
    ("bpy.types.texturenodeimage*", "editors/texture_node/types/input/image.html#bpy-types-texturenodeimage"),
    ("bpy.types.texturenodescale*", "editors/texture_node/types/distort/scale.html#bpy-types-texturenodescale"),
    ("bpy.types.userassetlibrary*", "editors/preferences/file_paths.html#bpy-types-userassetlibrary"),
    ("bpy.types.window.workspace*", "interface/window_system/workspaces.html#bpy-types-window-workspace"),
    ("bpy.types.world.lightgroup*", "render/cycles/world_settings.html#bpy-types-world-lightgroup"),
    ("bpy.ops.action.select_all*", "editors/dope_sheet/editing.html#bpy-ops-action-select-all"),
    ("bpy.ops.armature.dissolve*", "animation/armatures/bones/editing/delete.html#bpy-ops-armature-dissolve"),
    ("bpy.ops.armature.separate*", "animation/armatures/bones/editing/separate_bones.html#bpy-ops-armature-separate"),
    ("bpy.ops.asset.catalog_new*", "files/asset_libraries/catalogs.html#bpy-ops-asset-catalog-new"),
    ("bpy.ops.clip.clean_tracks*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-clean-tracks"),
    ("bpy.ops.clip.delete_track*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-delete-track"),
    ("bpy.ops.clip.select_lasso*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-select-lasso"),
    ("bpy.ops.clip.solve_camera*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-solve-camera"),
    ("bpy.ops.constraint.delete*", "animation/constraints/interface/header.html#bpy-ops-constraint-delete"),
    ("bpy.ops.curve.select_less*", "modeling/curves/selecting.html#bpy-ops-curve-select-less"),
    ("bpy.ops.curve.select_more*", "modeling/curves/selecting.html#bpy-ops-curve-select-more"),
    ("bpy.ops.curve.select_next*", "modeling/curves/selecting.html#bpy-ops-curve-select-next"),
    ("bpy.ops.curve.smooth_tilt*", "modeling/curves/editing/control_points.html#bpy-ops-curve-smooth-tilt"),
    ("bpy.ops.curves.select_all*", "modeling/curves_new/selecting.html#bpy-ops-curves-select-all"),
    ("bpy.ops.curves.tilt_clear*", "modeling/curves_new/editing/control_points.html#bpy-ops-curves-tilt-clear"),
    ("bpy.ops.file.reset_recent*", "editors/file_browser.html#bpy-ops-file-reset-recent"),
    ("bpy.ops.fluid.bake_guides*", "physics/fluid/type/domain/guides.html#bpy-ops-fluid-bake-guides"),
    ("bpy.ops.fluid.free_guides*", "physics/fluid/type/domain/guides.html#bpy-ops-fluid-free-guides"),
    ("bpy.ops.font.style_toggle*", "modeling/texts/editing.html#bpy-ops-font-style-toggle"),
    ("bpy.ops.graph.easing_type*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-easing-type"),
    ("bpy.ops.graph.handle_type*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-handle-type"),
    ("bpy.ops.graph.match_slope*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-match-slope"),
    ("bpy.ops.graph.select_less*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-less"),
    ("bpy.ops.graph.select_more*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-more"),
    ("bpy.ops.graph.time_offset*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-time-offset"),
    ("bpy.ops.marker.select_all*", "animation/markers.html#bpy-ops-marker-select-all"),
    ("bpy.ops.mask.copy_splines*", "movie_clip/masking/editing.html#bpy-ops-mask-copy-splines"),
    ("bpy.ops.mask.parent_clear*", "movie_clip/masking/editing.html#bpy-ops-mask-parent-clear"),
    ("bpy.ops.mask.select_lasso*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-lasso"),
    ("bpy.ops.mesh.bevel.vertex*", "modeling/meshes/editing/vertex/bevel_vertices.html#bpy-ops-mesh-bevel-vertex"),
    ("bpy.ops.mesh.delete_loose*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-delete-loose"),
    ("bpy.ops.mesh.face_shading*", "modeling/meshes/editing/face/shading.html#bpy-ops-mesh-face-shading"),
    ("bpy.ops.mesh.flip_normals*", "modeling/meshes/editing/mesh/normals.html#bpy-ops-mesh-flip-normals"),
    ("bpy.ops.mesh.select_loose*", "modeling/meshes/selecting/all_by_trait.html#bpy-ops-mesh-select-loose"),
    ("bpy.ops.mesh.vert_connect*", "modeling/meshes/editing/vertex/connect_vertex_pairs.html#bpy-ops-mesh-vert-connect"),
    ("bpy.ops.nla.soundclip_add*", "editors/nla/strips.html#bpy-ops-nla-soundclip-add"),
    ("bpy.ops.nla.tracks_delete*", "editors/nla/editing/track.html#bpy-ops-nla-tracks-delete"),
    ("bpy.ops.node.group_insert*", "interface/controls/nodes/groups.html#bpy-ops-node-group-insert"),
    ("bpy.ops.object.lightprobe*", "render/eevee/light_probes/index.html#bpy-ops-object-lightprobe"),
    ("bpy.ops.object.make_local*", "files/linked_libraries/link_append.html#bpy-ops-object-make-local"),
    ("bpy.ops.object.origin_set*", "scene_layout/object/origin.html#bpy-ops-object-origin-set"),
    ("bpy.ops.object.parent_set*", "scene_layout/object/editing/parent.html#bpy-ops-object-parent-set"),
    ("bpy.ops.object.pointcloud*", "modeling/point_cloud/index.html#bpy-ops-object-pointcloud"),
    ("bpy.ops.object.select_all*", "scene_layout/object/selecting.html#bpy-ops-object-select-all"),
    ("bpy.ops.object.shade_flat*", "scene_layout/object/editing/shading.html#bpy-ops-object-shade-flat"),
    ("bpy.ops.paintcurve.select*", "sculpt_paint/brush/stroke.html#bpy-ops-paintcurve-select"),
    ("bpy.ops.pointcloud.delete*", "modeling/point_cloud/editing.html#bpy-ops-pointcloud-delete"),
    ("bpy.ops.pose.paths_update*", "animation/motion_paths.html#bpy-ops-pose-paths-update"),
    ("bpy.ops.preferences.addon*", "editors/preferences/extensions.html#bpy-ops-preferences-addon"),
    ("bpy.ops.screen.actionzone*", "interface/window_system/areas.html#bpy-ops-screen-actionzone"),
    ("bpy.ops.screen.area_dupli*", "interface/window_system/areas.html#bpy-ops-screen-area-dupli"),
    ("bpy.ops.screen.area_split*", "interface/window_system/areas.html#bpy-ops-screen-area-split"),
    ("bpy.ops.screen.frame_jump*", "editors/timeline.html#bpy-ops-screen-frame-jump"),
    ("bpy.ops.screen.screenshot*", "interface/window_system/topbar.html#bpy-ops-screen-screenshot"),
    ("bpy.ops.sculpt.symmetrize*", "sculpt_paint/sculpting/tool_settings/symmetry.html#bpy-ops-sculpt-symmetrize"),
    ("bpy.ops.uv.align_rotation*", "modeling/meshes/uv/editing.html#bpy-ops-uv-align-rotation"),
    ("bpy.ops.uv.remove_doubles*", "modeling/meshes/uv/editing.html#bpy-ops-uv-remove-doubles"),
    ("bpy.ops.uv.select_similar*", "editors/uv/selecting.html#bpy-ops-uv-select-similar"),
    ("bpy.ops.uv.sphere_project*", "modeling/meshes/editing/uv.html#bpy-ops-uv-sphere-project"),
    ("bpy.ops.view3d.view_orbit*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-view-orbit"),
    ("bpy.ops.wm.alembic_export*", "files/import_export/alembic.html#bpy-ops-wm-alembic-export"),
    ("bpy.ops.wm.alembic_import*", "files/import_export/alembic.html#bpy-ops-wm-alembic-import"),
    ("bpy.ops.wm.previews_clear*", "files/blend/previews.html#bpy-ops-wm-previews-clear"),
    ("bpy.types.arealight.shape*", "render/lights/light_object.html#bpy-types-arealight-shape"),
    ("bpy.types.armature.layers*", "animation/armatures/bones/bone_collections.html#bpy-types-armature-layers"),
    ("bpy.types.armature.rigify*", "addons/rigging/rigify/basics.html#bpy-types-armature-rigify"),
    ("bpy.types.bone.use_deform*", "animation/armatures/bones/properties/deform.html#bpy-types-bone-use-deform"),
    ("bpy.types.booleanmodifier*", "modeling/modifiers/generate/booleans.html#bpy-types-booleanmodifier"),
    ("bpy.types.brush.direction*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-direction"),
    ("bpy.types.brush.mask_tool*", "sculpt_paint/sculpting/brushes/mask.html#bpy-types-brush-mask-tool"),
    ("bpy.types.camera.latitude*", "render/cycles/object_settings/cameras.html#bpy-types-camera-latitude"),
    ("bpy.types.constraint.mute*", "animation/constraints/interface/header.html#bpy-types-constraint-mute"),
    ("bpy.types.constraint.name*", "animation/constraints/interface/header.html#bpy-types-constraint-name"),
    ("bpy.types.curve.eval_time*", "modeling/curves/properties/path_animation.html#bpy-types-curve-eval-time"),
    ("bpy.types.curve.fill_mode*", "modeling/curves/properties/shape.html#bpy-types-curve-fill-mode"),
    ("bpy.types.driver.use_self*", "animation/drivers/drivers_panel.html#bpy-types-driver-use-self"),
    ("bpy.types.editbone.length*", "animation/armatures/bones/properties/transform.html#bpy-types-editbone-length"),
    ("bpy.types.editbone.parent*", "animation/armatures/bones/properties/relations.html#bpy-types-editbone-parent"),
    ("bpy.types.explodemodifier*", "modeling/modifiers/physics/explode.html#bpy-types-explodemodifier"),
    ("bpy.types.fcurvemodifiers*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fcurvemodifiers"),
    ("bpy.types.floorconstraint*", "animation/constraints/relationship/floor.html#bpy-types-floorconstraint"),
    ("bpy.types.fmodifiercycles*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifiercycles"),
    ("bpy.types.fmodifierlimits*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifierlimits"),
    ("bpy.types.imagepaint.mode*", "sculpt_paint/texture_paint/tool_settings/texture_slots.html#bpy-types-imagepaint-mode"),
    ("bpy.types.keymapitem.type*", "editors/preferences/keymap.html#bpy-types-keymapitem-type"),
    ("bpy.types.latticemodifier*", "modeling/modifiers/deform/lattice.html#bpy-types-latticemodifier"),
    ("bpy.types.light.normalize*", "render/lights/light_object.html#bpy-types-light-normalize"),
    ("bpy.types.materiallineart*", "render/materials/line_art.html#bpy-types-materiallineart"),
    ("bpy.types.musgravetexture*", "render/materials/legacy_textures/types/musgrave.html#bpy-types-musgravetexture"),
    ("bpy.types.object.location*", "scene_layout/object/properties/transforms.html#bpy-types-object-location"),
    ("bpy.types.object.rotation*", "scene_layout/object/properties/transforms.html#bpy-types-object-rotation"),
    ("bpy.types.particlehairkey*", "physics/particles/emitter/physics/keyed.html#bpy-types-particlehairkey"),
    ("bpy.types.pivotconstraint*", "animation/constraints/relationship/pivot.html#bpy-types-pivotconstraint"),
    ("bpy.types.preferencesedit*", "editors/preferences/editing.html#bpy-types-preferencesedit"),
    ("bpy.types.preferencesview*", "editors/preferences/interface.html#bpy-types-preferencesview"),
    ("bpy.types.rigidbodyobject*", "physics/rigid_body/index.html#bpy-types-rigidbodyobject"),
    ("bpy.types.scene.frame_end*", "render/output/properties/frame_range.html#bpy-types-scene-frame-end"),
    ("bpy.types.scene.sync_mode*", "editors/timeline.html#bpy-types-scene-sync-mode"),
    ("bpy.types.screen.use_play*", "editors/timeline.html#bpy-types-screen-use-play"),
    ("bpy.types.scriptdirectory*", "editors/preferences/file_paths.html#bpy-types-scriptdirectory"),
    ("bpy.types.shadernodebevel*", "render/shader_nodes/input/bevel.html#bpy-types-shadernodebevel"),
    ("bpy.types.shadernodeclamp*", "render/shader_nodes/converter/clamp.html#bpy-types-shadernodeclamp"),
    ("bpy.types.shadernodegamma*", "render/shader_nodes/color/gamma.html#bpy-types-shadernodegamma"),
    ("bpy.types.shadernodegroup*", "render/shader_nodes/groups.html#bpy-types-shadernodegroup"),
    ("bpy.types.shadernodeuvmap*", "render/shader_nodes/input/uv_map.html#bpy-types-shadernodeuvmap"),
    ("bpy.types.shadernodevalue*", "render/shader_nodes/input/value.html#bpy-types-shadernodevalue"),
    ("bpy.types.spaceclipeditor*", "movie_clip/index.html#bpy-types-spaceclipeditor"),
    ("bpy.types.spaceproperties*", "editors/properties_editor.html#bpy-types-spaceproperties"),
    ("bpy.types.spacetexteditor*", "editors/text_editor.html#bpy-types-spacetexteditor"),
    ("bpy.types.subsurfmodifier*", "modeling/modifiers/generate/subdivision_surface.html#bpy-types-subsurfmodifier"),
    ("bpy.types.sunlight.energy*", "render/lights/light_object.html#bpy-types-sunlight-energy"),
    ("bpy.types.text.use_module*", "editors/text_editor.html#bpy-types-text-use-module"),
    ("bpy.types.textcurve.shear*", "modeling/texts/properties.html#bpy-types-textcurve-shear"),
    ("bpy.types.texturenodemath*", "editors/texture_node/types/converter/math.html#bpy-types-texturenodemath"),
    ("bpy.types.volume.filepath*", "modeling/volumes/properties.html#bpy-types-volume-filepath"),
    ("bpy.ops.asset.tag_remove*", "editors/asset_browser.html#bpy-ops-asset-tag-remove"),
    ("bpy.ops.clip.join_tracks*", "movie_clip/tracking/clip/editing/track.html#bpy-ops-clip-join-tracks"),
    ("bpy.ops.constraint.apply*", "animation/constraints/interface/header.html#bpy-ops-constraint-apply"),
    ("bpy.ops.curve.select_all*", "modeling/curves/selecting.html#bpy-ops-curve-select-all"),
    ("bpy.ops.curve.select_nth*", "modeling/curves/selecting.html#bpy-ops-curve-select-nth"),
    ("bpy.ops.curve.select_row*", "modeling/surfaces/selecting.html#bpy-ops-curve-select-row"),
    ("bpy.ops.curve.tilt_clear*", "modeling/curves/editing/control_points.html#bpy-ops-curve-tilt-clear"),
    ("bpy.ops.curve.vertex_add*", "modeling/curves/editing/other.html#bpy-ops-curve-vertex-add"),
    ("bpy.ops.curves.subdivide*", "modeling/curves_new/editing/segments.html#bpy-ops-curves-subdivide"),
    ("bpy.ops.fluid.bake_noise*", "physics/fluid/type/domain/gas/noise.html#bpy-ops-fluid-bake-noise"),
    ("bpy.ops.fluid.free_noise*", "physics/fluid/type/domain/gas/noise.html#bpy-ops-fluid-free-noise"),
    ("bpy.ops.font.move_select*", "modeling/texts/selecting.html#bpy-ops-font-move-select"),
    ("bpy.ops.font.text_insert*", "modeling/texts/editing.html#bpy-ops-font-text-insert"),
    ("bpy.ops.font.textbox_add*", "modeling/texts/properties.html#bpy-ops-font-textbox-add"),
    ("bpy.ops.gpencil.annotate*", "interface/annotate_tool.html#bpy-ops-gpencil-annotate"),
    ("bpy.ops.graph.frame_jump*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-frame-jump"),
    ("bpy.ops.graph.select_all*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-all"),
    ("bpy.ops.graph.select_box*", "editors/graph_editor/introduction.html#bpy-ops-graph-select-box"),
    ("bpy.ops.marker.duplicate*", "animation/markers.html#bpy-ops-marker-duplicate"),
    ("bpy.ops.mask.select_less*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-less"),
    ("bpy.ops.mask.select_more*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-more"),
    ("bpy.ops.mesh.convex_hull*", "modeling/meshes/editing/mesh/convex_hull.html#bpy-ops-mesh-convex-hull"),
    ("bpy.ops.mesh.edge_rotate*", "modeling/meshes/editing/edge/rotate_edge.html#bpy-ops-mesh-edge-rotate"),
    ("bpy.ops.mesh.loop_select*", "modeling/meshes/selecting/loops.html#bpy-ops-mesh-loop-select"),
    ("bpy.ops.mesh.select_less*", "modeling/meshes/selecting/more_less.html#bpy-ops-mesh-select-less"),
    ("bpy.ops.mesh.select_mode*", "modeling/meshes/selecting/introduction.html#bpy-ops-mesh-select-mode"),
    ("bpy.ops.mesh.select_more*", "modeling/meshes/selecting/more_less.html#bpy-ops-mesh-select-more"),
    ("bpy.ops.mesh.unsubdivide*", "modeling/meshes/editing/edge/unsubdivide.html#bpy-ops-mesh-unsubdivide"),
    ("bpy.ops.mesh.uvs_reverse*", "modeling/meshes/uv/editing.html#bpy-ops-mesh-uvs-reverse"),
    ("bpy.ops.node.hide_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-hide-toggle"),
    ("bpy.ops.node.mute_toggle*", "interface/controls/nodes/editing.html#bpy-ops-node-mute-toggle"),
    ("bpy.ops.object.hide_view*", "scene_layout/object/editing/show_hide.html#bpy-ops-object-hide-view"),
    ("bpy.ops.object.quick_fur*", "physics/introduction.html#bpy-ops-object-quick-fur"),
    ("bpy.ops.object.track_set*", "scene_layout/object/editing/track.html#bpy-ops-object-track-set"),
    ("bpy.ops.paint.weight_set*", "sculpt_paint/weight_paint/editing.html#bpy-ops-paint-weight-set"),
    ("bpy.ops.pose.paths_clear*", "animation/motion_paths.html#bpy-ops-pose-paths-clear"),
    ("bpy.ops.pose.scale_clear*", "animation/armatures/posing/editing/clear.html#bpy-ops-pose-scale-clear"),
    ("bpy.ops.scene.view_layer*", "render/layers/introduction.html#bpy-ops-scene-view-layer"),
    ("bpy.ops.screen.area_join*", "interface/window_system/areas.html#bpy-ops-screen-area-join"),
    ("bpy.ops.screen.area_move*", "interface/window_system/areas.html#bpy-ops-screen-area-move"),
    ("bpy.ops.screen.area_swap*", "interface/window_system/areas.html#bpy-ops-screen-area-swap"),
    ("bpy.ops.screen.redo_last*", "interface/undo_redo.html#bpy-ops-screen-redo-last"),
    ("bpy.ops.sculpt.mask_init*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-sculpt-mask-init"),
    ("bpy.ops.sequencer.delete*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-delete"),
    ("bpy.ops.sequencer.reload*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-reload"),
    ("bpy.ops.sequencer.unlock*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-unlock"),
    ("bpy.ops.sequencer.unmute*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-unmute"),
    ("bpy.ops.transform.mirror*", "scene_layout/object/editing/mirror.html#bpy-ops-transform-mirror"),
    ("bpy.ops.transform.resize*", "scene_layout/object/editing/transform/scale.html#bpy-ops-transform-resize"),
    ("bpy.ops.transform.rotate*", "scene_layout/object/editing/transform/rotate.html#bpy-ops-transform-rotate"),
    ("bpy.ops.uv.lightmap_pack*", "modeling/meshes/editing/uv.html#bpy-ops-uv-lightmap-pack"),
    ("bpy.ops.uv.smart_project*", "modeling/meshes/editing/uv.html#bpy-ops-uv-smart-project"),
    ("bpy.ops.uv.snap_selected*", "modeling/meshes/uv/editing.html#bpy-ops-uv-snap-selected"),
    ("bpy.ops.view3d.localview*", "editors/3dview/navigate/local_view.html#bpy-ops-view3d-localview"),
    ("bpy.ops.view3d.view_axis*", "editors/3dview/navigate/viewpoint.html#bpy-ops-view3d-view-axis"),
    ("bpy.ops.view3d.view_roll*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-view-roll"),
    ("bpy.ops.wm.open_mainfile*", "files/blend/open_save.html#bpy-ops-wm-open-mainfile"),
    ("bpy.ops.wm.owner_disable*", "interface/window_system/workspaces.html#bpy-ops-wm-owner-disable"),
    ("bpy.ops.wm.read_userpref*", "editors/preferences/introduction.html#bpy-ops-wm-read-userpref"),
    ("bpy.ops.wm.save_homefile*", "interface/window_system/topbar.html#bpy-ops-wm-save-homefile"),
    ("bpy.ops.wm.save_mainfile*", "files/blend/open_save.html#bpy-ops-wm-save-mainfile"),
    ("bpy.ops.wm.save_userpref*", "editors/preferences/introduction.html#bpy-ops-wm-save-userpref"),
    ("bpy.types.arealight.size*", "render/lights/light_object.html#bpy-types-arealight-size"),
    ("bpy.types.attributegroup*", "modeling/meshes/properties/object_data.html#bpy-types-attributegroup"),
    ("bpy.types.bone.show_wire*", "animation/armatures/bones/properties/display.html#bpy-types-bone-show-wire"),
    ("bpy.types.bonecollection*", "animation/armatures/properties/bone_collections.html#bpy-types-bonecollection"),
    ("bpy.types.brush.hardness*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-hardness"),
    ("bpy.types.brush.strength*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-strength"),
    ("bpy.types.curves.surface*", "modeling/curves_new/properties.html#bpy-types-curves-surface"),
    ("bpy.types.curvesmodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-curvesmodifier"),
    ("bpy.types.editbone.color*", "animation/armatures/bones/properties/display.html#bpy-types-editbone-color"),
    ("bpy.types.ffmpegsettings*", "render/output/properties/output.html#bpy-types-ffmpegsettings"),
    ("bpy.types.fmodifier.mute*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifier-mute"),
    ("bpy.types.fmodifier.name*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifier-name"),
    ("bpy.types.fmodifiernoise*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifiernoise"),
    ("bpy.types.image.filepath*", "editors/image/image_settings.html#bpy-types-image-filepath"),
    ("bpy.types.keyframe.co_ui*", "editors/graph_editor/fcurves/properties.html#bpy-types-keyframe-co-ui"),
    ("bpy.types.light.exposure*", "render/lights/light_object.html#bpy-types-light-exposure"),
    ("bpy.types.mask.frame_end*", "movie_clip/masking/sidebar.html#bpy-types-mask-frame-end"),
    ("bpy.types.material.paint*", "sculpt_paint/texture_paint/index.html#bpy-types-material-paint"),
    ("bpy.types.mirrormodifier*", "modeling/modifiers/generate/mirror.html#bpy-types-mirrormodifier"),
    ("bpy.types.movieclipproxy*", "editors/clip/sidebar.html#bpy-types-movieclipproxy"),
    ("bpy.types.object.up_axis*", "scene_layout/object/properties/relations.html#bpy-types-object-up-axis"),
    ("bpy.types.particlesystem*", "physics/particles/index.html#bpy-types-particlesystem"),
    ("bpy.types.particletarget*", "physics/particles/emitter/physics/keyed.html#bpy-types-particletarget"),
    ("bpy.types.posebone.color*", "animation/armatures/bones/properties/display.html#bpy-types-posebone-color"),
    ("bpy.types.remeshmodifier*", "modeling/modifiers/generate/remesh.html#bpy-types-remeshmodifier"),
    ("bpy.types.rendersettings*", "render/index.html#bpy-types-rendersettings"),
    ("bpy.types.rigidbodyworld*", "physics/rigid_body/world.html#bpy-types-rigidbodyworld"),
    ("bpy.types.sculpt.gravity*", "sculpt_paint/sculpting/tool_settings/options.html#bpy-types-sculpt-gravity"),
    ("bpy.types.shaderfxshadow*", "grease_pencil/visual_effects/shadow.html#bpy-types-shaderfxshadow"),
    ("bpy.types.shadernodebump*", "render/shader_nodes/vector/bump.html#bpy-types-shadernodebump"),
    ("bpy.types.shadernodemath*", "render/shader_nodes/converter/math.html#bpy-types-shadernodemath"),
    ("bpy.types.shapekey.frame*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-frame"),
    ("bpy.types.shapekey.value*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-value"),
    ("bpy.types.smoothmodifier*", "modeling/modifiers/deform/smooth.html#bpy-types-smoothmodifier"),
    ("bpy.types.sound.use_mono*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-sound-use-mono"),
    ("bpy.types.spline.order_u*", "modeling/curves/properties/active_spline.html#bpy-types-spline-order-u"),
    ("bpy.types.striptransform*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-striptransform"),
    ("bpy.types.sunlight.angle*", "render/lights/light_object.html#bpy-types-sunlight-angle"),
    ("bpy.types.textcurve.size*", "modeling/texts/properties.html#bpy-types-textcurve-size"),
    ("bpy.types.timelinemarker*", "animation/markers.html#bpy-types-timelinemarker"),
    ("bpy.types.usersolidlight*", "editors/preferences/lights.html#bpy-types-usersolidlight"),
    ("bpy.types.uvwarpmodifier*", "modeling/modifiers/modify/uv_warp.html#bpy-types-uvwarpmodifier"),
    ("bpy.types.viewlayer.name*", "render/layers/introduction.html#bpy-types-viewlayer-name"),
    ("bpy.types.voronoitexture*", "render/materials/legacy_textures/types/voronoi.html#bpy-types-voronoitexture"),
    ("bpy.types.walknavigation*", "editors/preferences/navigation.html#bpy-types-walknavigation"),
    ("bpy.ops.*.select_circle*", "interface/selecting.html#bpy-ops-select-circle"),
    ("bpy.ops.armature.delete*", "animation/armatures/bones/editing/delete.html#bpy-ops-armature-delete"),
    ("bpy.ops.clip.select_all*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-select-all"),
    ("bpy.ops.clip.select_box*", "movie_clip/tracking/clip/selecting.html#bpy-ops-clip-select-box"),
    ("bpy.ops.clip.set_origin*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-set-origin"),
    ("bpy.ops.constraint.copy*", "animation/constraints/interface/header.html#bpy-ops-constraint-copy"),
    ("bpy.ops.curve.subdivide*", "modeling/curves/editing/segments.html#bpy-ops-curve-subdivide"),
    ("bpy.ops.curves.separate*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-separate"),
    ("bpy.ops.ed.undo_history*", "interface/undo_redo.html#bpy-ops-ed-undo-history"),
    ("bpy.ops.file.select_all*", "editors/file_browser.html#bpy-ops-file-select-all"),
    ("bpy.ops.file.select_box*", "editors/file_browser.html#bpy-ops-file-select-box"),
    ("bpy.ops.file.unpack_all*", "files/blend/packed_data.html#bpy-ops-file-unpack-all"),
    ("bpy.ops.fluid.bake_data*", "physics/fluid/type/domain/settings.html#bpy-ops-fluid-bake-data"),
    ("bpy.ops.fluid.bake_mesh*", "physics/fluid/type/domain/liquid/mesh.html#bpy-ops-fluid-bake-mesh"),
    ("bpy.ops.fluid.free_data*", "physics/fluid/type/domain/settings.html#bpy-ops-fluid-free-data"),
    ("bpy.ops.fluid.free_mesh*", "physics/fluid/type/domain/liquid/mesh.html#bpy-ops-fluid-free-mesh"),
    ("bpy.ops.font.select_all*", "modeling/texts/selecting.html#bpy-ops-font-select-all"),
    ("bpy.ops.font.text_paste*", "modeling/texts/editing.html#bpy-ops-font-text-paste"),
    ("bpy.ops.graph.bake_keys*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-bake-keys"),
    ("bpy.ops.graph.breakdown*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-breakdown"),
    ("bpy.ops.graph.push_pull*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-push-pull"),
    ("bpy.ops.image.clipboard*", "editors/image/editing.html#bpy-ops-image-clipboard"),
    ("bpy.ops.mask.parent_set*", "movie_clip/masking/editing.html#bpy-ops-mask-parent-set"),
    ("bpy.ops.mask.select_all*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-all"),
    ("bpy.ops.mask.select_box*", "movie_clip/masking/selecting.html#bpy-ops-mask-select-box"),
    ("bpy.ops.mesh.edge_split*", "modeling/meshes/editing/mesh/split.html#bpy-ops-mesh-edge-split"),
    ("bpy.ops.mesh.fill_holes*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-fill-holes"),
    ("bpy.ops.mesh.mark_sharp*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-mesh-mark-sharp"),
    ("bpy.ops.mesh.select_nth*", "modeling/meshes/selecting/checker_deselect.html#bpy-ops-mesh-select-nth"),
    ("bpy.ops.mesh.symmetrize*", "modeling/meshes/editing/mesh/symmetrize.html#bpy-ops-mesh-symmetrize"),
    ("bpy.ops.mesh.uvs_rotate*", "modeling/meshes/uv/editing.html#bpy-ops-mesh-uvs-rotate"),
    ("bpy.ops.nla.apply_scale*", "editors/nla/editing/strip.html#bpy-ops-nla-apply-scale"),
    ("bpy.ops.nla.clear_scale*", "editors/nla/editing/strip.html#bpy-ops-nla-clear-scale"),
    ("bpy.ops.nla.meta_remove*", "editors/nla/editing/strip.html#bpy-ops-nla-meta-remove"),
    ("bpy.ops.nla.mute_toggle*", "editors/nla/editing/strip.html#bpy-ops-nla-mute-toggle"),
    ("bpy.ops.node.group_edit*", "interface/controls/nodes/groups.html#bpy-ops-node-group-edit"),
    ("bpy.ops.node.group_make*", "interface/controls/nodes/groups.html#bpy-ops-node-group-make"),
    ("bpy.ops.node.join_named*", "interface/controls/nodes/frame.html#bpy-ops-node-join-named"),
    ("bpy.ops.node.links_mute*", "interface/controls/nodes/editing.html#bpy-ops-node-links-mute"),
    ("bpy.ops.node.parent_set*", "interface/controls/nodes/frame.html#bpy-ops-node-parent-set"),
    ("bpy.ops.node.select_all*", "interface/controls/nodes/selecting.html#bpy-ops-node-select-all"),
    ("bpy.ops.object.armature*", "animation/armatures/index.html#bpy-ops-object-armature"),
    ("bpy.ops.object.join_uvs*", "scene_layout/object/editing/link_transfer/link_data.html#bpy-ops-object-join-uvs"),
    ("bpy.ops.object.mode_set*", "editors/3dview/modes.html#bpy-ops-object-mode-set"),
    ("bpy.ops.outliner.delete*", "editors/outliner/editing.html#bpy-ops-outliner-delete"),
    ("bpy.ops.paint.hide_show*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-paint-hide-show"),
    ("bpy.ops.paintcurve.draw*", "sculpt_paint/brush/stroke.html#bpy-ops-paintcurve-draw"),
    ("bpy.ops.pose.relax_rest*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-relax-rest"),
    ("bpy.ops.pose.select_all*", "animation/armatures/posing/selecting.html#bpy-ops-pose-select-all"),
    ("bpy.ops.rigidbody.world*", "physics/rigid_body/world.html#bpy-ops-rigidbody-world"),
    ("bpy.ops.sculpt.optimize*", "sculpt_paint/sculpting/editing/sculpt.html#bpy-ops-sculpt-optimize"),
    ("bpy.ops.sequencer.split*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-split"),
    ("bpy.ops.text.run_script*", "editors/text_editor.html#bpy-ops-text-run-script"),
    ("bpy.ops.transform.shear*", "modeling/meshes/editing/mesh/transform/shear.html#bpy-ops-transform-shear"),
    ("bpy.ops.uv.cube_project*", "modeling/meshes/editing/uv.html#bpy-ops-uv-cube-project"),
    ("bpy.ops.uv.pack_islands*", "modeling/meshes/uv/editing.html#bpy-ops-uv-pack-islands"),
    ("bpy.ops.uv.select_split*", "modeling/meshes/uv/editing.html#bpy-ops-uv-select-split"),
    ("bpy.ops.view3d.cursor3d*", "editors/3dview/3d_cursor.html#bpy-ops-view3d-cursor3d"),
    ("bpy.ops.view3d.navigate*", "editors/3dview/navigate/index.html#bpy-ops-view3d-navigate"),
    ("bpy.ops.view3d.view_all*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-view-all"),
    ("bpy.ops.view3d.view_pan*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-view-pan"),
    ("bpy.ops.wm.app_template*", "advanced/app_templates.html#bpy-ops-wm-app-template"),
    ("bpy.ops.wm.batch_rename*", "files/blend/rename.html#bpy-ops-wm-batch-rename"),
    ("bpy.ops.wm.owner_enable*", "interface/window_system/workspaces.html#bpy-ops-wm-owner-enable"),
    ("bpy.ops.wm.redraw_timer*", "advanced/operators.html#bpy-ops-wm-redraw-timer"),
    ("bpy.ops.wm.splash_about*", "interface/window_system/topbar.html#bpy-ops-wm-splash-about"),
    ("bpy.types.armature.show*", "animation/armatures/properties/display.html#bpy-types-armature-show"),
    ("bpy.types.armaturebones*", "animation/armatures/bones/index.html#bpy-types-armaturebones"),
    ("bpy.types.arraymodifier*", "modeling/modifiers/generate/array.html#bpy-types-arraymodifier"),
    ("bpy.types.assetmetadata*", "editors/asset_browser.html#bpy-types-assetmetadata"),
    ("bpy.types.bevelmodifier*", "modeling/modifiers/generate/bevel.html#bpy-types-bevelmodifier"),
    ("bpy.types.brush.density*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-density"),
    ("bpy.types.brush.spacing*", "sculpt_paint/brush/stroke.html#bpy-types-brush-spacing"),
    ("bpy.types.brush.wet_mix*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-wet-mix"),
    ("bpy.types.buildmodifier*", "modeling/modifiers/generate/build.html#bpy-types-buildmodifier"),
    ("bpy.types.clothmodifier*", "physics/cloth/index.html#bpy-types-clothmodifier"),
    ("bpy.types.clothsettings*", "physics/cloth/settings/index.html#bpy-types-clothsettings"),
    ("bpy.types.cloudstexture*", "render/materials/legacy_textures/types/clouds.html#bpy-types-cloudstexture"),
    ("bpy.types.colorsequence*", "video_editing/edit/montage/strips/color.html#bpy-types-colorsequence"),
    ("bpy.types.crosssequence*", "video_editing/edit/montage/strips/transitions/cross.html#bpy-types-crosssequence"),
    ("bpy.types.curve.extrude*", "modeling/curves/properties/geometry.html#bpy-types-curve-extrude"),
    ("bpy.types.curvemodifier*", "modeling/modifiers/deform/curve.html#bpy-types-curvemodifier"),
    ("bpy.types.editbone.head*", "animation/armatures/bones/properties/transform.html#bpy-types-editbone-head"),
    ("bpy.types.editbone.lock*", "animation/armatures/bones/properties/transform.html#bpy-types-editbone-lock"),
    ("bpy.types.editbone.roll*", "animation/armatures/bones/properties/transform.html#bpy-types-editbone-roll"),
    ("bpy.types.editbone.tail*", "animation/armatures/bones/properties/transform.html#bpy-types-editbone-tail"),
    ("bpy.types.fieldsettings*", "physics/forces/force_fields/index.html#bpy-types-fieldsettings"),
    ("bpy.types.imagesequence*", "video_editing/edit/montage/strips/image.html#bpy-types-imagesequence"),
    ("bpy.types.key.eval_time*", "animation/shape_keys/shape_keys_panel.html#bpy-types-key-eval-time"),
    ("bpy.types.keyingsetpath*", "animation/keyframes/keying_sets.html#bpy-types-keyingsetpath"),
    ("bpy.types.marbletexture*", "render/materials/legacy_textures/types/marble.html#bpy-types-marbletexture"),
    ("bpy.types.modifier.name*", "modeling/modifiers/introduction.html#bpy-types-modifier-name"),
    ("bpy.types.modifier.show*", "modeling/modifiers/introduction.html#bpy-types-modifier-show"),
    ("bpy.types.moviesequence*", "video_editing/edit/montage/strips/movie.html#bpy-types-moviesequence"),
    ("bpy.types.movietracking*", "movie_clip/tracking/index.html#bpy-types-movietracking"),
    ("bpy.types.nlastrip.mute*", "editors/nla/sidebar.html#bpy-types-nlastrip-mute"),
    ("bpy.types.nodesmodifier*", "modeling/modifiers/generate/geometry_nodes.html#bpy-types-nodesmodifier"),
    ("bpy.types.object.parent*", "scene_layout/object/editing/parent.html#bpy-types-object-parent"),
    ("bpy.types.objectlineart*", "scene_layout/object/properties/line_art.html#bpy-types-objectlineart"),
    ("bpy.types.oceanmodifier*", "modeling/modifiers/physics/ocean.html#bpy-types-oceanmodifier"),
    ("bpy.types.particlebrush*", "physics/particles/mode.html#bpy-types-particlebrush"),
    ("bpy.types.raytraceeevee*", "render/eevee/render_settings/raytracing.html#bpy-types-raytraceeevee"),
    ("bpy.types.scene.gravity*", "physics/forces/gravity.html#bpy-types-scene-gravity"),
    ("bpy.types.scenesequence*", "video_editing/edit/montage/strips/scene.html#bpy-types-scenesequence"),
    ("bpy.types.screwmodifier*", "modeling/modifiers/generate/screw.html#bpy-types-screwmodifier"),
    ("bpy.types.shaderfxswirl*", "grease_pencil/visual_effects/swirl.html#bpy-types-shaderfxswirl"),
    ("bpy.types.shadernodemix*", "render/shader_nodes/converter/mix.html#bpy-types-shadernodemix"),
    ("bpy.types.shadernodergb*", "render/shader_nodes/input/rgb.html#bpy-types-shadernodergb"),
    ("bpy.types.shapekey.mute*", "animation/shape_keys/shape_keys_panel.html#bpy-types-shapekey-mute"),
    ("bpy.types.soundsequence*", "video_editing/edit/montage/strips/sound.html#bpy-types-soundsequence"),
    ("bpy.types.spaceoutliner*", "editors/outliner/index.html#bpy-types-spaceoutliner"),
    ("bpy.types.spacetimeline*", "editors/timeline.html#bpy-types-spacetimeline"),
    ("bpy.types.spaceuveditor*", "editors/uv/index.html#bpy-types-spaceuveditor"),
    ("bpy.types.strip.channel*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-channel"),
    ("bpy.types.stripmodifier*", "editors/video_sequencer/sequencer/sidebar/modifiers.html#bpy-types-stripmodifier"),
    ("bpy.types.stuccitexture*", "render/materials/legacy_textures/types/stucci.html#bpy-types-stuccitexture"),
    ("bpy.types.textbox.width*", "modeling/texts/properties.html#bpy-types-textbox-width"),
    ("bpy.types.texturenodeat*", "editors/texture_node/types/distort/at.html#bpy-types-texturenodeat"),
    ("bpy.types.view3doverlay*", "editors/3dview/display/overlays.html#bpy-types-view3doverlay"),
    ("bpy.types.viewlayer.use*", "render/layers/view_layer.html#bpy-types-viewlayer-use"),
    ("bpy.types.volumedisplay*", "modeling/volumes/properties.html#bpy-types-volumedisplay"),
    ("bpy.types.windowmanager*", "interface/index.html#bpy-types-windowmanager"),
    ("bpy.types.worldlighting*", "render/cycles/world_settings.html#bpy-types-worldlighting"),
    ("bpy.ops.*.select_lasso*", "interface/selecting.html#bpy-ops-select-lasso"),
    ("bpy.ops.armature.align*", "animation/armatures/bones/editing/transform.html#bpy-ops-armature-align"),
    ("bpy.ops.armature.split*", "animation/armatures/bones/editing/split.html#bpy-ops-armature-split"),
    ("bpy.ops.clip.set_plane*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-set-plane"),
    ("bpy.ops.clip.set_scale*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-set-scale"),
    ("bpy.ops.curve.decimate*", "modeling/curves/editing/curve.html#bpy-ops-curve-decimate"),
    ("bpy.ops.curve.separate*", "modeling/curves/editing/curve.html#bpy-ops-curve-separate"),
    ("bpy.ops.fluid.bake_all*", "physics/fluid/type/domain/cache.html#bpy-ops-fluid-bake-all"),
    ("bpy.ops.fluid.free_all*", "physics/fluid/type/domain/cache.html#bpy-ops-fluid-free-all"),
    ("bpy.ops.font.text_copy*", "modeling/texts/editing.html#bpy-ops-font-text-copy"),
    ("bpy.ops.graph.decimate*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-decimate"),
    ("bpy.ops.material.paste*", "render/materials/assignment.html#bpy-ops-material-paste"),
    ("bpy.ops.mesh.fill_grid*", "modeling/meshes/editing/face/grid_fill.html#bpy-ops-mesh-fill-grid"),
    ("bpy.ops.mesh.intersect*", "modeling/meshes/editing/face/intersect_knife.html#bpy-ops-mesh-intersect"),
    ("bpy.ops.mesh.mark_seam*", "modeling/meshes/editing/edge/edge_data.html#bpy-ops-mesh-mark-seam"),
    ("bpy.ops.mesh.polybuild*", "modeling/meshes/tools/poly_build.html#bpy-ops-mesh-polybuild"),
    ("bpy.ops.mesh.subdivide*", "modeling/meshes/editing/edge/subdivide.html#bpy-ops-mesh-subdivide"),
    ("bpy.ops.mesh.wireframe*", "modeling/meshes/editing/face/wireframe.html#bpy-ops-mesh-wireframe"),
    ("bpy.ops.nla.tracks_add*", "editors/nla/editing/track.html#bpy-ops-nla-tracks-add"),
    ("bpy.ops.node.add_image*", "addons/node/node_wrangler.html#bpy-ops-node-add-image"),
    ("bpy.ops.node.link_make*", "interface/controls/nodes/editing.html#bpy-ops-node-link-make"),
    ("bpy.ops.node.links_cut*", "interface/controls/nodes/editing.html#bpy-ops-node-links-cut"),
    ("bpy.ops.object.convert*", "scene_layout/object/editing/convert.html#bpy-ops-object-convert"),
    ("bpy.ops.object.speaker*", "render/output/audio/speaker.html#bpy-ops-object-speaker"),
    ("bpy.ops.paintcurve.new*", "sculpt_paint/brush/stroke.html#bpy-ops-paintcurve-new"),
    ("bpy.ops.pose.breakdown*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-breakdown"),
    ("bpy.ops.pose.loc_clear*", "animation/armatures/posing/editing/clear.html#bpy-ops-pose-loc-clear"),
    ("bpy.ops.pose.propagate*", "animation/armatures/posing/editing/propagate.html#bpy-ops-pose-propagate"),
    ("bpy.ops.pose.push_rest*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-push-rest"),
    ("bpy.ops.pose.rot_clear*", "animation/armatures/posing/editing/clear.html#bpy-ops-pose-rot-clear"),
    ("bpy.ops.sequencer.lock*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-lock"),
    ("bpy.ops.sequencer.mute*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-mute"),
    ("bpy.ops.sequencer.slip*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-slip"),
    ("bpy.ops.sequencer.snap*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-snap"),
    ("bpy.ops.sequencer.swap*", "video_editing/edit/montage/editing.html#bpy-ops-sequencer-swap"),
    ("bpy.ops.transform.bend*", "modeling/meshes/editing/mesh/transform/bend.html#bpy-ops-transform-bend"),
    ("bpy.ops.transform.tilt*", "modeling/curves/editing/control_points.html#bpy-ops-transform-tilt"),
    ("bpy.ops.uv.select_mode*", "editors/uv/selecting.html#bpy-ops-uv-select-mode"),
    ("bpy.ops.uv.snap_cursor*", "modeling/meshes/uv/editing.html#bpy-ops-uv-snap-cursor"),
    ("bpy.ops.wm.search_menu*", "interface/operators.html#bpy-ops-wm-search-menu"),
    ("bpy.types.bakesettings*", "render/cycles/baking.html#bpy-types-bakesettings"),
    ("bpy.types.blendtexture*", "render/materials/legacy_textures/types/blend.html#bpy-types-blendtexture"),
    ("bpy.types.brush.height*", "sculpt_paint/sculpting/brushes/layer.html#bpy-types-brush-height"),
    ("bpy.types.brush.jitter*", "sculpt_paint/brush/stroke.html#bpy-types-brush-jitter"),
    ("bpy.types.castmodifier*", "modeling/modifiers/deform/cast.html#bpy-types-castmodifier"),
    ("bpy.types.curve.offset*", "modeling/curves/properties/geometry.html#bpy-types-curve-offset"),
    ("bpy.types.drivertarget*", "animation/drivers/drivers_panel.html#bpy-types-drivertarget"),
    ("bpy.types.geometrynode*", "modeling/geometry_nodes/index.html#bpy-types-geometrynode"),
    ("bpy.types.glowsequence*", "video_editing/edit/montage/strips/effects/glow.html#bpy-types-glowsequence"),
    ("bpy.types.gpencillayer*", "interface/annotate_tool.html#bpy-types-gpencillayer"),
    ("bpy.types.hookmodifier*", "modeling/modifiers/deform/hooks.html#bpy-types-hookmodifier"),
    ("bpy.types.image.source*", "editors/image/image_settings.html#bpy-types-image-source"),
    ("bpy.types.imagetexture*", "render/materials/legacy_textures/types/image_movie.html#bpy-types-imagetexture"),
    ("bpy.types.latticepoint*", "animation/lattice.html#bpy-types-latticepoint"),
    ("bpy.types.light.shadow*", "render/eevee/light_settings.html#bpy-types-light-shadow"),
    ("bpy.types.magictexture*", "render/materials/legacy_textures/types/magic.html#bpy-types-magictexture"),
    ("bpy.types.maskmodifier*", "modeling/modifiers/generate/mask.html#bpy-types-maskmodifier"),
    ("bpy.types.masksequence*", "video_editing/edit/montage/strips/mask.html#bpy-types-masksequence"),
    ("bpy.types.materialslot*", "render/materials/assignment.html#bpy-types-materialslot"),
    ("bpy.types.metasequence*", "video_editing/edit/montage/meta.html#bpy-types-metasequence"),
    ("bpy.types.object.color*", "scene_layout/object/properties/display.html#bpy-types-object-color"),
    ("bpy.types.object.delta*", "scene_layout/object/properties/transforms.html#bpy-types-object-delta"),
    ("bpy.types.object.empty*", "modeling/empties.html#bpy-types-object-empty"),
    ("bpy.types.object.scale*", "scene_layout/object/properties/transforms.html#bpy-types-object-scale"),
    ("bpy.types.palettecolor*", "interface/controls/templates/color_palette.html#bpy-types-palettecolor"),
    ("bpy.types.particleedit*", "physics/particles/mode.html#bpy-types-particleedit"),
    ("bpy.types.scene.camera*", "scene_layout/scene/properties.html#bpy-types-scene-camera"),
    ("bpy.types.selectionset*", "animation/armatures/properties/selection_sets.html#bpy-types-selectionset"),
    ("bpy.types.sequencecrop*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-sequencecrop"),
    ("bpy.types.shaderfxblur*", "grease_pencil/visual_effects/blur.html#bpy-types-shaderfxblur"),
    ("bpy.types.shaderfxflip*", "grease_pencil/visual_effects/flip.html#bpy-types-shaderfxflip"),
    ("bpy.types.shaderfxglow*", "grease_pencil/visual_effects/glow.html#bpy-types-shaderfxglow"),
    ("bpy.types.shaderfxwave*", "grease_pencil/visual_effects/wave_distortion.html#bpy-types-shaderfxwave"),
    ("bpy.types.skinmodifier*", "modeling/modifiers/generate/skin.html#bpy-types-skinmodifier"),
    ("bpy.types.spaceconsole*", "editors/python_console.html#bpy-types-spaceconsole"),
    ("bpy.types.textsequence*", "video_editing/edit/montage/strips/text.html#bpy-types-textsequence"),
    ("bpy.types.unitsettings*", "scene_layout/scene/properties.html#bpy-types-unitsettings"),
    ("bpy.types.vertexcolors*", "sculpt_paint/vertex_paint/index.html#bpy-types-vertexcolors"),
    ("bpy.types.vertexgroups*", "modeling/meshes/properties/vertex_groups/index.html#bpy-types-vertexgroups"),
    ("bpy.types.view3dcursor*", "editors/3dview/3d_cursor.html#bpy-types-view3dcursor"),
    ("bpy.types.volumerender*", "modeling/volumes/properties.html#bpy-types-volumerender"),
    ("bpy.types.warpmodifier*", "modeling/modifiers/deform/warp.html#bpy-types-warpmodifier"),
    ("bpy.types.wavemodifier*", "modeling/modifiers/deform/wave.html#bpy-types-wavemodifier"),
    ("bpy.types.weldmodifier*", "modeling/modifiers/generate/weld.html#bpy-types-weldmodifier"),
    ("bpy.types.wipesequence*", "video_editing/edit/montage/strips/transitions/wipe.html#bpy-types-wipesequence"),
    ("bpy.ops.armature.fill*", "animation/armatures/bones/editing/fill_between_joints.html#bpy-ops-armature-fill"),
    ("bpy.ops.asset.tag_add*", "editors/asset_browser.html#bpy-ops-asset-tag-add"),
    ("bpy.ops.clip.prefetch*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-prefetch"),
    ("bpy.ops.clip.set_axis*", "movie_clip/tracking/clip/editing/reconstruction.html#bpy-ops-clip-set-axis"),
    ("bpy.ops.curves.delete*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-delete"),
    ("bpy.ops.file.pack_all*", "files/blend/packed_data.html#bpy-ops-file-pack-all"),
    ("bpy.ops.file.previous*", "editors/file_browser.html#bpy-ops-file-previous"),
    ("bpy.ops.font.case_set*", "modeling/texts/editing.html#bpy-ops-font-case-set"),
    ("bpy.ops.font.text_cut*", "modeling/texts/editing.html#bpy-ops-font-text-cut"),
    ("bpy.ops.grease_pencil*", "grease_pencil/index.html#bpy-ops-grease-pencil"),
    ("bpy.ops.image.project*", "sculpt_paint/texture_paint/tool_settings/options.html#bpy-ops-image-project"),
    ("bpy.ops.image.replace*", "editors/image/editing.html#bpy-ops-image-replace"),
    ("bpy.ops.image.save_as*", "editors/image/editing.html#bpy-ops-image-save-as"),
    ("bpy.ops.marker.delete*", "animation/markers.html#bpy-ops-marker-delete"),
    ("bpy.ops.marker.select*", "animation/markers.html#bpy-ops-marker-select"),
    ("bpy.ops.material.copy*", "render/materials/assignment.html#bpy-ops-material-copy"),
    ("bpy.ops.mesh.decimate*", "modeling/meshes/editing/mesh/cleanup.html#bpy-ops-mesh-decimate"),
    ("bpy.ops.mesh.dissolve*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-dissolve"),
    ("bpy.ops.mesh.rip_move*", "modeling/meshes/editing/vertex/rip_vertices.html#bpy-ops-mesh-rip-move"),
    ("bpy.ops.mesh.separate*", "modeling/meshes/editing/mesh/separate.html#bpy-ops-mesh-separate"),
    ("bpy.ops.mesh.solidify*", "modeling/meshes/editing/face/solidify_faces.html#bpy-ops-mesh-solidify"),
    ("bpy.ops.nla.duplicate*", "editors/nla/editing/strip.html#bpy-ops-nla-duplicate"),
    ("bpy.ops.nla.move_down*", "editors/nla/editing/strip.html#bpy-ops-nla-move-down"),
    ("bpy.ops.object.*clear*", "scene_layout/object/editing/clear.html#bpy-ops-object-clear"),
    ("bpy.ops.object.delete*", "scene_layout/object/editing/delete.html#bpy-ops-object-delete"),
    ("bpy.ops.render.opengl*", "editors/3dview/viewport_render.html#bpy-ops-render-opengl"),
    ("bpy.ops.screen.header*", "interface/window_system/regions.html#bpy-ops-screen-header"),
    ("bpy.ops.script.reload*", "advanced/operators.html#bpy-ops-script-reload"),
    ("bpy.ops.sculpt.expand*", "sculpt_paint/sculpting/editing/expand.html#bpy-ops-sculpt-expand"),
    ("bpy.ops.sculpt_curves*", "sculpt_paint/curves_sculpting/index.html#bpy-ops-sculpt-curves"),
    ("bpy.ops.sound.mixdown*", "render/output/audio/introduction.html#bpy-ops-sound-mixdown"),
    ("bpy.ops.ui.eyedropper*", "interface/controls/buttons/eyedropper.html#bpy-ops-ui-eyedropper"),
    ("bpy.ops.uv.select_all*", "editors/uv/selecting.html#bpy-ops-uv-select-all"),
    ("bpy.ops.view3d.select*", "editors/3dview/selecting.html#bpy-ops-view3d-select"),
    ("bpy.ops.wm.debug_menu*", "advanced/operators.html#bpy-ops-wm-debug-menu"),
    ("bpy.ops.wm.obj_export*", "files/import_export/obj.html#bpy-ops-wm-obj-export"),
    ("bpy.ops.wm.obj_import*", "files/import_export/obj.html#bpy-ops-wm-obj-import"),
    ("bpy.ops.wm.ply_export*", "files/import_export/ply.html#bpy-ops-wm-ply-export"),
    ("bpy.ops.wm.ply_import*", "files/import_export/ply.html#bpy-ops-wm-ply-import"),
    ("bpy.ops.wm.properties*", "files/custom_properties.html#bpy-ops-wm-properties"),
    ("bpy.ops.wm.stl_import*", "files/import_export/stl.html#bpy-ops-wm-stl-import"),
    ("bpy.ops.wm.usd_export*", "files/import_export/usd.html#bpy-ops-wm-usd-export"),
    ("bpy.ops.workspace.add*", "interface/window_system/workspaces.html#bpy-ops-workspace-add"),
    ("bpy.types.addsequence*", "video_editing/edit/montage/strips/effects/add.html#bpy-types-addsequence"),
    ("bpy.types.brush.cloth*", "sculpt_paint/sculpting/brushes/cloth.html#bpy-types-brush-cloth"),
    ("bpy.types.brush.color*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-color"),
    ("bpy.types.brush.curve*", "sculpt_paint/brush/falloff.html#bpy-types-brush-curve"),
    ("bpy.types.camera.show*", "render/cameras.html#bpy-types-camera-show"),
    ("bpy.types.consoleline*", "editors/python_console.html#bpy-types-consoleline"),
    ("bpy.types.curve.bevel*", "modeling/curves/properties/geometry.html#bpy-types-curve-bevel"),
    ("bpy.types.driver.type*", "animation/drivers/drivers_panel.html#bpy-types-driver-type"),
    ("bpy.types.light.color*", "render/lights/light_object.html#bpy-types-light-color"),
    ("bpy.types.mesh.remesh*", "modeling/meshes/retopology.html#bpy-types-mesh-remesh"),
    ("bpy.types.meshstatvis*", "modeling/meshes/mesh_analysis.html#bpy-types-meshstatvis"),
    ("bpy.types.nodesetting*", "interface/controls/nodes/parts.html#bpy-types-nodesetting"),
    ("bpy.types.object.lock*", "scene_layout/object/properties/transforms.html#bpy-types-object-lock"),
    ("bpy.types.object.show*", "scene_layout/object/properties/display.html#bpy-types-object-show"),
    ("bpy.types.particlekey*", "physics/particles/emitter/physics/keyed.html#bpy-types-particlekey"),
    ("bpy.types.pathcompare*", "editors/preferences/save_load.html#bpy-types-pathcompare"),
    ("bpy.types.posebone.ik*", "animation/armatures/posing/bone_constraints/inverse_kinematics/introduction.html#bpy-types-posebone-ik"),
    ("bpy.types.preferences*", "editors/preferences/index.html#bpy-types-preferences"),
    ("bpy.types.renderlayer*", "render/layers/passes.html#bpy-types-renderlayer"),
    ("bpy.types.sculpt.lock*", "sculpt_paint/sculpting/tool_settings/symmetry.html#bpy-types-sculpt-lock"),
    ("bpy.types.shaderfxrim*", "grease_pencil/visual_effects/rim.html#bpy-types-shaderfxrim"),
    ("bpy.types.spaceview3d*", "editors/3dview/index.html#bpy-types-spaceview3d"),
    ("bpy.types.uipopupmenu*", "interface/controls/buttons/menus.html#bpy-types-uipopupmenu"),
    ("bpy.types.vertexpaint*", "sculpt_paint/vertex_paint/index.html#bpy-types-vertexpaint"),
    ("bpy.types.volumegrids*", "modeling/volumes/properties.html#bpy-types-volumegrids"),
    ("bpy.types.woodtexture*", "render/materials/legacy_textures/types/wood.html#bpy-types-woodtexture"),
    ("bpy.types.world.color*", "render/lights/world.html#bpy-types-world-color"),
    ("bpy.ops.*.select_box*", "interface/selecting.html#bpy-ops-select-box"),
    ("bpy.ops.curve.delete*", "modeling/curves/editing/curve.html#bpy-ops-curve-delete"),
    ("bpy.ops.curve.reveal*", "modeling/curves/editing/curve.html#bpy-ops-curve-reveal"),
    ("bpy.ops.curve.smooth*", "modeling/curves/editing/control_points.html#bpy-ops-curve-smooth"),
    ("bpy.ops.curves.split*", "modeling/curves_new/editing/curves.html#bpy-ops-curves-split"),
    ("bpy.ops.file.execute*", "editors/file_browser.html#bpy-ops-file-execute"),
    ("bpy.ops.file.refresh*", "editors/file_browser.html#bpy-ops-file-refresh"),
    ("bpy.ops.fluid.preset*", "physics/fluid/type/domain/liquid/diffusion.html#bpy-ops-fluid-preset"),
    ("bpy.ops.graph.delete*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-delete"),
    ("bpy.ops.graph.mirror*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-mirror"),
    ("bpy.ops.graph.reveal*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-reveal"),
    ("bpy.ops.graph.smooth*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-smooth"),
    ("bpy.ops.image.invert*", "editors/image/editing.html#bpy-ops-image-invert"),
    ("bpy.ops.image.reload*", "editors/image/editing.html#bpy-ops-image-reload"),
    ("bpy.ops.image.resize*", "editors/image/editing.html#bpy-ops-image-resize"),
    ("bpy.ops.image.unpack*", "editors/image/editing.html#bpy-ops-image-unpack"),
    ("bpy.ops.material.new*", "render/materials/assignment.html#bpy-ops-material-new"),
    ("bpy.ops.nla.meta_add*", "editors/nla/editing/strip.html#bpy-ops-nla-meta-add"),
    ("bpy.ops.object.align*", "scene_layout/object/editing/transform/align_objects.html#bpy-ops-object-align"),
    ("bpy.ops.object.empty*", "modeling/empties.html#bpy-ops-object-empty"),
    ("bpy.ops.object.quick*", "physics/introduction.html#bpy-ops-object-quick"),
    ("bpy.ops.scene.delete*", "scene_layout/scene/introduction.html#bpy-ops-scene-delete"),
    ("bpy.ops.text.replace*", "editors/text_editor.html#bpy-ops-text-replace"),
    ("bpy.ops.text.save_as*", "editors/text_editor.html#bpy-ops-text-save-as"),
    ("bpy.ops.uv.mark_seam*", "modeling/meshes/uv/unwrapping/seams.html#bpy-ops-uv-mark-seam"),
    ("bpy.ops.view3d.dolly*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-dolly"),
    ("bpy.ops.view3d.ruler*", "editors/3dview/toolbar/measure.html#bpy-ops-view3d-ruler"),
    ("bpy.types.areaspaces*", "interface/window_system/areas.html#bpy-types-areaspaces"),
    ("bpy.types.bone.color*", "animation/armatures/bones/properties/display.html#bpy-types-bone-color"),
    ("bpy.types.bpy_struct*", "files/custom_properties.html#bpy-types-bpy-struct"),
    ("bpy.types.brush.flow*", "sculpt_paint/sculpting/brushes/paint.html#bpy-types-brush-flow"),
    ("bpy.types.brush.rate*", "sculpt_paint/brush/stroke.html#bpy-types-brush-rate"),
    ("bpy.types.brush.size*", "sculpt_paint/brush/brush_settings.html#bpy-types-brush-size"),
    ("bpy.types.collection*", "scene_layout/collections/collections.html#bpy-types-collection"),
    ("bpy.types.compositor*", "compositing/index.html#bpy-types-compositor"),
    ("bpy.types.constraint*", "animation/constraints/index.html#bpy-types-constraint"),
    ("bpy.types.imagepaint*", "sculpt_paint/texture_paint/index.html#bpy-types-imagepaint"),
    ("bpy.types.keyingsets*", "animation/keyframes/keying_sets.html#bpy-types-keyingsets"),
    ("bpy.types.keymapitem*", "editors/preferences/keymap.html#bpy-types-keymapitem"),
    ("bpy.types.light.type*", "render/lights/light_object.html#bpy-types-light-type"),
    ("bpy.types.lightprobe*", "render/eevee/light_probes/index.html#bpy-types-lightprobe"),
    ("bpy.types.maskparent*", "movie_clip/masking/sidebar.html#bpy-types-maskparent"),
    ("bpy.types.maskspline*", "movie_clip/masking/sidebar.html#bpy-types-maskspline"),
    ("bpy.types.motionpath*", "animation/motion_paths.html#bpy-types-motionpath"),
    ("bpy.types.node.color*", "interface/controls/nodes/sidebar.html#bpy-types-node-color"),
    ("bpy.types.node.label*", "interface/controls/nodes/sidebar.html#bpy-types-node-label"),
    ("bpy.types.nodesocket*", "interface/controls/nodes/parts.html#bpy-types-nodesocket"),
    ("bpy.types.paint.tile*", "sculpt_paint/texture_paint/tool_settings/tiling.html#bpy-types-paint-tile"),
    ("bpy.types.pointcache*", "physics/baking.html#bpy-types-pointcache"),
    ("bpy.types.pointlight*", "render/lights/light_object.html#bpy-types-pointlight"),
    ("bpy.types.renderview*", "render/output/properties/stereoscopy/index.html#bpy-types-renderview"),
    ("bpy.types.sceneeevee*", "render/eevee/index.html#bpy-types-sceneeevee"),
    ("bpy.types.strip.name*", "editors/video_sequencer/sequencer/sidebar/strip.html#bpy-types-strip-name"),
    ("bpy.types.stripproxy*", "editors/video_sequencer/sequencer/sidebar/proxy.html#bpy-types-stripproxy"),
    ("bpy.types.vectorfont*", "modeling/texts/index.html#bpy-types-vectorfont"),
    ("bpy.ops.asset.clear*", "files/asset_libraries/introduction.html#bpy-ops-asset-clear"),
    ("bpy.ops.clip.reload*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-reload"),
    ("bpy.ops.curve.split*", "modeling/curves/editing/curve.html#bpy-ops-curve-split"),
    ("bpy.ops.file.cancel*", "editors/file_browser.html#bpy-ops-file-cancel"),
    ("bpy.ops.file.delete*", "editors/file_browser.html#bpy-ops-file-delete"),
    ("bpy.ops.file.parent*", "editors/file_browser.html#bpy-ops-file-parent"),
    ("bpy.ops.file.rename*", "editors/file_browser.html#bpy-ops-file-rename"),
    ("bpy.ops.font.delete*", "modeling/texts/editing.html#bpy-ops-font-delete"),
    ("bpy.ops.graph.clean*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-clean"),
    ("bpy.ops.graph.paste*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-paste"),
    ("bpy.ops.graph.shear*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-shear"),
    ("bpy.ops.marker.move*", "animation/markers.html#bpy-ops-marker-move"),
    ("bpy.ops.mask.delete*", "movie_clip/masking/editing.html#bpy-ops-mask-delete"),
    ("bpy.ops.mesh.bisect*", "modeling/meshes/editing/mesh/bisect.html#bpy-ops-mesh-bisect"),
    ("bpy.ops.mesh.delete*", "modeling/meshes/editing/mesh/delete.html#bpy-ops-mesh-delete"),
    ("bpy.ops.nla.move_up*", "editors/nla/editing/strip.html#bpy-ops-nla-move-up"),
    ("bpy.ops.node.delete*", "interface/controls/nodes/editing.html#bpy-ops-node-delete"),
    ("bpy.ops.node.detach*", "interface/controls/nodes/frame.html#bpy-ops-node-detach"),
    ("bpy.ops.object.bake*", "render/cycles/baking.html#bpy-ops-object-bake"),
    ("bpy.ops.object.hook*", "modeling/meshes/editing/vertex/hooks.html#bpy-ops-object-hook"),
    ("bpy.ops.object.join*", "scene_layout/object/editing/join.html#bpy-ops-object-join"),
    ("bpy.ops.object.text*", "modeling/texts/index.html#bpy-ops-object-text"),
    ("bpy.ops.preferences*", "editors/preferences/index.html#bpy-ops-preferences"),
    ("bpy.ops.spreadsheet*", "editors/spreadsheet.html#bpy-ops-spreadsheet"),
    ("bpy.ops.text.reload*", "editors/text_editor.html#bpy-ops-text-reload"),
    ("bpy.ops.uv.rip_move*", "modeling/meshes/uv/editing.html#bpy-ops-uv-rip-move"),
    ("bpy.ops.view3d.snap*", "scene_layout/object/editing/snap.html#bpy-ops-view3d-snap"),
    ("bpy.ops.view3d.walk*", "editors/3dview/navigate/walk_fly.html#bpy-ops-view3d-walk"),
    ("bpy.ops.view3d.zoom*", "editors/3dview/navigate/navigation.html#bpy-ops-view3d-zoom"),
    ("bpy.types.arealight*", "render/lights/light_object.html#bpy-types-arealight"),
    ("bpy.types.blenddata*", "files/data_blocks.html#bpy-types-blenddata"),
    ("bpy.types.bone.hide*", "animation/armatures/bones/properties/display.html#bpy-types-bone-hide"),
    ("bpy.types.bonecolor*", "animation/armatures/bones/properties/display.html#bpy-types-bonecolor"),
    ("bpy.types.colorramp*", "interface/controls/templates/color_ramp.html#bpy-types-colorramp"),
    ("bpy.types.dopesheet*", "editors/dope_sheet/index.html#bpy-types-dopesheet"),
    ("bpy.types.fmodifier*", "editors/graph_editor/fcurves/modifiers.html#bpy-types-fmodifier"),
    ("bpy.types.freestyle*", "render/freestyle/index.html#bpy-types-freestyle"),
    ("bpy.types.masklayer*", "movie_clip/masking/sidebar.html#bpy-types-masklayer"),
    ("bpy.types.movieclip*", "movie_clip/index.html#bpy-types-movieclip"),
    ("bpy.types.node.name*", "interface/controls/nodes/sidebar.html#bpy-types-node-name"),
    ("bpy.types.nodeframe*", "interface/controls/nodes/frame.html#bpy-types-nodeframe"),
    ("bpy.types.nodegroup*", "interface/controls/nodes/groups.html#bpy-types-nodegroup"),
    ("bpy.types.spotlight*", "render/lights/light_object.html#bpy-types-spotlight"),
    ("bpy.types.textbox.x*", "modeling/texts/properties.html#bpy-types-textbox-x"),
    ("bpy.types.textbox.y*", "modeling/texts/properties.html#bpy-types-textbox-y"),
    ("bpy.types.textcurve*", "modeling/texts/index.html#bpy-types-textcurve"),
    ("bpy.types.udimtiles*", "modeling/meshes/uv/workflows/udims.html#bpy-types-udimtiles"),
    ("bpy.types.uipiemenu*", "interface/controls/buttons/menus.html#bpy-types-uipiemenu"),
    ("bpy.types.uipopover*", "interface/controls/buttons/menus.html#bpy-types-uipopover"),
    ("bpy.types.viewlayer*", "render/layers/introduction.html#bpy-types-viewlayer"),
    ("bpy.ops.asset.mark*", "files/asset_libraries/introduction.html#bpy-ops-asset-mark"),
    ("bpy.ops.collection*", "scene_layout/collections/collections.html#bpy-ops-collection"),
    ("bpy.ops.constraint*", "animation/constraints/index.html#bpy-ops-constraint"),
    ("bpy.ops.curve.draw*", "modeling/curves/tools/draw.html#bpy-ops-curve-draw"),
    ("bpy.ops.curve.hide*", "modeling/curves/editing/curve.html#bpy-ops-curve-hide"),
    ("bpy.ops.curve.spin*", "modeling/surfaces/editing/surface.html#bpy-ops-curve-spin"),
    ("bpy.ops.curves.add*", "modeling/curves_new/primitives.html#bpy-ops-curves-add"),
    ("bpy.ops.graph.copy*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-copy"),
    ("bpy.ops.graph.ease*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-ease"),
    ("bpy.ops.graph.hide*", "editors/graph_editor/channels/editing.html#bpy-ops-graph-hide"),
    ("bpy.ops.graph.snap*", "editors/graph_editor/fcurves/editing.html#bpy-ops-graph-snap"),
    ("bpy.ops.image.flip*", "editors/image/editing.html#bpy-ops-image-flip"),
    ("bpy.ops.image.open*", "editors/image/editing.html#bpy-ops-image-open"),
    ("bpy.ops.image.pack*", "editors/image/editing.html#bpy-ops-image-pack"),
    ("bpy.ops.image.save*", "editors/image/editing.html#bpy-ops-image-save"),
    ("bpy.ops.image.tile*", "modeling/meshes/uv/workflows/udims.html#bpy-ops-image-tile"),
    ("bpy.ops.marker.add*", "animation/markers.html#bpy-ops-marker-add"),
    ("bpy.ops.mesh.bevel*", "modeling/meshes/editing/edge/bevel.html#bpy-ops-mesh-bevel"),
    ("bpy.ops.mesh.inset*", "modeling/meshes/editing/face/inset_faces.html#bpy-ops-mesh-inset"),
    ("bpy.ops.mesh.knife*", "modeling/meshes/editing/mesh/knife_topology_tool.html#bpy-ops-mesh-knife"),
    ("bpy.ops.mesh.merge*", "modeling/meshes/editing/mesh/merge.html#bpy-ops-mesh-merge"),
    ("bpy.ops.mesh.screw*", "modeling/meshes/editing/edge/screw.html#bpy-ops-mesh-screw"),
    ("bpy.ops.mesh.split*", "modeling/meshes/editing/mesh/split.html#bpy-ops-mesh-split"),
    ("bpy.ops.nla.delete*", "editors/nla/editing/strip.html#bpy-ops-nla-delete"),
    ("bpy.ops.paint.mask*", "sculpt_paint/sculpting/editing/mask.html#bpy-ops-paint-mask"),
    ("bpy.ops.pose.paste*", "animation/armatures/posing/editing/copy_paste.html#bpy-ops-pose-paste"),
    ("bpy.ops.pose.relax*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-relax"),
    ("bpy.ops.safe_areas*", "render/cameras.html#bpy-ops-safe-areas"),
    ("bpy.ops.view3d.fly*", "editors/3dview/navigate/walk_fly.html#bpy-ops-view3d-fly"),
    ("bpy.ops.wm.toolbar*", "interface/tool_system.html#bpy-ops-wm-toolbar"),
    ("bpy.types.aov.type*", "render/layers/passes.html#bpy-types-aov-type"),
    ("bpy.types.armature*", "animation/armatures/index.html#bpy-types-armature"),
    ("bpy.types.editbone*", "animation/armatures/bones/editing/index.html#bpy-types-editbone"),
    ("bpy.types.keyframe*", "animation/keyframes/index.html#bpy-types-keyframe"),
    ("bpy.types.linesets*", "render/freestyle/view_layer/line_set.html#bpy-types-linesets"),
    ("bpy.types.material*", "render/materials/index.html#bpy-types-material"),
    ("bpy.types.metaball*", "modeling/metas/index.html#bpy-types-metaball"),
    ("bpy.types.modifier*", "modeling/modifiers/index.html#bpy-types-modifier"),
    ("bpy.types.nlastrip*", "editors/nla/strips.html#bpy-types-nlastrip"),
    ("bpy.types.nlatrack*", "editors/nla/tracks.html#bpy-types-nlatrack"),
    ("bpy.types.nodelink*", "interface/controls/nodes/parts.html#bpy-types-nodelink"),
    ("bpy.types.nodetree*", "interface/controls/nodes/parts.html#bpy-types-nodetree"),
    ("bpy.types.particle*", "physics/particles/index.html#bpy-types-particle"),
    ("bpy.types.shaderfx*", "grease_pencil/visual_effects/index.html#bpy-types-shaderfx"),
    ("bpy.types.shapekey*", "animation/shape_keys/index.html#bpy-types-shapekey"),
    ("bpy.types.spacenla*", "editors/nla/index.html#bpy-types-spacenla"),
    ("bpy.types.sunlight*", "render/lights/light_object.html#bpy-types-sunlight"),
    ("bpy.ops.clip.open*", "movie_clip/tracking/clip/editing/clip.html#bpy-ops-clip-open"),
    ("bpy.ops.curve.pen*", "modeling/curves/tools/pen.html#bpy-ops-curve-pen"),
    ("bpy.ops.file.next*", "editors/file_browser.html#bpy-ops-file-next"),
    ("bpy.ops.image.new*", "editors/image/editing.html#bpy-ops-image-new"),
    ("bpy.ops.mesh.fill*", "modeling/meshes/editing/face/fill.html#bpy-ops-mesh-fill"),
    ("bpy.ops.mesh.poke*", "modeling/meshes/editing/face/poke_faces.html#bpy-ops-mesh-poke"),
    ("bpy.ops.mesh.spin*", "modeling/meshes/tools/spin.html#bpy-ops-mesh-spin"),
    ("bpy.ops.nla.split*", "editors/nla/editing/strip.html#bpy-ops-nla-split"),
    ("bpy.ops.node.join*", "interface/controls/nodes/frame.html#bpy-ops-node-join"),
    ("bpy.ops.pose.copy*", "animation/armatures/posing/editing/copy_paste.html#bpy-ops-pose-copy"),
    ("bpy.ops.pose.push*", "animation/armatures/posing/editing/in_betweens.html#bpy-ops-pose-push"),
    ("bpy.ops.rigidbody*", "physics/rigid_body/index.html#bpy-ops-rigidbody"),
    ("bpy.ops.sequencer*", "video_editing/index.html#bpy-ops-sequencer"),
    ("bpy.ops.text.find*", "editors/text_editor.html#bpy-ops-text-find"),
    ("bpy.ops.text.open*", "editors/text_editor.html#bpy-ops-text-open"),
    ("bpy.ops.text.save*", "editors/text_editor.html#bpy-ops-text-save"),
    ("bpy.ops.transform*", "scene_layout/object/editing/transform/index.html#bpy-ops-transform"),
    ("bpy.ops.uv.reveal*", "modeling/meshes/uv/editing.html#bpy-ops-uv-reveal"),
    ("bpy.ops.uv.select*", "editors/uv/selecting.html#bpy-ops-uv-select"),
    ("bpy.ops.uv.stitch*", "modeling/meshes/uv/editing.html#bpy-ops-uv-stitch"),
    ("bpy.ops.uv.unwrap*", "modeling/meshes/editing/uv.html#bpy-ops-uv-unwrap"),
    ("bpy.ops.wm.append*", "files/linked_libraries/link_append.html#bpy-ops-wm-append"),
    ("bpy.ops.wm.splash*", "interface/window_system/splash.html#bpy-ops-wm-splash"),
    ("bpy.ops.workspace*", "interface/window_system/workspaces.html#bpy-ops-workspace"),
    ("bpy.types.lattice*", "animation/lattice.html#bpy-types-lattice"),
    ("bpy.types.library*", "files/linked_libraries/index.html#bpy-types-library"),
    ("bpy.types.speaker*", "render/output/audio/speaker.html#bpy-types-speaker"),
    ("bpy.types.textbox*", "modeling/texts/properties.html#bpy-types-textbox"),
    ("bpy.types.texture*", "render/materials/legacy_textures/index.html#bpy-types-texture"),
    ("bpy.ops.armature*", "animation/armatures/index.html#bpy-ops-armature"),
    ("bpy.ops.geometry*", "modeling/index.html#bpy-ops-geometry"),
    ("bpy.ops.material*", "render/materials/index.html#bpy-ops-material"),
    ("bpy.ops.nla.bake*", "editors/nla/editing/strip.html#bpy-ops-nla-bake"),
    ("bpy.ops.nla.snap*", "editors/nla/editing/strip.html#bpy-ops-nla-snap"),
    ("bpy.ops.nla.swap*", "editors/nla/editing/strip.html#bpy-ops-nla-swap"),
    ("bpy.ops.outliner*", "editors/outliner/index.html#bpy-ops-outliner"),
    ("bpy.ops.particle*", "physics/particles/index.html#bpy-ops-particle"),
    ("bpy.ops.text.new*", "editors/text_editor.html#bpy-ops-text-new"),
    ("bpy.ops.uv.align*", "modeling/meshes/uv/editing.html#bpy-ops-uv-align"),
    ("bpy.ops.uv.paste*", "modeling/meshes/uv/editing.html#bpy-ops-uv-paste"),
    ("bpy.ops.uv.reset*", "modeling/meshes/editing/uv.html#bpy-ops-uv-reset"),
    ("bpy.ops.wm.addon*", "editors/preferences/extensions.html#bpy-ops-wm-addon"),
    ("bpy.types.action*", "animation/actions.html#bpy-types-action"),
    ("bpy.types.camera*", "render/cameras.html#bpy-types-camera"),
    ("bpy.types.cycles*", "render/cycles/index.html#bpy-types-cycles"),
    ("bpy.types.driver*", "animation/drivers/index.html#bpy-types-driver"),
    ("bpy.types.fcurve*", "editors/graph_editor/fcurves/index.html#bpy-types-fcurve"),
    ("bpy.types.header*", "interface/window_system/regions.html#bpy-types-header"),
    ("bpy.types.object*", "scene_layout/object/index.html#bpy-types-object"),
    ("bpy.types.region*", "interface/window_system/regions.html#bpy-types-region"),
    ("bpy.types.render*", "render/index.html#bpy-types-render"),
    ("bpy.types.screen*", "interface/index.html#bpy-types-screen"),
    ("bpy.types.sculpt*", "sculpt_paint/sculpting/index.html#bpy-types-sculpt"),
    ("bpy.types.shader*", "render/shader_nodes/shader/index.html#bpy-types-shader"),
    ("bpy.types.spline*", "modeling/curves/properties/active_spline.html#bpy-types-spline"),
    ("bpy.types.volume*", "modeling/volumes/index.html#bpy-types-volume"),
    ("bpy.types.window*", "interface/index.html#bpy-types-window"),
    ("bpy.ops.buttons*", "interface/controls/buttons/buttons.html#bpy-ops-buttons"),
    ("bpy.ops.console*", "editors/python_console.html#bpy-ops-console"),
    ("bpy.ops.ed.redo*", "interface/undo_redo.html#bpy-ops-ed-redo"),
    ("bpy.ops.ed.undo*", "interface/undo_redo.html#bpy-ops-ed-undo"),
    ("bpy.ops.lattice*", "animation/lattice.html#bpy-ops-lattice"),
    ("bpy.ops.palette*", "interface/controls/templates/color_palette.html#bpy-ops-palette"),
    ("bpy.ops.poselib*", "animation/armatures/posing/editing/pose_library.html#bpy-ops-poselib"),
    ("bpy.ops.ptcache*", "physics/baking.html#bpy-ops-ptcache"),
    ("bpy.ops.surface*", "modeling/surfaces/index.html#bpy-ops-surface"),
    ("bpy.ops.texture*", "render/materials/legacy_textures/index.html#bpy-ops-texture"),
    ("bpy.ops.uv.copy*", "modeling/meshes/uv/editing.html#bpy-ops-uv-copy"),
    ("bpy.ops.uv.hide*", "modeling/meshes/uv/editing.html#bpy-ops-uv-hide"),
    ("bpy.ops.uv.weld*", "modeling/meshes/uv/editing.html#bpy-ops-uv-weld"),
    ("bpy.ops.wm.link*", "files/linked_libraries/link_append.html#bpy-ops-wm-link"),
    ("bpy.ops.wm.tool*", "interface/tool_system.html#bpy-ops-wm-tool"),
    ("bpy.types.addon*", "editors/preferences/extensions.html#bpy-types-addon"),
    ("bpy.types.brush*", "sculpt_paint/brush/introduction.html#bpy-types-brush"),
    ("bpy.types.curve*", "modeling/curves/index.html#bpy-types-curve"),
    ("bpy.types.image*", "files/media/image_formats.html#bpy-types-image"),
    ("bpy.types.itasc*", "animation/armatures/posing/bone_constraints/inverse_kinematics/introduction.html#bpy-types-itasc"),
    ("bpy.types.light*", "render/lights/light_object.html#bpy-types-light"),
    ("bpy.types.nodes*", "interface/controls/nodes/index.html#bpy-types-nodes"),
    ("bpy.types.paint*", "sculpt_paint/index.html#bpy-types-paint"),
    ("bpy.types.panel*", "interface/window_system/tabs_panels.html#bpy-types-panel"),
    ("bpy.types.scene*", "scene_layout/scene/index.html#bpy-types-scene"),
    ("bpy.types.sound*", "render/output/audio/index.html#bpy-types-sound"),
    ("bpy.types.space*", "editors/index.html#bpy-types-space"),
    ("bpy.types.strip*", "video_editing/index.html#bpy-types-strip"),
    ("bpy.types.theme*", "editors/preferences/themes.html#bpy-types-theme"),
    ("bpy.types.world*", "render/lights/world.html#bpy-types-world"),
    ("bpy.ops.action*", "animation/actions.html#bpy-ops-action"),
    ("bpy.ops.camera*", "render/cameras.html#bpy-ops-camera"),
    ("bpy.ops.cycles*", "render/cycles/index.html#bpy-ops-cycles"),
    ("bpy.ops.dpaint*", "physics/dynamic_paint/index.html#bpy-ops-dpaint"),
    ("bpy.ops.export*", "files/import_export/index.html#bpy-ops-export"),
    ("bpy.ops.import*", "files/import_export/index.html#bpy-ops-import"),
    ("bpy.ops.marker*", "animation/markers.html#bpy-ops-marker"),
    ("bpy.ops.object*", "scene_layout/object/index.html#bpy-ops-object"),
    ("bpy.ops.render*", "render/index.html#bpy-ops-render"),
    ("bpy.ops.script*", "advanced/scripting/index.html#bpy-ops-script"),
    ("bpy.ops.sculpt*", "sculpt_paint/sculpting/index.html#bpy-ops-sculpt"),
    ("bpy.ops.uv.pin*", "modeling/meshes/uv/editing.html#bpy-ops-uv-pin"),
    ("bpy.ops.uv.rip*", "modeling/meshes/uv/tools/rip.html#bpy-ops-uv-rip"),
    ("bpy.ops.view3d*", "editors/3dview/index.html#bpy-ops-view3d"),
    ("bpy.types.area*", "interface/window_system/areas.html#bpy-types-area"),
    ("bpy.types.boid*", "physics/particles/emitter/physics/boids.html#bpy-types-boid"),
    ("bpy.types.bone*", "animation/armatures/bones/index.html#bpy-types-bone"),
    ("bpy.types.mask*", "movie_clip/masking/index.html#bpy-types-mask"),
    ("bpy.types.menu*", "interface/controls/buttons/menus.html#bpy-types-menu"),
    ("bpy.types.mesh*", "modeling/meshes/index.html#bpy-types-mesh"),
    ("bpy.types.node*", "interface/controls/nodes/index.html#bpy-types-node"),
    ("bpy.types.pose*", "animation/armatures/posing/index.html#bpy-types-pose"),
    ("bpy.types.text*", "editors/text_editor.html#bpy-types-text"),
    ("bpy.ops.brush*", "sculpt_paint/brush/introduction.html#bpy-ops-brush"),
    ("bpy.ops.cloth*", "physics/cloth/index.html#bpy-ops-cloth"),
    ("bpy.ops.curve*", "modeling/curves/index.html#bpy-ops-curve"),
    ("bpy.ops.graph*", "editors/graph_editor/index.html#bpy-ops-graph"),
    ("bpy.ops.image*", "files/media/image_formats.html#bpy-ops-image"),
    ("bpy.ops.mball*", "modeling/metas/index.html#bpy-ops-mball"),
    ("bpy.ops.paint*", "sculpt_paint/index.html#bpy-ops-paint"),
    ("bpy.ops.scene*", "scene_layout/scene/index.html#bpy-ops-scene"),
    ("bpy.ops.sound*", "render/output/audio/index.html#bpy-ops-sound"),
    ("bpy.types.aov*", "render/layers/passes.html#bpy-types-aov"),
    ("bpy.types.key*", "animation/shape_keys/index.html#bpy-types-key"),
    ("bpy.ops.anim*", "animation/index.html#bpy-ops-anim"),
    ("bpy.ops.boid*", "physics/particles/emitter/physics/boids.html#bpy-ops-boid"),
    ("bpy.ops.clip*", "movie_clip/index.html#bpy-ops-clip"),
    ("bpy.ops.file*", "editors/file_browser.html#bpy-ops-file"),
    ("bpy.ops.font*", "modeling/texts/index.html#bpy-ops-font"),
    ("bpy.ops.mask*", "movie_clip/masking/index.html#bpy-ops-mask"),
    ("bpy.ops.mesh*", "modeling/meshes/index.html#bpy-ops-mesh"),
    ("bpy.ops.node*", "interface/controls/nodes/index.html#bpy-ops-node"),
    ("bpy.ops.pose*", "animation/armatures/posing/index.html#bpy-ops-pose"),
    ("bpy.ops.text*", "editors/text_editor.html#bpy-ops-text"),
    ("bpy.ops.time*", "editors/timeline.html#bpy-ops-time"),
    ("bpy.types.id*", "files/data_blocks.html#bpy-types-id"),
    ("bpy.ops.nla*", "editors/nla/index.html#bpy-ops-nla"),
    ("bpy.ops.ed*", "interface/undo_redo.html#bpy-ops-ed"),
    ("bpy.ops.ui*", "interface/index.html#bpy-ops-ui"),
    ("bpy.ops.wm*", "interface/index.html#bpy-ops-wm"),
)

# autopep8: on
